/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/* Test the new conditional-write protocol */

#include <math.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_error.h"
#include "pasis_timer.h"

#include "pasisio.h"
#include "pasisio_config_parse.h"
#include "pasisio_md_protocol.h"
#include "pasisio_drives.h"
#include "pasisio_select.h"

#include "pasis_md_prot.h"


static pasis_config_server_list_t md_server_list;
static pasis_md_protocol_t        md_protocol;
static pasis_md_linkage_t         md_linkages[PASIS_MD_MAX_LINKAGES];



void
do_update (int fid1, int fid2,
           int modified_chunks1, int modified_chunks2,
           p_uint32 server_override_bitmap)
{
    pasis_md_fid_t *fid_array[2], fid_1, fid_2;
    p_uint32 modified_chunks[2];
    int arg_size = 512, res_size = 128, chunk_size = 128, compute_time = 0;
    pasis_error_t error;
    
    /* Setup FIDs */
    fid_array[0] = &fid_1;
    fid_array[1] = &fid_2;
    memset (fid_array[0]->data, fid1, PASISIO_FILEID_SIZE);
    memset (fid_array[1]->data, fid2, PASISIO_FILEID_SIZE);

    modified_chunks[0] = modified_chunks1;
    modified_chunks[1] = modified_chunks2;

    error = PasisIO_MDSynthUpdate (&md_protocol,
                                   fid_array, 
                                   2,
                                   md_linkages,
                                   md_server_list.num_servers,
                                   arg_size,
                                   res_size,
                                   chunk_size,
                                   modified_chunks,
                                   compute_time,
                                   1, /* TS increment */
                                   1, /* use cached history */
                                   server_override_bitmap,
                                   NULL);

    fprintf(stderr, "PasisIO_SynthUpdate returned %d\n", error);
}


void
do_single_update_data (int fid1,
                       int modified_chunks1,
                       int use_cached_history,
                       /* For testing, useful for client to force server to
                        * increment logical TS by more than 1 */
                       int ts_increment, /* Usually 1 */
                       p_uint32 server_override_bitmap,
                       p_uint32 arg_size,
                       p_uint32 res_size,
                       p_uint32 chunk_size,
                       p_byte *data)
{
    pasis_md_fid_t *fid_array[1], fid_1;
    p_uint32 modified_chunks[1];
    pasis_error_t error;
    
    /* Setup FIDs */
    fid_array[0] = &fid_1;
    memset (fid_array[0]->data, fid1, PASISIO_FILEID_SIZE);

    modified_chunks[0] = modified_chunks1;

    error = PasisIO_MDSynthUpdate (&md_protocol,
                                   fid_array, 
                                   1,
                                   md_linkages,
                                   md_server_list.num_servers,
                                   arg_size,
                                   res_size,
                                   chunk_size,
                                   modified_chunks,
                                   0, /* compute_time, */
                                   ts_increment, /* TS increment */
                                   use_cached_history,
                                   server_override_bitmap,
                                   data);

    if (error != 0) {
        fprintf(stderr, "PasisIO_SynthUpdate returned %d\n", error);
    }
}



void
do_single_update (int fid1,
                  int modified_chunks1,
                  int use_cached_history,
                  /* For testing, useful for client to force server to
                   * increment logical TS by more than 1 */
                  int ts_increment, /* Usually 1 */
                  p_uint32 server_override_bitmap)
{
    int arg_size = 512, res_size = 128, chunk_size = 128;
    do_single_update_data(fid1, modified_chunks1, use_cached_history,
                          ts_increment, server_override_bitmap,
                          arg_size, res_size, chunk_size, NULL);
}


void
do_single_query_data (int fid1,
                      int read_chunks1,
                      p_uint32 server_override_bitmap,
                      p_uint32 arg_size,
                      p_uint32 res_size,
                      p_uint32 chunk_size,
                      p_byte *data)
{
    pasis_md_fid_t *fid_array[2], fid_1;
    p_uint32 read_chunks[2];
    pasis_error_t error;
    
    /* Setup FIDs */
    fid_array[0] = &fid_1;
    memset (fid_array[0]->data, fid1, PASISIO_FILEID_SIZE); 

    read_chunks[0] = read_chunks1;

    error = PasisIO_MDSynthQuery (&md_protocol,
                                  fid_array, 
                                  1,
                                  md_linkages,
                                  md_server_list.num_servers,
                                  arg_size,
                                  res_size,
                                  chunk_size,
                                  read_chunks,
                                  0, /* compute_time, */
                                  0,  /* use cached history */
                                  server_override_bitmap,
                                  data);

    if (error != 0) {
        fprintf(stderr, "PasisIO_SynthQuery returned %d\n", error);
    }
}


void
do_single_query (int fid1,
                 int read_chunks1,
                 p_uint32 server_override_bitmap)
{
    int arg_size = 512, res_size = 128, chunk_size = 128;
    do_single_query_data(fid1, read_chunks1, server_override_bitmap,
                         arg_size, res_size, chunk_size, NULL);
}


static void
cleanup (void)
{
}


static int
init (void)
{
    pasis_error_t error;

    st_init ();

    Pasis_TimeInit ();

    /* Initialize PASIS */
    error = PasisIO_Init (1, 16384 /* BLOCK_SIZE */);
    if (error) {
        PASIS_PRINT_ERR ("PasisIO_Init error: %s\n", PASIS_STRERROR (error));
        return -1;
    }

    /* Add storage nodes to PASIS drives module */
    error = PasisIO_ConnectToStorageNodes (&md_server_list, 
                                           PASISIO_DRIVE_TYPE_MD_INDEX,
                                           md_linkages, NULL);
    if (error) {
        PASIS_PRINT_ERR ("IOInit Error while connecting to md drives\n");
        cleanup ();
        exit (1);
    }

    return 0;
}


static void
usage (char *name)
{
    PASIS_PRINT_INFO ("Usage: %s -r <file>\n", name);
    PASIS_PRINT_INFO (" -r <file> root config file (for protocol/linkages)\n");
}


int
main (int argc, char *argv[])
{
    char   *md_cfg_file = NULL;
    char buffer1[8], buffer2[8];
    char    c;

    pasis_debug_print_enable = 0;

    while ((c = getopt (argc, argv, "r:")) != -1) {
        switch (c) {
        case 'r':
            md_cfg_file = optarg;
            break;
        default:
            usage (argv[0]);
            exit (1);
        }
    }

    if (md_cfg_file == NULL) {
        usage (argv[0]);
        exit (1);
    }

    /* Parse config file, which contains protocol/scheme/storage node list */
    if (PasisIO_ParseConfigurationFile (md_cfg_file, &md_protocol, &md_server_list,
                                        NULL, NULL)) {
        PASIS_PRINT_ERR ("Error: configuration file parsing failed\n");
        exit (1);
    }

    if (init ()) {
        cleanup ();
        exit (1);
    }

    STRESS_TESTING = 1;


    memset(buffer1, 6, sizeof(buffer1));
    do_single_update_data (1,   /* fid 1 */
                           0x1, /* modify chunk 1 */
                           1,   /* use cached history */
                           1,   /* TS increment */
                           0,  /* send to all servers */
                           sizeof(buffer1),
                           0, /* result */
                           sizeof(buffer1),
                           (p_byte*)buffer1);

    memset(buffer2, 0, sizeof(buffer2));
    do_single_query_data (1,   /* fid 1 */
                          0x1, /* read chunk 1 */
                          0,  /* send to all servers */
                          0, /* arg size */
                          sizeof(buffer1),
                          sizeof(buffer1),
                          (p_byte*)buffer2);

    PASIS_DASSERT(memcmp(buffer1, buffer2, sizeof(buffer1)) == 0);

#if 0
    printf("------------------------------------------------------------\n");
    printf("Issuing complete, reading complete\n");
    printf("------------------------------------------------------------\n");
    printf("Issuing a complete update to object 1 \n");
    do_single_update (1,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      0);  /* send to all servers */
    printf("Issuing another complete update to object 1 \n");
    do_single_update (1,   /* fid 1 */
                      0x2, /* modify chunk 2 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      0);  /* send to all servers */

    printf("Issuing a query to object 1 \n");
    do_single_query (1,   /* fid 1 */
                     0x3, /* read chunks 1,3 */
                     0);  /* send to all servers */

#endif



#if 0
    printf("------------------------------------------------------------\n");
    printf("Issuing complete, reading complete, issuing incomplete, reading "
           "earlier complete\n");
    printf("------------------------------------------------------------\n");
    printf("Issuing a complete update to object 1 \n");
    do_single_update (1,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      0);  /* send to all servers */
    printf("Issuing a query to object 1 \n");
    do_single_query (1,   /* fid 1 */
                     0x1, /* read chunk 1 */
                     0);  /* send to all servers */
    
    printf("Issuing an incomplete update to object 1 to servers 1,2 \n");
    do_single_update (1,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2);  /* send to servers 1,2 */

    printf("Issuing a query to object 1 (should read version 1)\n");
    do_single_query (1,   /* fid 1 */
                     0x1, /* read chunk 1 */
                     0);  /* send to all servers */
#endif
    

#if 0
    printf("------------------------------------------------------------\n");
    printf("Issuing complete, repairable, and query (which fast-repairs repairable)\n");
    printf("------------------------------------------------------------\n");
    printf("Issuing a complete update to object 2 \n");
    do_single_update (2,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      0);  /* send to all servers */
    printf("Issuing a query to object 2 \n");
    do_single_query (2,   /* fid 1 */
                     0x1, /* read chunk 1 */
                     0);  /* send to all servers */

    printf("Issuing a repairable update to object 2 to servers 1,2,3 \n");
    do_single_update (2,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2|4);  /* send to all servers 1,2,4 */

    printf("Issuing a query to object 2 (should fast-repair version 2 "
           "and return it)\n");
    do_single_query (2,   /* fid 1 */
                     0x1, /* read chunk 1 */
                     0);  /* send to all servers */
#endif


#if 0
    printf("------------------------------------------------------------\n");
    printf("Issuing complete, repairable, and update (which fast-repairs repairable)\n");
    printf("------------------------------------------------------------\n");
    printf("Issuing a complete update to object 12 \n");
    do_single_update (12,   /* fid 1 */
                      0x3, /* modify chunks 1,2 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      0);  /* send to all servers */

    printf("Issuing a repairable update to object 12 to servers 1,2,3,4,5 \n");
    do_single_update (12,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2|4|8|16);  /* send to all servers 1,2,3,4,5 */

    printf("Issuing a completeup date to object 12 (should fast-repair version 2 "
           "and return it)\n");
    do_single_update (12,   /* fid 1 */
                      0x2, /* modify chunk 2 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      0);  /* send to all servers */
#endif


#if 0
    printf("------------------------------------------------------------\n");
    printf("Issuing complete, querying different quorum, causing object syncing\n");
    printf("------------------------------------------------------------\n");
    printf("Issuing a complete update to object 3 \n");
    do_single_update (3,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2|4|8|16);  /* send to servers 1,2,3,4,5 */
    printf("Issuing a complete update to object 3 to a different quorum "
           "(should cause object syncing)\n");
    do_single_update (3,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      2|4|8|16|32);  /* send to all servers */
#endif


#if 0
    printf("------------------------------------------------------------\n");
    printf("Issuing incomplete, issuing complete (first doing barrier write)\n");
    printf("------------------------------------------------------------\n");
    printf("Issuing an incomplete update to object 4 \n");
    do_single_update (4,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2);  /* send to |incomplete| set of servers */
    printf("Issuing a complete update to object 4 \n");
    do_single_update (4,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      0);  /* send to all servers */
#endif



#if 0
    printf("------------------------------------------------------------\n");
    printf("Issuing repairable, incomplete, and query (which does slow repair)\n");
    printf("------------------------------------------------------------\n");
    /* Tricky tricky tricky...
     * Ordinarily, we can't write an incomplete _after_ a repairable, since
     * during execution of the incomplete the client will have the OHS showing a
     * repair and so it would repair it.
     * The way to get around this (in order to test the code) is to add a global
     * variable, do_not_update_history, that we set before executing the
     * repairable operation.*/
    printf("Issuing a complete update to object 6 \n");
    do_single_update (6,   /* fid 1 */
                      0x3, /* modify chunks 1,2 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2|4|8|16|32);  /* send to all */
    printf("Issuing a repairable update to object 6 \n");
    DO_NOT_UPDATE_HISTORY = 1; /* IGNORE REPAIRABLE OP HISTORY */
    do_single_update (6,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2|4|8);  /* send to |repairable| set of servers */
    DO_NOT_UPDATE_HISTORY = 0; /* back to normal */

    printf("Issuing an incomplete update to object 6 \n");
    do_single_update (6,   /* fid 1 */
                      0x2, /* modify chunk 1,2 */
                      1,   /* use cached history */
                      2,   /* Increment TS by 2 to get this in front of repairable */
                      16);  /* send to |incomplete| set of servers */

    printf("Issuing a query to object 6, causing slow repair \n");
    do_single_query (6,   /* fid 1 */
                     0x1, /* read chunk 1 */
                     0);  /* send to all servers */
#endif


#if 0
    printf("------------------------------------------------------------\n");
    printf("Issuing repairable, incomplete, and update (which does slow repair)\n");
    printf("------------------------------------------------------------\n");
    /* Tricky tricky tricky...
     * Ordinarily, we can't write an incomplete _after_ a repairable, since
     * during execution of the incomplete the client will have the OHS showing a
     * repair and so it would repair it.
     * The way to get around this (in order to test the code) is to add a global
     * variable, do_not_update_history, that we set before executing the
     * repairable operation.*/
    printf("Issuing a complete update to object 6 \n");
    do_single_update (6,   /* fid 1 */
                      0x7, /* modify chunks 1,2,3 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2|4|8|16|32);  /* send to all */
    printf("Issuing a repairable update to object 6 \n");
    DO_NOT_UPDATE_HISTORY = 1; /* IGNORE REPAIRABLE OP HISTORY */
    do_single_update (6,   /* fid 1 */
                      0x1, /* modify chunk 1 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2|4|8);  /* send to |repairable| set of servers */

    printf("Issuing an incomplete update to object 6 \n");
    do_single_update (6,   /* fid 1 */
                      0x2, /* modify chunk 1,2 */
                      1,   /* use cached history */
                      2,   /* Increment TS by 2 to get this in front of repairable */
                      16);  /* send to |incomplete| set of servers */
    DO_NOT_UPDATE_HISTORY = 0; /* back to normal */

    printf("Issuing an update to object 6, causing slow repair \n");
    do_single_update (6,   /* fid 1 */
                      0x4, /* modify chunk 4 */
                      1,   /* use cached history */
                      1,   /* TS increment */
                      1|2|4|8|16|32);  /* send to all set of servers */
#endif
    


    pasis_info_print_enable = 0;
    PasisIO_Shutdown ();

    return 0;
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
