/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/* Test the new conditional-write protocol */

#include <math.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_error.h"
#include "pasis_timer.h"
#include "pasis_macosx_compat.h"

#include "pasisio.h"
#include "pasisio_config_parse.h"
#include "pasisio_md_protocol.h"
#include "pasisio_drives.h"
#include "pasisio_select.h"

#include "pasis_md_prot.h"


#define                        BLOCK_SIZE   16384


static pasis_config_server_list_t md_server_list, data_server_list;
static pasis_md_protocol_t        md_protocol;
static pasis_md_protocol_t        data_protocol;
static pasis_md_linkage_t         md_linkages[PASIS_MD_MAX_LINKAGES];
static pasis_md_linkage_t         data_linkages[PASIS_MD_MAX_LINKAGES];

static unsigned int              shm_region_size; 
static void                     *shm_region;

static p_byte                   *read_buffer;
static p_byte                   *write_buffer;

static pasis_md_fid_t            test_fid;



static void
do_create(int i, char* name)
{
    pasis_md_attr_t name_attr_in, parent_attr_out, name_attr_out;
    pasis_md_fid_t test_fid;
    pasis_error_t error;
    
    memset (test_fid.data, i, PASISIO_FILEID_SIZE); 

    /* Setup create attributes */
    name_attr_in.mode = MD_ATTR_MODE_REG;
    name_attr_in.ref_cnt = 1;
    name_attr_in.uid = 314;
    name_attr_in.gid = 159;
    name_attr_in.fileid = 20202020;
    memset(&name_attr_in.mtime, 0, sizeof(name_attr_in.mtime));
    memset(&name_attr_in.atime, 0, sizeof(name_attr_in.atime));
    memset(&name_attr_in.ctime, 0, sizeof(name_attr_in.ctime));
    name_attr_in.blocks = 0;
    name_attr_in.size = 0;

    error = PasisIO_MDCreate (&md_protocol,
                              md_linkages,
                              md_server_list.num_servers,
                              (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C,
                              &test_fid,
                              name,
                              NULL, /* symbolic link */
                              &name_attr_in,
                              &data_protocol,
                              data_linkages,
                              data_server_list.num_servers,
                              &parent_attr_out,
                              &name_attr_out);
    fprintf(stderr, "PasisIO_MDCreate returned %d\n", error);
}


static void
do_getattr(int i)
{
    pasis_md_attr_t attr_out;
    pasis_md_loc_t distribution_out;
    pasis_md_fid_t test_fid;
    pasis_error_t error;
    
    memset (test_fid.data, i, PASISIO_FILEID_SIZE); 

    error = PasisIO_MDGetattr (&md_protocol,
                               md_linkages,
                               md_server_list.num_servers,
                               &test_fid,
                               &attr_out,
                               &distribution_out);

    fprintf(stderr, "PasisIO_GetAttr returned %d\n", error);
}


static void
do_lookup(char* name, int *fid_out_out)
{
    pasis_md_attr_t dir_attr_out;
    pasis_md_fid_t fid_out;
    pasis_md_loc_t distribution_out;
    pasis_error_t error;
    
    error = PasisIO_MDLookup (&md_protocol,
                              md_linkages,
                              md_server_list.num_servers,
                              (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C,
                              name,
                              &fid_out,
                              &dir_attr_out,
                              &distribution_out);

    if (fid_out_out) {
        *fid_out_out = fid_out.data[0];
    }

    fprintf(stderr, "PasisIO_Lookup returned %d\n", error);
}


static void
do_link(int from, int to_dir, char* name)
{
    pasis_md_attr_t dir_attr_out, file_attr_out;
    pasis_md_fid_t from_fid, to_dir_fid;
    pasis_error_t error;
    
    memset (from_fid.data, from, PASISIO_FILEID_SIZE); 
    memset (to_dir_fid.data, to_dir, PASISIO_FILEID_SIZE); 

    error = PasisIO_MDLink (&md_protocol,
                            md_linkages,
                            md_server_list.num_servers,
                            &from_fid,
                            &to_dir_fid,
                            name,
                            &file_attr_out,
                            &dir_attr_out);

    fprintf(stderr, "PasisIO_Link returned %d\n", error);
}


static void
do_unlink(int parent, int file, char* name)
{
    pasis_md_attr_t dir_attr_out;
    pasis_md_fid_t parent_fid, file_fid;
    pasis_error_t error;
    
    memset (parent_fid.data, parent, PASISIO_FILEID_SIZE); 
    memset (file_fid.data, file, PASISIO_FILEID_SIZE); 

    error = PasisIO_MDUnlink (&md_protocol,
                              md_linkages,
                              md_server_list.num_servers,
                              &parent_fid, 
                              &file_fid,
                              name,
                              &dir_attr_out);

    fprintf(stderr, "PasisIO_Unlink returned %d\n", error);
}


static void
do_rename(int from_dir, int to_dir, int from_fid, int to_fid,
          char* from_name, char* to_name, int to_name_valid)
{
    pasis_md_attr_t from_dir_attr_out, to_dir_attr_out;
    pasis_md_fid_t from_dir_fid, to_dir_fid, from_fid_fid, to_fid_fid;
    pasis_error_t error;
    
    memset (from_dir_fid.data, from_dir, PASISIO_FILEID_SIZE); 
    memset (to_dir_fid.data, to_dir, PASISIO_FILEID_SIZE); 
    memset (from_fid_fid.data, from_fid, PASISIO_FILEID_SIZE); 
    memset (to_fid_fid.data, to_fid, PASISIO_FILEID_SIZE); 

    error = PasisIO_MDRename (&md_protocol,
                              md_linkages,
                              md_server_list.num_servers,
                              &from_dir_fid,
                              &to_dir_fid,
                              &from_fid_fid,
                              &to_fid_fid,
                              from_name,
                              to_name,
                              to_name_valid,
                              &from_dir_attr_out,
                              &to_dir_attr_out);

    fprintf(stderr, "PasisIO_Rename returned %d\n", error);
}


static void
do_setattr(int i)
{
    pasis_md_sattr_t sattr;
    pasis_md_attr_t attr_out;
    p_uint32 guard_ctime[2] = {0, 0};
    pasis_error_t error;
    pasis_md_fid_t test_fid;
    
    memset (test_fid.data, i, PASISIO_FILEID_SIZE); 

    sattr.set_uid = sattr.set_gid = sattr.set_size = 1;
    sattr.set_mode = sattr.set_atime = sattr.set_mtime = 0;
    sattr.uid = 271;
    sattr.gid = 828;
    sattr.size = 123456789;

    error = PasisIO_MDSetattr (&md_protocol,
                               md_linkages,
                               md_server_list.num_servers,
                               &test_fid,
                               &sattr,
                               guard_ctime,
                               &attr_out);

    fprintf(stderr, "PasisIO_MDSetattr returned %d\n", error);
}


static void
do_create_symlink(int i, char* sym_name, char* target_name)
{
    pasis_md_attr_t name_attr_in, parent_attr_out, name_attr_out;
    pasis_md_fid_t fid;
    pasis_error_t error;
    
    /* Setup new FID  */
    memset (fid.data, i, PASISIO_FILEID_SIZE); 

    /* Setup create attributes */
    name_attr_in.mode = MD_ATTR_MODE_LNK;
    name_attr_in.ref_cnt = 1;
    name_attr_in.uid = 602;
    name_attr_in.gid = 1023;
    name_attr_in.fileid = 50505050;
    memset(&name_attr_in.mtime, 0, sizeof(name_attr_in.mtime));
    memset(&name_attr_in.atime, 0, sizeof(name_attr_in.atime));
    memset(&name_attr_in.ctime, 0, sizeof(name_attr_in.ctime));
    name_attr_in.blocks = 0;
    name_attr_in.size = 5;

    error = PasisIO_MDCreate (&md_protocol,
                              md_linkages,
                              md_server_list.num_servers,
                              (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C,
                              &fid,
                              sym_name,
                              target_name,
                              &name_attr_in,
                              &data_protocol,
                              data_linkages,
                              data_server_list.num_servers,
                              &parent_attr_out,
                              &name_attr_out);

    fprintf(stderr, "PasisIO_MDCreate (symlink) returned %d\n", error);
}


static void
do_readlink(int i)
{
    pasis_md_attr_t fid_attr_out;
    pasis_md_fid_t fid;
    pasis_error_t error;
    char          path[1024];
    
    /* Setup symbolic link FID  */
    memset (fid.data, i, PASISIO_FILEID_SIZE); 
    
    error = PasisIO_MDReadlink (&md_protocol,
                                md_linkages,
                                md_server_list.num_servers,
                                &fid,
                                path,
                                &fid_attr_out);

    fprintf(stderr, "PasisIO_ReadLink returned %d\n", error);
}

static void
do_readdir(int dir, int block)
{
    pasis_md_attr_t dir_attr_out;
    pasis_md_fid_t fid;
    pasis_error_t error;
    p_byte        buffer[4096];
    
    memset (fid.data, dir, PASISIO_FILEID_SIZE); 
    
    error = PasisIO_MDReaddir (&md_protocol,
                               md_linkages,
                               md_server_list.num_servers,
                               &fid,
                               block,
                               buffer,
                               &dir_attr_out);
    
    fprintf(stderr, "PasisIO_ReadDir returned %d\n", error);
}


static void
do_remove_symlink(void)
{
    pasis_md_attr_t dir_attr_out;
    pasis_md_fid_t fid;
    pasis_error_t error;
    
    /* Setup new FID  */
    memset (fid.data, 3, PASISIO_FILEID_SIZE); 

    error = PasisIO_MDUnlink (&md_protocol,
                              md_linkages,
                              md_server_list.num_servers,
                              (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C,
                              &fid,
                              "famous quote",
                              &dir_attr_out);

    fprintf(stderr, "PasisIO_Unlink returned %d\n", error);
}


static void
cleanup (void)
{
    munmap (shm_region, shm_region_size);
}


static int
init (void)
{
    pasis_error_t error;

    st_init ();

    Pasis_TimeInit ();

    /* Read/write buffers must be in shared memory, in case ED is running
     * in different process */
    shm_region_size = PAGE_ALIGN (2 * BLOCK_SIZE);
    shm_region = mmap (0, shm_region_size, PROT_READ | PROT_WRITE,
                       MAP_ANONYMOUS | MAP_SHARED, -1, 0);
    if (shm_region == MAP_FAILED) {
        perror ("Mmap failed");
        return -1;
    }
    read_buffer  = shm_region;
    write_buffer = shm_region + BLOCK_SIZE;

    /* Initialize PASIS */
    error = PasisIO_Init (1, BLOCK_SIZE);
    if (error) {
        PASIS_PRINT_ERR ("PasisIO_Init error: %s\n", PASIS_STRERROR (error));
        return -1;
    }

    /* Add storage nodes to PASIS drives module */
    error = PasisIO_ConnectToStorageNodes (&md_server_list, 
                                           PASISIO_DRIVE_TYPE_MD_INDEX,
                                           md_linkages, NULL);
    if (error) {
        PASIS_PRINT_ERR ("IOInit Error while connecting to md drives\n");
        cleanup ();
        exit (1);
    }

#if 0
    error = PasisIO_ConnectToStorageNodes (&data_server_list, 
                                           PASISIO_DRIVE_TYPE_S4_INDEX,
                                           data_linkages, NULL);
    if (error) {
        PASIS_PRINT_ERR ("IOInit Error while connecting to data drives\n");
        cleanup ();
        exit (1);
    }
#endif
    
    return 0;
}


static void
usage (char *name)
{
    PASIS_PRINT_INFO ("Usage: %s -r <file> -d <file>\n", name);
    PASIS_PRINT_INFO (" -r <file> root config file (for protocol/linkages)\n");
    PASIS_PRINT_INFO (" -d <file> data config file (for protocol/linkages)\n");
}


int
main (int argc, char *argv[])
{
    char   *md_cfg_file = NULL, *data_cfg_file = NULL;
    char    c;

    pasis_debug_print_enable = 0;

    while ((c = getopt (argc, argv, "r:d:")) != -1) {
        switch (c) {
        case 'r':
            md_cfg_file = optarg;
            break;
        case 'd':
            data_cfg_file = optarg;
            break;
        default:
            usage (argv[0]);
            exit (1);
        }
    }

    if (md_cfg_file == NULL || data_cfg_file == NULL) {
        usage (argv[0]);
        exit (1);
    }

    /* Parse config file, which contains protocol/scheme/storage node list */
    if (PasisIO_ParseConfigurationFile (md_cfg_file, &md_protocol,
                                        &md_server_list, NULL, NULL)) {
        PASIS_PRINT_ERR ("Error: configuration file parsing failed\n");
        exit (1);
    }

    /* Parse config file, which contains protocol/scheme/storage node list */
    if (PasisIO_ParseConfigurationFile (data_cfg_file, &data_protocol,
                                        &data_server_list, NULL, NULL)) {
                                        
        PASIS_PRINT_ERR ("Error: configuration file parsing failed\n");
        exit (1);
    }

    if (init ()) {
        cleanup ();
        exit (1);
    }

    do_create(2, "one");

    do_getattr(2);

    do_lookup("one", NULL);

    do_unlink(0xff /* root dir */, 2, "one");

    printf("%s: following lookup should fail\n", __FUNCTION__);
    do_lookup("one", NULL);


    printf("%s: following lookup should fail\n", __FUNCTION__);
    do_lookup("two-clone", NULL);

    do_create(4, "two");
    do_link(4, 0xff /* root dir */, "two-clone");
    do_lookup("two-clone", NULL);


    do_rename(0xff /* root dir */, 0xff /* root dir */, 4, 0 /* unused */,
              "two-clone", "two-renamed", 0 /* to_name  VALID? */);
    printf("%s: following lookup should fail\n", __FUNCTION__);
    do_lookup("two-clone", NULL);
    int fid_out1;
    do_lookup("two-renamed", &fid_out1);
    PASIS_DASSERT(fid_out1 == 4);

    
    do_create(55, "two-to-be-renamed-over");
    do_lookup("two-to-be-renamed-over", NULL);
    do_rename(0xff /* root dir */, 0xff /* root dir */, 4, 55,
              "two-renamed", "two-to-be-renamed-over", 1 /* to_name VALID? */);
    int fid_out2;
    do_lookup("two-to-be-renamed-over", &fid_out2);
    PASIS_DASSERT(fid_out2 == 4);


    do_create(6, "three");
    do_setattr(6);
    do_getattr(6);

    do_readdir(0xff, 0);

    do_create_symlink(7, "is a charm", "three");
    do_readlink(7);


#if 0
    do_lookup("three small step for mankind...");

    do_getattr((pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C);

    do_getattr(&test_fid);

    do_unlink();

    do_lookup("famous quote");


    do_lookup("famous quote");


    do_remove_symlink();

    do_lookup("famous quote");
#endif

    cleanup ();

    pasis_info_print_enable = 0;
    PasisIO_Shutdown ();

    return 0;
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
