/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_error.h"
#include "pasis_timer.h"

#include "pasisio.h"
#include "pasisio_config_parse.h"
#include "pasisio_md_protocol.h"
#include "pasis_md_prot.h"
#include "pasisio_drives.h"
#include "pasisio_select.h"

#include <math.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>



#define WORKER_PHASE_WAIT      0x00
#define WORKER_PHASE_WARMUP    0x01
#define WORKER_PHASE_CALIB     0x02
#define WORKER_PHASE_MEASURE   0x03
#define WORKER_PHASE_SLOWDOWN  0x04
#define WORKER_PHASE_SHUTDOWN  0x05

#define RESP_STATS_BINS 9 /* always make this odd */
#define RESP_STATS_BIN_MEAN (RESP_STATS_BINS/2)


struct resp_stats_t;
struct Operation;
struct Object;
struct WorkerState;


static resp_stats_t           *stats;

static unsigned int            client_id=0;

static unsigned int            use_history_cache = 1;

static st_utime_t              execution_rate_period=0;
static unsigned int            current_execution_rate_period=0;
static unsigned int            max_num_execution_rate_periods=0;
static unsigned int           *execution_rates = NULL;
static st_utime_t              current_period_start_time;

static unsigned int            num_operation_types=0;
static unsigned int            total_relative_frequencies=0;
static Operation*             *operations;

static unsigned int            chunk_size;

static unsigned int            num_objects=0;
static unsigned int            randomly_select_fids=0;
static unsigned int            fid_start;
static Object                 *objects;

static unsigned int            num_outstanding;

static WorkerState            *workers;
static unsigned int            worker_phase;
static st_cond_t               worker_cond;
static st_thread_t             controller;

static pasis_md_linkage_t      linkages[PASIS_MAX_NUM_SERVERS];
static pasis_md_protocol_t     protocol;

static unsigned int            warmup_time;
static unsigned int            run_time;
static unsigned int            slowdown_time;
static unsigned int            calib_time;

static int                     read_retries;

static int                     num_servers;
static pasisio_drive_instance_t *instances[PASIS_MAX_NUM_SERVERS];


struct resp_stats_t {
    unsigned int  bins[RESP_STATS_BINS];
    unsigned int  min;
    unsigned int  max;

    unsigned int  mean;
    unsigned int  stdv;

    unsigned int  calib_samples;
    unsigned int  calib_failures;

    unsigned int  samples;
    unsigned int  failures;

    p_uint64      total_resp;
    p_uint64      total_sq_resp;

};

struct WorkerState {
    WorkerState () {
        samples = new unsigned int[num_operation_types];
        ts = new pasis_timespec_t[num_operation_types];
        memset (samples, 0, num_operation_types * sizeof (unsigned int));
        memset (ts, 0, num_operation_types * sizeof (pasis_timespec_t));
    }

    ~WorkerState () {
        delete samples;
        delete ts;
    }

    st_thread_t        thread_id;
    unsigned int      *samples;
    pasis_timespec_t  *ts;
};


struct Object {
    Object () : busy (0), op_count(0) {
        memset (&fid, 0, sizeof(pasis_md_fid_t));
    }

    pasis_md_fid_t fid;
    int            busy;
    int            op_count;
};


struct Operation {
    Operation (char* operation_name, bool isQueryOp,
               bool isSlowRepairOp, bool isFastRepairOp,
               int num_op_objects, int arg_size, int res_size,
               int num_affected_chunks,
               int client_compute_time, int server_compute_time,
               int relative_frequency);

    pasis_error_t execute (int *object_indices);
    pasis_error_t setup_state (int *object_indices);

    char operation_name[64];
    bool isQueryOp;
    bool isSlowRepairOp;
    bool isFastRepairOp;
    int num_op_objects;
    int arg_size;
    int res_size;
    int num_affected_chunks;
    int client_compute_time;
    int server_compute_time;
    int relative_frequency;
};



static int line_num=0;

static inline char *
read_line (FILE *file, char *buffer)
{
    while (1) {
        line_num++;
        if (fgets (buffer, 255, file) == NULL) {
            return NULL;
        }
        if (buffer[0] == '#' || buffer[0] == '\n') continue;
        break;
    }
    return buffer;
}

#define STR(s) #s

#define PARSE_INPUT(file, buffer, name, format, var)    \
do {                                                    \
    int rc;                                             \
    if (read_line (file, buffer) == NULL) return -1;    \
    rc = sscanf (buffer, STR(name=format), var);        \
    if (rc == EOF || rc == 0) goto error;               \
} while (0)


#define PARSE_LINE(file, buffer)                             \
do {                                                         \
    char *tmp;                                               \
    if (read_line (file, buffer) == NULL) return -1;         \
    if ((tmp = strchr (buffer, '\n')) != NULL) *tmp = '\0';  \
} while (0)


extern unsigned int protocol_rereads;


static int
parse_file (char *filename, pasis_config_server_list_t *server_list)
{
    unsigned int  u_tmp;
    char          buffer[255];
    int           err;
    unsigned int  i;
    FILE         *file;
    long          file_offset;

    /* Parse the server list, protocol, and scheme sections */
    if (PasisIO_ParseConfigurationFile (filename, &protocol, server_list,
                                        &file_offset, &line_num)) {
        PASIS_PRINT_ERR ("Error: configuration file parsing failed\n");
        return -1;
    }
    num_servers = protocol.N;

    /* The standard pasisio config parse function closes the file after it's
     * done; reopen the file since we have more to parse */
    file = fopen (filename, "r");
    if (file == NULL) {
        PASIS_PRINT_ERR ("%s: error: file open failed (%s, %s)\n",
                         __FUNCTION__, filename, PASIS_STRERROR (errno));
        return -1;
    }

    /* Seek so we skip the lines that's been parsed already */
    err = fseek (file, file_offset, SEEK_SET);
    if (err == -1) {
        PASIS_PRINT_ERR ("%s: error: file seek failed (%s, %ld, %s)\n",
                         __FUNCTION__, filename, file_offset,
                         PASIS_STRERROR (err));
        return -1;
    }

    /* Parse the definition of the operations */
    PARSE_INPUT (file, buffer, NUM_OPERATION_TYPES, %u, &u_tmp);
    num_operation_types = u_tmp;
    operations = (Operation**)malloc (num_operation_types * sizeof(Operation*));

    for (i = 0; i < num_operation_types; ++i) {
        char operation_name[64], operation_type[64];
        int num_op_objects, arg_size, res_size, client_compute_time;
        int server_compute_time;
        int num_affected_chunks, relative_frequency, isQueryOp;
        int isSlowRepairOp, isFastRepairOp;

        PARSE_INPUT (file, buffer, OPERATION_NAME, %s, operation_name);
        PARSE_INPUT (file, buffer, OPERATION_TYPE, %s, operation_type);
        if (!strncasecmp (operation_type, "query", 63)) {
            isQueryOp = 1;
            isSlowRepairOp = 0;
            isFastRepairOp = 0;
        } else if (!strncasecmp (operation_type, "update", 63)) {
            isQueryOp = 0;
            isSlowRepairOp = 0;
            isFastRepairOp = 0;
        } else if (!strncasecmp (operation_type, "slow-repair", 63)) {
            isQueryOp = 0;
            isSlowRepairOp = 1;
            isFastRepairOp = 0;
        } else if (!strncasecmp (operation_type, "fast-repair", 63)) {
            isQueryOp = 0;
            isSlowRepairOp = 0;
            isFastRepairOp = 1;
        } else {
            PASIS_PRINT_ERR ("%s: error: invalid operation type: \"%s\"\n",
                             __FUNCTION__, operation_type);
            goto error;
        }

        PARSE_INPUT (file, buffer, NUM_OBJECTS, %u, &u_tmp);
        num_op_objects = u_tmp;

        PARSE_INPUT (file, buffer, ARG_SIZE, %u, &u_tmp);
        arg_size = u_tmp;

        PARSE_INPUT (file, buffer, RES_SIZE, %u, &u_tmp);
        res_size = u_tmp;

        if (isQueryOp) {
            PARSE_INPUT (file, buffer, NUM_READ_CHUNKS, %u, &u_tmp);
        } else {
            PARSE_INPUT (file, buffer, NUM_WRITTEN_CHUNKS, %u, &u_tmp);
        }
        num_affected_chunks = u_tmp;

        PARSE_INPUT (file, buffer, CLIENT_COMPUTE_TIME, %u, &u_tmp);
        client_compute_time = u_tmp;
        
        PARSE_INPUT (file, buffer, SERVER_COMPUTE_TIME, %u, &u_tmp);
        server_compute_time = u_tmp;

        PARSE_INPUT (file, buffer, RELATIVE_FREQUENCY, %u, &u_tmp);
        relative_frequency = u_tmp;
        
        total_relative_frequencies += relative_frequency;

        operations[i] = new Operation(operation_name, isQueryOp,
                                      isSlowRepairOp, isFastRepairOp,
                                      num_op_objects,
                                      arg_size, res_size,
                                      num_affected_chunks,
                                      client_compute_time,
                                      server_compute_time,
                                      relative_frequency);
    }

    /* Parse the chunk size */
    PARSE_INPUT (file, buffer, CHUNK_SIZE, %u, &u_tmp);
    chunk_size = u_tmp;

    /* Parse the number of objects */
    PARSE_INPUT (file, buffer, NUM_OBJECTS, %u, &u_tmp);
    num_objects = u_tmp;

    /* Parse the number of outstanding requests */
    PARSE_INPUT (file, buffer, NUM_OUTSTANDING, %u, &u_tmp);
    num_outstanding = u_tmp;

    /* Parse whether to randomly select object FID's or not */
    PARSE_INPUT (file, buffer, RANDOMLY_SELECT_FID, %u, &u_tmp);
    randomly_select_fids = u_tmp;

    /* Parse whether to randomly select object FID's or not */
    PARSE_INPUT (file, buffer, FID_START, %x, &u_tmp);
    fid_start = u_tmp;

    if (num_objects < num_outstanding) {
        PASIS_PRINT_ERR ("%s: num_objects (%u) can't be less than "
                         "num_oustanding (%u)\n", __FUNCTION__,
                         num_objects, num_outstanding);
        exit(-1);
    }


    PARSE_INPUT (file, buffer, READ_RETRIES, %u, &read_retries);

    /* Warmup, run, and slow down times */
    PARSE_INPUT (file, buffer, WARMUP_TIME, %u, &warmup_time);
    PARSE_INPUT (file, buffer, CALIB_TIME, %u, &calib_time);
    PARSE_INPUT (file, buffer, RUN_TIME, %u, &run_time);
    PARSE_INPUT (file, buffer, SLOWDOWN_TIME, %u, &slowdown_time);


    return 0;

 error:
  
    buffer[strlen(buffer)-1] = '\0';
    fprintf (stderr, "Error: parse error at line %d: %s\n", line_num, buffer);

    return -1;
}


static inline char *
dump_fid (pasis_md_fid_t *f)
{
    static char buf[PASISIO_FILEID_SIZE*2 + 1];
    int i;
    char *bp = buf;
    for (i=0; i<PASISIO_FILEID_SIZE; i++) {
        snprintf(bp, 3, "%02X", (unsigned char)f->data[i]);
        bp +=2;
    }

    return buf;
}


pasis_error_t
do_single_query (pasis_md_fid_t *fid, int arg_size, int res_size, int chunk_size,
                 unsigned int read_chunks, int server_compute_time)
{
    int retries = read_retries + 1;
    pasis_error_t error;
    
    do {
        if (retries != read_retries + 1) {
            printf ("Retrying read %d\n", 
                    read_retries + 1 - retries);
        }

        error = PasisIO_MDSynthQuery (&protocol,
                                      &fid,
                                      1,
                                      linkages,
                                      num_servers,
                                      arg_size,
                                      res_size,
                                      chunk_size,
                                      &read_chunks,
                                      server_compute_time,
                                      1,  /* use cached history */
                                      0,  /* server_override_bitmap */
                                      NULL);

        retries--;
    } while (retries > 0 && error != PASIS_SUCCESS);

    if (error) {
        printf("PasisIO_SynthQuery returned %d\n", error);
    }

    return error;
}


pasis_error_t
do_single_update (pasis_md_fid_t* fid, int arg_size, int res_size,
                  int chunk_size, unsigned int modified_chunks,
                  int server_compute_time,
                  p_uint32 server_override_bitmap)
{
    pasis_error_t error;
    
    error = PasisIO_MDSynthUpdate (&protocol,
                                   &fid, 
                                   1,
                                   linkages,
                                   num_servers,
                                   arg_size,
                                   res_size,
                                   chunk_size,
                                   &modified_chunks,
                                   server_compute_time,
                                   1, /* TS increment */
                                   use_history_cache,
                                   server_override_bitmap,
                                   NULL);
                                   //1|2|4|8|16|32|64|128|256|512|1024|2048|4096|8192|16384|32768|65536); /* server_override_bitmap */p

    if (error) {
//         printf("PasisIO_SynthUpdate returned %d\n", error);
    }

    return error;
}


pasis_error_t
do_update (pasis_md_fid_t** fids, int num_fids,
           int arg_size, int res_size, int chunk_size,
           unsigned int modified_chunks, int server_compute_time,
           int use_cached_history)
{
    pasis_error_t error;
    
    error = PasisIO_MDSynthUpdate (&protocol,
                                   fids, 
                                   num_fids,
                                   linkages,
                                   num_servers,
                                   arg_size,
                                   res_size,
                                   chunk_size,
                                   &modified_chunks,
                                   server_compute_time,
                                   1, /* TS increment */
                                   use_cached_history,
                                   0, /* server_override_bitmap */
                                   NULL);
//1|2|4|8|16|32|64|128|256|512|1024|2048|4096|8192|16384|32768|65536); /* server_override_bitmap */

    if (error) {
//         printf("PasisIO_SynthUpdate returned %d\n", error);
    }

    return error;
}


Operation::Operation(char* _operation_name, bool _isQueryOp,
                     bool _isSlowRepairOp, bool _isFastRepairOp,
                     int _num_op_objects, int _arg_size, int _res_size,
                     int _num_affected_chunks, int _client_compute_time,
                     int _server_compute_time,
                     int _relative_frequency)
{
    strcpy (operation_name, _operation_name);
    isQueryOp = _isQueryOp;
    isSlowRepairOp = _isSlowRepairOp;
    isFastRepairOp = _isFastRepairOp;
    num_op_objects = _num_op_objects;
    arg_size = _arg_size;
    res_size = _res_size;
    num_affected_chunks = _num_affected_chunks;
    client_compute_time = _client_compute_time;
    server_compute_time = _server_compute_time;
    relative_frequency = _relative_frequency;
}

pasis_error_t Operation::execute (int *object_indices)
{
    pasis_error_t error;
    unsigned int affected_chunks;
    static int arg_size_addition = 0;

    affected_chunks = 0;
    for (int i = 0; i < num_affected_chunks; ++i) {
        affected_chunks = (affected_chunks << 1) | 1;
    }

    if (isQueryOp) {
        PASIS_DASSERT (num_op_objects == 1);
        error = do_single_query (&objects[object_indices[0]].fid,
                                 arg_size, res_size, chunk_size,
                                 affected_chunks, server_compute_time);
    } else {
        pasis_md_fid_t* fids[16];
        PASIS_DASSERT (num_op_objects < 16);
        for (int i = 0; i < num_op_objects; ++i) {
            fids[i] = &objects[object_indices[i]].fid;
        }
        /* if benchmarking repair, then don't use cached history,
         * to simulate a client having a stale OHS */
        int local_use_history_cache = isSlowRepairOp == false;
        if (local_use_history_cache && !use_history_cache) {
            // global history cache setting overrides this
            local_use_history_cache = false;
        }
        arg_size_addition = (arg_size_addition + 1) % 10;
        error = do_update (fids, num_op_objects,
                           arg_size + arg_size_addition, res_size, chunk_size,
                           affected_chunks, server_compute_time,
                           local_use_history_cache);
    }

    st_usleep(client_compute_time * 1000);

    return error;
}


pasis_error_t Operation::setup_state (int *object_indices)
{
    pasis_error_t error;
    unsigned int affected_chunks;

    if (!isSlowRepairOp) {
        /* no state to setup for non-repair operations */
        return PASIS_SUCCESS;
    }

    PASIS_DASSERT (num_op_objects == 1);

    affected_chunks = 0;
    for (int i = 0; i < num_affected_chunks; ++i) {
        affected_chunks = (affected_chunks << 1) | 1;
    }

    error = do_single_update (&objects[object_indices[0]].fid,
                              arg_size, res_size, chunk_size,
                              affected_chunks, 0, 0);

    return error;
}


/* Selects which operation to execute.  It does this based on the "relative
 * frequency" parameter associated with each operation. */
static inline int
select_operation (void)
{
    p_int32 operation_seq, cumulative_operation_seq = 0;

    operation_seq = rand () % total_relative_frequencies;
    
    for (unsigned int i = 0; i < num_operation_types; ++i) {
        cumulative_operation_seq += operations[i]->relative_frequency;
        if (operation_seq < cumulative_operation_seq) {
            return i;
        }
    }

    PASIS_DASSERT (0);
    return 0;
}


static inline void
select_objects (int num_objects_requested, int object_indices[])
{
    /* iterate over objects for load balancing */
    static int next_object_search_index = 0;
    int original_num_objects_requested = num_objects_requested;
    for (unsigned int i = 0; i < num_objects && num_objects_requested > 0; ++i) {
        int object_num = (next_object_search_index + i) % num_objects;
        if (objects[object_num].busy == 0) {
            objects[object_num].busy = 1;
            objects[object_num].op_count++;
            object_indices[--num_objects_requested] = object_num;
        }
    }
    PASIS_DASSERT (num_objects_requested == 0);
    next_object_search_index =
        (next_object_search_index + original_num_objects_requested) %
        num_objects;
}


static inline void
free_objects (int num_objects_requested, int object_indices[])
{
    for (int i = 0; i < num_objects_requested; ++i) {
        PASIS_DASSERT (objects[object_indices[i]].busy == 1);
        objects[object_indices[i]].busy = 0;
    }
}


static void
worker_bin_time (resp_stats_t *stats, unsigned int time)
{
    int bin = time - stats->mean;
    
    if (stats->stdv == 0) {
        /* something arbitrary to avoid divide by 0 */
        stats->stdv = 1;
    }
 
    bin /= (int)stats->stdv;

    bin += RESP_STATS_BIN_MEAN;

    if (bin >= RESP_STATS_BINS) bin = RESP_STATS_BINS - 1;
    if (bin < 0) bin = 0;

    if (stats->min == 0) {
        stats->min = stats->max = time;
    }
    if (time < stats->min) stats->min = time;
    if (time > stats->max) stats->max = time;
    
    stats->bins[bin]++;
}


static void
track_execution_rate (void)
{
    st_utime_t current_time;

    if (execution_rate_period == 0) {
        return;
    }

    /* Incrememnt request count for current execution period */
    ++execution_rates[current_execution_rate_period];

    /* Check if we're in a new execution period */
    current_time = st_utime ();
    if (current_time - current_period_start_time >= execution_rate_period) {
        current_period_start_time = current_time;
        ++current_execution_rate_period;
        PASIS_DASSERT (current_execution_rate_period <=
                       max_num_execution_rate_periods);
    }
}


static void
issue_req (WorkerState *worker)
{
    pasis_timer_t     timer;
    pasis_timespec_t  ts;
    pasis_error_t     status;

    int               operation_index;
    int              *object_indices;

    // Select which operation to execute
    operation_index = select_operation ();

    // Select objects for that operation
    object_indices = (int*)malloc (operations[operation_index]->num_op_objects *
                                   sizeof (int));
    select_objects (operations[operation_index]->num_op_objects, object_indices);
    
    // For repair operations, setup state before measuring operation response time
    status = operations[operation_index]->setup_state (object_indices);

    PASIS_TIMER_START_TM (&timer);

    // Execute operation
    status = operations[operation_index]->execute (object_indices);

    if (status != PASIS_SUCCESS) {
        PASIS_PRINT_ERR ("%s: received status=%#x from %s call\n",
                         __FUNCTION__, status,
                         operations[operation_index]->operation_name);
    }

    PASIS_TIMER_STOP_TM (&timer);
    PASIS_TIMER_ELAPSED_TS (&timer, &ts);

    // Keep statistics
    if (worker_phase == WORKER_PHASE_MEASURE || 
        worker_phase == WORKER_PHASE_CALIB) {
        p_uint64 time;
        resp_stats_t *req_stats = &stats[operation_index];

        time = ts.ts_sec * 1000000 + ts.ts_nsec / 1000;
        req_stats->total_resp += time;
        req_stats->total_sq_resp += (time * time);

        if (worker_phase == WORKER_PHASE_MEASURE) {
            worker_bin_time (req_stats, time);
            
            PASIS_TIMER_ADD_TS (&ts, &worker->ts[operation_index],
                                &worker->ts[operation_index]);
            worker->samples[operation_index]++;

            track_execution_rate ();
        }

        req_stats->samples++;

        if (status != PASIS_SUCCESS) {
            req_stats->failures++;
        }
    }
    
    free_objects (operations[operation_index]->num_op_objects, object_indices);
    free (object_indices);
}


static void *
worker_thread (void *arg)
{
    struct timeval tv1;
    WorkerState *worker = (WorkerState *) arg;

    while (worker_phase == WORKER_PHASE_WAIT) {
        st_cond_wait (worker_cond);
    }

    gettimeofday (&tv1, NULL);
    printf ("Thread entering startup phase (%lu.%06"PRIu64")\n", 
            tv1.tv_sec, (p_uint64)tv1.tv_usec);

    /* Warmup phase */    
    while (worker_phase == WORKER_PHASE_WARMUP) {
        issue_req (worker);
    }

    gettimeofday (&tv1, NULL); 
    printf ("Thread entering calibration phase (prev: %lu.%06"PRIu64")\n",
            tv1.tv_sec, (p_uint64)tv1.tv_usec);

    /* Calibration phase */   
    while (worker_phase == WORKER_PHASE_CALIB) {
        issue_req (worker);
    }

    gettimeofday (&tv1, NULL); 
    printf ("Thread entering measurement phase (prev: %lu.%06"PRIu64")\n",
            tv1.tv_sec, (p_uint64)tv1.tv_usec);

    /* Measure phase */
    while (worker_phase == WORKER_PHASE_MEASURE) {
        issue_req (worker);
    }

    gettimeofday (&tv1, NULL);
    printf ("Thread entering slowdown phase (prev: %lu.%06"PRIu64")\n",
            tv1.tv_sec, (p_uint64)tv1.tv_usec);

    /* Slowdown phase */
    while (worker_phase == WORKER_PHASE_SLOWDOWN) {
        issue_req (worker);
    }

    gettimeofday (&tv1, NULL);
    printf ("Thread done (prev: %lu.%06"PRIu64")\n",
            tv1.tv_sec, (p_uint64)tv1.tv_usec);

    return NULL;
}

static inline unsigned int
calc_stdv (double sum, double sq_sum, double samples)
{
    double stdv;

    stdv = sqrt (sq_sum/samples - ((sum/samples) * (sum/samples)));

    return (unsigned int)stdv;
}
    

static void
calibrate (resp_stats_t *rstats)
{
    if (rstats->samples > 0) {
        rstats->stdv = calc_stdv (rstats->total_resp, rstats->total_sq_resp,
                                  rstats->samples);

        rstats->mean = rstats->total_resp / rstats->samples;

        if (rstats->stdv == 0) {
            /* something arbitrary to avoid divide by 0 */
            rstats->stdv = 1;
        }
    }

    rstats->calib_samples = rstats->samples;
    rstats->calib_failures = rstats->failures;

    rstats->samples = 0;
    rstats->failures = 0;
    rstats->total_resp = 0;
    rstats->total_sq_resp = 0;
}

struct timeval control_runtime;

/* Controls changing worker phases */
static void *
control_thread (void *arg)
{
    unsigned int i;
    struct timeval tv;

    worker_phase = WORKER_PHASE_WARMUP;
    st_cond_broadcast (worker_cond);
    
    st_sleep (warmup_time);
    worker_phase = WORKER_PHASE_CALIB;

    st_sleep (calib_time);

    for (i = 0; i < num_operation_types; ++i) {
        calibrate (&stats[i]);
    }

    current_period_start_time = st_utime ();
    worker_phase = WORKER_PHASE_MEASURE;
    st_cond_broadcast (worker_cond);
    gettimeofday (&control_runtime, NULL);

    st_sleep (run_time);
    worker_phase = WORKER_PHASE_SLOWDOWN;
    gettimeofday (&tv, NULL);

    timersub (&tv, &control_runtime, &control_runtime);

    st_sleep (slowdown_time);
    worker_phase = WORKER_PHASE_SHUTDOWN;
    
    for (i = 0; i < num_outstanding; i++) {
        st_thread_join (workers[i].thread_id, NULL);
    }

    return NULL;
}


static int
do_format (void)
{
    unsigned int i;
    int max_num_read_chunks = 0;
    
    /* Must perform a write for every chunk that can be read.
     * First, determine the highest chunk number that can be affected */
    for (i = 0; i < num_operation_types; i++) {
        if (operations[i]->isQueryOp  &&
            operations[i]->num_affected_chunks > max_num_read_chunks) {
            max_num_read_chunks = operations[i]->num_affected_chunks;
        }
    }

    /* Make sure we do at least a single write */
    if (max_num_read_chunks == 0) {
        max_num_read_chunks = 1;
    }

    /* Second, create every object and write the max # of read chunks */
    for (i = 0; i < num_objects; i++) {
        do_single_update (&objects[i].fid,
                          0,  /* arg size */
                          0,  /* res size */
                          chunk_size,
                          max_num_read_chunks,
                          0, /* compute time */
                          0); /* server override bitmap */
    }

    return 0;
}



static int
init (pasis_config_server_list_t *server_list)
{
    pasis_error_t status;
    unsigned int i, j;
    int fd, ret;

#if 1
    ret = st_set_eventsys(ST_EVENTSYS_ALT); /* use epoll instead of poll */
    printf("%s: using the %s StateThread event notification system\n",
           __FUNCTION__, st_get_eventsys_name());
#endif

    st_init ();

    srand (time (NULL));

    fd = open ("/dev/urandom", O_RDONLY);
    if (fd < 0) {
        fprintf (stderr, "Error: opening /dev/urandom failed (%s)\n",
                 PASIS_STRERROR (errno));
        return -1;
    }

    objects = new Object [num_objects];

    if (randomly_select_fids) {
        /* Assign the FID's from /dev/urandom */
        for (i = 0; i < num_objects; ++i) {
            read (fd, &objects[i].fid, sizeof(pasis_md_fid_t));
        }
    } else {
        /* Assign the FID's sequentially.  But put random bits in the top 16
         * bits of the first word, to help tree quorums select quorums */

        /* Usually we want the clients to use different objects, to avoid
         * concurrency.  But for the concurrency experiment, use #if 0 in the
         * next line */
#if 1
        j = client_id;
#else
        j = 0;
#endif

        fid_start += client_id * num_objects * (protocol.Qc+protocol.B); /* |Q| */

        for (i = 0; i < num_objects; ++i) {
#if 1
            *(unsigned int*)&objects[i].fid =
                fid_start +                /* start from here */
                i * (protocol.Qc+protocol.B);/* load balance*/
#else
            int first_level_shuffle, second_level_shuffle, shuffle;

            /* Lots of shuffling in the object ID's, needed for threshold/tree
             * quorums which select quorums based on the object ID's */
            *((unsigned int*)&objects[i].fid + 1) =
                fid_start + num_objects * client_id;
            *(unsigned int*)&objects[i].fid = j << 16 | (i + client_id);

            shuffle = *(unsigned int*)&objects[i].fid;
            first_level_shuffle = (unsigned int)(shuffle >> 16) % protocol.N;
            second_level_shuffle = (unsigned int)shuffle % protocol.N;

            if (i && i % protocol.N == 0) {
                ++j;
            }
#endif
        }
    }

    close (fd);

    Pasis_TimeInit ();

    status = PasisIO_Init (num_outstanding, 16384); /* block size unused */
    if (status) {
        PASIS_PRINT_ERR ("PasisIO_Init error: %s\n", PASIS_STRERROR (status));
        return -1;
    }

    /* Add storage nodes to PASIS drives module */
    status = PasisIO_ConnectToStorageNodes (server_list, 
                                            PASISIO_DRIVE_TYPE_MD_INDEX,
                                            linkages, instances);
    if (status) {
        PASIS_PRINT_ERR ("IOInit Error while connecting to drives\n");
        return -1;
    }

    worker_cond = st_cond_new ();
    workers = new WorkerState[num_outstanding];

    stats = (resp_stats_t*)malloc (num_operation_types * sizeof (resp_stats_t));
    memset (stats, 0, num_operation_types * sizeof (resp_stats_t));

    if (execution_rate_period > 0) {
        current_execution_rate_period = 0;
        max_num_execution_rate_periods =
            (run_time * 1000000) / execution_rate_period + 1;
        execution_rates = (unsigned int*) calloc (max_num_execution_rate_periods,
                                                  sizeof(unsigned int));
    }

    return 0;
}

static int
start_threads (void)
{
    unsigned int i;

    worker_phase = WORKER_PHASE_WAIT;

    /* Set up worker threads */
    for (i = 0; i < num_outstanding; i++) {
        WorkerState *worker = &workers[i];

        worker->thread_id = st_thread_create (worker_thread, worker, 1, 0);
    }

    controller = st_thread_create (control_thread, NULL, 1, 0);

    return 0;
}

static inline void
dump_stats (resp_stats_t *rstats)
{
    int i, min_bucket, max_bucket;
    unsigned int meas_stdv;

    printf ("Calibration: samples=%u, failures=%u, mean=%.4fms, stdv=%.4fms\n",
            rstats->calib_samples, rstats->calib_failures,
            (double)rstats->mean / 1000.0,
            (double)rstats->stdv / 1000.0);
    
    meas_stdv = calc_stdv (rstats->total_resp, rstats->total_sq_resp,
                           rstats->samples);

    printf ("Measurement: samples=%u, failures=%u, mean=%.4fms, stdv=%.4fms\n",
            rstats->samples, rstats->failures,
            (double)rstats->total_resp/(double)rstats->samples/1000.0,
            (double)meas_stdv / 1000.0);

    printf ("\nHistogram: ");
    min_bucket = max_bucket = 0;
    for (i = 0; i < RESP_STATS_BINS; i++) {
        printf ("| %d ", rstats->bins[i]);
        if (min_bucket == 0 && rstats->bins[i] > 0) {
            min_bucket = i;
        }
        if (rstats->bins[i] > 0) {
            max_bucket = i;
        }
    }
    printf ("\n");
    printf ("           %d buckets, min=%.4fms, max=%.4fms\n",
            RESP_STATS_BINS,
            ((double)rstats->mean + (double)(min_bucket - RESP_STATS_BIN_MEAN) *
            rstats->stdv) / 1000.0,
            ((double)rstats->mean + (double)(max_bucket - RESP_STATS_BIN_MEAN) *
            rstats->stdv) / 1000.0);
    printf ("           min value = %.4fms, max value = %.4fms\n",
            (double)rstats->min / 1000.0, (double)rstats->max / 1000.0);
    printf ("\n\n");
}


static void
print_execution_rates (void)
{
    unsigned int i;

    if (execution_rate_period == 0) {
        return;
    }

    printf ("\n\n\n-------------------------\n");
    printf ("Execution rates:  (in %d bins of %lld seconds each)\n",
            current_execution_rate_period,
            execution_rate_period / 1000000);

    printf("[%d", execution_rates[0]); /* Matlab-friendly output */
    for (i = 1; i <= current_execution_rate_period; ++i) {
        printf(" %d", execution_rates[i]);
    }
    printf("]\n");
    printf ("-------------------------\n");
}


static void
analyze (void)
{
    double total_time, data_size, weighted_data_size, tput, total_tput;
    unsigned int i, j;

    printf ("\nResults:\n");
    printf ("--------\n\n");

    printf ("Run Time        :  %u s (%lu.%06"PRIu64")\n", run_time, 
            control_runtime.tv_sec, (p_uint64)control_runtime.tv_usec);
    printf ("Max Outstanding :  %u\n\n", num_outstanding);

    total_time = 0;
    for (i = 0; i < num_outstanding; i++) {
        WorkerState *worker = &workers[i];

        for (j = 0; j < num_operation_types; ++j) {
            total_time += (double) worker->ts[j].ts_sec +
                ((double) worker->ts[j].ts_nsec / 1000000000.0);
        }
    }

    weighted_data_size = 0;
    total_tput = 0;
    for (i = 0; i < num_operation_types; ++i) {
        if (stats[i].samples > 0) {
            printf ("\n\n-------------------------\n");
            printf (operations[i]->operation_name);
            printf ("\n-------------------------\n\n");
        
            dump_stats (&stats[i]);

            tput = (double) stats[i].samples / (total_time) * 
                (double)num_outstanding;
            total_tput += tput;

            data_size = operations[i]->arg_size + operations[i]->res_size;
            weighted_data_size += data_size * operations[i]->relative_frequency;
            printf ("Throughput :  %.3f req/s,  %.3f MB/s\n", 
                    tput, tput * (double)data_size / 1024.0 / 1024.0);
        }
    }

    weighted_data_size /= total_relative_frequencies;
    printf ("\n\n\n-------------------------\n");
    printf ("Total Throughput:  %.3f req/s,  %.3f MB/s\n",
            total_tput, total_tput * (double) weighted_data_size / 1024.0 / 1024.0);
    printf ("-------------------------\n");

    printf ("\n");

    print_execution_rates ();
}


static void
cleanup (void)
{
    unsigned int i;

    delete[] objects;
    delete[] workers;
    for (i = 0; i < num_operation_types; ++i) {
        delete operations[i];
    }
    free (operations);
    free (stats);
    st_cond_destroy (worker_cond);
}


static void
usage (char *name)
{
    PASIS_PRINT_INFO ("Usage: %s [-fF] config_filename [-c client_id]\n", name);
    PASIS_PRINT_INFO ("  -f             Format, do writes to all blocks\n");
    PASIS_PRINT_INFO ("  -F             Format, followed by execution\n");
    PASIS_PRINT_INFO ("  -c <client ID> Client ID\n");
    PASIS_PRINT_INFO ("  -H             Do _not_ use a History cache.\n");
    PASIS_PRINT_INFO ("  -R             Use random quorums (instead of preferred quorums)\n");
    PASIS_PRINT_INFO ("  -p <p>         Print execution rates in <p> periods\n");
}


int
main (int argc, char *argv[])
{
    char                      *filename=NULL;
    char                       c;
    int                        format=0, use_random_quorums = 0;
    pasis_config_server_list_t server_list;

    while ((c = getopt (argc, argv, "c:fFhHRp:s:")) != -1) {
        switch (c) {
        case 'c':
            sscanf (optarg, "%u", &client_id);
            break;
        case 'f':
            format = 1;
            break;
        case 'F':
            format = 2;
            break;
        case 'H':
            use_history_cache = 0;
            break;
        case 'R':
            use_random_quorums = 1;
            break;
        case 'p':
            execution_rate_period = atoi (optarg);
            execution_rate_period *= 1000000; /* convert to us */
            break;
        case 'h':
        default:
            usage (argv[0]);
            exit (1);
        }
    }

    if (optind < argc) {
        filename = argv[optind];
    }

    if (filename == NULL) {
        usage (argv[0]);
        exit (1);
    }

    if (parse_file (filename, &server_list)) {
        exit (1);
    }

    if (init (&server_list)) {
        exit (1);
    }

    if (format) {
        do_format ();
        printf ("Format Done\n");

        if (format == 1) {
            PasisIO_Shutdown ();
            cleanup ();
            exit (0);
        }
    }

    if (use_random_quorums) {
        PasisIO_UseRandomQuorums ();
    }
    if (start_threads ()) {
        exit (1);
    }
    st_thread_join (controller, NULL);

    PasisIO_Shutdown ();

    analyze ();

    cleanup ();

    return 0;
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
