/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/*
 * This file defines a FIFO & LRU cache.
 * It can be used to cache any data as long as the cache's key type has
 * the operator "<" defined for it.
 * And maybe eventually supporting RANDOM replacement
 *
 * CHANGELOG:
 * 2001/11/13 - jstrunk - Added ability to do LRU in addition to just FIFO
 * 2001/09/03 - jstrunk - Adapted this from my S4 in-time cache code.  Removed
 *              the in-time caching aspects.
 */

#ifndef CACHE_H
#define CACHE_H

#include <map>
#include <string>
#include <iostream>

using namespace std;


template<class Tag, class Data> class Cache {
public:
    // The possible cache replacement policies
    enum replacement_t {
        LRU,                              // Least recently used
        FIFO                              // First in, first out
    };

private:
    typedef unsigned long time_index_t;   // monotonic time counter
    time_index_t timeIndex;               // time counter
    struct data_node_t;                   // holds actual cache data
    const unsigned maxEntries;            // maximum cache entries
    unsigned long tries, hits;            // cache stats
    string cacheName;                     // name of the cache
    const replacement_t policy;           // policy for the cache

    // keep entries indexed by time
    typedef map<time_index_t, Tag> timemap_t;
    timemap_t timeList;
    // keep entries indexed by cache tag
    typedef map<Tag, data_node_t> tagmap_t;
    tagmap_t dataList;

    // delete an entry using an iterator into either tree
    inline void
	DelEntry(typename timemap_t::iterator i)
        throw();
    inline void
	DelEntry(typename tagmap_t::iterator i)
        throw();

public:
    Cache(unsigned mE,                    // max cache entries
          string name = "",               // optional name of the cache
          replacement_t rAlg = LRU)       // replacement algorithm to use
        throw();
    ~Cache();

    // Insert a new entry into the cache.  This will overwrite an
    // already-existing entry with the same tag.
    // Provide the tag and data
    inline void
    Insert(const Tag &t,
           const Data &d)
        throw();

    // Delete an entry from the cache.
    // Provide the Tag
    inline void
    Delete(const Tag &t)
        throw();

    // Find an entry in the cache.
    inline Data *
    Lookup(const Tag &t)
        throw();

    // Find newest (most recently used, in the case of LRU) entry in
    // the cache.
    inline Tag *
    Newest()
        throw();

    // Flush the entire cache.
    void
    Flush()
        throw();

    // Return the number of entries in the cache
    unsigned
    Size() const
        throw();
};



template<class Tag, class Data> class Cache<Tag, Data>::data_node_t {
    time_index_t tIndex;                  // index so we can find in timeList
    Data data;                            // cache data
};



/*
 * Inline member functions...
 */

template<class Tag, class Data> inline void
Cache<Tag, Data>::Insert(const Tag &t,
                         const Data &d)
    throw()
{
    // make sure it's not already here
    DelEntry(dataList.find(t));

    // add new entry
    data_node_t &newData = dataList[t];
    newData.tIndex = timeIndex++;
    newData.data = d;

    timeList[newData.tIndex] = t;

    // make sure we don't have too many entries
    if (timeList.size() > maxEntries) {
        DelEntry(timeList.begin());
    }

    // if timeIndex wraps around, flush the cache because bad things will
    // happen
    if (timeIndex == 0) {
        Flush();
    }
}

template<class Tag, class Data> inline void
Cache<Tag, Data>::Delete(const Tag &t)
    throw()
{
    DelEntry(dataList.find(t));
}

template<class Tag, class Data> inline Data *
Cache<Tag, Data>::Lookup(const Tag &t)
    throw()
{
    tries++;

    typename tagmap_t::iterator i;
    i = dataList.find(t);
    if (i == dataList.end()) {  // no entry for this tag
        return 0;
    }

    if (policy == LRU) {

        timeList.erase(i->second.tIndex);
        i->second.tIndex = timeIndex++;
        timeList[i->second.tIndex] = i->first;

        // Just like Insert()... watch the wraparound
        if (timeIndex == 0) {
	    Flush();
        }
    }

    hits++;
    return &i->second.data;
}

template<class Tag, class Data> inline Tag *
Cache<Tag, Data>::Newest()
    throw()
{
    typename timemap_t::reverse_iterator i = timeList.rbegin();
    if (i != timeList.rend())
        return &i->second;
    else
        return 0;
}


/*
 * Implementation below here....
 *
 * This is only here because g++ doesn't support templates from a
 * different compilation unit.
 */


template<class Tag, class Data>
Cache<Tag, Data>::Cache(unsigned mE,
                        string name,
                        replacement_t rAlg)
    throw() :
    maxEntries(mE), policy(rAlg)
{
    cacheName = name;
    timeIndex = 0;
    tries = 0;
    hits = 0;
}

template<class Tag, class Data>
Cache<Tag, Data>::~Cache()
{
    cout << "Cache: " << cacheName
              << " -- maxEntries:" << maxEntries
              << " hits:" << hits
              << " tries:" << tries << endl;
}

template<class Tag, class Data> void
Cache<Tag, Data>::Flush()
    throw()
{
    timeList.clear();
    dataList.clear();
}

template<class Tag, class Data> unsigned
Cache<Tag, Data>::Size() const
    throw()
{
    return timeList.size();
}


/*
 * Private member functions
 */

template<class Tag, class Data> inline void
Cache<Tag, Data>::DelEntry(typename timemap_t::iterator i)
    throw()
{
    if (i == timeList.end()) {
        return;
    }

    dataList.erase(i->second);
    timeList.erase(i);
}

template<class Tag, class Data> inline void
Cache<Tag, Data>::DelEntry(typename tagmap_t::iterator i)
    throw()
{
    if (i == dataList.end()) {
        return;
    }

    timeList.erase(i->second.tIndex);
    dataList.erase(i);
}


#endif // !CACHE_H
