/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_debug.h"

#include "pasis_freelist.h"
#include "pasis_cache.h"

static char *
internalLookup (pasis_cache_t *cache,
                void *tag,
                unsigned int hash)
{
    char *ptr;
    pasis_cache_obj_t *obj;

    PASIS_DPRINTF (PASIS_DEBUG_CACHE, 5, "%s: hash=%#x, tag=%#x\n", 
                   __FUNCTION__, hash, *(unsigned*)tag);

    obj = PASIS_Q_FIRST (cache->hash_array[hash]);
    while (obj) {
        ptr = (char *)obj;
        if (cache->compare_fp ((void *)(ptr + sizeof (pasis_cache_obj_t)), 
            tag)) {
            PASIS_Q_REMOVE (cache->lru_queue, lru_q, obj);
            PASIS_Q_ADD_HEAD (cache->lru_queue, lru_q, obj);
            PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, "%s: cache hit\n", 
                           __FUNCTION__);
            return ptr;
        }
        obj = PASIS_Q_NEXT (chain, obj);
    }
    PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, "%s: cache miss\n", __FUNCTION__);
    return NULL;
}

static void
internalEvict (pasis_cache_t *cache)
{
    pasis_cache_obj_t *obj;
    
    obj = PASIS_Q_LAST (cache->lru_queue);
    while (obj->pin_count > 0) {
        /* Assuming cache isn't of size 1! */
        PASIS_DASSERT (obj != PASIS_Q_FIRST (cache->lru_queue));
        obj = PASIS_LL_PREV (lru_q, obj);
    }

    PASIS_Q_REMOVE (cache->lru_queue, lru_q, obj);
    PASIS_Q_REMOVE (cache->hash_array[obj->hash], chain, obj);

    PasisFreelist_Put (cache->freelist, (void *)obj);

    cache->size--;
}


void
PasisCache_PinData (pasis_cache_t *cache,
                    void          *data)
{
    char *ptr;
    pasis_cache_obj_t *obj;

    ptr = ((char *)data) - cache->tag_size - sizeof (pasis_cache_obj_t);
    obj = (pasis_cache_obj_t *)ptr;

    ptr = ((char *)data) - cache->tag_size;
    PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, "Pinning: tag=%#x, obj=%p\n", 
                   *(unsigned int *)ptr, obj);

    PASIS_DASSERT (obj->pin_count >= 0);
    obj->pin_count++;
}


void
PasisCache_UnpinData (pasis_cache_t *cache,
                      void          *data)
{
    char *ptr;
    pasis_cache_obj_t *obj;

    ptr = ((char *)data) - cache->tag_size - sizeof (pasis_cache_obj_t);
    obj = (pasis_cache_obj_t *)ptr;

    ptr = ((char *)data) - cache->tag_size;
    PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, "Unpinning: tag=%#x, obj=%p\n", 
                   *(unsigned int *)ptr, obj);

    PASIS_DASSERT (obj->pin_count > 0);

    /* Move to head of LRU queue */
    PASIS_Q_REMOVE (cache->lru_queue, lru_q, obj);
    PASIS_Q_ADD_HEAD (cache->lru_queue, lru_q, obj);

    obj->pin_count--;
}


void
PasisCache_Pin (pasis_cache_t *cache,
                void *tag)
{
    pasis_cache_obj_t *obj;
    unsigned int hash;

    hash = cache->hash_fp (tag, cache->buckets);
    PASIS_DASSERT (hash < cache->buckets);

    obj = (pasis_cache_obj_t *)internalLookup (cache, tag, hash);
    if (obj) {
        PASIS_DASSERT (obj->pin_count >= 0);
        obj->pin_count++;
    }

    return;   
}

void
PasisCache_Unpin (pasis_cache_t *cache,
                  void *tag)
{
    pasis_cache_obj_t *obj;
    unsigned int hash;

    hash = cache->hash_fp (tag, cache->buckets);
    PASIS_DASSERT (hash < cache->buckets);

    obj = (pasis_cache_obj_t *)internalLookup (cache, tag, hash);
    if (obj) {
        PASIS_DASSERT (obj->pin_count >= 0);
        obj->pin_count--;
    }

    return;
}

void
PasisCache_Remove (pasis_cache_t *cache,
                   void *tag)
{
    pasis_cache_obj_t *obj;
    unsigned int hash;

    hash = cache->hash_fp (tag, cache->buckets);
    PASIS_DASSERT (hash < cache->buckets);

    obj = (pasis_cache_obj_t *)internalLookup (cache, tag, hash);

    if (!obj) {
        return;
    }

    if (obj->pin_count > 0) {
        obj->pin_count--;
        if (obj->pin_count > 0) {
            return;
        }
    }

    PASIS_Q_REMOVE (cache->lru_queue, lru_q, obj);
    PASIS_Q_REMOVE (cache->hash_array[obj->hash], chain, obj);
    PasisFreelist_Put (cache->freelist, (void *)obj);

    cache->size--;
}

void *
PasisCache_Lookup (pasis_cache_t *cache,
                   void *tag,
                   int pin)
{
    unsigned int hash;
    char *ptr;

    hash = cache->hash_fp (tag, cache->buckets);
    PASIS_DASSERT (hash < cache->buckets);

    ptr = internalLookup (cache, tag, hash);

    if (ptr != NULL) {
        if (pin) {
            pasis_cache_obj_t *obj = (pasis_cache_obj_t *)ptr;
            obj->pin_count++;
        }
        PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, "%s: cache hit\n", __FUNCTION__);
        return (ptr + cache->tag_size + sizeof (pasis_cache_obj_t));
    }

    PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, "%s: cache miss\n", __FUNCTION__);
    return NULL;
}

void
PasisCache_Insert (pasis_cache_t *cache,
                   void *tag,
                   void *data)
{
    char *ptr;
    unsigned int hash;
    pasis_cache_obj_t *obj;

    hash = cache->hash_fp (tag, cache->buckets);
    PASIS_DASSERT (hash < cache->buckets);

    if ((ptr = internalLookup (cache, tag, hash)) != NULL) {
        memcpy (ptr + sizeof (pasis_cache_obj_t), tag, cache->tag_size);
        memcpy (ptr + cache->tag_size + sizeof (pasis_cache_obj_t), data,
                cache->data_size);
        return;
    }

    if (cache->fixed_flag && cache->size >= cache->expected_size) {
        internalEvict (cache);
    }
    
    ptr = PasisFreelist_Get (cache->freelist);
    PASIS_DASSERT (ptr != NULL);

    memcpy (ptr + sizeof (pasis_cache_obj_t), tag, cache->tag_size);
    memcpy (ptr + cache->tag_size + sizeof (pasis_cache_obj_t), data,
            cache->data_size);
    
    obj = (pasis_cache_obj_t *)ptr;
    obj->hash = hash;
    obj->pin_count = 0;

    PASIS_Q_ADD_HEAD (cache->hash_array[hash], chain, obj);
    PASIS_Q_ADD_HEAD (cache->lru_queue, lru_q, (pasis_cache_obj_t *)ptr);

    cache->size++;

    return;
}

/* Similar to insert, but either gets an empty (evicted) entry or the entry
 * that matches the tag.  It is inserted into the cache, so it is up to the
 * user to zero the entry (or update it) before blocking (when it becomes
 * visible to others) */
void *
PasisCache_GetEmpty (pasis_cache_t *cache,
                     void          *tag)
{
    char *ptr;
    unsigned int hash;
    pasis_cache_obj_t *obj;
 
    /* Sanity check first */
    hash = cache->hash_fp (tag, cache->buckets);
    PASIS_DASSERT (hash < cache->buckets);

    if ((ptr = internalLookup (cache, tag, hash)) != NULL) {
        obj = (pasis_cache_obj_t *)ptr;

        PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, 
                       "%s: tag=%#x, obj=%p (pin_count=%d)\n", __FUNCTION__, 
                       *(unsigned int *)tag, obj, obj->pin_count);

        PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, "%s: cache hit\n", __FUNCTION__);

        PASIS_DASSERT (obj->pin_count == 0);
        return (ptr + cache->tag_size + sizeof (pasis_cache_obj_t));
    }

    PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, "%s: cache miss\n", __FUNCTION__);

    if (cache->fixed_flag && cache->size >= cache->expected_size) {
        internalEvict (cache);
    }
    
    ptr = PasisFreelist_Get (cache->freelist);
    PASIS_DASSERT (ptr != NULL);

    memcpy (ptr + sizeof (pasis_cache_obj_t), tag, cache->tag_size);
     
    obj = (pasis_cache_obj_t *)ptr;
    obj->hash = hash;
    obj->pin_count = 0;

    PASIS_DPRINTF (PASIS_DEBUG_CACHE, 4, "%s: returning obj=%p\n", 
                   __FUNCTION__, obj);

    PASIS_Q_ADD_HEAD (cache->hash_array[hash], chain, obj);
    PASIS_Q_ADD_HEAD (cache->lru_queue, lru_q, (pasis_cache_obj_t *)ptr);

    cache->size++;
    return (ptr + cache->tag_size + sizeof (pasis_cache_obj_t));
}


void
PasisCache_Dump (pasis_cache_t *cache, 
                 void (*dump_fp)(void *tag, void *data))
{
    pasis_cache_obj_t *obj;
    obj = PASIS_Q_FIRST (cache->lru_queue);
    while (obj) {
        dump_fp ((void *)((char *)(obj) + sizeof(pasis_cache_obj_t)),
                 (void *)((char *)(obj) + sizeof(pasis_cache_obj_t) + 
                          cache->tag_size));
        obj = PASIS_Q_NEXT (lru_q, obj);
    }
}

pasis_cache_t *
PasisCache_Create (int expected_size,
                   char fixed_flag,
                   unsigned int tag_size, 
                   unsigned int data_size,
                   unsigned int (* hash_fp)(void *tag, unsigned int buckets),
                   int (* compare_fp)(void *tag, void *tag2))
{
    pasis_cache_t *cache;
    unsigned int elems, elem_size;
    int i;

    cache = (pasis_cache_t *) malloc (sizeof (pasis_cache_t));
    cache->expected_size = expected_size;
    cache->fixed_flag = fixed_flag;
    cache->size = 0;
    cache->tag_size = tag_size;
    cache->data_size = data_size;
    cache->hash_fp = hash_fp;
    cache->compare_fp = compare_fp;

    cache->buckets = (expected_size >> 3) - 1;
    if (cache->buckets == 0) {
        cache->buckets = expected_size;
    }

    elems = cache->buckets >> 2;
    if (elems == 0) {
        elems = cache->buckets;
    }

    elem_size = tag_size + data_size + sizeof (pasis_cache_obj_t);

    if (fixed_flag) {
        cache->freelist = PasisFreelist_CreateFixed (elem_size, expected_size);
    } else {
        cache->freelist = PasisFreelist_Create (elem_size, elems, 
                                                elems/2, elems); 
    }
    
    cache->hash_array = (pasis_cache_queue_t *) malloc (cache->buckets * 
                                                 sizeof (pasis_cache_queue_t));

    for (i = 0; i < cache->buckets; i++) {
        PASIS_Q_INIT (cache->hash_array[i]);
    }

    PASIS_Q_INIT (cache->lru_queue);

    return cache;
}

void 
PasisCache_Shutdown (pasis_cache_t *cache)
{
    while (cache->size > 0) {
        internalEvict (cache);
    }

    free (cache->hash_array);

    PasisFreelist_Destroy (cache->freelist);
    
    free (cache);
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
