#ifndef __RPC_COMMON_H__
#define __RPC_COMMON_H__

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_freelist.h"
#include "pasis_timer.h"

#if USE_KERBEROS
#include "pasisio.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif
#if 0 
} 
#endif

#define RPC_RETRIES             2          /* Number of RPC retries */
#define RPC_TIMEOUT_BASE        1000000LL  /* RPC base timeout in useconds */
#define RPC_TIMEOUT_MULT        2          /* RPC timeout multiplier */
#define RPC_TRANSMIT_HIGH_WATER 4          /* Per client outstanding rpcs */

#define RPC_ACCEPT_TIMEOUT      5000000LL  /* TCP accept timeout in usecs */
#define RPC_REQ_TIMEOUT        60000000LL  /* TCP rpc request timeout usecs */
#define RPC_CONNECT_TIMEOUT    10000000LL  /* TCP connect timeout in usecs */

#define RPC_MAX_IOVEC_LEN   (18)
#define RPC_REQ_HDR_SIZE    (320 >> 2) /* 24 hdr bytes, 296 max auth bytes */
#define RPC_REPLY_HDR_SIZE  (400 >> 2)
#define RPC_MAX_DATA_SIZE   (128 << 10) /* Maximum pure data size */
#define RPC_XDR_PAD_SIZE    (800)       /* XDR padding size -- pretty big! */

#define RPC_MAX_PAYLOAD     (RPC_MAX_DATA_SIZE + RPC_XDR_PAD_SIZE) 

#define RPC_MAX_CONNECTIONS 256        /* Maximum number of tcp connections */

#define RPC_AUTH_NONE       0               /* no authentication */
#define RPC_AUTH_NULL       0               /* backward compatibility */
#define RPC_AUTH_SYS        1               /* unix style (uid, gids) */
#define RPC_AUTH_UNIX       RPC_AUTH_SYS
#define RPC_AUTH_SHORT      2               /* short hand unix style */

/* RPC error codes -- only really used by tcp client (right now) */
#define RPC_STATUS_SUCCESS     0   /**< RPC Success */
#define RPC_STATUS_ERROR      -1   /**< RPC Generic error */
#define RPC_STATUS_TIMEOUT    -2   /**< RPC Timeout */
#define RPC_STATUS_TERM       -3   /**< RPC Terminated */
#define RPC_STATUS_CONN       -4   /**< RPC Connection Broken */
#define RPC_STATUS_AUTH       -5   /**< RPC Authentication error */

typedef enum rpc_msg_type {
        RPC_CALL=0,
        RPC_REPLY=1
} rpc_msg_type_t;

typedef enum rpc_reply_stat {
        RPC_MSG_ACCEPTED=0,
        RPC_MSG_DENIED=1
} rpc_reply_stat_t;

typedef enum rpc_accept_stat {
        RPC_ACCEPT_SUCCESS=0,
        RPC_PROG_UNAVAIL=1,
        RPC_PROG_MISMATCH=2,
        RPC_PROC_UNAVAIL=3,
        RPC_GARBAGE_ARGS=4,
        RPC_SYSTEM_ERR=5
} rpc_accept_stat_t;

typedef enum rpc_reject_stat {
        RPC_REJECT_MISMATCH=0,
        RPC_AUTH_ERROR=1
} rpc_reject_stat_t;

typedef enum rpc_auth_stat {
        RPC_AUTH_OK=0,
        /* failed at remote end */
        RPC_AUTH_BADCRED=1,             /* bogus credentials (seal broken) */
        RPC_AUTH_REJECTEDCRED=2,        /* client should begin new session */
        RPC_AUTH_BADVERF=3,             /* bogus verifier (seal broken) */
        RPC_AUTH_REJECTEDVERF=4,        /* verifier expired or was replayed */
        RPC_AUTH_TOOWEAK=5,             /* rejected due to security reasons */
        /* failed locally */
        RPC_AUTH_INVALIDRESP=6,         /* bogus response verifier */
        RPC_AUTH_FAILED=7               /* some unknown reason */
} rpc_auth_stat_t;


typedef struct rpc_auth_s {
    u_int32_t          uid;
    u_int32_t          gid;
} rpc_auth_t;


typedef struct rpc_tcp_reply_hdr_s {
#if USE_KERBEROS
    p_byte    signature[PASIS_HASH_SIZE];
    p_uint32  hash_region_len;          /* Data region covered by signature */
                                        /* excluding header (data args only) */
#endif
    p_uint32  xid;
    p_uint32  msg_type;                 /* Should be RPC_REPLY */
    p_uint32  reply_status;
    p_uint32  accepted_status;          
    p_uint32  auth_status;              
    p_uint32  payload_len;
} rpc_tcp_reply_hdr_t;

typedef struct rpc_tcp_call_hdr_s {
#if USE_KERBEROS
    p_byte    signature[PASIS_HASH_SIZE];
    p_uint32  hash_region_len;          /* Data region covered by signature */
                                        /* excluding header (data args only) */
#endif
    p_uint32  xid;
    p_uint32  msg_type;                 /* Should be RPC_CALL */
    p_uint32  prog;
    p_uint32  vers;
    p_uint32  proc;
    p_uint32  payload_len;
} rpc_tcp_call_hdr_t;


typedef enum rpc_handle_type_e {
    RPC_HANDLE_CLIENT = 0,
    RPC_HANDLE_SERVER = 1
} rpc_handle_type_t;


typedef struct rpc_req_s {
    u_int32_t        *rpc_hdr;
    u_int32_t         rpc_xid;
    u_int32_t         rpc_prog;
    u_int32_t         rpc_vers;
    u_int32_t         rpc_proc;
    rpc_auth_t        rpc_auth;

    p_byte           *rpc_buffer;

    int               wait_status;

    rpc_reply_stat_t  reply_status; /* 0 - ACCEPTED, 1 - DENIED */
    /* For denied reply statuses */
    rpc_reject_stat_t reply_rejected_status; /* 0 - RPC, 1 - AUTH */
    rpc_auth_stat_t   reply_rejected_auth_status; /* If rejected stat == 1 */
    /* For accepted reply statuses */
    rpc_accept_stat_t reply_accepted_status; /* 0 - SUCCESS, 1-5 - FAILURES */ 

    struct iovec      req_iov[RPC_MAX_IOVEC_LEN];
    unsigned int      req_iov_len;

    struct iovec      reply_iov[RPC_MAX_IOVEC_LEN];
    unsigned int      reply_iov_len;
    int               reply_msg_len;

    int               req_iov_cur;
    int               reply_iov_cur;

    int               wait_flag;
    int               transmit_flag;
    st_cond_t         wait_cond;

    struct sockaddr_in reply_addr;  /* Used by udp server to reply only */
    st_netfd_t         reply_netfd; /* Used by tcp server to reply only */
    p_uint32           reply_ip;    /* Used by tcp server to reply only */
    
    struct rpc_req_table_s *req_entry; /* Used by tcp client only */

    p_byte           *data_buffer;

    struct rpc_req_s *trans_next;
    struct rpc_req_s *trans_prev;

    struct rpc_req_s *next; 
    struct rpc_req_s *prev;
} rpc_req_t;


typedef enum rpc_req_state_e {
    RPC_REQ_STATE_INVALID = 0,
    RPC_REQ_STATE_VALID = 1,
    RPC_REQ_STATE_TRANSMIT = 2,
    RPC_REQ_STATE_WAIT = 3,
    RPC_REQ_STATE_RECEIVE = 4
} rpc_req_state_t;


typedef struct rpc_req_table_s {
    rpc_req_state_t    state;
    rpc_req_t         *req;
    st_thread_t        thread_id;
    int                terminate;
} rpc_req_table_t;


typedef struct rpc_handle_s {
    rpc_handle_type_t  handle_type;

    unsigned int       max_outstanding;
    
    st_netfd_t         sock_fd;
    struct sockaddr_in sock_addr;
    st_mutex_t         sock_lock;

    p_uint32           local_ip;
    p_uint32           remote_ip;

    pasis_freelist_t  *req_freelist;
    /* if threads are waiting for a free entry */
    st_cond_t          req_freelist_cond;

    rpc_req_t         *inprog_queue_hd;
    rpc_req_t         *inprog_queue_tl;

    rpc_req_t         *trans_queue_hd;
    rpc_req_t         *trans_queue_tl;
    unsigned int       trans_inprog_count;

    int                req_poll_shutdown;
    st_thread_t        req_poll_thread;

    unsigned long      proto;

    /* Statistics */
    struct timeval     create_time;
    p_uint64           bytes_read;
    p_uint64           bytes_written;
    p_uint32           num_requests;

    pasis_timer_state_t *timers;

    /* Used by server only */
    pasis_freelist_t  *buffer_freelist;
    unsigned long      program;
    unsigned long      version;
    void             (*callback)(rpc_req_t *req, p_byte *data, 
                                 unsigned int len);

    /* Used by tcp client */
    void              *disconnect_arg;
    void             (*disconnect_handler)(void *);
    st_cond_t          reconnect_cond;
    rpc_req_table_t  *req_table;
    /* if threads are waiting for a free entry */
    st_cond_t          req_table_cond;
    
    /* Used by udp client */
    char             *local_hostname;
    unsigned int      local_hostlen;

} rpc_handle_t;

/* To shutdown cleanly:
 * 1. call rpc_signal_shutdown for all handles
 * 2. make sure all outstanding reqs have completed
 * 3. call rpc_server/client_destroy for all handles
 * 4. call rpc_shutdown
 */

int rpc_init (void);

int rpc_auth_client_init (void);

int rpc_auth_server_init (void);

void rpc_shutdown (void);

void rpc_auth_client_shutdown (void);

void rpc_auth_server_shutdown (void);

p_uint32 rpc_get_host_ip (void);

void rpc_signal_shutdown (rpc_handle_t *); 

rpc_req_t *rpc_req_get (rpc_handle_t *);

void rpc_req_put (rpc_handle_t *, rpc_req_t *);

u_int32_t rpc_req_xid_get (void);

void rpc_encode_tcp_call (rpc_handle_t *, rpc_req_t *req, 
                          rpc_tcp_call_hdr_t *hdr);

void rpc_encode_tcp_reply (rpc_handle_t *, rpc_req_t *req, rpc_tcp_reply_hdr_t *hdr);

int rpc_decode_tcp_reply (rpc_tcp_reply_hdr_t *hdr);

int rpc_decode_tcp_call (rpc_req_t *req, rpc_tcp_call_hdr_t *rpc_hdr, 
                         p_uint32 remote_ip);
                         

u_int32_t *rpc_encode_udp_call (rpc_handle_t *client, 
                                rpc_req_t *req, u_int32_t *rpc_hdr);

u_int32_t *rpc_decode_udp_call (rpc_req_t *req, u_int32_t *rpc_hdr);
                                

int rpc_decode_udp_reply (rpc_req_t *req, u_int32_t *rpc_hdr);

int rpc_req_udp_wait (rpc_handle_t *client, rpc_req_t *req);

int rpc_transmit_udp_req (rpc_handle_t *client, rpc_req_t *req);

int rpc_transmit_tcp_req (rpc_handle_t *client, rpc_req_t *req);

int rpc_transmit_udp_reply (rpc_handle_t *server, rpc_req_t *req);

int rpc_transmit_tcp_reply (rpc_handle_t *server, rpc_req_t *req);

int rpc_client_reconnect (rpc_handle_t *client);

rpc_handle_t * rpc_client_create_common (p_uint32 server_ip,
                                         unsigned int max_outstanding,
                                         unsigned long prog,
                                         unsigned long vers,
                                         unsigned long proto);

rpc_handle_t *rpc_client_create_tcp (p_uint32 server_ip,
                                     unsigned int max_outstanding,
                                     unsigned long prog, unsigned long vers,
                                     void *arg,
                                     void (*disconnect_handler)(void *));

rpc_handle_t *rpc_client_create_udp (p_uint32 server_ip,
                                     unsigned int max_outstanding, 
                                     unsigned long prog, unsigned long vers);

void rpc_client_destroy (rpc_handle_t *client);


rpc_handle_t *rpc_server_create_common (unsigned int max_oustanding,
                                        unsigned short port, 
                                        unsigned long prog,
                                        unsigned long vers,
                                        unsigned long proto,
                                        void (*callback)(rpc_req_t *, p_byte *,
                                                         unsigned int));

rpc_handle_t *rpc_server_create_udp (unsigned int max_oustanding,
                                     unsigned short port, unsigned long prog,
                                     unsigned long vers,
                                     void (*callback)(rpc_req_t *, p_byte *,
                                                         unsigned int));

rpc_handle_t *rpc_server_create_tcp (unsigned int max_oustanding,
                                     unsigned short port, unsigned long prog, 
                                     unsigned long vers,
                                     void (*callback)(rpc_req_t *, p_byte *,
                                                      unsigned int));
                                                   

void rpc_server_destroy (rpc_handle_t *server);


void rpc_early_terminate (rpc_handle_t *handle, st_thread_t   thread_id);
                          

/**** Inline Helper Functions ****/

static inline void
transmit_req_queue (rpc_handle_t *handle, 
                    rpc_req_t *req)
{
    /* Add to tail */
    if (handle->trans_queue_tl == NULL) {
        handle->trans_queue_hd = handle->trans_queue_tl = req;
        req->trans_next = req->trans_prev = NULL;
    } else {
        handle->trans_queue_tl->trans_next = req;
        req->trans_prev = handle->trans_queue_tl;
        req->trans_next = NULL;
        handle->trans_queue_tl = req;
    }
}


static inline void
transmit_req_dequeue (rpc_handle_t *handle,
                      rpc_req_t *req)
{
    if (handle->trans_queue_hd == req) {
        handle->trans_queue_hd = req->trans_next;
    } else {
        req->trans_prev->trans_next = req->trans_next;
    }
    if (handle->trans_queue_tl == req) {
        handle->trans_queue_tl = req->trans_prev;
    } else {
        req->trans_next->trans_prev = req->trans_prev;
    }
}


static inline void
inprog_req_queue (rpc_handle_t *handle, 
                  rpc_req_t *req)
{
    if (handle->inprog_queue_tl == NULL) {
        handle->inprog_queue_hd = handle->inprog_queue_tl = req;
        req->next = req->prev = NULL;
    } else {
        handle->inprog_queue_tl->next = req;
        req->prev = handle->inprog_queue_tl;
        req->next = NULL;
        handle->inprog_queue_tl = req;
    }
}


static inline void
inprog_req_dequeue (rpc_handle_t *handle,
                    rpc_req_t *req)
{
    PASIS_DASSERT (handle->inprog_queue_hd != NULL);
    PASIS_DASSERT (handle->inprog_queue_tl != NULL);

    if (handle->inprog_queue_hd == req) {
        handle->inprog_queue_hd = req->next;
    } else {
        req->prev->next = req->next;
    }
    if (handle->inprog_queue_tl == req) {
        handle->inprog_queue_tl = req->prev;
    } else {
        req->next->prev = req->prev;
    }
}


#ifdef __cplusplus
}
#endif

#endif /* __RPC_COMMON_H__ */


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */

