#include <rpc/rpc.h>
#include <rpc/auth.h>
#include <rpc/auth_unix.h>
#include <rpc/pmap_clnt.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_debug.h"
#include "pasis_timer.h"

#if USE_KERBEROS
#include "krb_box.h"
#endif

#include "rpc_common.h"

#include <netdb.h>

#define HOSTNAME_LEN 128


p_byte                rpc_dummy_buffer[RPC_MAX_PAYLOAD];

static unsigned int   rpc_inited = 0;
static u_int32_t      xid;
static p_uint32       host_ip=0;
static char           hostname[HOSTNAME_LEN];
static unsigned int   hostname_len=0;


static inline void
rpc_req_init (void *arg)
{
    rpc_req_t *req = (rpc_req_t *)arg;

    memset (req, 0, sizeof (rpc_req_t));
    req->wait_cond = st_cond_new ();
    req->rpc_buffer = malloc (RPC_MAX_PAYLOAD);
}


static inline void
rpc_req_destroy (void *arg)
{
    rpc_req_t *req = (rpc_req_t *)arg;

    st_cond_destroy (req->wait_cond);
    free (req->rpc_buffer);
}


static pasis_error_t
rpc_set_host_ip (void)
{
    struct hostent *hp;
    int rc, off=0;
    
    rc = gethostname (hostname, 128);
    if (rc) {
        PASIS_PRINT_ERR ("Error: gethostname failed (%d)\n", errno);
        return PASIS_ERROR;
    }
    hostname_len = rc;

    hp = gethostbyname (hostname);
    if (hp == NULL || hp->h_length != 4) {
        PASIS_PRINT_ERR ("Error: gethostbyname failed (%d, len=%d)\n",
                         errno, hp->h_length);
        return PASIS_ERROR;
    }

    do {
        /* h_addr is in network byte order */
        host_ip = *(p_uint32 *)hp->h_addr_list[off];
        off++;
    } while (host_ip == htonl (INADDR_LOOPBACK) && hp->h_addr_list[off]);

#if defined(__APPLE__)
    /* Don't do the following check if I'm running on my Mac. */
#else
    if (host_ip == htonl (INADDR_LOOPBACK)) {
        PASIS_PRINT_ERR ("Error: loopback address found as hostname, check /etc/hosts.\n");
        return PASIS_ERROR;
    }
#endif
    
    return PASIS_SUCCESS;
}


p_uint32
rpc_get_host_ip (void)
{
    return host_ip;
}


int
rpc_init (void)
{
    int fd;

    rpc_inited++;
    if (rpc_inited != 1) {
        return 0;
    }

    if (rpc_set_host_ip ()) {
        rpc_inited--;
        return -1;
    }

    fd = open ("/dev/urandom", O_RDONLY);
    if (fd < 0) {
        xid = 12345;
    } else {
        read (fd, (char *)&xid, sizeof (u_int32_t));
        close (fd);
    }

    return 0;
}


void 
rpc_shutdown (void)
{
    rpc_inited--;
    if (rpc_inited != 0) {
        return;
    }
    
}


int
rpc_auth_client_init (void)
{
#if USE_KERBEROS
    krb_init_keycache_cli();
#endif
    return 0;
}


int
rpc_auth_server_init (void)
{
#if USE_KERBEROS
    krb_init_keycache_srv();
#endif
    return 0;
}


void
rpc_auth_client_shutdown (void)
{
#if USE_KERBEROS
    krb_shutdown_cli();
#endif
}


void
rpc_auth_server_shutdown (void)
{
#if USE_KERBEROS
    krb_shutdown_srv();
#endif
}


u_int32_t
rpc_req_xid_get (void)
{
    u_int32_t req_xid;

    req_xid = xid;
    xid++;

    return req_xid;
}


rpc_req_t *
rpc_req_get (rpc_handle_t *handle)
{
    rpc_req_t *req;

    while (1) {

        req = PasisFreelist_Get (handle->req_freelist);
        if (req) {
            break;
        }

        st_cond_wait (handle->req_freelist_cond);
    }

    PASIS_DASSERT (req != NULL);

    req->rpc_hdr = NULL;
    req->rpc_auth.uid = -1;
    req->rpc_auth.gid = -1;
    req->req_iov_len = 0;
    req->reply_iov_len = 0;
    req->req_iov_cur = 0;
    req->wait_flag = 0;
    req->wait_status = RPC_STATUS_SUCCESS;
    req->transmit_flag = 0;
    req->trans_next = NULL;
    req->trans_prev = NULL;
    req->next = NULL;
    req->prev = NULL;
    req->req_entry = NULL;
    req->data_buffer = NULL;

    return req;
}


void
rpc_req_put (rpc_handle_t *handle,
             rpc_req_t *req)
{
    if (handle->handle_type == RPC_HANDLE_SERVER && req->data_buffer) {
        PasisFreelist_Put (handle->buffer_freelist, req->data_buffer);
    }

    PasisFreelist_Put (handle->req_freelist, req);

    st_cond_signal (handle->req_freelist_cond);
}



void
rpc_early_terminate (rpc_handle_t *handle,
                     st_thread_t   thread_id)
{
    rpc_req_table_t *entry;
    int i;

    /* No request table, early termination not supported */
    if (handle->req_table == NULL) return;
    
    for (i = 0; i < handle->max_outstanding; i++) {
        entry = &handle->req_table[i];

        /* Find entry */
        if (entry->state != RPC_REQ_STATE_INVALID &&
            entry->thread_id == thread_id) {
            entry->terminate = 1;

            PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "%s: found thread, state=%u\n", 
                           __FUNCTION__, entry->state);

            /* If it is in the wait state, send termination */
            if (entry->state == RPC_REQ_STATE_WAIT) {
                entry->req->wait_flag = 0;
                st_cond_signal (entry->req->wait_cond);
            }
            return;
        }
    }
}


void
rpc_signal_shutdown (rpc_handle_t *handle)
{
    handle->req_poll_shutdown = 1;    

    if (handle->handle_type == RPC_HANDLE_SERVER) {
        pmap_unset (handle->program, handle->version);
    }

    st_thread_interrupt (handle->req_poll_thread);
    st_thread_join (handle->req_poll_thread, NULL);
}


static inline void
rpc_handle_destroy (rpc_handle_t *handle)
{
    struct timeval tv;

    if (handle == NULL) {
        return;
    }

    if (handle->inprog_queue_hd != NULL) {
        rpc_req_t *req = handle->inprog_queue_hd;

        PASIS_PRINT_INFO ("%s: handle inprog queue not empty\n", 
                          __FUNCTION__);

        while (req) {
            PASIS_PRINT_INFO ("XID=%u ", req->rpc_xid);
            req = req->next;
        }
        PASIS_PRINT_INFO ("\n");
        PASIS_DASSERT (0);
    }

    gettimeofday (&tv, NULL);
    timersub (&tv, &handle->create_time, &tv);

    if (handle->proto == IPPROTO_TCP) {
        st_mutex_destroy (handle->sock_lock);
    }

    /* May be NULL if connection is closed */
    if (handle->sock_fd != NULL) {
        st_netfd_close (handle->sock_fd);
    }

    if (handle->req_table != NULL) {
        free (handle->req_table);
    }

    PasisFreelist_Destroy (handle->req_freelist);

    PASIS_PRINT_INFO ("\nStatistics: RPC Handle (%s, %s)\n", 
                      (handle->handle_type == RPC_HANDLE_CLIENT) ? "Client" :
                      "Server", pasis_ip_to_hostname (handle->remote_ip));
    PASIS_PRINT_INFO ("---------------------------------------\n");
    PASIS_PRINT_INFO ("Time         :  %lu.%09lu\n", tv.tv_sec,
                      (long unsigned int)tv.tv_usec);
    PASIS_PRINT_INFO ("Bytes Written:  %"PRIu64"\n", handle->bytes_written);
    PASIS_PRINT_INFO ("Bytes Read   :  %"PRIu64"\n\n", handle->bytes_read);
    PASIS_PRINT_INFO ("Num requests :  %u\n\n", handle->num_requests);

    if (handle->timers) {
        Pasis_TimerShutdown (handle->timers);
    }
    
    free (handle);
}


void
rpc_client_destroy (rpc_handle_t *handle)
{
    if (handle->proto == IPPROTO_TCP) {
        st_cond_destroy (handle->reconnect_cond);
    }
    
    rpc_handle_destroy (handle);
}


void
rpc_server_destroy (rpc_handle_t *handle)
{
    PasisFreelist_Destroy (handle->buffer_freelist);
    rpc_handle_destroy (handle);
}




/* Common client create shared between tcp and udp creates */
rpc_handle_t *
rpc_client_create_common (p_uint32 server_ip,
                          unsigned int max_outstanding,
                          unsigned long program,
                          unsigned long version,
                          unsigned long proto)
{
    rpc_handle_t *client;
    unsigned short pport;

    client = (rpc_handle_t *) malloc (sizeof (rpc_handle_t));

    client->sock_addr.sin_family = AF_INET;
    memcpy(&(client->sock_addr.sin_addr.s_addr), &server_ip, 4);

    client->local_hostname = hostname;
    client->local_hostlen = hostname_len;
    client->local_ip = host_ip;

    pport = pmap_getport (&client->sock_addr, program, version, proto);
    if (pport == 0) {
        PASIS_PRINT_ERR ("Error: rpc_client_create pmap_getport failed\n");
        free (client);
        return NULL;
    }

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 5, "PMAP: got port %d for %lu, %lu\n", 
                   pport, program, version);

    client->sock_addr.sin_port = htons (pport);
    client->remote_ip = server_ip;
    client->callback = NULL;

    client->inprog_queue_hd = client->inprog_queue_tl = NULL;
    client->trans_queue_hd = client->trans_queue_tl = NULL;
    client->trans_inprog_count = 0;

    client->max_outstanding = max_outstanding;

    client->req_table = NULL;

    /* Allocate request buffers */
    client->req_freelist = PasisFreelist_CreateInitFixed (sizeof (rpc_req_t), 
                                                          max_outstanding,
                                                          rpc_req_init, 
                                                          rpc_req_destroy);
    client->req_freelist_cond = st_cond_new ();

    client->req_poll_shutdown = 0;

    gettimeofday (&client->create_time, NULL);
    client->bytes_read = 0;
    client->bytes_written = 0;
    client->num_requests = 0;
    client->timers = NULL;

    return client;
}


rpc_handle_t *
rpc_server_create_common (unsigned int max_outstanding,
			  unsigned short port,
			  unsigned long  program,
			  unsigned long  version,
			  unsigned long  proto,
			  void (*callback)(rpc_req_t *, p_byte *, unsigned))
                                    
{
    rpc_handle_t *server;

    server = (rpc_handle_t *) malloc (sizeof (rpc_handle_t));

    server->sock_addr.sin_family = AF_INET;
    server->sock_addr.sin_addr.s_addr = 0;
    server->sock_addr.sin_port = htons (port);

#if defined(__APPLE__)
    if (getuid() != 0) {
        fprintf(stderr, "%s: warning: pmap_set is slow/not function under "
                "Mac OS X under non-root.  Try running as root user.\n",
                __FUNCTION__);
    }
#endif
    if (pmap_set (program, version, proto, port) != 1) {
        PASIS_PRINT_ERR ("Error: rpc_server_create, pmap_set failed\n");
        free (server);
        return NULL;
    }
    
    server->handle_type = RPC_HANDLE_SERVER;

    server->callback = callback;

    server->inprog_queue_hd = server->inprog_queue_tl = NULL;
    server->trans_queue_hd = server->trans_queue_tl = NULL;
    server->trans_inprog_count = 0;

    server->local_ip = host_ip;
    server->remote_ip = 0;

    /* Allocate request buffers */
    /* Allocate one extra since recieve thread will only block in server
     * dispatch code, thus one extra buffer required for blocked request */

    server->req_freelist = PasisFreelist_CreateInitFixed (sizeof (rpc_req_t), 
                                                          max_outstanding+1,
                                                          rpc_req_init, 
                                                          rpc_req_destroy);
    server->req_freelist_cond = st_cond_new ();

    server->buffer_freelist = PasisFreelist_CreateFixed (RPC_MAX_PAYLOAD,
                                                         max_outstanding+1);

    server->max_outstanding = max_outstanding;

    server->program = program;
    server->version = version;
    server->proto = proto;

    /* Start completion handling thread */
    server->req_poll_shutdown = 0;

    gettimeofday (&server->create_time, NULL);
    server->bytes_read = 0;
    server->bytes_written = 0;
    server->num_requests = 0;
    server->timers = NULL;

    return server;
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
