/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"

#include "fe.h"

#include "nfs_prot.h"
#include "nfs3_prot.h"
#include "nfs3_fe.h"

#include "xdr_nfs3.h"

#include "pasis_debug.h"
#include "pasis_timer.h"
#include "pasis_error.h"

#include "fe_rpc.h"
#include "pasisfs_data.h"
#include "pasisfs_fid_lock.h"

#include "pasis_md_prot.h"
#include "pasis_md_dir.h"
#include "pasis_md_attr.h"

static struct timeval        pasisfs_startup_time;
static int                   rand_fd;

/* Root metadata nodes and protocol */
pasis_md_protocol_t root_md_protocol;
unsigned int        root_num_nodes;
pasis_md_linkage_t *root_nodes;

/* Default data nodes and protocol */
static pasis_md_protocol_t data_protocol;
static unsigned int        data_num_nodes;
static pasis_md_linkage_t *data_nodes;


#define NFS_MAX_DATA_SIZE    RPC_MAX_PAYLOAD

enum create_type {
    CREATE = 0,
    MKDIR = 1,
    MKNOD = 2,
    SYMLINK = 3,
};
typedef enum create_type create_type_t;

typedef struct create_data_s {
    create_type_t type;
    union {
        createmode3 create_mode;
        nfspath3 sym_path;
        mknoddata3 *mknod_data;
    } create_data_u;
} create_t;


#define FID_CAST(fh3) ((pasis_md_fid_t *)((fh3)->data.data_val))

#define FA3_TIME_SET(fa3_time, time)       \
do {                                       \
    (fa3_time)->seconds = time[0];         \
    (fa3_time)->nseconds = time[1] * 1000; \
} while (0)

#define INVALIDATE_WCC(wcc)                       \
do {                                              \
    (wcc)->before.present = 0;                    \
    (wcc)->after.present = 0;                     \
} while (0)

static inline char *
dump_fid (pasis_md_fid_t *f)
{
#if PASIS_DEBUG > 0
    static char buf[PASISIO_FILEID_SIZE*2 + 1];
    int i;
    char *bp = buf;
    for (i=0; i<PASISIO_FILEID_SIZE; i++) {
        snprintf(bp, 3, "%02X", (unsigned char)f->data[i]);
        bp +=2;
    }

    return buf;
#else 
    return NULL;
#endif
}


static inline void
SetCreateAttr (pasis_md_attr_t *cattr,
               sattr3          *sa3, 
               unsigned         uid,
               unsigned         gid,
               pasis_md_fid_t  *fid)
{
    memset (cattr, 0, sizeof(pasis_md_attr_t));

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, 
                   "%s: m:%o(%d) u:%d(%d) g:%d(%d) s:%"PRIu64"(%d)\n",
                   __FUNCTION__, sa3->mode.set_uint32_t_u.val,
                   sa3->mode.set, sa3->uid.set_uint32_t_u.val,
                   sa3->uid.set, sa3->gid.set_uint32_t_u.val,
                   sa3->gid.set, sa3->size.set_uint64_t_u.val, sa3->size.set);

    /* Check that uid and gid are set */
    if (sa3->uid.set != 1) {
        cattr->uid = uid;
    } else {
        cattr->uid = sa3->uid.set_uint32_t_u.val;
    }

    if (sa3->gid.set != 1) {
        cattr->gid = gid;
    } else {
        cattr->gid = sa3->gid.set_uint32_t_u.val;
    }

    if (sa3->size.set != 1) {
        cattr->size = 0;
    } else {
        cattr->size = sa3->size.set_uint64_t_u.val;
    }

    if (sa3->mode.set != 1) {
        cattr->mode = 0;
    } else {
        cattr->mode = sa3->mode.set_uint32_t_u.val;
    }

    cattr->fileid = *(p_uint32 *)(fid->data);
}


static inline void
SetFH3 (nfs_fh3 *dest, const pasis_md_fid_t *src)
{
    memcpy (dest->data.data_val, src->data, NFS_FHSIZE); 
    dest->data.data_len = NFS_FHSIZE;
}


static inline void
SetAttrOut (pasis_md_attr_t *attr,
            fattr3          *fa3)
{
    switch (attr->mode & MD_ATTR_MODE_MASK) {
    case MD_ATTR_MODE_DIR: 
        fa3->type = NF3DIR;
        break;
    case MD_ATTR_MODE_CHR:
        fa3->type = NF3CHR;
        break;
    case MD_ATTR_MODE_BLK:
        fa3->type = NF3BLK;
        break;
    case MD_ATTR_MODE_REG:
        fa3->type = NF3REG;
        break;
    case MD_ATTR_MODE_LNK:
        fa3->type = NF3LNK;
        break;
    case MD_ATTR_MODE_SOCK:
        fa3->type = NF3SOCK;
        break;
    case MD_ATTR_MODE_FIFO:
        fa3->type = NF3FIFO;
        break;
    }

    fa3->mode = attr->mode & ~0770000;
    fa3->nlink = attr->ref_cnt;
    fa3->uid = attr->uid;
    fa3->gid = attr->gid;
    fa3->size = attr->size;
    fa3->used = attr->size;
    fa3->rdev.minor = 0;    /* XXX Fix these to support devices */
    fa3->rdev.major = 0;
    fa3->fsid = 0;
    fa3->fileid = attr->fileid;

    FA3_TIME_SET (&fa3->atime, attr->atime);
    FA3_TIME_SET (&fa3->mtime, attr->mtime);
    FA3_TIME_SET (&fa3->ctime, attr->ctime);
}

static inline void
SetAttrIn (sattr3           *sa3,
           pasis_md_sattr_t *sattr)
{
    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, 
                   "%s: m:%o(%d) u:%d(%d) g:%d(%d) s:%"PRIu64"(%d)\n",
                   __FUNCTION__, sa3->mode.set_uint32_t_u.val,
                   sa3->mode.set, sa3->uid.set_uint32_t_u.val,
                   sa3->uid.set, sa3->gid.set_uint32_t_u.val,
                   sa3->gid.set, sa3->size.set_uint64_t_u.val, sa3->size.set);

    sattr->set_blocks = 0;

    sattr->set_mode = sa3->mode.set;
    sattr->mode = sa3->mode.set_uint32_t_u.val;

    sattr->set_uid = sa3->uid.set;
    sattr->uid = sa3->uid.set_uint32_t_u.val;

    sattr->set_gid = sa3->gid.set;
    sattr->gid = sa3->gid.set_uint32_t_u.val;

    sattr->set_size = sa3->size.set;
    sattr->size = sa3->size.set_uint64_t_u.val;

    sattr->set_atime = 1;
    if (sa3->atime.set == DONT_CHANGE) {
        sattr->set_atime = 0;
    } else if (sa3->atime.set == SET_TO_SERVER_TIME) {
        pasis_md_attr_set_time (sattr->atime);
    } else {
        sattr->atime[0] = sa3->atime.set_time_u.time.seconds;
        sattr->atime[1] = sa3->atime.set_time_u.time.nseconds / 1000;
    }

    sattr->set_mtime = 1;
    if (sa3->mtime.set == DONT_CHANGE) {
        sattr->set_mtime = 0;
    } else if (sa3->mtime.set == SET_TO_SERVER_TIME) {
        pasis_md_attr_set_time (sattr->mtime);
    } else {
        sattr->mtime[0] = sa3->mtime.set_time_u.time.seconds;
        sattr->mtime[1] = sa3->mtime.set_time_u.time.nseconds / 1000;
    }
    
}


static inline void
SetPostOpAttr (post_op_attr    *pa3,
               pasis_md_attr_t *attr)
{
    pa3->present = PASIS_TRUE;
    SetAttrOut (attr, &pa3->post_op_attr_u.attributes);
}


static inline void
SetPreOpAttrFA3 (pre_op_attr *pa3, fattr3 *fa3)
{
    pa3->present = PASIS_TRUE;
    pa3->pre_op_attr_u.attributes.size = fa3->size;
    memcpy (&pa3->pre_op_attr_u.attributes.mtime, &fa3->mtime, 
            sizeof (nfstime3));
    memcpy (&pa3->pre_op_attr_u.attributes.ctime, &fa3->ctime, 
            sizeof (nfstime3));
}


static inline void
SetCreateFID (pasis_md_fid_t *fid)
{
#if 1
    int rc;
    rc = read (rand_fd, fid->data, PASISIO_FILEID_SIZE);
    PASIS_DASSERT (rc == PASISIO_FILEID_SIZE);
#else
    static unsigned int count = 1;

    memset (fid, 0, PASISIO_FILEID_SIZE);
    *(unsigned int*)fid = count++;
#endif
}


int
PasisFS_NFS3FeInit (pasis_md_protocol_t *md_proto, 
                    unsigned int         md_num_linkages,
                    pasis_md_linkage_t  *md_linkages,
                    pasis_md_protocol_t *data_proto,
                    unsigned int         data_num_linkages,
                    pasis_md_linkage_t  *data_linkages)
{
    gettimeofday (&pasisfs_startup_time, NULL);

    rand_fd = open ("/dev/urandom", O_RDONLY);
    PASIS_DASSERT (rand_fd > 0);

    memmove(&root_md_protocol, md_proto, sizeof(pasis_md_protocol_t));
    root_nodes = md_linkages;
    root_num_nodes = md_num_linkages;

    memmove(&data_protocol, data_proto, sizeof(pasis_md_protocol_t));
    data_nodes = data_linkages;
    data_num_nodes = data_num_linkages;

    pasis_debug_print_enable = 0; /* PASIS_DEBUG_FS_NFS; */

    return 0;
}


void
PasisFS_NFS3FeShutdown (void)
{
    return;
}


nfsstat3
PasisFS_NFS3GetRoot (unsigned int uid, 
                     nfspath path,
                     nfs_fh3 *fh_out)
{
    char *c_strt, *c_end;
    nfs_fh3 fh1, fh2, *parent_fh, *out_fh, *tmp_fh;
    char    fhd1[NFS_FHSIZE], fhd2[NFS_FHSIZE];
    fattr3 fattr;
    post_op_attr p_attr;
    nfsstat3 status;

    fh1.data.data_val = fhd1;
    fh1.data.data_len = NFS_FHSIZE;

    fh2.data.data_val = fhd2;
    fh2.data.data_len = NFS_FHSIZE;

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                   st_thread_self());

    /* Most often just "/" will be mounted */
    if (path[0] == '\0') {
        SetFH3 (fh_out, &PASIS_MD_ROOT_FID_C);
        return NFS3_OK;
    } else {
        SetFH3 (&fh1,  &PASIS_MD_ROOT_FID_C);
    }

    parent_fh = &fh1;
    out_fh = &fh2;
    
    /* Go through the string looking up each component of the name */
    c_strt = path + 1;
    do { 
        c_end = strchr (c_strt, '/');
        if (c_end) {
            *c_end = '\0';
        }
        
        /* Lookup using the last found or root fh */
        status = PasisFS_NFS3Lookup (uid, parent_fh, c_strt, out_fh, 
                                     &p_attr, &p_attr, &p_attr);
        if (status != NFS3_OK) {
            /* Map error codes to good mount error codes */
            if (status == NFS3ERR_STALE) {
                status = NFS3ERR_INVAL;
            }
            return status;
        }

        /* Check that the fh is a directory */
        PasisFS_NFS3GetAttr (uid, out_fh, &fattr);
        if (fattr.type != NF3DIR) {
            return NFS3ERR_NOTDIR;
        }

        tmp_fh = parent_fh;
        parent_fh = out_fh;
        out_fh = tmp_fh;

        c_strt = c_end + 1;
    } while (c_end);
    
    /* Parent was just set to point to the result from PasisFS_NFS3Lookup */
    memcpy (fh_out->data.data_val, parent_fh->data.data_val, NFS_FHSIZE);
    fh_out->data.data_len = NFS_FHSIZE;            

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3GetAttr (unsigned int uid,
                     nfs_fh3 *fh,
                     fattr3 *fa_out)
{
    pasis_md_fid_t  *fid;
    pasis_md_attr_t  attr;
    pasis_error_t    error;
   
    fid = FID_CAST (fh);

    PASIS_DPRINTF (PASIS_DEBUG_FS_GETATTR, 3, "\n%s IN (%p)\n", __FUNCTION__, 
                   st_thread_self());
    PASIS_DPRINTF (PASIS_DEBUG_FS_GETATTR, 5, "%s\n", dump_fid (fid));

    /* Lock the FID */
    pasisfs_fid_lock (1, &fid, 1);

    error = PasisIO_MDGetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               fid, &attr, NULL);
    if (error) {
        pasisfs_fid_unlock (1, &fid);
        return error;
    }

    /*
    printf ("%s: current_time=%lu (ctime=%u, mtime=%u, atime=%u)\n",
            __FUNCTION__, time (NULL), fa_out->ctime.seconds, 
            fa_out->mtime.seconds, fa_out->atime.seconds);
    */

    SetAttrOut (&attr, fa_out);

    pasisfs_fid_unlock (1, &fid);

    PASIS_DPRINTF (PASIS_DEBUG_FS_GETATTR, 3, "\n%s OUT (%p)\n", __FUNCTION__, 
                   st_thread_self());

    return NFS3_OK;
}


nfsstat3 
PasisFS_NFS3SetAttr (unsigned int uid,
                     nfs_fh3 *fh, 
                     sattr3 *sa_in,
                     sattrguard3 *guard, 
                     wcc_data *wcc)
{
    pasis_md_fid_t  *fid;
    pasis_md_attr_t  attr;
    pasis_md_sattr_t sattr;
    p_uint32         guard_ctime[2] ={0,0};
    pasis_error_t    error;

    fid = FID_CAST (fh);

#if 0
    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                   st_thread_self());
    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 5, "%s\n", dump_fid (fid));
#endif

    INVALIDATE_WCC (wcc);

    pasisfs_fid_lock (1, &fid, 0);

    SetAttrIn (sa_in, &sattr);
    
    if (guard->check) {
        guard_ctime[0] = guard->sattrguard3_u.ctime.nseconds / 1000;
        guard_ctime[1] = guard->sattrguard3_u.ctime.seconds;
    }

    error = PasisIO_MDSetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               fid, &sattr, guard_ctime, &attr);
    if (error) {
        pasisfs_fid_unlock (1, &fid);

        printf ("%s: error=%d\n", __FUNCTION__, error);

        if (error == PASIS_ERROR_FILE_ISDIR) {
            return NFS3ERR_ISDIR;
        }
        return NFS3ERR_INVAL;
    }

    SetPostOpAttr(&wcc->after, &attr);  

    pasisfs_fid_unlock (1, &fid);

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3Lookup (unsigned int uid, 
                    nfs_fh3 *dir, 
                    filename3 file, 
                    nfs_fh3 *file_out, 
                    post_op_attr *obj_attr_out,
                    post_op_attr *dir_attr_out,
                    post_op_attr *fail_attr)
{
    pasis_md_fid_t  *parent_fid;
    pasis_md_fid_t  *file_fid;
    pasis_md_attr_t  obj_attr;
    pasis_md_attr_t  dir_attr;
    pasis_error_t    error;
    pasis_md_loc_t   distribution;

    parent_fid = FID_CAST (dir);
    file_fid   = FID_CAST (file_out);
    file_out->data.data_len = PASISIO_FILEID_SIZE;

    PASIS_DPRINTF (PASIS_DEBUG_FS_LOOKUP, 3, "\n%s IN (%p): %s\n",
                   __FUNCTION__, st_thread_self(), file);
    PASIS_DPRINTF (PASIS_DEBUG_FS_LOOKUP, 5, "Dir %s\n", dump_fid (parent_fid));

    if (strlen (file) > NFS_MAXNAMLEN) {
        fail_attr->present = 0;
        return NFS3ERR_NAMETOOLONG;
    }

    pasisfs_fid_lock (1, &parent_fid, 1);

    obj_attr_out->present = 0;
    dir_attr_out->present = 0;

    if (strcmp (file, ".") == 0) {
        SetFH3 (file_out, FID_CAST (dir));

        /* If get attr fails, oh well, we don't set post_op_attrs... */
        error = PasisIO_MDGetattr (&root_md_protocol, root_nodes, root_num_nodes,
                                    parent_fid, &dir_attr, NULL);
        if (error == PASIS_SUCCESS) {
            SetPostOpAttr (obj_attr_out, &dir_attr);
            SetPostOpAttr (dir_attr_out, &dir_attr);
        }

        pasisfs_fid_unlock (1, &parent_fid);
        
        PASIS_DPRINTF (PASIS_DEBUG_FS_LOOKUP, 3, "\n%s 1 OUT (%p)\n",
                       __FUNCTION__, st_thread_self());

        return NFS3_OK;
    }
    
    /* Do lookup */
    error = PasisIO_MDLookup (&root_md_protocol, root_nodes, root_num_nodes,
                              parent_fid, file, file_fid, &dir_attr,
                              &distribution);
    if (error) {
        pasisfs_fid_unlock (1, &parent_fid);

        PASIS_DPRINTF (PASIS_DEBUG_FS_LOOKUP, 3, "\n%s 2 OUT (%p)\n",
                       __FUNCTION__, st_thread_self());

        if (error == NFS3ERR_NOENT) {
            fail_attr->present = 1;
            SetPostOpAttr (fail_attr, &dir_attr);
            return NFS3ERR_NOENT;
        }
        
        fail_attr->present = 0;
        return error;
    }

    pasisfs_fid_unlock (1, &parent_fid);

    PASIS_DPRINTF (PASIS_DEBUG_FS_LOOKUP, 3, "%s: found file %s as FID=%s\n",
                   __FUNCTION__, file, dump_fid (file_fid));

    SetPostOpAttr (dir_attr_out, &dir_attr);

    pasisfs_fid_lock (1, &file_fid, 1);

    /* Get object's attrs */
    error = PasisIO_MDGetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               file_fid, &obj_attr, NULL);
    if (error) {
        fail_attr->present = 1;
        SetPostOpAttr (fail_attr, &dir_attr);
        pasisfs_fid_unlock (1, &file_fid);

        PASIS_DPRINTF (PASIS_DEBUG_FS_LOOKUP, 3, "\n%s 3 OUT (%p)\n",
                       __FUNCTION__, st_thread_self());

        return error;
    }

    SetPostOpAttr (obj_attr_out, &obj_attr);

    pasisfs_fid_unlock (1, &file_fid);

    PASIS_DPRINTF (PASIS_DEBUG_FS_LOOKUP, 3, "\n%s 4 OUT (%p)\n",
                   __FUNCTION__, st_thread_self());

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3Access (unsigned int uid,
                    nfs_fh3 *fh,
                    uint32_t access_in,
                    post_op_attr *obj_attr_out,
                    uint32_t *access_out)
{
    pasis_md_fid_t  *fid;
    pasis_md_attr_t  attr;
    pasis_error_t    error;

    fid = FID_CAST (fh);

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 3, "Access bits: %#x\n", access_in);

    pasisfs_fid_lock (1, &fid, 1);

    error = PasisIO_MDGetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               fid, &attr, NULL);
    if (error) {
        pasisfs_fid_unlock (1, &fid);
        
        /* Do something */
        obj_attr_out->present = 0;

        return error;
    }

    SetPostOpAttr (obj_attr_out, &attr);

    pasisfs_fid_unlock (1, &fid);

    *access_out = access_in;  /* XXX check this */

    return NFS3_OK;
}



nfsstat3
PasisFS_NFS3Readlink (unsigned int uid,
                      nfs_fh3 *fh, 
                      nfspath3 path_out,
                      post_op_attr *sym_attr_out)
{
    pasis_md_fid_t  *fid;
    pasis_md_attr_t  attr;
    pasis_error_t    error;

    fid = FID_CAST (fh);

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                   st_thread_self());

    pasisfs_fid_lock (1, &fid, 1);

    sym_attr_out->present = 0;

    error = PasisIO_MDReadlink (&root_md_protocol, root_nodes, root_num_nodes,
                                fid, path_out, &attr);
    if (error) {
        pasisfs_fid_unlock (1, &fid);
        return NFS3ERR_NXIO;
    }

    SetPostOpAttr (sym_attr_out, &attr);
    
    pasisfs_fid_unlock (1, &fid);

    return NFS3_OK;
}


static inline nfsstat3
CreateCommon (unsigned int uid,
              unsigned int gid,
              nfs_fh3 *dir,
              filename3 name,
              sattr3 *sa_in,
              diropres3ok *dirres,
              create_t *create)
{
    pasis_md_fid_t *dir_fid, *new_fid, *fid_array[2];
    pasis_md_attr_t cattr, new_attr, parent_attr;
    pasis_md_protocol_t *new_proto;
    pasisio_linkage_t  *new_linkages;
    pasis_error_t   error;
    nfspath3        sym = NULL;

    dir_fid = FID_CAST (dir);
    new_fid = FID_CAST (&dirres->obj.post_op_fh3_u.handle);
    dirres->obj.post_op_fh3_u.handle.data.data_len = PASISIO_FILEID_SIZE;

    INVALIDATE_WCC (&dirres->dir_wcc);
    dirres->obj_attributes.present = 0;

    SetCreateFID (new_fid);
    SetCreateAttr (&cattr, sa_in, uid, gid, new_fid);
    
    /* Set and Check Create Attrs */
    if ((cattr.mode & 0770000) == 0) {
        if (create->type == CREATE || create->type == SYMLINK) {
            cattr.mode |= NFSMODE_REG;
        } else if (create->type == MKDIR) {
            cattr.mode |= NFSMODE_DIR;
        } else {
            if (create->create_data_u.mknod_data->type == NF3CHR) {
                cattr.mode |= NFSMODE_CHR;
            } else if (create->create_data_u.mknod_data->type == NF3SOCK) {
                cattr.mode |= NFSMODE_SOCK;
            } else if (create->create_data_u.mknod_data->type == NF3FIFO) {
                cattr.mode |= NFSMODE_FIFO;
            } else if (create->create_data_u.mknod_data->type == NF3BLK) {
                cattr.mode |= NFSMODE_BLK;
            } else {
                return NFS3ERR_INVAL;
            }
        }
    } else {
        int mismatch=0;
        if (create->type == CREATE || create->type == SYMLINK) {
            mismatch = (cattr.mode | NFSMODE_REG);
        } else if (create->type == MKDIR) {
            mismatch = (cattr.mode | NFSMODE_DIR);
        } else {
            mismatch = ((cattr.mode | NFSMODE_SOCK) || 
                        (cattr.mode | NFSMODE_CHR) || 
                        (cattr.mode | NFSMODE_FIFO) ||
                        (cattr.mode | NFSMODE_BLK));
        }
        if (mismatch == 0) {
            return NFS3ERR_INVAL;
        }
    }

    if (create->type == SYMLINK) {
        sym = create->create_data_u.sym_path;
    }

    /* Set new protocol and linkages */
    if (cattr.mode & NFSMODE_DIR) {
        new_proto = &root_md_protocol;
        new_linkages = root_nodes;
    } else {
        new_proto = &data_protocol;
        new_linkages = data_nodes;
    }

    fid_array[0] = dir_fid;
    fid_array[1] = new_fid;
    pasisfs_fid_lock (2, fid_array, 0);

    error = PasisIO_MDCreate (&root_md_protocol, root_nodes, root_num_nodes,
                              dir_fid, new_fid, name, sym, &cattr, new_proto,
                              new_linkages, data_num_nodes, &parent_attr, &new_attr);
    if (error) {
        pasisfs_fid_unlock (2, fid_array);
        return error;
    }
    
    dirres->obj.present = 1;

    SetPostOpAttr (&dirres->obj_attributes, &new_attr);
    SetPostOpAttr (&dirres->dir_wcc.after, &parent_attr);

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 5, "created: %s\n", dump_fid (new_fid));
 
    pasisfs_fid_unlock (2, fid_array);

    return NFS3_OK;    
}


nfsstat3
PasisFS_NFS3Create (unsigned int uid,
                    unsigned int gid,
                    nfs_fh3 *dir,
                    filename3 name, 
                    createmode3 mode,
                    sattr3 *sa_in, 
                    diropres3ok *dirres)
{
    create_t create;

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p): %s mode=%d\n", 
                   __FUNCTION__, st_thread_self(), name, mode);
    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 5, "Dir %s\n", dump_fid (FID_CAST (dir)));

    /* XXX If mode == EXCLUSIVE, 
     * Exclusivity is not exactly guaranteed -- currently no atomicity between
     * directory lookup and insertion, best effort provided
     */

    create.type = CREATE;
    create.create_data_u.create_mode = mode;

    return CreateCommon (uid, gid, dir, name, sa_in, dirres, &create);
}



nfsstat3 
PasisFS_NFS3Mkdir (unsigned int uid,
                   unsigned int gid,
                   nfs_fh3 *dir,
                   filename3 name, 
                   sattr3 *sa_in,
                   diropres3ok *dirres)
{
    create_t create;

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                   st_thread_self());

    if (strcmp (name, ".") == 0 || strcmp (name, "..") == 0) {
        return NFS3ERR_EXIST;
    }

    create.type = MKDIR;
    
    return CreateCommon (uid, gid, dir, name, sa_in, dirres, &create);
}


nfsstat3
PasisFS_NFS3Symlink (unsigned int uid,
                     unsigned int gid,
                     nfs_fh3 *dir, 
                     filename3 name,
                     nfspath3 link, 
                     sattr3 *sa_in,
                     diropres3ok *dirres)
{
    create_t create;

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                   st_thread_self());

    create.type = SYMLINK;
    create.create_data_u.sym_path = link;

    return CreateCommon (uid, gid, dir, name, sa_in, dirres, &create);
}


nfsstat3 
PasisFS_NFS3Mknod (unsigned int uid,
                   unsigned int gid,
                   nfs_fh3 *dir,
                   filename3 name,
                   mknoddata3 *mknod_data,
                   diropres3ok *dirres)
{
    create_t create;

     PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                    st_thread_self());

     create.type = MKNOD;
     create.create_data_u.mknod_data = mknod_data;

     //return CreateCommon (uid, gid, dir, name, sa_in, dirres, &create);

     return NFS3ERR_NOTSUPP;
}


nfsstat3
PasisFS_NFS3Link (unsigned int uid,
                  nfs_fh3 *from_fh,
                  nfs_fh3 *to_dir,
                  filename3 to_name,
                  post_op_attr *post_file_attr,
                  wcc_data *linkdir_wcc)
{
    pasis_md_fid_t  *from_fid, *to_dir_fid, *fid_array[2];
    pasis_md_attr_t  from_attr, to_dir_attr;
    pasis_error_t    error;

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                   st_thread_self());

    INVALIDATE_WCC (linkdir_wcc);
    post_file_attr->present = 0;

    if (strcmp (to_name, ".") == 0 || strcmp (to_name, "..") == 0) {
        return NFSERR_INVAL;
    }

    from_fid = FID_CAST (from_fh);
    to_dir_fid = FID_CAST (to_dir);

    fid_array[0] = from_fid;
    fid_array[1] = to_dir_fid;
    pasisfs_fid_lock (2, fid_array, 0);
    
    error = PasisIO_MDLink (&root_md_protocol, root_nodes, root_num_nodes,
                            from_fid, to_dir_fid, to_name, &from_attr, 
                            &to_dir_attr);
    if (error) {
        pasisfs_fid_unlock (2, fid_array);
        return error;
    }

    SetPostOpAttr (post_file_attr, &from_attr);
    SetPostOpAttr (&linkdir_wcc->after, &to_dir_attr);

    pasisfs_fid_unlock (2, fid_array);

    return NFS3_OK;
}


nfsstat3 
PasisFS_NFS3Unlink (unsigned int uid,
                    nfs_fh3 *dir,
                    filename3 name,
                    wcc_data *wcc)
{
    pasis_md_fid_t *dir_fid;
    pasis_md_fid_t  from_fid;
    pasis_md_attr_t dir_attr;
    pasis_md_loc_t  distribution;
    pasis_error_t   error;

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                   st_thread_self());

    INVALIDATE_WCC (wcc);

    dir_fid = FID_CAST (dir);

    pasisfs_fid_lock (1, &dir_fid, 0);

    error = PasisIO_MDLookup (&root_md_protocol, root_nodes, root_num_nodes,
                              dir_fid, name, &from_fid, &dir_attr,
                              &distribution);
    if (error) {
        pasisfs_fid_unlock (1, &dir_fid);
        return error;
    }

    error = PasisIO_MDUnlink (&root_md_protocol, root_nodes, root_num_nodes,
                              dir_fid, &from_fid, name, &dir_attr);
    if (error) {
        pasisfs_fid_unlock (1, &dir_fid);
        return error;
    }

    SetPostOpAttr (&wcc->after, &dir_attr);

    pasisfs_fid_unlock (1, &dir_fid);

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3Rmdir (unsigned int uid, 
                   nfs_fh3 *dir,
                   filename3 name,
                   wcc_data *wcc)
{
    pasis_md_fid_t *dir_fid;
    pasis_md_fid_t  from_fid;
    pasis_md_attr_t dir_attr;
    pasis_md_loc_t  distribution;
    pasis_error_t   error;


    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                   st_thread_self());

    dir_fid = FID_CAST (dir);

    INVALIDATE_WCC (wcc);

    if (strcmp (name, "..") == 0) {
        return NFS3ERR_NOTEMPTY;
    }

    if (strcmp (name, ".") == 0) {
        return NFS3ERR_INVAL;
    }

    dir_fid = FID_CAST (dir);

    pasisfs_fid_lock (1, &dir_fid, 0);

    error = PasisIO_MDLookup (&root_md_protocol, root_nodes, root_num_nodes,
                              dir_fid, name, &from_fid, &dir_attr,
                              &distribution);
    if (error) {
        pasisfs_fid_unlock (1, &dir_fid);
        return error;
    }

    error = PasisIO_MDUnlink (&root_md_protocol, root_nodes, root_num_nodes,
                              dir_fid, &from_fid, name, &dir_attr);
    if (error) {
        pasisfs_fid_unlock (1, &dir_fid);
        return error;
    }

    SetPostOpAttr (&wcc->after, &dir_attr);

    pasisfs_fid_unlock (1, &dir_fid);

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3Rename (unsigned int uid,
                    nfs_fh3 *from_dir, 
                    filename3 from_name,
                    nfs_fh3 *to_dir, 
                    filename3 to_name,
                    wcc_data *fromdir_wcc, 
                    wcc_data *todir_wcc)
{
    pasis_md_fid_t *from_dir_fid, *to_dir_fid, *fid_array[4];
    pasis_md_fid_t  from_fid, to_fid;
    pasis_md_attr_t from_dir_attr, to_dir_attr;
    pasis_md_loc_t  distribution;
    pasis_error_t   error;
    int             to_name_valid;

    from_dir_fid = FID_CAST (from_dir);
    to_dir_fid   = FID_CAST (to_dir);

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, "\n%s (%p)\n", __FUNCTION__, 
                   st_thread_self());

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 5, "From Dir %s\n", 
                   dump_fid (from_dir_fid));
    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 5, "To Dir %s\n", dump_fid (to_dir_fid));
    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 5, "From name %s\n", from_name);
    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 5, "To name %s\n", to_name);

    INVALIDATE_WCC (fromdir_wcc);
    INVALIDATE_WCC (todir_wcc);

    /* Lock in order as to prevent potential deadlock */
    fid_array[0] = from_dir_fid;
    fid_array[1] = to_dir_fid;
    pasisfs_fid_lock (2, fid_array, 1);
    
    /*
    SetPreOpAttr (&fromdir_wcc->before, &fa);
    SetPreOpAttr (&todir_wcc->before, &fa);
    */

    error = PasisIO_MDLookup (&root_md_protocol, root_nodes, root_num_nodes,
                              from_dir_fid, from_name, &from_fid, 
                              &from_dir_attr, &distribution);
                              
    if (error) {
        SetPostOpAttr (&fromdir_wcc->after, &from_dir_attr);
        pasisfs_fid_unlock (2, fid_array);
        return error;
    }

    error = PasisIO_MDLookup (&root_md_protocol, root_nodes, root_num_nodes,
                              to_dir_fid, to_name, &to_fid, 
                              &to_dir_attr, &distribution);
    if (error == NFS3_OK) {
        to_name_valid = 1;
    } else {
        to_name_valid = 0;
    }

    /* release first locks so we can lock all FID's in order */
    pasisfs_fid_unlock (2, fid_array);

    fid_array[2] = &from_fid;
    fid_array[3] = &to_fid;

    pasisfs_fid_lock (to_name_valid ? 4 : 3, fid_array, 0);

    error = PasisIO_MDRename (&root_md_protocol, root_nodes, root_num_nodes,
                              from_dir_fid, to_dir_fid, &from_fid, &to_fid,
                              from_name, to_name, to_name_valid, 
                              &from_dir_attr, &to_dir_attr);

    pasisfs_fid_unlock (to_name_valid ? 4 : 3, fid_array);

    SetPostOpAttr (&fromdir_wcc->after, &from_dir_attr);
    SetPostOpAttr (&todir_wcc->after, &to_dir_attr);

    if (error) {
        return error;
    }

    return NFS3_OK;
}



typedef struct dirent {
    pasis_md_fid_t  *fid;
    char            *name;
    p_uint8          name_len;
    p_uint32         fileid;
    p_uint64         cookie;
} md_dirent_t;


static int
insert_entry3 (struct iovec *iov,
               md_dirent_t  *dent)
{
    u_int32_t *p, *new_p;
    entry3 new_entry;
    char name[NFS_MAXNAMLEN];
    int len;

    p = (u_int32_t *)(iov->iov_base + iov->iov_len);

    new_entry.cookie = dent->cookie;
    
    /* Setup new entry */
    /* Need to null terminate string name for xdr -- could be optimized */
    memcpy (name, dent->name, dent->name_len);
    name[dent->name_len] = '\0';

    PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 3,
                   "%s: name=%s, len=%u, fileid=%u, cookie=%"PRIu64"\n",
                   __FUNCTION__, name, dent->name_len, dent->fileid,
                   new_entry.cookie);

    new_entry.name = name;
    new_entry.fileid = dent->fileid;
    new_entry.nextentry = NULL;
    
    /* Encode the entry */
    new_p = xdr_encode_entry (p, &new_entry);

    /* Update the length */
    len = (new_p - p) << 2;
    iov->iov_len += len;
    
    return len;
}

static int
insert_entryplus3 (struct iovec *iov,
                   md_dirent_t *dent)
{
    u_int32_t *p, *new_p;
    entryplus3 new_entry;
    char name[NFS_MAXNAMLEN];
    char fh[NFS3_FHSIZE];
    int len;
    
    p = (u_int32_t *)(iov->iov_base + iov->iov_len);

    new_entry.cookie = dent->cookie;
    
    /* Setup new entry */
    /* Need to null terminate string name for xdr -- could be optimized */
    memcpy (name, dent->name, dent->name_len);
    name[dent->name_len] = '\0';

    PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 3,
                   "%s: name=%s, len=%u, fileid=%u, cookie=%"PRIu64"\n", 
                   __FUNCTION__, name, dent->name_len, dent->fileid,
                   new_entry.cookie);

    new_entry.name = name;
    new_entry.fileid = dent->fileid;
    new_entry.nextentry = NULL;
    new_entry.name_handle.present = 1;
    new_entry.name_handle.post_op_fh3_u.handle.data.data_val = fh;
    new_entry.name_handle.post_op_fh3_u.handle.data.data_len = NFS3_FHSIZE;
    SetFH3 (&new_entry.name_handle.post_op_fh3_u.handle, dent->fid);

    new_entry.name_attributes.present = 0;

    /* Encode the entry */
    new_p = xdr_encode_entryplus (p, &new_entry);

    /* Update the length */
    len = (new_p - p) << 2;
    iov->iov_len += len;
    
    return len;
}



static nfsstat3
readdir_common (unsigned int uid,
                nfs_fh3 *dir, 
                uint32_t dircount,
                uint32_t count,
                uint64_t cookie,
                cookieverf3 cookieverf_in, 
                post_op_attr *post_dir_attr, 
                cookieverf3 cookieverf_out,
                struct iovec *iov,
                unsigned int entry_size,
                int (*insert_entry_fp)(struct iovec *iov, md_dirent_t *dent))
{
    pasis_md_fid_t       *dir_fid;
    pasis_md_attr_t       attr;
    p_uint64              blkno;
    pasis_error_t         error;

    pasis_md_dir_hdr_t   *hdr;
    pasis_md_dir_entry_t *entry;

    md_dirent_t           dent;
    p_uint32              next_offset;

    p_uint32              last_mtime[2];

    p_byte               *blk;
    
    int eof, size = count, dir_size = dircount, done=0;
    u_int32_t *p, *new_p;


    dir_fid = FID_CAST (dir);

    PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 3, "\n%s (%p): cookie=%"PRIu64"\n",
                   __FUNCTION__, st_thread_self(), cookie);
    PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 5, "Dir %s\n", dump_fid (dir_fid));

    blk = malloc (PASIS_MD_DIR_BLK_SIZE);

    pasisfs_fid_lock (1, &dir_fid, 1);

    iov->iov_len = 0;

    /* Adjust count for xdr of post_op_attrs and cookieverf */
    size -= (4 + sizeof (post_op_attr) + NFS3_COOKIEVERFSIZE);

    post_dir_attr->present = 0;
    eof = PASIS_FALSE;

    blkno = cookie >> PASIS_MD_DIR_BLK_SHIFT;

    error = PasisIO_MDReaddir (&root_md_protocol, root_nodes, root_num_nodes,
                               dir_fid, blkno, blk, &attr);
    if (error) {
        free (blk);
        pasisfs_fid_unlock (1, &dir_fid);

        PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 5, "%s: 1 OUT, Dir %s\n",
                       __FUNCTION__, dump_fid (dir_fid));
        return NFS3ERR_BAD_COOKIE;
    }

    PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 3,
                   "readdir: attr.blocks = %"PRIu64"\n", attr.blocks);

    SetPostOpAttr (post_dir_attr, &attr);

    /* Check the cookieverf */
    if (cookie != 0) {
        memcpy (&last_mtime[0], &cookieverf_in[0], 4);
        memcpy (&last_mtime[1], &cookieverf_in[4], 4);

        if (last_mtime[0] != attr.mtime[0] ||
            last_mtime[1] != attr.mtime[1]) {
            free (blk);
            pasisfs_fid_unlock (1, &dir_fid);

            PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 5, "%s: 2 OUT, Dir %s\n",
                           __FUNCTION__, dump_fid (dir_fid));
            return NFS3ERR_BAD_COOKIE;
        }
    }

    /* Set up new cookieverf */
    memcpy (&cookieverf_out[0], &attr.mtime[0], 4);
    memcpy (&cookieverf_out[4], &attr.mtime[1], 4);

    next_offset = cookie % PASIS_MD_DIR_BLK_SIZE;

    /* Need to handle "." dir entry */
    if (cookie == 0) {
        char *dot = ".";

        if (size >= (2 + entry_size)) { 
            dent.fid = dir_fid;
            dent.fileid = attr.fileid;
            dent.name_len = 1;
            dent.name = dot;
            next_offset = PASIS_MD_DIR_HDR_SIZE;
            dent.cookie = next_offset;
            
            size -= insert_entry_fp (iov, &dent);
            dir_size -= (dent.name_len + 16);
        }
    }

    hdr = (pasis_md_dir_hdr_t *)blk;
    PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 3, " - free_offset=%d\n",
                   hdr->free_offset);

    /* Check if block is empty */
    if (hdr->free_offset == PASIS_MD_DIR_HDR_SIZE) {
        /* Insert an empty entry, so client gets cookie */
        md_dirent_t empty;
        memset (&empty, 0, sizeof(md_dirent_t));
        ++blkno;
        empty.cookie = ((blkno << PASIS_MD_DIR_BLK_SHIFT) +
                       PASIS_MD_DIR_HDR_SIZE);
        insert_entry_fp (iov, &empty);
        done = 1;
    }

    while (!done && size > 0 && dir_size  > 0) {
        entry = (pasis_md_dir_entry_t *)(blk + next_offset);
        PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 3,
                       " - entry: offset=%d, name_len=%d\n",
                       next_offset, entry->name_len);
        
        dent.fid = &entry->fid;
        dent.fileid = entry->fileid;
        dent.name_len = entry->name_len;
        dent.name = (char *)(blk + next_offset+ PASIS_MD_DIR_ENT_SIZE);
        
        next_offset += PASIS_MD_DIR_ENT_SIZE + dent.name_len;
        
        PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 3, " - next: offset=%d\n",
                       next_offset);

        if (next_offset >= hdr->free_offset) {
            next_offset = PASIS_MD_DIR_HDR_SIZE;
            done = 1;
            ++blkno;
        }
        
        dent.cookie = ((blkno << PASIS_MD_DIR_BLK_SHIFT) + next_offset);

        PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 5, 
                       "blkno=%"PRIu64", next_offset=%u\n",
                       blkno, next_offset);
        
        size -= insert_entry_fp (iov, &dent);
        dir_size -= dent.name_len + 16;
    }

    if (blkno >= attr.blocks) {
        eof = PASIS_TRUE;
        PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 3, "End of directory listing\n");
    } else {
        PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 3, "More dir. entries left\n");
    }

    pasisfs_fid_unlock (1, &dir_fid);
    free (blk);
    

    /* Encode end of readdir */
    p = (u_int32_t *)(iov->iov_base + iov->iov_len);

    new_p = xdr_encode_readdir_end (p, eof);  /* Insert end of list and EOF */

    iov->iov_len += (new_p - p) << 2;         /* Update length */

    PASIS_DPRINTF (PASIS_DEBUG_FS_READDIR, 5, "%s: 3 OUT, Dir %s, eof=%d\n",
                   __FUNCTION__, dump_fid (dir_fid), eof);

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3Readdir (unsigned int uid,
                     nfs_fh3 *dir, 
                     uint32_t count, 
                     uint64_t cookie,
                     cookieverf3 cookieverf_in, 
                     post_op_attr *post_dir_attr, 
                     cookieverf3 cookieverf_out,
                     struct iovec *iov)
{
    return readdir_common (uid, dir, count, count, cookie, cookieverf_in, 
                           post_dir_attr, cookieverf_out, iov, 
                           sizeof (entry3), insert_entry3);
}


nfsstat3
PasisFS_NFS3Readdirplus (unsigned int uid, 
                         nfs_fh3 *dir, 
                         uint32_t count,
                         uint32_t dircount, 
                         uint64_t cookie,
                         cookieverf3 cookieverf_in, 
                         post_op_attr *post_dir_attr, 
                         cookieverf3 cookieverf_out, 
                         struct iovec *iov)
{
    return readdir_common (uid, dir, dircount, count, cookie, 
                           cookieverf_in, post_dir_attr, cookieverf_out, 
                           iov, sizeof (entryplus3) - sizeof (fattr3), 
                           insert_entryplus3);
}


nfsstat3
PasisFS_NFS3FSstat (unsigned int uid,
                    nfs_fh3 *fh, 
                    fsstat3resok *stat)
{
    pasis_md_fid_t  *fid;
    pasis_md_attr_t  attr;
    pasis_error_t    error;

    fid = FID_CAST (fh);

    memset (stat, 0, sizeof (fsstat3resok));

    stat->tbytes = stat->fbytes = stat->abytes = 1000000000;
    stat->tfiles = stat->ffiles = stat->afiles = 10000;

    pasisfs_fid_lock (1, &fid, 1);

    error = PasisIO_MDGetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               fid, &attr, NULL);
    if (error) {
        pasisfs_fid_unlock (1, &fid);
        return error;
    }

    SetPostOpAttr (&stat->obj_attributes, &attr);

    pasisfs_fid_unlock (1, &fid);

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3FSinfo (unsigned int uid,
                    nfs_fh3 *fh,
                    fsinfo3resok *info)
{
    pasis_md_fid_t  *fid;
    pasis_md_attr_t  attr;
    pasis_error_t    error;

    fid = FID_CAST (fh);

    info->wtmax = info->rtmax = NFS_MAX_DATA_SIZE;
    info->wtpref = info->rtpref = NFS_MAX_DATA_SIZE;
    /* Linux uses the max (wtmult, rtmult) as the blocksize */
    info->wtmult = info->rtmult = PASIS_NFS_DATA_BLKSIZE; 
    info->dtpref = NFS_MAX_DATA_SIZE;
    info->maxfilesize = 2147483648U;
    info->time_delta.seconds = 0;
    info->time_delta.nseconds = 1000;
    info->properties = FSF3_LINK | FSF3_SYMLINK | FSF3_HOMOGENEOUS | 
                       FSF3_CANSETTIME;

    pasisfs_fid_lock (1, &fid, 1);
    
    error = PasisIO_MDGetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               fid, &attr, NULL);
    if (error) {
        info->obj_attributes.present = 0;
        pasisfs_fid_unlock (1, &fid);
        return error;
    }

    SetPostOpAttr (&info->obj_attributes, &attr);

    pasisfs_fid_unlock (1, &fid);

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3Pathconf (unsigned int uid, 
                      nfs_fh3 *fh, 
                      pathconf3resok *pathconf)
{
    pasis_md_fid_t  *fid;
    pasis_md_attr_t  attr;
    pasis_error_t    error;

    fid = FID_CAST (fh);

    pathconf->linkmax = 65535;
    pathconf->name_max = NFS_MAXNAMLEN;
    pathconf->no_trunc = PASIS_FALSE;
    pathconf->chown_restricted = PASIS_FALSE;
    pathconf->case_insensitive = PASIS_FALSE;
    pathconf->case_preserving = PASIS_TRUE;

    pasisfs_fid_lock (1, &fid, 1);
    
    error = PasisIO_MDGetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               fid, &attr, NULL);
    if (error) {
        pasisfs_fid_unlock (1, &fid);
        pathconf->obj_attributes.present = 0;
        return error;
    }

    SetPostOpAttr (&pathconf->obj_attributes, &attr);

    pasisfs_fid_unlock (1, &fid);

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3Read (unsigned int uid,
                  nfs_fh3 *fh,
                  uint64_t offset,
                  uint32_t len,
                  char *buffer,
                  uint32_t *len_out, 
                  bool_t *eof,
                  post_op_attr *file_attr_out)
{
    pasis_md_fid_t   *fid;
    pasis_md_attr_t   attr;
    pasis_md_loc_t    loc;
    pasis_error_t     error;
    int               lock;

    *len_out = 0;
    *eof = 0;

    //printf ("\n%s: offset=%"PRIu64", length=%u\n", __FUNCTION__, offset, len);

    fid = FID_CAST (fh);

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, 
                   "\n%s (%p): offset: %"PRIu64", length: %u\n",
                   __FUNCTION__, st_thread_self(), offset, len);

    file_attr_out->present = 0;

    pasisfs_fid_lock (1, &fid, 1);

    error = PasisIO_MDGetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               fid, &attr, &loc);
    if (error != PASIS_SUCCESS) {
        pasisfs_fid_unlock (1, &fid);
        return error;
    }

    //printf ("\n%s: got file attrs\n", __FUNCTION__);

    error = PasisFS_FileRead (fid, &loc, &attr, len, offset, (p_byte*)buffer, 
                              len_out);
    /*
    printf ("\n%s: performed read (error=%d), len_out=%d/%"PRIu64"\n", 
            __FUNCTION__, error, *len_out, attr.size);
    */

    if (error) {
        pasisfs_fid_unlock (1, &fid);
        return NFS3ERR_INVAL;
    }

    PASIS_DASSERT (*len_out <= PASIS_NFS_DATA_BLKSIZE);

    SetPostOpAttr (file_attr_out, &attr);

    /* Check for EOF */
    if (attr.size == (*len_out + offset)) {
        *eof = PASIS_TRUE;
    } else {
        *eof = PASIS_FALSE;
    }

    pasisfs_fid_unlock (1, &fid);

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3Write (unsigned int uid,
                   nfs_fh3 *fh,
                   uint64_t offset, 
                   uint32_t len,
                   char *buffer,
                   stable_how stable,  
                   uint32_t *len_out,
                   wcc_data *wcc,
                   stable_how *committed,
                   writeverf3 verf_cookie)
{
    pasis_md_fid_t   *fid;
    pasis_md_attr_t   attr;
    pasis_md_loc_t    loc;
    pasis_error_t     error;
    pasis_md_sattr_t  sattr;
    p_uint32          guard_ctime[2]={0,0};

    //printf ("\n%s: offset=%"PRIu64", length=%u\n", __FUNCTION__, offset, len);

    fid = FID_CAST (fh);
    
    //    printf("%s: IN\n", __FUNCTION__);

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, 
                   "\n%s (%p): offset=%"PRIu64", length=%u\n",
                   __FUNCTION__, st_thread_self(), offset, len);

    /* Set write verf cookie */
    memcpy (&verf_cookie[0], &pasisfs_startup_time.tv_sec, 4);
    memcpy (&verf_cookie[4], &pasisfs_startup_time.tv_usec, 4);    

    INVALIDATE_WCC (wcc);

    pasisfs_fid_lock (1, &fid, 0);

    error = PasisIO_MDGetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               fid, &attr, &loc);

    if (error == PASIS_SUCCESS) {
        //SetPreOpAttr (&wcc->before, &attr);
    }

    error = PasisFS_FileWrite (fid, &loc, &attr, len, offset, (p_byte*)buffer, 
                               len_out);
    if (error) {
        PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 5, "Write: error=%d\n", error);
        pasisfs_fid_unlock (1, &fid);

        //        printf("%s: 111 OUT\n", __FUNCTION__);

        return NFS3ERR_ISDIR;
    }

    /* Update attributes */
    sattr.set_mode = sattr.set_uid = sattr.set_gid = sattr.set_atime = 0;
    sattr.set_blocks = 1;
    sattr.blocks = attr.blocks;
    sattr.set_size = 1;
    sattr.size = attr.size;
    sattr.set_mtime = 1;
    sattr.mtime[0] = attr.mtime[0];
    sattr.mtime[1] = attr.mtime[1];

    error = PasisIO_MDSetattr (&root_md_protocol, root_nodes, root_num_nodes,
                               fid, &sattr, guard_ctime, &attr);
    if (error) {
        PASIS_PRINT_ERR ("Error: %s setting attributes\n", __FUNCTION__);
        pasisfs_fid_unlock (1, &fid);

        return error;
    }

    SetPostOpAttr (&wcc->after, &attr);

    pasisfs_fid_unlock (1, &fid);

#if 0
    error = PasisFS_FileGetAttr (inode, &fa);
    if (error == PASIS_SUCCESS) {
        SetPostOpAttrFA2 (&wcc->after, &fa);
    }

    if (stable == UNSTABLE) {
        PasisFS_FileUnlockAll (inode);
        *committed = UNSTABLE;
    } else {
        PasisFS_FileSyncUnlockAll (inode);
        *committed = FILE_SYNC;
    }
#endif

    PASIS_DPRINTF (PASIS_DEBUG_FS_FE, 3, 
                   "Wrote: %u/%u at offset=%"PRIu64"x` (committed=%#x)\n",
                   *len_out, len, offset, *committed);

    //    printf("%s: OUT successfully \n", __FUNCTION__);

    return NFS3_OK;
}


nfsstat3
PasisFS_NFS3Commit (unsigned int uid, 
                    nfs_fh3 *fh, 
                    uint64_t offset, 
                    uint32_t count, 
                    wcc_data *file_wcc, 
                    writeverf3 writeverf_out)
{
    /* Set write verf cookie */
    memcpy (&writeverf_out[0], &pasisfs_startup_time.tv_sec, 4);
    memcpy (&writeverf_out[4], &pasisfs_startup_time.tv_usec, 4);

    INVALIDATE_WCC (file_wcc);

    return NFS3_OK;  /* Should change if data_cache is not write through */
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
