/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_error.h"
#include "pasis_lists.h"

#include <rpc/xdr.h>
#include <rpc/pmap_clnt.h>

/* XXX Not supported yet */
//#include "nfs_prot.h"
#include "mount1_prot.h"

#include "nfs3_prot.h"
#include "nfs3_svc.h"
#include "mount3_prot.h"

#include "nfs_xact_cache.h"
#include "fe_rpc.h"

#ifndef NFS_PORT
#define NFS_PORT 2049
#endif

#define RPC_ZERO_OUT_LEN(rpc_state) ((rpc_state)->out_iov[1].iov_len = 0)

#define STATS_INCRV(var, val) (statistics.var += val)
#define STATS_UPDATE(var, val) (statistics.var = val)
#define STATS_INCR(var) (statistics.var++)
#define STATS_MAX(var, val) ((val > statistics.var) ? (statistics.var=val) : 0)

/* Internal types */
typedef void (*dispatch_func_t)(rpc_state_t *);

typedef struct worker_stats_s {
    unsigned int  total_rpcs;
    unsigned int  total_outstanding;
    unsigned int  avg_outstanding;
    unsigned int  max_outstanding;
    unsigned int  rpc_retransmits;
    unsigned int  dispatch_blocked;
} worker_stats_t;

typedef struct rpc_dispatch_s {
    unsigned long prognum;
    unsigned long version;
    unsigned long proto;
    dispatch_func_t dispatch_func;
    PASIS_LL_PTRS (list, struct rpc_dispatch_s);
} rpc_dispatch_t;

typedef struct worker_thread_s {
    st_thread_t thread_id;
    st_cond_t   cond;
    int         work;

    rpc_state_t rpc_state;

    PASIS_Q_PTRS(struct worker_thread_s) queue;
} worker_thread_t;

typedef PASIS_Q_STRUCT(worker_q_s, worker_thread_t) 
     worker_queue_t;

/* File descriptor specific data */
typedef struct netfd_data_s {
    int        in_poll_set;
    int        op_inprog;
    st_mutex_t write_lock;
} netfd_data_t;


/* Internal globals */
static worker_thread_t *worker_thread_array;
static unsigned int worker_thread_num;
static worker_queue_t worker_queue;
static st_cond_t dispatch_cond;
static int worker_shutdown = 0;
static int workers_working = 0;
static unsigned in_progress = 0;
static unsigned dispatch_waiting = 0;
static worker_stats_t statistics;

/* RPC event variables */
static struct pollfd  *poll_fds;
static unsigned int    poll_num=0;
static int             rpc_udp_fd;
static int             rpc_tcp_fd;
static int             rpc_event_thread_shutdown=0;
static st_thread_t     rpc_event_thread;


PASIS_LL_HEAD (static rpc_dispatch_t, rpc_dispatch_hd);

/* Dispatch function prototypes */
void nfs_program_2(rpc_state_t *rpc_state);
void mount_program_1(rpc_state_t *rpc_state);
void nfs_program_3(rpc_state_t *rpc_state);
void mount_program_3(rpc_state_t *rpc_state);


static inline int
poll_fd_handle_error (st_netfd_t fd)
{
    netfd_data_t *data = st_netfd_getspecific (fd);

    /* Check the inprog count of the rpc fd and make sure that this fd is
     * not present in the poll_fds array -- then close fd.
     * Otherwise let event loop handle closing of socket.
     */
    if (data->op_inprog == 1 && data->in_poll_set == 0) {
        st_netfd_close (fd);
        return 0;
    }

    return 1;
}



static void *
WorkerThread(void *arg)
{
    worker_thread_t *worker = (worker_thread_t *)arg;
    rpc_dispatch_t *dispatch;
    rpc_act_entry_t this_req;
    dispatch_func_t dispatch_func;
    netfd_data_t *fd_data;
    PASIS_DPRINTF(PASIS_DEBUG_FS_NFS, 4, "%s -- WORKER START\n", __FUNCTION__);

    workers_working++;

    while (1) {

        /* Wait for work or shutdown */
        while (worker->work == 0 && worker_shutdown == 0) {
            st_cond_wait (worker->cond);
        }
        PASIS_DPRINTF(PASIS_DEBUG_FS_NFS, 4, "%s: got work\n", __FUNCTION__);
        PASIS_DPRINTF(PASIS_DEBUG_FS_NFS, 4, "%s: worker release (%d)\n",
                      __FUNCTION__, in_progress);

        /* Check for shutdown */
        if (worker_shutdown) {
            break;
        }
        
        /* Get file descriptor specific data structure */
        fd_data = st_netfd_getspecific (worker->rpc_state.rpc_fd);

        /* local copies of the request structures */
        this_req.trans_id = worker->rpc_state.rpc_xid;
        this_req.prog_num = worker->rpc_state.rpc_prog;
        this_req.version = worker->rpc_state.rpc_vers;

        PASIS_DPRINTF(PASIS_DEBUG_FS_NFS, 4, 
                      "%s: prog=%lu, ver=%lu, proc=%u id=%lu\n",
                   __FUNCTION__, this_req.prog_num, this_req.version,
                   worker->rpc_state.rpc_proc, this_req.trans_id);

        /* Find dispatch function -- this should be a short list... */
        dispatch = PASIS_LL_GET_HEAD (rpc_dispatch_hd);
        dispatch_func = NULL;

        while (dispatch != NULL) {
            if (dispatch->prognum == worker->rpc_state.rpc_prog && 
                dispatch->version == worker->rpc_state.rpc_vers) {
                dispatch_func = dispatch->dispatch_func;
                break;
            }
            dispatch = PASIS_LL_NEXT (list, dispatch);
        }
        
        if (dispatch == NULL) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 4, "%s: null dispatch\n",
                           __FUNCTION__);
            goto work_done;
        }
        
        in_progress++;
        PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 4, "In Progress=%d\n", 
                       in_progress);

        PASIS_DASSERT (dispatch_func);

        if (NFSXactCacheLookup (&this_req) == 0) {
            rpc_state_t *rpc_state = &worker->rpc_state;
            pasis_error_t error;

            /* some accounting */
            STATS_INCR (total_rpcs);
            STATS_MAX (max_outstanding, in_progress);
            STATS_INCRV (total_outstanding, in_progress);

            /* add entry to xact cache */
            NFSXactCacheInsert (&this_req);

            /* setup output argument buffers */
            rpc_state->out_iov[0].iov_base = rpc_state->out_buffer;
            rpc_state->out_iov[0].iov_len = RPC_MAX_HDR;

            rpc_state->out_iov[1].iov_base = rpc_state->out_buffer + 
                RPC_MAX_HDR;
            rpc_state->out_iov[1].iov_len = RPC_MAX_PAYLOAD;           
            rpc_state->out_iov[2].iov_len = 0;

            /* issue service request */
            if( rpc_state->reply_status == RPC_MSG_ACCEPTED && 
                rpc_state->reply_accepted_status == RPC_ACCEPT_SUCCESS ) {
                dispatch_func (rpc_state);
            } else {
                RPC_ZERO_OUT_LEN(rpc_state);
            }

            /* encode reply */
            rpc_state->reply_status = RPC_MSG_ACCEPTED;
            rpc_state->reply_accepted_status = RPC_ACCEPT_SUCCESS;
            
            rpc_encode_reply (rpc_state);
                              
            /* issue reply */
            if (rpc_state->rpc_proto == IPPROTO_UDP) {
                error = rpc_send_udp_reply (rpc_state, fd_data->write_lock);
            } else {
                error = rpc_send_tcp_reply (rpc_state, fd_data->write_lock);
            }
            
            /* Handle any errors that may have occurred during transmission */
            if (error) {
                PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 3, 
                               "%s: Error in reply transmission\n", 
                               __FUNCTION__);
                if (poll_fd_handle_error (rpc_state->rpc_fd) == 0) {
                    /* FD was successfully closed */
                    fd_data = NULL;
                }
            }

            /* delete entry */
            NFSXactCacheRemove (&this_req);

        } else {
            PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 4,
                           "%s: detected retransmit (%lu)... drop.\n",
                           __FUNCTION__, this_req.trans_id);
            STATS_INCR (rpc_retransmits);
        }

        in_progress--;

    work_done:

        /* Decrement use count on the file descriptor, if not NULL */
        if (fd_data) fd_data->op_inprog--;

        PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 4, "Worker done: In Progress=%d\n", 
                       in_progress);

        worker->work = 0;

        PASIS_Q_ADD_TAIL (worker_queue, queue, worker);

        /* Check if the dispatch is blocked waiting for a worker */
        if (dispatch_waiting) {
            dispatch_waiting = 0;
            st_cond_signal (dispatch_cond);
        }

    }

    /* Shutdown */
    workers_working--;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 4, "%s -- WORKER EXIT\n", __FUNCTION__);

    return NULL;
}


/* Dispatch function called in the context of RPCEventThread */
static void
Dispatch(st_netfd_t netfd, unsigned int ipproto)
{
    pasis_error_t    error;
    worker_thread_t *worker;

    worker = PASIS_Q_FIRST (worker_queue);
    while (worker == NULL) {
        PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 4, "%s: no free worker threads\n",
                       __FUNCTION__);

        STATS_INCR (dispatch_blocked);
        dispatch_waiting = 1;

        while (dispatch_waiting) {
            st_cond_wait (dispatch_cond);
        }

        worker = PASIS_Q_FIRST (worker_queue);
    }

    PASIS_Q_REMOVE (worker_queue, queue, worker);

    worker->work = 1;
    worker->rpc_state.rpc_proto = ipproto;
    worker->rpc_state.rpc_fd = netfd;

    /* Read in RPC request off of socket -- puts all data+hdr into in_buffer */
    if (ipproto == IPPROTO_UDP) {
        /* UDP request */
        PASIS_DPRINTF(PASIS_DEBUG_FS_NFS, 4, "%s: reading UDP request\n", __FUNCTION__);
        error = rpc_read_udp_request (&worker->rpc_state);
    } else {
        /* TCP request */
        printf ("%s: reading TCP request\n", __FUNCTION__);
        error = rpc_read_tcp_request (&worker->rpc_state);
    }

    PASIS_DPRINTF(PASIS_DEBUG_FS_NFS, 4, "%s: read error=%d\n", __FUNCTION__, error);

    /* Handle read errors */
    if (error) {
        if (poll_fd_handle_error (worker->rpc_state.rpc_fd) != 0) {
            /* FD not closed, decr inprog count */
            netfd_data_t *data;
            data = st_netfd_getspecific (worker->rpc_state.rpc_fd);
            data->op_inprog--;
        }
        /* Add worker back to queue and return */
        worker->work = 0;
        PASIS_Q_ADD_TAIL (worker_queue, queue, worker);
        return;
    }

    /* Decode rpc hdr, fills in rpc_state rpc vars, sets in_data ptr and len */
    error = rpc_decode_call (&worker->rpc_state);
    PASIS_DASSERT (error == PASIS_SUCCESS);

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 4, "\n%s: signal worker\n", 
                   __FUNCTION__);
    
    st_cond_signal (worker->cond);
}

static void
poll_destroy_fdspecific (void *mem)
{
    netfd_data_t *data = (netfd_data_t *)mem;
    st_mutex_destroy (data->write_lock);
    free (data);
}

static inline void
poll_init_fd (st_netfd_t fd)
{
    netfd_data_t *data;

    data = malloc (sizeof (netfd_data_t));
    data->in_poll_set = 1;
    data->op_inprog = 0;
    data->write_lock = st_mutex_new ();

    st_netfd_setspecific (fd, (void *)data, poll_destroy_fdspecific);
}

/*
 * This is the function watching for the events on the RPC fd.
 */
static void *
RPCEventThread (void *arg)
{
    st_netfd_t  *net_fds;
    int          i;

    net_fds = malloc (RPC_MAX_CONNECTIONS * sizeof (st_netfd_t));
    poll_fds = malloc (RPC_MAX_CONNECTIONS * sizeof (struct pollfd));

    /* Setup UDP and TCP sockets in poll array */
    poll_fds[0].fd = rpc_tcp_fd;
    poll_fds[0].events = POLLIN;

    poll_fds[1].fd = rpc_udp_fd;
    poll_fds[1].events = POLLIN;

    /* Setup UDP and TCP sockets as st_netfd_ts */
    net_fds[0] = st_netfd_open_socket (rpc_tcp_fd);
    net_fds[1] = st_netfd_open_socket (rpc_udp_fd);

    poll_init_fd (net_fds[1]);

    poll_num = 2;

    /* Main loop of event thread */
    while (!rpc_event_thread_shutdown) {
        int rc;
        
        rc = st_poll (poll_fds, poll_num, -1);
        PASIS_DPRINTF(PASIS_DEBUG_FS_NFS, 4, "%s: rc=%d\n", __FUNCTION__, rc);
        if (rc == -1) {
            if (errno == EINTR) { /* Thread interrupted for shutdown */
                continue;
            }
            PASIS_PRINT_ERR ("Error: nfs_worker st_poll (%s)\n", 
                             PASIS_STRERROR (errno));
        }
        
        /* Iterate through list of active sockets, ignoring tcp listening fd */
        for (i = 0; i < poll_num && rc > 0; i++) {

            /* Have a dispatchable event */
            if (poll_fds[i].revents & POLLIN) {
                PASIS_DPRINTF(PASIS_DEBUG_FS_NFS, 4,"%s: Found event on: %d\n",
                              __FUNCTION__, poll_fds[i].fd);

                if (poll_fds[i].fd == rpc_tcp_fd) {
                    /* Accept new connection */
                    net_fds[poll_num] = st_accept (net_fds[i], NULL, NULL,
                                                   RPC_ACCEPT_TIMEOUT);
                    /* Init new connection */
                    if (net_fds[poll_num] != NULL) {
                        poll_fds[poll_num].fd = 
                            st_netfd_fileno (net_fds[poll_num]);
                        poll_fds[poll_num].events = POLLIN;
                        poll_init_fd (net_fds[poll_num]);
                        poll_num++;
                    }
                } else {
                    /* Set op in prog flag */
                    netfd_data_t *data = st_netfd_getspecific (net_fds[i]);
                    data->op_inprog++;

                    /* Dispatch request */
                    Dispatch (net_fds[i], (poll_fds[i].fd == rpc_udp_fd) ? 
                              IPPROTO_UDP : IPPROTO_TCP);
                }

                rc--;

            } else if (poll_fds[i].revents & (POLLERR | POLLHUP | POLLNVAL)) {
                netfd_data_t *data = st_netfd_getspecific (net_fds[i]);

                if (poll_fds[i].fd == rpc_tcp_fd || 
                    poll_fds[i].fd == rpc_udp_fd) {
                    PASIS_PRINT_ERR ("%s: warning closing %s socket\n",
                                     __FUNCTION__, 
                                     (poll_fds[i].fd == rpc_tcp_fd) ? 
                                     "tcp listening" : "udp");
                }

                /* Remove fd from poll set and net fd set */
                if (i != poll_num - 1) {
                    memmove (&poll_fds[i], &poll_fds[i+1], (poll_num - i + 1) *
                             sizeof (struct pollfd));
                    memmove (&net_fds[i], &net_fds[i+1], (poll_num - i + 1) *
                             sizeof (st_netfd_t));
                    i--;
                }
                poll_num--;

                /* If there is currently an operation in progress that is
                 * going to be using this socket, don't close it here,
                 * let using thread detect error and close it once inprog
                 * count has dropped to 0.
                 */
                if (data->op_inprog) {
                    /* Error close connection */
                    st_netfd_close (net_fds[i]);                    
                } else {
                    /* Mark descriptor as removed from poll set */
                    data->in_poll_set = 0;
                }

                rc--;
            } /* else if */
        }  /* for */
    } /* while */

    for (i = 0; i < poll_num; i++) {
        PASIS_DASSERT (net_fds[i] != NULL);
        st_netfd_close (net_fds[i]);
    }

    free (poll_fds);
    free (net_fds);

    return NULL;
}


static pasis_error_t
WorkerStartup(unsigned int num_workers)
{
    int i;

    worker_thread_num = num_workers;
    worker_thread_array = malloc (num_workers * sizeof (worker_thread_t));
                                  

    /* Initialize dispatch waiting condition variable */
    if ((dispatch_cond = st_cond_new ()) == NULL) {
        PASIS_PRINT_ERR("Unable to init condition variable\n");
        PASIS_DASSERT (0);
    }

    /* Initialize and startup all worker threads */
    PASIS_Q_INIT (worker_queue);

    memset (&statistics, 0, sizeof (worker_stats_t));

    for (i = 0; i < worker_thread_num; i++) {
        worker_thread_t *worker;

        worker = &worker_thread_array[i];
        
        if ((worker->cond = st_cond_new ()) == NULL) {
            PASIS_PRINT_ERR ("Unable to init condition variable\n");
            PASIS_DASSERT (0);
        }

        worker->work = 0;

        worker->rpc_state.in_buffer = malloc (RPC_MAX_PACKET);
        worker->rpc_state.out_buffer = malloc (RPC_MAX_PACKET);

        worker->thread_id = st_thread_create(WorkerThread, worker, 1, 0);
        PASIS_DASSERT (worker->thread_id != NULL);

        PASIS_Q_ADD_TAIL (worker_queue, queue, worker);
    }

    return PASIS_SUCCESS;
}


static void
WorkerShutdown(void)
{
    int i;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s\n", __FUNCTION__);

    worker_shutdown = 1;

    /* Shutdown and cleanup all worker threads */
    for (i = 0; i < worker_thread_num; i++) {
        worker_thread_t *worker;
        worker = &worker_thread_array[i];

        st_cond_signal (worker->cond);
        st_thread_join (worker->thread_id, NULL);
        
        free (worker->rpc_state.in_buffer);
        free (worker->rpc_state.out_buffer);

        st_cond_destroy (worker->cond);
    }

    free (worker_thread_array);

    PASIS_DASSERT (workers_working == 0);
    
    PASIS_PRINT_INFO ("---- PasisFS Worker Shutdown ----\n");
    PASIS_PRINT_INFO ("Total RPCs       :  %u\n", statistics.total_rpcs);
    PASIS_PRINT_INFO ("Max Outstanding  :  %u\n", statistics.max_outstanding);
    PASIS_PRINT_INFO ("Avg Outstanding  :  %f\n", 
                      (double)statistics.total_outstanding / 
                      (double)statistics.total_rpcs);
    PASIS_PRINT_INFO ("RPC Retransmits  :  %u\n", statistics.rpc_retransmits);
    PASIS_PRINT_INFO ("Dispatch Blocked :  %u\n\n", 
                      statistics.dispatch_blocked);
}


static inline pasis_error_t
DispatchRegister(unsigned long prognum,
                 unsigned long version,
                 dispatch_func_t dispatch_func,
                 unsigned long proto,
                 unsigned short port)
{
    rpc_dispatch_t *dispatch; 

    if (dispatch_func == NULL) {
        PASIS_PRINT_ERR("%s: Null function!\n", __FUNCTION__);
        return PASIS_ERROR;
    }

    /* Register with portmapper */
    if (pmap_set (prognum, version, proto, port) != 1) {
        PASIS_PRINT_ERR ("%s: error pmap_set (prog=%lu, ver=%lu, proto=%lu)\n",
                         __FUNCTION__, prognum, version, proto);

        return PASIS_ERROR;
    }

    dispatch = (rpc_dispatch_t *) malloc (sizeof (rpc_dispatch_t));
                                                  
    dispatch->prognum = prognum;
    dispatch->version = version;
    dispatch->proto = proto;
    dispatch->dispatch_func = dispatch_func;
    
    PASIS_LL_ADD_NODE_BEFORE (list, dispatch, rpc_dispatch_hd, 
                              rpc_dispatch_hd);

    return PASIS_SUCCESS;
}


static void
DispatchUnregister(unsigned long prognum,
                   unsigned long version)
{
    rpc_dispatch_t *dispatch;

    /* Go through list and find entry to remove */
    dispatch = PASIS_LL_GET_HEAD (rpc_dispatch_hd);
    while (dispatch != NULL) {
        if (dispatch->prognum == prognum && dispatch->version == version) {
            PASIS_LL_REMOVE_NODE (list, dispatch, rpc_dispatch_hd);

            /* Unregister from portmapper */
            pmap_unset (prognum, version);

            free (dispatch);
            break;
        }
        dispatch = PASIS_LL_NEXT (list, dispatch);
    }

    return;
}


static void
RPCShutdown(void)
{
    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s\n", __FUNCTION__);

    /* Cleanup rpc thread */
    rpc_event_thread_shutdown = 1;
    st_thread_interrupt (rpc_event_thread);
    st_thread_join (rpc_event_thread, NULL);

    /* XXX NFSv2 and Mountv1 currently unsupported */
    /*
    DispatchUnregister(MOUNTPROG, MOUNTVERS);
    DispatchUnregister(NFS_PROGRAM, NFS_VERSION);
    */
    DispatchUnregister(MOUNT_PROGRAM, MOUNT_V3);
    DispatchUnregister(NFS_PROGRAM, NFS_V3);
}


static pasis_error_t
RPCStartup(void)
{
    int flag = 1;
    struct sockaddr_in addr;

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = 0;
    addr.sin_port = htons (NFS_PORT);

    /* Create UDP socket */
    rpc_udp_fd = socket (AF_INET, SOCK_DGRAM, 0);
    if (rpc_udp_fd < 0) {
        PASIS_PRINT_ERR ("%s: error creating socket (%d)\n", 
                         __FUNCTION__, errno);
        return errno;
    }
    setsockopt (rpc_udp_fd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof (int));

    if (bind (rpc_udp_fd, (struct sockaddr *)&addr, sizeof (addr)) < 0) {
        close (rpc_udp_fd);
        PASIS_PRINT_ERR ("%s: error binding udp socket to port %d\n",
                         __FUNCTION__, NFS_PORT);
        return errno;
    }

    /* Create TCP socket */
    rpc_tcp_fd = socket (AF_INET, SOCK_STREAM, 0);
    if (rpc_tcp_fd < 0) {
        close (rpc_udp_fd);
        PASIS_PRINT_ERR ("%s: error creating socket (%d)\n", 
                         __FUNCTION__, errno);
        return errno;
    }
    setsockopt (rpc_tcp_fd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof (int));

    if (bind (rpc_tcp_fd, (struct sockaddr *)&addr, sizeof (addr)) < 0) {
        close (rpc_udp_fd);
        close (rpc_tcp_fd);
        PASIS_PRINT_ERR ("%s: error binding tcp socket to port %d\n",
                         __FUNCTION__, NFS_PORT);
        return errno;
    }

    if (listen (rpc_tcp_fd, 5) < 0) {
        close (rpc_udp_fd);
        close (rpc_tcp_fd);
        PASIS_PRINT_ERR ("%s: error listening on tcp socket\n", __FUNCTION__);
        return errno;
    }

    PASIS_LL_HEAD_INIT (rpc_dispatch_hd);

    /* XXX NFSv2 and Mountv1 not currently supported */
    if (DispatchRegister(MOUNTPROG, MOUNTVERS, mount_program_1, IPPROTO_UDP, 
                         NFS_PORT)) {
                         
        PASIS_PRINT_ERR("%s: can't register (MOUNTPROG,MOUNTVERS,udp).\n",
                        __FUNCTION__);
        close (rpc_udp_fd);
        close (rpc_tcp_fd);

        return PASIS_ERROR;
    }

#if 0
    if (DispatchRegister(NFS_PROGRAM, NFS_VERSION, nfs_program_2,
                         IPPROTO_UDP, NFS_PORT)) {
        PASIS_PRINT_ERR("%s: can't register (NFS_PROGRAM,NFS_VERSION,udp)\n",
                        __FUNCTION__);
        DispatchUnregister(MOUNTPROG, MOUNTVERS);

        close (rpc_udp_fd);
        close (rpc_tcp_fd);

        return PASIS_ERROR;
    }
#endif

    /* XXX could also register TCP versions of these protocols */

    /* Register NFSv3 and Mountv3 */
    if (DispatchRegister(NFS_PROGRAM, NFS_V3, nfs_program_3,
                         IPPROTO_UDP, NFS_PORT)) {
        PASIS_PRINT_ERR("%s: can't register (NFS_PROGRAM,NFS_VERSION,udp)\n",
                        __FUNCTION__);
        /*
        DispatchUnregister(MOUNTPROG, MOUNTVERS);
        DispatchUnregister(NFS_PROGRAM, NFS_VERSION);
        */

        close (rpc_udp_fd);
        close (rpc_tcp_fd);

        return PASIS_ERROR;
    }

    if (DispatchRegister(MOUNT_PROGRAM, MOUNT_V3, mount_program_3,
                         IPPROTO_UDP, NFS_PORT)) {
        PASIS_PRINT_ERR("%s: can't register (NFS_PROGRAM,NFS_VERSION,udp)\n",
                        __FUNCTION__);
        /*
        DispatchUnregister(MOUNTPROG, MOUNTVERS);
        DispatchUnregister(NFS_PROGRAM, NFS_VERSION);
        */
        DispatchUnregister(NFS_PROGRAM, NFS_V3);

        close (rpc_udp_fd);
        close (rpc_tcp_fd);

        return PASIS_ERROR;
    }

    /* Startup event thread */
    rpc_event_thread = st_thread_create (RPCEventThread, NULL, 1, 0);
    if (rpc_event_thread == NULL) {
        /*
        DispatchUnregister(MOUNTPROG, MOUNTVERS);
        DispatchUnregister(NFS_PROGRAM, NFS_VERSION);
        */
        DispatchUnregister(NFS_PROGRAM, NFS_V3);
        DispatchUnregister(MOUNT_PROGRAM, MOUNT_V3);
        close (rpc_udp_fd);
        close (rpc_tcp_fd);

        return PASIS_ERROR;
    }

    return PASIS_SUCCESS;
}


void
FE_Shutdown (void)
{   
    RPCShutdown ();
    WorkerShutdown ();   
    NFS3Shutdown ();
}


pasis_error_t
FE_Startup (unsigned int nfs_workers)
{
    pasis_error_t error;
    int error_level = 0;

    error = RPCStartup ();
    if (error) {
        PASIS_PRINT_ERR ("RPCStartup Error: %s\n", PASIS_STRERROR(error));
        goto error;
    }

    error = WorkerStartup (nfs_workers);
    if (error) {
        PASIS_PRINT_ERR ("WorkerStartup Error: %s\n", PASIS_STRERROR(error));
        error_level = 1;
        goto error;
    }

    return PASIS_SUCCESS;

 error:
    switch (error_level) {
    case 1:
        RPCShutdown ();
    }

    return error;
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
