/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_lists.h"

#include "pasis_cache.h"
#include "pasisfs_buf_cache.h"

static inline char *
print_queue(pasisfs_buf_t *bp,
           pasisfs_buf_queue_t *queue)
{
    static char name[64];

    if (&bp->b_cache->free_buf_queue == queue) {
        sprintf (name, "free");
    } else if (&bp->b_cache->clean_buf_queue == queue) {
        sprintf (name, "clean");
    } else if (&bp->b_cache->dirty_buf_queue == queue) {
        sprintf (name, "dirty");
    } else if (&bp->b_cache->locked_buf_queue == queue) {
        sprintf (name, "locked");
    }

    return name;
}

#define PASISFS_B_Q_MASK \
(PASISFS_B_Q_NONE | PASISFS_B_Q_FREE | PASISFS_B_Q_CLEAN | PASISFS_B_Q_DIRTY | PASISFS_B_Q_LOCKED)

#define PASISFS_B_Q_IS_ON(bp, q) \
  (((bp)->b_flags & (q)) ? PASIS_TRUE : PASIS_FALSE)

#define PASISFS_B_Q_FLAGS_CLEAR(bp)       \
do {                                      \
    (bp)->b_flags &= ~(PASISFS_B_Q_MASK); \
    (bp)->b_flags |= PASISFS_B_Q_NONE;    \
} while (0)

#define PASISFS_B_Q_ADD(bp, q_head, q_flag)               \
do {                                                      \
    (bp)->b_flags &= ~(PASISFS_B_Q_MASK);                 \
    (bp)->b_flags |= q_flag;                              \
    PASIS_Q_ADD_TAIL (q_head, b_queue, bp);               \
    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, "BufQueueAdd: bp=%p (%d), flags=%#x, queue=%s\n", bp, bp->b_id, bp->b_flags, print_queue (bp, &q_head)); \
} while (0);

#define PASISFS_B_Q_REMOVE(bp, q_head)                   \
do {                                                     \
    PASIS_DASSERT (((bp)->b_flags & PASISFS_B_IO) == 0); \
    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, "BufQueueRemove: bp=%p (%d), flags=%#x, queue=%s\n", bp, bp->b_id, bp->b_flags, print_queue (bp, &q_head)); \
    PASISFS_B_Q_FLAGS_CLEAR (bp);                        \
    PASIS_Q_REMOVE (q_head, b_queue, bp);                \
} while (0);


#define PASISFS_B_Q_ADD_FREE(bp)                                  \
do {                                                              \
    (bp)->b_flags &= ~(PASISFS_B_Q_MASK);                         \
    ((bp)->b_flags) |= PASISFS_B_Q_FREE;                          \
    PASIS_Q_ADD_TAIL (bp->b_cache->free_buf_queue, b_queue, bp);  \
} while (0)

#define PASISFS_B_Q_REMOVE_FREE(bp)                             \
do {                                                            \
    PASISFS_B_Q_FLAGS_CLEAR (bp);                               \
    PASIS_Q_REMOVE (bp->b_cache->free_buf_queue, b_queue, bp);  \
} while (0)


#define PASISFS_B_Q_ADD_DIRTY(bp)                                         \
do {                                                                      \
    PASIS_DASSERT ((bp)->b_flags & PASISFS_B_DIRTY);                      \
    PASIS_DASSERT ((bp)->b_flags & PASISFS_B_Q_NONE);                     \
    PASISFS_B_Q_ADD (bp, bp->b_cache->dirty_buf_queue, PASISFS_B_Q_DIRTY) \
} while (0)

#define PASISFS_B_Q_REMOVE_DIRTY( bp) \
  PASISFS_B_Q_REMOVE (bp, bp->b_cache->dirty_buf_queue)


#define PASISFS_B_Q_ADD_CLEAN(bp)                                         \
do {                                                                      \
    PASIS_DASSERT (((bp)->b_flags & PASISFS_B_DIRTY) == 0);               \
    PASIS_DASSERT ((bp)->b_flags & PASISFS_B_Q_NONE);                     \
    PASISFS_B_Q_ADD (bp, bp->b_cache->clean_buf_queue, PASISFS_B_Q_CLEAN) \
} while (0)

#define PASISFS_B_Q_REMOVE_CLEAN(bp) \
  PASISFS_B_Q_REMOVE (bp, bp->b_cache->clean_buf_queue)


#define PASISFS_B_Q_ADD_LOCKED(bp)                                          \
do {                                                                        \
    PASIS_DASSERT (PASISFS_B_IS_PINNED (bp));                               \
    PASIS_DASSERT ((bp)->b_flags & PASISFS_B_Q_NONE);                       \
    PASISFS_B_Q_ADD (bp, bp->b_cache->locked_buf_queue, PASISFS_B_Q_LOCKED) \
} while (0)

#define PASISFS_B_Q_REMOVE_LOCKED(bp) \
  PASISFS_B_Q_REMOVE (bp, bp->b_cache->locked_buf_queue)


/* Debug list validation */
#define dump_free_list(cache, verbose) \
  dump_buf_list(cache, PASISFS_B_Q_FREE, verbose)
#define dump_clean_list(cache, verbose) \
  dump_buf_list(cache, PASISFS_B_Q_CLEAN, verbose)
#define dump_dirty_list(cache, verbose) \
  dump_buf_list(cache, PASISFS_B_Q_DIRTY, verbose)
#define dump_locked_list(cache, verbose) \
  dump_buf_list(cache, PASISFS_B_Q_LOCKED, verbose)

static int dump_buf_list (pasisfs_bcache_t *cache, unsigned int queue, int ver);
static void dump_all_buffers (pasisfs_bcache_t *cache);

#if 0 /* Currently unused, but implemented */
static void dump_all_lists (pasisfs_bcache_t *cache);
static void verify_all_lists (pasisfs_bcache_t *cache);
#endif

static inline pasisfs_buf_t *
BufCacheLookup (pasisfs_bcache_t *cache, void *tag)
{
    unsigned int hash;
    pasisfs_buf_t *bp;

    hash = cache->hash_fp (tag, cache->buckets);
    PASIS_DASSERT (hash < cache->buckets);

    bp = PASIS_Q_FIRST (cache->hash_array[hash]);
    while (bp) {
        if (cache->compare_fp (bp->b_tag, tag)) {
            return bp;
        }
        bp = PASIS_Q_NEXT (b_hash_chain, bp);
    }

    return NULL;
}


static inline void
BufCacheInsert (pasisfs_buf_t *bp)
{
    unsigned int hash;
    pasisfs_buf_t *b;

    hash = bp->b_cache->hash_fp (bp->b_tag, bp->b_cache->buckets);
    PASIS_DASSERT (hash < bp->b_cache->buckets);

    b = BufCacheLookup (bp->b_cache, bp->b_tag);
    if (b != NULL) {
        PASIS_DASSERT (b == bp);
    }

    bp->b_hash = hash;
    PASIS_Q_ADD_HEAD (bp->b_cache->hash_array[hash], b_hash_chain, bp);
    bp->b_flags |= PASISFS_B_HASHED; 

    return;
}


static inline void
BufCacheRemove (pasisfs_buf_t *bp)
{
    pasisfs_buf_t *b;

    if (PASISFS_B_IS_HASHED (bp)) {
        PASIS_Q_REMOVE (bp->b_cache->hash_array[bp->b_hash], b_hash_chain, bp);
        bp->b_flags &= ~PASISFS_B_HASHED; 
    }

    b = BufCacheLookup (bp->b_cache, bp->b_tag);
   
    PASIS_DASSERT (b == NULL);
}


pasisfs_bcache_t *
PasisFS_BufCacheInit (char *cache_name,
                      unsigned int count,
                      unsigned int tag_size, 
                      unsigned int (* hash_fp)(void *tag, unsigned int num),
                      int (* compare_fp)(void *tag, void *tag2),
                      pasis_error_t (* write_fp)(pasisfs_buf_t *),
                      void (* evict_fp)(pasisfs_buf_t *),
                      void *(* alloc_fp)(unsigned int idx),
                      void (* dealloc_fp)(void *))
{
    pasisfs_bcache_t *cache;
    p_byte *metadata;
    int i;

    /* Allocate and init cache structure */
    cache = (pasisfs_bcache_t *) malloc (sizeof (pasisfs_bcache_t));
    cache->name = malloc (strlen (cache_name) + 1);
    strcpy (cache->name, cache_name);
    cache->count = count;
    cache->tag_size = tag_size;

    PASIS_DASSERT (hash_fp != NULL);
    PASIS_DASSERT (compare_fp != NULL);
    PASIS_DASSERT (write_fp != NULL);

    if (alloc_fp) {
        PASIS_DASSERT (dealloc_fp != NULL);
    }

    cache->hash_fp = hash_fp;
    cache->compare_fp = compare_fp;
    cache->write_fp = write_fp;
    cache->evict_fp = evict_fp;
    cache->dealloc_fp = dealloc_fp;

    cache->buckets = (count >> 3);
    if (cache->buckets == 0) {
        cache->buckets = count;
    } else {
        cache->buckets -= 1;
    }

    PASIS_Q_INIT (cache->free_buf_queue);
    PASIS_Q_INIT (cache->clean_buf_queue);
    PASIS_Q_INIT (cache->dirty_buf_queue);
    PASIS_Q_INIT (cache->locked_buf_queue);

    /* Allocate metadata (tags and header) memory region */
    cache->metadata_region = (p_byte *) malloc ((sizeof (pasisfs_buf_t) + 
                                                 tag_size) * count);
    metadata = cache->metadata_region;

    /* Init each buffer and add to free queue */
    for (i = 0; i < count; i++) {
        pasisfs_buf_t *bp;

        bp = (pasisfs_buf_t *)metadata;
        memset (bp, 0, sizeof (pasisfs_buf_t));

        bp->b_id = i;
        bp->b_cache = cache;
        bp->b_tag = (metadata + sizeof (pasisfs_buf_t));
        bp->b_io_cond = st_cond_new ();

        if (alloc_fp) {
            bp->b_data = alloc_fp (bp->b_id);
        }

        PASISFS_B_Q_ADD_FREE (bp);

        metadata += (tag_size + sizeof (pasisfs_buf_t));

        /*
        PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 6, "[%d] tag=%p, data=%p\n", 
                       bp->b_id, bp->b_tag, bp->b_data);
        */
    }
    
    /* Init hash buckets */
    cache->hash_array = (pasisfs_buf_queue_t *) malloc (cache->buckets * 
                                                sizeof (pasisfs_buf_queue_t));

    for (i = 0; i < cache->buckets; i++) {
        PASIS_Q_INIT (cache->hash_array[i]);
    }

    return cache;
}

void 
PasisFS_BufCacheShutdown (pasisfs_bcache_t *cache)
{
    int i;
    pasisfs_buf_t *bp;

    /* Dump stats */
    PASIS_PRINT_INFO ("\n---- %s stats ----\n", cache->name);
    PASIS_PRINT_INFO ("In core hits    :   %u\n", cache->in_core_hits);
    PASIS_PRINT_INFO ("In core misses  :   %u\n", cache->in_core_misses);
    PASIS_PRINT_INFO ("Buffer releases :   %u\n", cache->releases);
    PASIS_PRINT_INFO ("Dirty releases  :   %u\n", cache->dirty_releases);
    PASIS_PRINT_INFO ("Free gets       :   %u\n", cache->free_gets);
    PASIS_PRINT_INFO ("Clean gets      :   %u\n", cache->clean_gets);
    PASIS_PRINT_INFO ("Dirty gets      :   %u\n", cache->dirty_gets);
    PASIS_PRINT_INFO ("Get retries     :   %u\n\n", cache->get_retries);

    /* Go through all buffers, evicting and syncing them */
    for (i = 0; i < cache->count; i++) {

        /* Forces sync and eviction of all buffers */
        bp = PasisFS_BufCacheGetEmpty (cache);
        
        st_cond_destroy (bp->b_io_cond);
        
        if (cache->dealloc_fp) {
            cache->dealloc_fp (bp->b_data);
        }
    }
    
    free (cache->name);
    free (cache->metadata_region);
    free (cache->hash_array);
    free (cache);
}


pasis_error_t
PasisFS_BufCachePin (pasisfs_buf_t *bp)
{
    bp->b_pin_count++;

    PASIS_DASSERT (bp->b_pin_count < 3);

    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 4, 
                   "%s: bp=%p (%d), flags=%#x, pin_count=%d\n", 
                   __FUNCTION__, bp, bp->b_id, bp->b_flags, bp->b_pin_count);

    /* This is done to ensure that pin is atomic with regards to a write...
       So pinned data will not be in danger of interferring with in progress 
       writes (such as cache evictions -- basically it is there to block until
       the io completes...)
    */
    /* XXX not sure about this */
    while (bp->b_flags & PASISFS_B_IO) {
        PASIS_DASSERT (bp->b_pin_count > 1);
        st_cond_wait (bp->b_io_cond);
    }

    PASIS_DASSERT ((bp->b_flags & PASISFS_B_IO) == 0);

    if (bp->b_pin_count == 1) {
        /* Shouldn't be on the free list */
        PASIS_DASSERT (!PASISFS_B_Q_IS_ON (bp, PASISFS_B_Q_FREE));

        /* Remove from the list it was on and add to locked list */
        switch (bp->b_flags & (PASISFS_B_Q_MASK)) {
        case PASISFS_B_Q_CLEAN:
            PASISFS_B_Q_REMOVE_CLEAN (bp);
            PASISFS_B_Q_ADD_LOCKED (bp);
        break;
        case PASISFS_B_Q_DIRTY:
            PASISFS_B_Q_REMOVE_DIRTY (bp);
            PASISFS_B_Q_ADD_LOCKED (bp);
            break;
        case PASISFS_B_Q_NONE:
            PASISFS_B_Q_ADD_LOCKED (bp);
            break;
        }
    }
        
    PASIS_DASSERT (PASISFS_B_Q_IS_ON (bp, PASISFS_B_Q_LOCKED));
    
    return PASIS_SUCCESS;
}


void
PasisFS_BufCacheUnpin (pasisfs_buf_t *bp)
{
    PASIS_DASSERT (PASISFS_B_Q_IS_ON (bp, PASISFS_B_Q_LOCKED));

    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 4, "%s: bp=%p (%d), pin_count=%d\n", 
                   __FUNCTION__, bp, bp->b_id, bp->b_pin_count - 1);

    PASIS_DASSERT (bp->b_pin_count != 0);

    bp->b_pin_count--;

    /* Remove from locked queue if no more references exist */
    if (bp->b_pin_count == 0) {
        PASISFS_B_Q_REMOVE_LOCKED (bp);

        /* Check which list it should be put on, dirty or clean */
        if (PASISFS_B_IS_DIRTY (bp)) {
            PASISFS_B_Q_ADD_DIRTY (bp);
        } else {
            PASISFS_B_Q_ADD_CLEAN (bp);
        }
    }

    return;
}


void
PasisFS_BufCacheMarkDirty (pasisfs_buf_t *bp)
{
    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 4, "%s: bp=%p (%d) flags=%#x\n",
                   __FUNCTION__, bp, bp->b_id, bp->b_flags);

    if (!(bp->b_flags & PASISFS_B_DIRTY)) {
        bp->b_flags |= PASISFS_B_DIRTY;
    }

    /* Put buffer on dirty queue if not locked or dirty, else leave alone */
    switch (bp->b_flags & PASISFS_B_Q_MASK) {
    case PASISFS_B_Q_FREE:
        PASISFS_B_Q_REMOVE_FREE (bp);
        PASISFS_B_Q_ADD_DIRTY (bp);
        break;
    case PASISFS_B_Q_CLEAN:
        PASISFS_B_Q_REMOVE_CLEAN (bp);
        PASISFS_B_Q_ADD_DIRTY (bp);
        break;
    case PASISFS_B_Q_NONE:
        PASISFS_B_Q_ADD_DIRTY (bp);
        break;
    }

    //verify_all_lists ();
    return;
}


void
PasisFS_BufCacheMarkClean (pasisfs_buf_t *bp)
{
    if (bp->b_flags & PASISFS_B_DIRTY) {
        bp->b_flags &= ~PASISFS_B_DIRTY;
    }

    /* If it was on the dirty queue, move it to the clean queue */
    if (PASISFS_B_Q_IS_ON (bp, PASISFS_B_Q_DIRTY)) {
        PASISFS_B_Q_REMOVE_DIRTY (bp);
        PASISFS_B_Q_ADD_CLEAN (bp);
    }

    //verify_all_lists ();
    return;
}


pasisfs_buf_t *
PasisFS_BufCacheFindInCore (pasisfs_bcache_t *cache,
                            void *tag)
{
    pasisfs_buf_t *bp;

    bp = BufCacheLookup (cache, tag);
    if (bp) {
        cache->in_core_hits++;
    } else {
        cache->in_core_misses++;
    }

    return bp;
}


pasis_error_t
PasisFS_BufCacheInsert (pasisfs_buf_t *bp)
{
    /* Insert the buffer into the cache */
    BufCacheInsert (bp);

#if DEBUG > 0
    verify_all_lists ();
#endif

    return PASIS_SUCCESS;
}


pasis_error_t
PasisFS_BufCacheRelease (pasisfs_buf_t *bp)
{
    PASIS_DASSERT (bp);
    PASIS_DASSERT ((bp->b_flags & PASISFS_B_Q_LOCKED) == 0);
    PASIS_DASSERT ((bp->b_flags & PASISFS_B_IO) == 0);  /* No pending I/O */
    PASIS_DASSERT (!PASISFS_B_IS_PINNED (bp));

    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 3, "%s: bp=%d(%p), flags=%#x\n",
                   __FUNCTION__, bp->b_id, bp, bp->b_flags);

    bp->b_cache->releases++;

    /* Dirty block needs to be written out first */
    if (PASISFS_B_IS_DIRTY(bp)) {

        PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 4, 
                       "%s: writing out dirty block\n", __FUNCTION__);
        
        /* Do the write */
        bp->b_cache->write_fp (bp);

        bp->b_cache->dirty_releases++;

        if (PASISFS_B_IS_PINNED (bp)) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 4, 
                           "%s: still pinned error inuse\n", __FUNCTION__);
            return PASIS_ERROR_BUF_INUSE;
        }
    }

    /* Make sure it is clean */
    PASIS_DASSERT ((bp->b_flags & PASISFS_B_DIRTY) == 0);

    /* Do eviction callback if there is one -- may not be non-blocking */
    if (bp->b_cache->evict_fp) {
        PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, "%s: doing eviction callback\n",
                       __FUNCTION__);
        bp->b_cache->evict_fp (bp);
    }

    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, 
                   "%s: eviction done bp=%d (%p), flags=%#x\n", 
                   __FUNCTION__, bp->b_id, bp, bp->b_flags);
    
    /* Remove it from the cache */
    BufCacheRemove (bp);

    /* Remove buffer from the lists it is on */
    PASIS_DASSERT (bp->b_pin_count == 0);

    /* Remove from clean list */
    PASIS_DASSERT (PASISFS_B_Q_IS_ON (bp, PASISFS_B_Q_CLEAN));
    PASISFS_B_Q_REMOVE_CLEAN (bp);

    /* Reset buffer */
    bp->b_flags = PASISFS_B_NONE;
    bp->b_hash = (unsigned int)-1;
    bp->b_list.next = NULL;
    bp->b_list.prev = NULL;

    /* Add buffer back to free list */
    PASISFS_B_Q_ADD_FREE (bp);

    /*
    bufs_in_use--;

    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, "%s: Buffers in use=%d\n",
                   __FUNCTION__, bufs_in_use);
    */

    return PASIS_SUCCESS;
}


pasisfs_buf_t *
PasisFS_BufCacheGetEmpty (pasisfs_bcache_t *cache)
{
    pasisfs_buf_t *bp;
    pasis_error_t  error;
    int            start_over;
    
    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 3, "%s\n", __FUNCTION__); 

    /* First try and get a block from the free list */
    do {
        start_over = 0;
        
        bp = PASIS_Q_FIRST (cache->free_buf_queue);
        if (bp == NULL) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, "Free list empty\n");
        }
        if (bp) {
            PASIS_DASSERT (PASISFS_B_Q_IS_ON (bp, PASISFS_B_Q_FREE));

            PASISFS_B_Q_REMOVE_FREE (bp);
            PASISFS_B_Q_ADD_CLEAN (bp);
            
            /*
            bufs_in_use++;
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, 
                           "%s: Buffers in use=%d, bp=%d(%p)\n",
                           __FUNCTION__, bufs_in_use, bp->b_id, bp);
            */
            
            cache->free_gets++;

            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, "%s free get bp=%p, (%d)\n", 
                           __FUNCTION__, bp, bp->b_id); 
            return bp;
        }

        /* Next search for a block from clean list */

        bp = PASIS_Q_FIRST (cache->clean_buf_queue);
        if (bp == NULL) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, "Clean list empty\n");
        }
        if (bp) {
            PASIS_DASSERT (!(PASISFS_B_IS_DIRTY (bp)));
 
            /* Release it from all it's lists -- it is put on free list */
            error = PasisFS_BufCacheRelease (bp);

            /* Can't be an error since it was clean...and won't block */
            PASIS_DASSERT (error == PASIS_SUCCESS);

            PASISFS_B_Q_REMOVE_FREE (bp);
            PASISFS_B_Q_ADD_CLEAN (bp);

            /*
            bufs_in_use++;
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, 
                           "%s: Buffers in use=%d, bp=%d(%p)\n",
                           __FUNCTION__, bufs_in_use, bp->b_id, bp);
            */

            cache->clean_gets++;
        
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, "%s clean get bp=%p (%d)\n", 
                           __FUNCTION__, bp, bp->b_id); 

            return bp;
        }

        /* Next search the dirty list */
        bp = PASIS_Q_FIRST (cache->dirty_buf_queue);

        if (bp == NULL) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, "Dirty list empty\n");
        }
        if (bp) {
            /* Release it from all it's lists -- placed on free list */
            error = PasisFS_BufCacheRelease (bp);
            
            /* Make sure this was successful, if not try again... */
            if (error == PASIS_SUCCESS) {
                PASISFS_B_Q_REMOVE_FREE (bp);
                PASISFS_B_Q_ADD_CLEAN (bp);
                
                /*
                bufs_in_use++;
                PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, 
                       "%s: Buffers in use=%d, bp=%d(%p)\n",
                       __FUNCTION__, bufs_in_use, bp->b_id, bp);
                */

                cache->dirty_gets++;
                
                PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 5, 
                               "%s dirty get bp=%p (%d)\n", 
                               __FUNCTION__, bp, bp->b_id); 

                return bp;
            } else {
                start_over = 1;
                cache->get_retries++;
            }
        }
        
    } while (start_over == 1);

    /* Dump out list of locked buffers... */
    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 6, "%s: Dumping locked buffers\n",
                   __FUNCTION__);

    dump_locked_list (cache, 1);
    dump_clean_list (cache, 1);
    dump_dirty_list (cache, 1);

    dump_all_buffers (cache);

    /* Can't end up here... Well we could, but it would be BAD */
    PASIS_DASSERT (0);
    
    return NULL;
}


static int
dump_buf_list (pasisfs_bcache_t *cache, unsigned int queue, int verbose)
{
    pasisfs_buf_queue_t *buf_queue;
    pasisfs_buf_t *bp;
    p_uint32 flags;
    unsigned int flag_mask = PASISFS_B_Q_MASK | PASISFS_B_DIRTY;
    int count = 0;

    switch (queue) {
    case PASISFS_B_Q_FREE:
        if (verbose) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 6, "Dumping Free List\n");
        }
        buf_queue = &cache->free_buf_queue;
        flags = PASISFS_B_Q_FREE;
        break;
    case PASISFS_B_Q_CLEAN:
        if (verbose) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 6, "Dumping Clean List\n");
        }
        buf_queue = &cache->clean_buf_queue;
        flags = PASISFS_B_Q_CLEAN;
        break;
    case PASISFS_B_Q_DIRTY:
        if (verbose) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 6, "Dumping Dirty List\n");
        }
        buf_queue = &cache->dirty_buf_queue;
        flags = PASISFS_B_Q_DIRTY | PASISFS_B_DIRTY;
        break;
    case PASISFS_B_Q_LOCKED:
        if (verbose) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 6, "Dumping Locked List\n");
        }
        buf_queue = &cache->locked_buf_queue;
        flag_mask = PASISFS_B_Q_MASK;
        flags = PASISFS_B_Q_LOCKED;
        break;
    default:
        return 0;
    }

    bp = PASIS_Q_FIRST (*buf_queue);
    if (bp == NULL) {
        if (verbose) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 6, "Buffer list is empty\n");
        }
    }

    while (bp) {
        if (verbose) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 6, 
                           "BP=%d (%p), dirty=%#x, queue=%#x\n",
                           bp->b_id, bp, bp->b_flags & PASISFS_B_DIRTY, 
                           bp->b_flags & PASISFS_B_Q_MASK); 
        }

        PASIS_DASSERT ((bp->b_flags & flag_mask) == flags);

        bp = PASIS_Q_NEXT (b_queue, bp);
        count++;
    }

    return count;
}


static inline void
verify_buffer_queue (pasisfs_bcache_t *cache,
                     pasisfs_buf_queue_t queue,
                     p_uint32 qflag)
{
    pasisfs_buf_t *bp;
    p_byte *cache_metadata_end;

    cache_metadata_end = cache->metadata_region + ((sizeof (pasisfs_buf_t) + 
                         cache->tag_size) * cache->count);
                                                      
    bp = PASIS_Q_FIRST (queue);
    while (bp) {
        PASIS_DASSERT ((p_byte *)bp >= cache->metadata_region &&
                       ((((p_byte *)bp) + sizeof (pasisfs_buf_t)) 
                        <= cache_metadata_end));

        /* Make sure flag for correct queue is set */
        PASIS_DASSERT (bp->b_flags & qflag);

        /* No other flags should be set -- other than hash flag */
        PASIS_DASSERT (!(bp->b_flags &
                         ~(PASISFS_B_IO | PASISFS_B_HASHED | qflag)));

        /* Validate fields of each buffer */
        /* Check pin count */
        if (!(qflag & PASISFS_B_Q_LOCKED)) {
            PASIS_DASSERT (bp->b_pin_count == 0);
        }

        /* Check tag is within memory bounds */
        PASIS_DASSERT (bp->b_tag >= cache->metadata_region &&
                       (bp->b_tag + cache->tag_size <= 
                        cache_metadata_end));

        /* If this entry is in the hash table */
        if (PASISFS_B_IS_HASHED (bp)) {
            pasisfs_buf_t *tmp_bp;
            int found = 0;
            PASIS_DASSERT (bp->b_hash < cache->buckets);
            tmp_bp = PASIS_Q_FIRST (cache->hash_array[bp->b_hash]);
            while (tmp_bp) {
                if (tmp_bp == bp) {
                    found = 1;
                    break;
                }
                tmp_bp = PASIS_Q_NEXT (b_hash_chain, tmp_bp);
            }
            PASIS_DASSERT (found == 1);
        }

        bp = PASIS_Q_NEXT (b_queue, bp);
    }
}


void
PasisFS_BufCacheVerify (pasisfs_bcache_t *cache,
                        int allow_locked)
{
    pasisfs_buf_t *bp;

    /* No locked or dirty buffers should exist */
    if (allow_locked == 0) {
        bp = PASIS_Q_FIRST (cache->locked_buf_queue);
        if (bp) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 3, "%s: found locked buffers\n",
                           __FUNCTION__);
            do {
                PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 3, 
                               "bp=%p (%d) pin_count=%d\n",
                               bp, bp->b_id, bp->b_pin_count);
            } while ((bp = PASIS_Q_NEXT (b_queue, bp)));
        }

        PASIS_DASSERT (PASIS_Q_FIRST (cache->locked_buf_queue) == NULL);
    } else {
        verify_buffer_queue (cache, cache->locked_buf_queue, 
                             PASISFS_B_Q_LOCKED | PASISFS_B_DIRTY);
    }

    /* Next check all queues and buffers */
    verify_buffer_queue (cache, cache->free_buf_queue, PASISFS_B_Q_FREE);
    verify_buffer_queue (cache, cache->clean_buf_queue, PASISFS_B_Q_CLEAN);
    verify_buffer_queue (cache, cache->dirty_buf_queue,
                         PASISFS_B_Q_DIRTY | PASISFS_B_DIRTY);
}


#if 0 /* Currently unused */
static void
dump_all_lists (pasisfs_bcache_t *cache)
{
    int count = 0;
    count += dump_free_list(cache, 1);
    count += dump_clean_list(cache, 1);
    count += dump_dirty_list(cache, 1);
    count += dump_locked_list(cache, 1);
}

static void
verify_all_lists (pasisfs_bcache_t *cache)
{
    int count = 0;
    count += dump_free_list(cache, 0);
    count += dump_clean_list(cache, 0);
    count += dump_dirty_list(cache, 0);
    count += dump_locked_list(cache, 0);
}
#endif

static void
dump_all_buffers (pasisfs_bcache_t *cache)
{
    pasisfs_buf_t *bp;
    int i;

    PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 3, "\nDumping all system buffers\n");

    for (i = 0; i < cache->buckets; i++) {
        while ((bp = PASIS_Q_FIRST (cache->hash_array[i]))) {
        
            PASIS_DPRINTF (PASIS_DEBUG_FS_BUF, 6, 
                           "[%d] BP=%d (%p), dirty=%#x, queue=%#x\n",
                           i, bp->b_id, bp, bp->b_flags & PASISFS_B_DIRTY, 
                           bp->b_flags & PASISFS_B_Q_MASK); 

            bp = PASIS_Q_NEXT (b_hash_chain, bp);
        }
    }
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
