/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"

#include "pasisio.h"
#include "pasisio_md_protocol.h"
#include "pasis_md_attr.h"
#include "pasis_md_prot.h"

#include "pasisfs_data_cache.h"
#include "pasisfs_buf_cache.h"
#include "pasisfs_data.h"

#include "fe.h"

#ifndef P_MIN
#define P_MIN(a,b) (((a) > (b)) ? (b) : (a))
#endif /* P_MIN */

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif


static inline char *
dump_fid (pasis_md_fid_t *f)
{
#if PASIS_DEBUG > 0
    static char buf[PASISIO_FILEID_SIZE*2 + 1];
    int i;
    char *bp = buf;
    for (i=0; i<PASISIO_FILEID_SIZE; i++) {
        snprintf(bp, 3, "%02X", (unsigned char)f->data[i]);
        bp +=2;
    }

    return buf;
#else 
    return NULL;
#endif
}

static pasisfs_buf_t *
FileCreateDataBlock (pasis_md_fid_t   *fid,
                     pasis_md_loc_t   *loc,
                     pasis_md_attr_t  *attr,
                     pasisfs_blkaddr_t blkno)
{
    pasisfs_buf_t *bp;
    pasisfs_dcache_data_t *data;

    /* Get empty buffer */
    bp = PasisFS_DataCacheGetEmpty ();

    PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 3, 
                   "%s: bp=%d blkno=%"PRIu64", fid=%s\n",
                   __FUNCTION__, bp->b_id, blkno, dump_fid (fid));

    /* Init data block data buffer */
    data = PASISFS_DCACHE_DATA (bp);

    /* Zero the block */
    memset (data->data, 0, PASIS_NFS_DATA_BLKSIZE);

    /* Mark buffer as dirty */
    PasisFS_BufCacheMarkDirty (bp);

    /* Update attributes */
    attr->blocks++;
    pasis_md_attr_set_time (attr->mtime);

    /* Insert buffer into the cache */
    PasisFS_DataCacheInsert (fid, loc, blkno, bp);

    return bp;
}


/* Gets a data block.
 * Returns the buffer pointer to the block, possibly creating it.
 */
static pasisfs_buf_t *
FileGetDataBlock (pasis_md_fid_t   *fid,
                  pasis_md_loc_t   *loc,
                  pasis_md_attr_t  *attr,
                  pasisfs_blkaddr_t blkno,
                  int               create_flag)
{
    pasisfs_buf_t *bp;
    pasisfs_blkaddr_t max_blk=0;

    PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 3, "%s: data block=%"PRIu64", create=%d\n",
                   __FUNCTION__, blkno, create_flag);

    /* If the block requested isn't off the end of the file issue a data cache
     * read.  Since no point in issuing read for block that can't exist.
     */
    max_blk = attr->size >> PASIS_NFS_DATA_BLOCK_SHIFT;
    if ((attr->size % PASIS_NFS_DATA_BLKSIZE) > 0) {
        max_blk++;
    }

    if (blkno < max_blk) {

        /* First check data cache for block */
        bp = PasisFS_DataCacheRead (fid, loc, blkno);
        if (bp) {
            PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 3, "%s: data cache read\n",
                           __FUNCTION__);
            return bp;
        }
    } else {
        PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 3, "%s: block doesn't exist\n",
                       __FUNCTION__);
    }

    if (create_flag) {
        PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 3, "%s: creating block\n",
                       __FUNCTION__);

        /* Check that the block isn't already cached -- it may be due to
         * a truncate that occured.  If it is in the cache, zero it and
         * return it.
         */
        if ((bp = PasisFS_DataCacheLookup (fid, blkno)) != NULL) {
            pasisfs_dcache_data_t *data;
            /* Init data block data buffer */
            data = PASISFS_DCACHE_DATA (bp);
            memset (data->data, 0, PASIS_NFS_DATA_BLKSIZE);

            return bp;
        }

        bp = FileCreateDataBlock (fid, loc, attr, blkno);
    } else {
        return NULL;
    }

    return bp;
}


pasis_error_t
PasisFS_FileWrite (pasis_md_fid_t *fid,
                   pasis_md_loc_t *loc,
                   pasis_md_attr_t *attr,
                   unsigned int len,
                   pasisfs_blkaddr_t offset,
                   p_byte *buffer,
                   unsigned int *len_out)
{
    pasisfs_buf_t *bp;
    pasisfs_blkaddr_t blkno;
    p_byte *buffer_ptr;
    unsigned int blk_off, blk_len, to_write=len;
    p_uint64 current_offset;

    if (MD_ATTR_IS_DIR (attr)) {
        return PASIS_ERROR_FILE_ISDIR;
    }

    if (!MD_ATTR_IS_REG (attr)) {
        return PASIS_ERROR_FILE_NOTREG;
    }

    current_offset = offset;
    buffer_ptr = buffer;
    blk_off = offset % PASIS_NFS_DATA_BLKSIZE;
    blkno = offset / PASIS_NFS_DATA_BLKSIZE;
 
    *len_out = to_write;

    while (to_write > 0) {

        blk_len = P_MIN (to_write, PASIS_NFS_DATA_BLKSIZE - blk_off);
        PASIS_DASSERT (blk_len <= len);
        
        PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 5, 
                       "%s: blk=%"PRIu64", off=%u, attr->size = %"PRIu64"\n",
                       __FUNCTION__, blkno, blk_off, attr->size);
        
        /* Need to get a buffer */
        /* Need to do a read even if on overwrite, so that we can tell if the
         * block already existed or not; if not, we have to increment block
         * count - which PasisFS_FileGetDataBlock does */
        bp = FileGetDataBlock (fid, loc, attr, blkno, 1);

        /* Copy data into data cache buffer */
        memcpy ((PASISFS_DCACHE_DATA(bp)->data) + blk_off, 
                buffer_ptr, blk_len);

        PasisFS_BufCacheMarkDirty (bp);

        /* Make it write through */
        PasisFS_DataCacheWrite (bp);

        current_offset += blk_len;
        attr->size = MAX (attr->size, current_offset);
        blk_off = 0;
        to_write -= blk_len;
        buffer_ptr += blk_len;
        blkno++;

    }
    
    pasis_md_attr_set_time (attr->mtime);

    return PASIS_SUCCESS;
}


pasis_error_t
PasisFS_FileRead (pasis_md_fid_t  *fid,
                  pasis_md_loc_t  *loc,
                  pasis_md_attr_t *attr,
                  unsigned int len,
                  pasisfs_blkaddr_t offset,
                  p_byte *buffer,
                  unsigned int *len_out)
{
    pasisfs_buf_t *bp;
    pasisfs_blkaddr_t blkno;
    p_byte *buffer_ptr;
    unsigned int blk_off;
    unsigned int blk_len;
    unsigned int to_read=len;

    if (MD_ATTR_IS_DIR (attr)) {
        return PASIS_ERROR_FILE_ISDIR;
    }

    if (!MD_ATTR_IS_REG (attr)) {
        return PASIS_ERROR_FILE_NOTREG;
    }

    if (attr->size <= offset) {
        *len_out = 0;
        return PASIS_SUCCESS;
    }

    if ((len + offset) > attr->size) {
        to_read = attr->size - offset;
    }

    buffer_ptr = buffer;
    blk_off = offset % PASIS_NFS_DATA_BLKSIZE;
    blkno = offset / PASIS_NFS_DATA_BLKSIZE;
 
    *len_out = to_read;

    while (to_read > 0) {
        blk_len = P_MIN (to_read, PASIS_NFS_DATA_BLKSIZE - blk_off);
        
        PASIS_DASSERT (blk_len <= len);

        /* Read and pin the buffer */
        bp = PasisFS_DataCacheRead (fid, loc, blkno);
        if (bp) {
            /* Copy data out of data cache */
            memcpy (buffer_ptr, 
                    (PASISFS_DCACHE_DATA(bp)->data) + blk_off, blk_len);
        } else {
            /* This is a sparse file; zero the data buffer. We're not polluting
             * the cache because PasisFS_DataCacheRead didn't create a block */
            memset (buffer_ptr, 0, blk_len);
        }
        
        blk_off = 0;
        to_read -= blk_len;
        buffer_ptr += blk_len;
        blkno++;
    }
    
    return PASIS_SUCCESS;
}




