/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include <map>
#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_types.h"
#include "pasisfs_fid_lock.h"


using namespace std;


struct LockData {
    LockData() : wait_count(0) {
        mutex = st_mutex_new();
        PASIS_DASSERT(mutex != NULL);
    }

    LockData(const LockData& other) : wait_count(0) {
        mutex = st_mutex_new();
        PASIS_DASSERT(mutex != NULL);
    }

    LockData& operator=(const LockData& other) {
        wait_count = 0;
        mutex = st_mutex_new();
        PASIS_DASSERT(mutex != NULL);
    }

    ~LockData() {
        PASIS_DASSERT(wait_count == 0);
        st_mutex_destroy(mutex);
    }

    st_mutex_t mutex;
    p_uint32   wait_count;
};


// For map
struct fid_lt
{
    bool operator()(const pasis_md_fid_t &f1, const pasis_md_fid_t &f2) const
    {
        return md_fid_compare(&f1, &f2) < 0;
    }
};


typedef map<pasis_md_fid_t, LockData, fid_lt> lock_map_t;
typedef lock_map_t::iterator                  lock_map_iter_t;


static lock_map_t lock_map;


static void
internal_lock(pasis_md_fid_t *fid,
              int             only_need_read_lock)
{
    LockData &ld = lock_map[*fid];
    ++ld.wait_count;
    st_mutex_lock(ld.mutex);
    // no fancy read/write locks for now
}


static void
internal_unlock(pasis_md_fid_t *fid)
{
    LockData &ld = lock_map[*fid];
    --ld.wait_count;
    st_mutex_unlock(ld.mutex);
    if (ld.wait_count == 0) {
        lock_map.erase(*fid);
    }
}



// For qsort
static int qsort_compar(const void *a, const void *b)
{
    pasis_md_fid_t *one = (pasis_md_fid_t*)a, *two = (pasis_md_fid_t*)b;
    return md_fid_compare(one, two);
}


/* Lock N objects.  Objects are sorted first, on OID, so we can do 2 phase
 * locking and avoid deadlocks. */
void
pasisfs_fid_lock (p_uint32            n,
                  pasis_md_fid_t    **fids,
                  int                 only_need_read_lock)
{
    pasis_md_fid_t *sorted_fids[16];

    /* Sort objects.  for consistency, don't change the order of the caller's
     * objects; hence, we need to work with our own copy of the pointers. */
    PASIS_DASSERT(n <= 16);
    memcpy(sorted_fids, fids, n * sizeof(pasis_md_fid_t*));
    qsort(sorted_fids, n, sizeof(pasis_md_fid_t*), qsort_compar);

    /* Lock each object */
    for (p_uint32 i = 0; i < n; ++i) {
        internal_lock(sorted_fids[i], only_need_read_lock);
    }
}


void
pasisfs_fid_unlock (p_uint32            n,
                    pasis_md_fid_t    **fids)
{
    for (p_uint32 i = 0; i < n; ++i) {
        internal_unlock(fids[i]);
    }
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
