/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_timer.h"

#include "pasisio.h"
#include "pasisio_md_protocol.h"
#include "pasis_md_prot.h"

#include "pasisfs_io.h"
#include "fe.h"

#include <sys/stat.h>
#include <sys/types.h>


#define TEST_DIR "/tmp/pasisfs"



/* Flag specifying if data should be stored on the local filesystem, as opposed
 * to using the PASIS I/O functions.  This is used during testing. */
static int store_data_locally_flag;

/* Flag specifying if data should be stored using Q/U synthetic objects */
static int store_data_using_qu;

typedef struct pasisfs_io_stats_s {
    unsigned int creates;
    unsigned int reads;
    unsigned int writes;
    unsigned int deletes;
    unsigned int read_blocks;
    unsigned int write_blocks;
    unsigned int max_blocks_outstanding;
    pasis_timer_state_t *timers;
} pasisfs_io_stats_t;


/* From nfs3_fe.c */
extern pasis_md_protocol_t  root_md_protocol;
extern unsigned int         root_num_nodes;
extern pasis_md_linkage_t  *root_nodes;


static pasisfs_io_stats_t statistics;


static pasis_error_t
io_read (pasis_md_fid_t *fid,
         p_uint64        blkno,
         p_byte         *buffer)
{
    char fid_fname[PASISIO_FILEID_SIZE * 2 + 1];
    off_t lseek_off;
    int fd, i, cnt;

    for (i = 0; i < PASISIO_FILEID_SIZE; i++) {
        snprintf (fid_fname+(2*i), 3, "%02X", (unsigned char)fid->data[i]);
    }

    fd = open (fid_fname, O_RDONLY);
    if (fd < 0) return errno;

    lseek_off = lseek (fd, 0, SEEK_END);
    PASIS_DPRINTF (PASIS_DEBUG_FS_IO, 5, "%s: eof at %"PRIu64", read at %"PRIu64"\n", 
                   __FUNCTION__, (p_uint64)lseek_off, 
                   blkno << PASIS_NFS_DATA_BLOCK_SHIFT);
    /* Check for reads past end of file -- return zeroed buffer */
    if (lseek_off <= (blkno << PASIS_NFS_DATA_BLOCK_SHIFT)) {
        memset (buffer, 0, PASIS_NFS_DATA_BLKSIZE);
        return PASIS_SUCCESS;
    }

    lseek_off = lseek (fd, blkno << PASIS_NFS_DATA_BLOCK_SHIFT, SEEK_SET);
    PASIS_DASSERT (lseek_off == blkno << PASIS_NFS_DATA_BLOCK_SHIFT);

    cnt = read (fd, buffer, PASIS_NFS_DATA_BLKSIZE);
    close (fd);

    if (cnt != PASIS_NFS_DATA_BLKSIZE) {
        PASIS_PRINT_ERR ("%s: Error: %s\n", __FUNCTION__, 
                         PASIS_STRERROR (errno));
        return errno;
    }

    return PASIS_SUCCESS;
}


static pasis_error_t
io_write (pasis_md_fid_t *fid,
          p_uint64        blkno,
          p_byte         *buffer)
{
    char fid_fname[PASISIO_FILEID_SIZE * 2 + 1];
    off_t lseek_off;
    int fd, i, cnt;

    for (i = 0; i < PASISIO_FILEID_SIZE; i++) {
        snprintf (fid_fname+(2*i), 3, "%02X", (unsigned char)fid->data[i]);
    }

    fd = open (fid_fname, O_WRONLY);
    if (fd < 0) {
        fd = open (fid_fname, O_CREAT | O_WRONLY | O_EXCL, 0666);
        if (fd < 0) {
            return errno;
        }
    }

    lseek_off = lseek (fd, blkno << PASIS_NFS_DATA_BLOCK_SHIFT, SEEK_SET);
    PASIS_DASSERT (lseek_off == blkno << PASIS_NFS_DATA_BLOCK_SHIFT);

    cnt = write (fd, buffer, PASIS_NFS_DATA_BLKSIZE);
    close (fd);

    if (cnt != PASIS_NFS_DATA_BLKSIZE) {
        PASIS_PRINT_ERR ("%s: Error: %s\n", __FUNCTION__, 
                         PASIS_STRERROR (errno));
        return errno;
    }

    return PASIS_SUCCESS;
}



static void
IOPrintStats (void)
{
    PASIS_PRINT_INFO ("---- IO Statistics ----\n");
    PASIS_PRINT_INFO ("Number of Create IOs  :  %u\n", statistics.creates);
    PASIS_PRINT_INFO ("Number of Read IOs    :  %u\n", statistics.reads);
    PASIS_PRINT_INFO ("Number of Write IOs   :  %u\n", statistics.writes);
    PASIS_PRINT_INFO ("Number of Delete IOs  :  %u\n", statistics.deletes);
    PASIS_PRINT_INFO ("Blocks Read           :  %u\n", statistics.read_blocks);
    PASIS_PRINT_INFO ("Blocks Written        :  %u\n", 
                      statistics.write_blocks);
    PASIS_PRINT_INFO ("Max Blocks Outstanding:  %u\n", 
                      statistics.max_blocks_outstanding);
}


pasis_error_t
PasisFS_IOInit (int _store_data_locally_flag, int _store_data_using_qu)
{
    memset (&statistics, 0, sizeof (pasisfs_io_stats_t));

    statistics.timers = Pasis_TimerInit ("PasisFS IO Timers", 2, 0);
    Pasis_TimerRegister (statistics.timers, 0, "Read");
    Pasis_TimerRegister (statistics.timers, 1, "Write");

    store_data_locally_flag = _store_data_locally_flag;
    store_data_using_qu = _store_data_using_qu;
    if (store_data_locally_flag) {
        mkdir (TEST_DIR, 0777);
        if (chdir (TEST_DIR) < 0) {
            PASIS_PRINT_ERR ("%s: error changing to testdir %s\n", 
                             __FUNCTION__, TEST_DIR);
            return PASIS_ERROR;
        }
    }

    return PASIS_SUCCESS;
}

void 
PasisFS_IOShutdown (void)
{
    IOPrintStats ();
    Pasis_TimerShutdown (statistics.timers);
}

/* Inode contains FID, attributes, current uid, and loc_t metadata */
pasis_error_t
PasisFS_IORead (pasis_md_fid_t  *fid,
                pasis_md_loc_t  *loc,
                p_uint64         blkno,
                p_byte          *buffer)
{
    pasis_error_t      error;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_FS_IO, 3, "%s: %"PRIu64"\n", __FUNCTION__, blkno);

    if (store_data_locally_flag) {
        error = io_read (fid, blkno, buffer);
    } else if (store_data_using_qu) {
        p_uint32 read_bitmap = 1 << blkno;
        pasis_md_fid_t new_fid, *new_fid_ptr = &new_fid;
        PASIS_DASSERT(blkno < 32);

        /* Use the synethtic objects for storage */
        /* need to use a different fid than the metadata op; i'm just gonna shit
         * everything to the left by one byte, and hope there's no
        collisions. -mabdelmalek, 8/12/2005 */
        memmove(new_fid.data, fid->data + 1, sizeof(pasis_md_fid_t) - 1);
        memset(new_fid.data + sizeof(pasis_md_fid_t) - 1, 0, 1);

        error = PasisIO_MDSynthQuery (&root_md_protocol, &new_fid_ptr, 1,
                                      root_nodes, root_num_nodes,
                                      0, PASIS_NFS_DATA_BLKSIZE,
                                      PASIS_NFS_DATA_BLKSIZE,
                                      &read_bitmap, 0, 1, 0, buffer);
    } else {
        /* Support for using PASIS R/W protocol not in Q/U release.*/
        PASIS_DASSERT(0);
    }

    statistics.reads++;
    statistics.read_blocks++;

    PASIS_TIMER_STOP (statistics.timers, 0);

    return error;
}

pasis_error_t 
PasisFS_IOWrite (pasis_md_fid_t      *fid,
                 pasis_md_loc_t      *loc,
                 p_uint64             blkno,
                 p_byte              *buffer)
{
    pasis_error_t      error;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_FS_IO, 3, "%s: %"PRIu64"\n", __FUNCTION__, blkno);

    if (store_data_locally_flag) {
        error = io_write (fid, blkno, buffer);
    } else if (store_data_using_qu) {
        p_uint32 write_bitmap = 1 << blkno;
        pasis_md_fid_t new_fid, *new_fid_ptr = &new_fid;

        PASIS_DASSERT(blkno < 32);

        /* Use the synethtic objects for storage */
        /* need to use a different fid than the metadata op; i'm just gonna shift
         * everything to the left by one byte, and hope there's no
        collisions. -mabdelmalek, 8/12/2005 */
        memmove(new_fid.data, fid->data + 1, sizeof(pasis_md_fid_t) - 1);
        memset(new_fid.data + sizeof(pasis_md_fid_t) - 1, 0, 1);

        /* Use the synethtic objects for storage */
        error = PasisIO_MDSynthUpdate (&root_md_protocol, &new_fid_ptr, 1,
                                       root_nodes, root_num_nodes,
                                       PASIS_NFS_DATA_BLKSIZE, 0,
                                       PASIS_NFS_DATA_BLKSIZE,
                                       &write_bitmap, 0, 1, 1, 0, buffer);
    } else {
        /* Support for using PASIS R/W protocol not in Q/U release.*/
        PASIS_DASSERT(0);
    }

    statistics.writes++;
    statistics.write_blocks++;

    PASIS_TIMER_STOP (statistics.timers, 1);

    return error;
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */ 
