/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#ifndef __PASISIO_DRIVES_H__
#define __PASISIO_DRIVES_H__

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_debug.h"
#include "pasis_lists.h"
#include "pasisio.h"
#include "pasisio_internal.h"

#ifdef __cplusplus
extern "C" {
#endif
#if 0
}
#endif


/* List of drive types and their index -- do not change, only add to */

#define PASISIO_DRIVE_TYPE_NUM            1

#define PASISIO_DRIVE_TYPE_ANY_INDEX      0xFF          /**< Reserved */

#define PASISIO_DRIVE_TYPE_MD_NAME        "md"
#define PASISIO_DRIVE_TYPE_MD_INDEX       0x00

/** Maximum number of servers per drive type */
#define PASISIO_DRIVE_MAXINSTANCES        1024

/** Maximum path name length of path in server list */
#define PASISIO_DRIVE_MAXPATHLEN          255

/** Timeout in seconds to check if drive instance is still alive */
#define PASISIO_DRIVE_ALIVE_TIMEOUT       1


/********************************* Internal **********************************/

/** Drive operation types */
typedef enum pasisio_drive_op_e {
    PASISIO_DRIVE_OP_INVALID = 0x00,
    PASISIO_DRIVE_OP_CREATE  = 0x01,
    PASISIO_DRIVE_OP_MKDIR   = 0x02,
    PASISIO_DRIVE_OP_READ    = 0x03,
    PASISIO_DRIVE_OP_WRITE   = 0x04,
    PASISIO_DRIVE_OP_OWRITE  = 0x05,
    PASISIO_DRIVE_OP_UNLINK  = 0x06,
    PASISIO_DRIVE_OP_RMDIR   = 0x07,

    /** Reserved OP codes -- used to pass drive type specific op codes */
    PASISIO_DRIVE_OP_RESV1   = 0x08,
    PASISIO_DRIVE_OP_RESV2   = 0x09,
    PASISIO_DRIVE_OP_RESV3   = 0x0a,
    PASISIO_DRIVE_OP_RESV4   = 0x0b,
    PASISIO_DRIVE_OP_RESV5   = 0x0c,
    PASISIO_DRIVE_OP_RESV6   = 0x0d,
    PASISIO_DRIVE_OP_RESV7   = 0x0e,
    PASISIO_DRIVE_OP_RESV8   = 0x0f,
} pasisio_drive_op_t;

/* Drive error codes */
#define PASISIO_DRIVE_REQ_SUCCESS  0x00 /**< Success */
#define PASISIO_DRIVE_REQ_SERVERR  0x01 /**< Server error -- unreachable */
#define PASISIO_DRIVE_REQ_OPERERR  0x02 /**< Operation error */
#define PASISIO_DRIVE_REQ_TERM     0x03 /**< Request terminated */

/** Drive statuses */
typedef enum pasisio_drive_status_e {
 PASISIO_DRIVE_STATUS_UP   = 0x00,  /**< Drive is available */
 PASISIO_DRIVE_STATUS_DOWN = 0x01   /**< Drive is unavailable */
} pasisio_drive_status_t;

struct pasisio_select_s; /* To avoid circular definitions... */

typedef struct pasisio_drive_req_state_s {
    st_thread_t   thread_id;
    int           early_terminate;
} pasisio_drive_req_state_t;

/** Drive request header structure -- tracks drive requests */
typedef struct pasisio_drive_req_hdr_s {
    
    pasisio_drive_req_state_t req_state[PASIS_MAX_NUM_SERVERS];
    unsigned int              early_term_threshold;

    unsigned int              successful;
    unsigned int              completed;
    unsigned int              issued;
    st_cond_t                 cond;
    int                       signal;

    /** Parent request */
    pasisio_request_t       *parent;

} pasisio_drive_req_hdr_t;


/** Drive request for a single share */
typedef struct pasisio_drive_req_s {
    pasisio_drive_op_t      op_type;
    pasis_error_t           status;

    /** Share information */
    unsigned int            share_index;            

    p_uint64                block_num;

    /** Higher level request */
    pasisio_drive_req_hdr_t *req_hdr;

    PASIS_Q_PTRS(struct pasisio_drive_req_s) queue;
} pasisio_drive_req_t;

typedef PASIS_Q_STRUCT(drive_req_q_s, pasisio_drive_req_t) 
     pasisio_drive_req_queue_t;


struct pasisio_drive_instance_s;


/** Drive type */
typedef struct pasisio_drive_type_s {
    char            *type_name;
    unsigned int     type_index;

    struct pasisio_select_s *select_instances;

    pasis_error_t (*req_handler) (struct pasisio_drive_instance_s *,
                                  pasisio_drive_req_t *);

    pasis_error_t (*create_instance) (p_uint32 ip, char *base_path,
                                      struct pasisio_drive_instance_s **);

    void (*destroy_instance) (struct pasisio_drive_instance_s *);

    void (*req_terminate) (struct pasisio_drive_instance_s *, st_thread_t);
} pasisio_drive_type_t;


/** Drive worker thread */
typedef struct pasisio_drive_thread_s {
    struct pasisio_drive_instance_s  *instance;
    pasisio_drive_req_t              *req;
    st_cond_t                         cond;
    st_thread_t                       thread_id;
    int                               id;
    int                               signal;
    PASIS_Q_PTRS(struct pasisio_drive_thread_s) queue;    
} pasisio_drive_thread_t;

typedef PASIS_Q_STRUCT(drive_thread_q_s, pasisio_drive_thread_t) 
     pasisio_drive_thread_queue_t;

/** Drive instance */
typedef struct pasisio_drive_instance_s {
    pasisio_drive_status_t     status;

    p_uint32                   ip_address;
    char                       base_path[PASISIO_DRIVE_MAXPATHLEN];
    void                      *priv;
    pasisio_drive_type_t      *drive;

    unsigned int               dispatch_shutdown;
    unsigned int               dispatch_count;
    pasisio_drive_req_queue_t  dispatch_req_queue;

    unsigned int                  threads_in_prog;
    unsigned int                  reqs_queued;

    pasisio_drive_thread_queue_t  thread_queue;
    pasisio_drive_thread_t       *thread_array;
    unsigned int                  thread_num;

    /* Used during initialization of server connection, to prevent race
     * condition.  Common pasisio_drive code initializes these variables, and
     * each drive code is responsible for reseting init_complete and
     * st_cond_broadcast()ing to the init_cond when it's finished with the
     * initialization.  */
    int                        init_complete;
    st_cond_t                  init_cond;

    PASIS_LL_PTRS (list, struct pasisio_drive_instance_s);
} pasisio_drive_instance_t;


void *
PasisIO_DriveQueueThread (void *arg);


int
PasisIO_DriveEnqueueShares (pasisio_drive_op_t op, pasisio_request_t *req, 
                            unsigned int to_issue, unsigned int early_term);


pasisio_drive_instance_t *
PasisIO_DriveAddInstance (unsigned int  type_index,
                          p_uint32      ip_address,
                          char         *base_path);

pasisio_drive_type_t *
PasisIO_DriveFindTypeName (char *type_name);


pasisio_drive_type_t *
PasisIO_DriveFindTypeIndex (unsigned int type_index);


pasisio_drive_instance_t *
PasisIO_DriveFindInstanceName (unsigned int  type_index,
                               char         *hostname,
                               char         *base_path);

pasisio_drive_instance_t *
PasisIO_DriveFindInstanceIP (unsigned int type_index,
                             p_uint32     ip_address,
                             char        *base_path);



struct pasisio_select_s *
PasisIO_DriveGetSelectInstances (unsigned int type);

pasisio_drive_type_t *
PasisIO_DriveCreateType (unsigned int type,
                         char *type_name,
                         pasis_error_t (*req_handler) 
                         (pasisio_drive_instance_t *, pasisio_drive_req_t *),
                         pasis_error_t (*create_instance) (p_uint32, char *,
                                                           pasisio_drive_instance_t **),
                         void (*destroy_instance) (pasisio_drive_instance_t *),
                         void (*req_terminate) (pasisio_drive_instance_t *,
                                                st_thread_t));
                         

void
PasisIO_DriveDestroyType (pasisio_drive_type_t *type);                         


pasisio_drive_instance_t *
PasisIO_DriveCreateInstance (pasisio_drive_type_t *type,
                             p_uint32 ip_address,
                             char *base_path,
                             unsigned int worker_threads);

void
PasisIO_DriveDestroyInstance (pasisio_drive_instance_t *instance);

pasis_error_t
PasisIO_DriveInit (unsigned int max_outstanding);


void
PasisIO_DriveShutdown (void);


#ifdef __cplusplus
}
#endif


#endif /* __PASISIO_DRIVES_H__ */


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */ 
