/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include "pasisio_config_parse.h"
#include "pasisio_select.h"
#include "pasisio_md_protocol.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


static int line_num;


static inline char *
read_line (FILE *file, char *buffer)
{
    while (1) {
        line_num++;
        if (fgets (buffer, 255, file) == NULL) {
            return NULL;
        }
        if (buffer[0] == '#' || buffer[0] == '\n') continue;
        break;
    }
    return buffer;
}

#define STR(s) #s

#define PARSE_INPUT(file, buffer, format_string, var)                \
do {                                                                 \
    int rc;                                                          \
    if (read_line (file, buffer) == NULL) {                          \
        buffer[0] = 0; /* signal end of file */                      \
        goto error;                                                  \
    }                                                                \
    rc = sscanf (buffer, format_string, var);                        \
    if (rc == EOF || rc == 0) {                                      \
        fprintf (stderr, "PARSE_INPUT: sscanf returned %d\n", rc);   \
        goto error;                                                  \
    }                                                                \
} while (0)



/* Input: B, T, m, sync, repair, cond are set in the input protocol */
/* Output: full protocol filled in */
static void
PasisIO_ProtocolParamsInit (pasis_md_protocol_t *protocol)
{
    int N, Qc;

    Qc = 2 * protocol->T + protocol->B + 1;
    N = Qc + protocol->T + protocol->B;

    /* For threshold quorums, delta is added to Qc and N in
     * PasisIO_QuorumParamsInit below. */

    protocol->N = N;
    protocol->Qc = Qc;
}


static pasis_error_t
parse_protocol (FILE *file, pasis_md_protocol_t *protocol)
{
    char         buffer[255], cond_type_str[64];
    unsigned int u_tmp;

    PARSE_INPUT (file, buffer, "B=%u", &u_tmp);
    protocol->B = u_tmp;

    PARSE_INPUT (file, buffer, "T=%u", &u_tmp);
    protocol->T = u_tmp;

    PasisIO_ProtocolParamsInit (protocol);

    return PASIS_SUCCESS;

 error:
    PASIS_PRINT_ERR ("%s: parsing error, line=%d\n", __FUNCTION__, line_num);
    return PASIS_ERROR;
}



/* Input: B, T, delta, quorum_type are set in the input protocol.
 * PasisIO_ProtocolParamsInit has already set Qc, N. */
/* Output: full protocol filled in */
static void
PasisIO_QuorumParamsInit (pasis_md_protocol_t *protocol)
{
    unsigned int repairable, i;

    switch (protocol->quorum_type) {
    case PASISIO_QUORUM_NONE:
        break;
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B_T:
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B:
        protocol->Qc += 2 * protocol->delta;
        protocol->N  += 3 * protocol->delta;

        repairable = PasisIO_GetRepairableThreshold (protocol);
        printf ("Actually, protocol using threshold quorums: "
                "N=%u, Qc=%u, |R| = %u, b=%u, t=%u, delta=%u\n", protocol->N,
                protocol->Qc, repairable, protocol->B, protocol->T, protocol->delta);
        break;
    case PASISIO_QUORUM_TREE:
        /* For t=b=1, N=6, Q=24.  Trust us for now; this needs to be cleaned up
         * later.
         * The current mess below is so we can fit in with the existing code
         * that sends to Qc+b S-Ns */
        protocol->Qc += protocol->B;
        for (i = 0; i < protocol->delta; ++i) {
            protocol->Qc *= protocol->Qc;
            protocol->N  *= protocol->N;
        }
        protocol->Qc -= protocol->B;
        printf ("Actually, protocol using tree quorums: "
                "N=%u, Qc=%u, b=%u, t=%u, delta=%u\n", protocol->N,
                protocol->Qc, protocol->B, protocol->T, protocol->delta);
        break;
    default:
        PASIS_DASSERT (0);
    }
}



static pasis_error_t
parse_quorums (FILE *file, pasis_md_protocol_t *protocol)
{
    char         buffer[255], quorum_type_str[64];
    long         file_offset;

    protocol->quorum_type = PASISIO_QUORUM_NONE;

    /* Remember the current file offset; we may have to rewind back if there was
     * no quorum section. */
    file_offset = ftell (file);

    PARSE_INPUT (file, buffer, "QUORUM=%s", quorum_type_str);
    if (!strncasecmp (quorum_type_str, "none", 63)) {
        protocol->quorum_type = PASISIO_QUORUM_NONE;
    } else if (!strncasecmp (quorum_type_str, "threshold_comp_qc_b_t", 63)) {
        protocol->quorum_type = PASISIO_QUORUM_THRESHOLD_COMP_QC_B_T;
    } else if (!strncasecmp (quorum_type_str, "threshold_comp_qc_b", 63)) {
        protocol->quorum_type = PASISIO_QUORUM_THRESHOLD_COMP_QC_B;
    } else if (!strncasecmp (quorum_type_str, "tree", 63)) {
        protocol->quorum_type = PASISIO_QUORUM_TREE;
    } else {
        goto error;
    }

    switch (protocol->quorum_type) {
    case PASISIO_QUORUM_NONE:
        break;
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B_T:
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B:
    case PASISIO_QUORUM_TREE: {
        unsigned int delta;
        PARSE_INPUT (file, buffer, "DELTA=%u", &delta);
        protocol->delta = delta;
        break;
    }
    default:
        PASIS_DASSERT (0);
    }

    PasisIO_QuorumParamsInit (protocol);

    return PASIS_SUCCESS;

 error:
    fseek (file, file_offset, SEEK_SET);

    // PASIS_PRINT_ERR ("%s: didn't find a QUORUM option; not using quorums\n",
    //                  __FUNCTION__);
    // always return success so we're backward-compatible with old PASIS MD
    // configuration files that don't have a quorum section.  For these
    // configuration files, we use the regular write-all, read-some threshold
    // quorum.
    protocol->quorum_type = PASISIO_QUORUM_NONE;
    return PASIS_SUCCESS;
}


static pasis_error_t
parse_server_list (FILE *file, pasis_config_server_list_t *server_list)
{
    char buffer[255];
    unsigned int i, num_servers;

    /* Read server list */
    PARSE_INPUT (file, buffer, "NUM_SERVERS=%u", &num_servers);

    server_list->num_servers = num_servers;
    server_list->servers = (pasis_config_server_t *)
        malloc (num_servers * sizeof(pasis_config_server_t));

    for (i = 0; i < num_servers; i++) {
        char *base_path;
        struct hostent *hp;

        /* Read server name and base path (i.e., partition) */
        PARSE_INPUT (file, buffer, "%254s", buffer);
        
        /* Extract base path */
        base_path = strchr (buffer, ':');
        if (base_path == NULL) {
            goto error;
        }
        *base_path = '\0';
        base_path++;

        /* Convert hostname to IP, and store in server list */
        hp = gethostbyname (buffer);
        if (hp == NULL) {
            PASIS_PRINT_ERR("%s: gethostbyname failed for host %s with error %d\n",
                            __FUNCTION__, buffer, h_errno);
            return PASIS_ERROR;
        }

        PASIS_DASSERT (hp->h_length == 4);

        server_list->servers[i].ip_address = *(p_uint32 *)hp->h_addr;

        /* Store base path in server list */
        strncpy (server_list->servers[i].base_path, base_path,
                 sizeof(server_list->servers[i].base_path));
    }

    return PASIS_SUCCESS;

 error:
    PASIS_PRINT_ERR ("%s: parsing error, line=%d\n", __FUNCTION__, line_num);
    return PASIS_ERROR;
}



pasis_error_t
PasisIO_ParseConfigurationFile (char                       *filename,
                                pasis_md_protocol_t        *protocol,
                                pasis_config_server_list_t *server_list,
                                long                       *file_offset,
                                int                        *_line_num)
{
    FILE          *file;
    pasis_error_t  error;

    file = fopen (filename, "r");
    if (file == NULL) {
        PASIS_PRINT_ERR ("%s: error: file open failed (%s, %s)\n",
                         __FUNCTION__, filename, PASIS_STRERROR (errno));
        exit (1);
    }

    line_num = 0;

    error = parse_server_list(file, server_list);
    if (error) {
        fclose (file);
        return error;
    }
    
    error = parse_protocol (file, protocol);
    if (error) {
        fclose (file);
        return error;
    }

    error = parse_quorums (file, protocol);
    if (error) {
        fclose (file);
        return error;
    }

    PASIS_DASSERT (server_list->num_servers >= protocol->N  &&
                   server_list->num_servers <= PASIS_MAX_NUM_SERVERS);

    if (file_offset) {
        *file_offset = ftell (file);
    }

    if (_line_num) {
        *_line_num = line_num;
    }

    fclose (file);
    return PASIS_SUCCESS;
}


pasis_error_t
PasisIO_ConnectToStorageNodes (pasis_config_server_list_t *server_list,
                               unsigned int                drive_type_index,
                               pasisio_linkage_t          *linkages,
                               pasisio_drive_instance_t  **instances_in)
{
    unsigned int i;
    pasisio_select_t select;
    pasisio_drive_instance_t *instances_local[PASIS_MAX_NUM_SERVERS];
    pasisio_drive_instance_t **instances;

    if (instances_in) {
        instances = instances_in;
    } else {
        instances = instances_local;
    }

    for (i = 0; i < server_list->num_servers; i++) {
        instances[i] = PasisIO_DriveFindInstanceIP (drive_type_index,
                                           server_list->servers[i].ip_address,
                                           server_list->servers[i].base_path);
                                                       
        if (instances[i] == NULL) {
            PASIS_PRINT_ERR ("%s: error: server not found (%s:%s)\n",
                     __FUNCTION__,
                     pasis_ip_to_hostname (server_list->servers[i].ip_address),
                     server_list->servers[i].base_path);
            return PASIS_ERROR;
        }
    }

    if (linkages) {
        select.count = server_list->num_servers;
        select.instances = instances;
        PasisIO_SelectEncodeLinkages (linkages, server_list->num_servers,
                                      &select);
    }

    return PASIS_SUCCESS;
}


void
PasisIO_ServerListToLinkageRecords (pasis_config_server_list_t *server_list,
                                    pasisio_linkage_t          *linkages)
{
    unsigned int i;
    
    for (i = 0; i < server_list->num_servers; i++) {
        linkages[i].ip_address = server_list->servers[i].ip_address;
        sscanf (server_list->servers[i].base_path, "/%u",
                &linkages[i].partition);
    }
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
