/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_lists.h"

#include "pasisio.h"
#include "pasisio_drives.h"
#include "pasisio_select.h"
#include "pasisio_config_parse.h"
#include "pasisio_md_protocol.h"

#define MAX_LOOP 50 /* To prevent infinite loops when selecting a server */



static inline pasisio_drive_instance_t *
selectRandom (pasisio_select_t *list)
{
    long int rnum;
    pasisio_drive_instance_t *instance;

    do {
        rnum = random ();
        rnum %= list->count;

        instance = list->instances [rnum];
    } while (instance->status != PASISIO_DRIVE_STATUS_UP);

    return instance;
}


static inline pasisio_drive_instance_t *
selectRoundRobin (pasisio_select_t *list)
{
    static int last_index = -1;
    pasisio_drive_instance_t *instance;

    do {
        last_index++;
        last_index %= list->count;
        
        instance = list->instances [last_index];

    } while (instance->status != PASISIO_DRIVE_STATUS_UP);
    
    return instance;
}


static inline int
selectCheckDuplicates (pasisio_drive_instance_t *instance,
                       pasisio_select_t *list,
                       int count)
{
    int i;

    for (i = 0; i < count; i++) {
        if (list->instances[i] == instance) {
            return 1;
        }
    }
    
    return 0;
}

int
PasisIO_ReselectServers (unsigned int      select_num,
                         int              *select_mask,
                         pasisio_select_t *select_list,
                         pasisio_select_mode_t type,
                         unsigned int instance_type)
{
    pasisio_select_t *instance_list;
    int i, reselected=0;

    instance_list = PasisIO_DriveGetSelectInstances (instance_type);

    for (i = 0; i < select_num; i++) {
        pasisio_drive_instance_t *instance = NULL;
        int check, loop_iters=0;

        /* Leave this entry alone */
        if (select_mask[i] == 0) {
            continue;
        }

        /* Otherwise pick new entry */
        do {
            switch (type & 0x0F) {
            case PASISIO_SELECT_RANDOM:
                instance = selectRandom(instance_list);
                break;
            case PASISIO_SELECT_RNDRBN:
                instance = selectRoundRobin (instance_list);
                break;
            }
            
            check = selectCheckDuplicates (instance, select_list, i);

            if (++loop_iters > MAX_LOOP) {
                /* Failed -- no more servers to pick from */
                PASIS_PRINT_ERR ("Error: server selection failed\n");
                return reselected;
            }

        } while (check == 1);
        
        instance_list->instances[i] = instance;
        instance_list->flags[i] = PASISIO_SELECT_IO_READY;

        reselected++;
    }

    return reselected;
}


int
PasisIO_SelectNewServers (unsigned int select_num,
                          pasisio_select_t *out_list,
                          pasisio_select_t *hint_list,
                          pasisio_select_mode_t type,
                          unsigned int instance_type)
{
    pasisio_select_t *instance_list;
    int i, iters=0, selected=0;

    instance_list = PasisIO_DriveGetSelectInstances (instance_type);

    for (i = 0; i < select_num; i++) {
        pasisio_drive_instance_t *instance = NULL;
        int check, hint_found=0, loop_iters=0;

        /* First check the hint list */
        if ((hint_list != NULL) && (hint_list->count > i) &&
            (hint_list->instances[i]->status == PASISIO_DRIVE_STATUS_UP)) {
            instance = hint_list->instances[i];
            hint_found = 1;
        }
            
        /* Make sure if a hint was found it isn't a duplicate, if no hint was
           found select new instance */
        do {

            if (!hint_found) {
                switch (type & 0x0F) {
                case PASISIO_SELECT_RANDOM:
                    instance = selectRandom(instance_list);
                    break;
                case PASISIO_SELECT_RNDRBN:
                    instance = selectRoundRobin (instance_list);
                    break;
                }
            }
            
            check = selectCheckDuplicates (instance, out_list, i);
            /* If we go through loop again select new instance -- regardless */
            hint_found = 0; 
            iters++;
 
            if (++loop_iters > MAX_LOOP) {
                /* Failed */
                PASIS_PRINT_ERR ("Error: server selection failed\n");
                out_list->count = selected;
                return selected;
            }

        } while (check == 1);
        
        out_list->instances[i] = instance;
        out_list->flags[i] = PASISIO_SELECT_IO_READY;

        selected++;
    }

    PASIS_DPRINTF (PASIS_DEBUG_IO_SEL, 4, 
                   "Server selection complete: %d iters for %d servers\n",
                   iters, i);

    out_list->count = selected;

    return selected;
}


int
PasisIO_SelectServers (int num,
                       pasisio_select_t *list)
{
    int i, count=0, shuffled_idx=0;
    char buf[255];

    /* For the servers we're not using, set the drive flags so that we don't
     * issue a request to them. */
    memset(list->flags, PASISIO_SELECT_IO_IGNORE,
           list->count * sizeof(pasisio_select_flags_t));

    for (i = 0; i < list->count && count < num; i++) {
        shuffled_idx = (i + list->shuffle) % list->count;

        if ((list->flags[shuffled_idx] == PASISIO_SELECT_IO_READY ||
             list->flags[shuffled_idx] == PASISIO_SELECT_IO_RETRY ||
             list->flags[shuffled_idx] == PASISIO_SELECT_IO_IGNORE) &&
            (list->instances[shuffled_idx] &&
             list->instances[shuffled_idx]->status == PASISIO_DRIVE_STATUS_UP)) {
            list->flags[shuffled_idx] = PASISIO_SELECT_IO_ISSUE;
            count++;
        }
    }

    PASIS_DPRINTF (PASIS_DEBUG_IO_SEL, 4, "Select: select shares, got %d/%d\n",
                   count, num);

    return count;
}


void
PasisIO_SelectDecodeLinkages (pasisio_linkage_t *linkages,
                              unsigned int linkage_count,
                              unsigned int type_index,
                              pasisio_select_t *list,
                              int shuffle)
{
    char base_path[64];
    unsigned int curr_instance, found=0;
    int i;
    
    curr_instance = shuffle % linkage_count;

    memset (list->flags, 0, sizeof (pasisio_select_flags_t) * PASIS_MAX_NUM_SERVERS);

    for (i = 0; i < linkage_count; i++) {
        if (curr_instance >= list->count) break;

        snprintf (base_path, 64, "/%u", linkages[i].partition);

        list->instances[curr_instance] = 
            PasisIO_DriveFindInstanceIP (type_index, linkages[i].ip_address, 
                                         base_path);
                                                                 
         if (list->instances[curr_instance] == NULL) {
            PASIS_DPRINTF (PASIS_DEBUG_IO_SEL, 4, 
                           "Select: decode linkage failed to find instance\n");
            continue;
         }
         
         list->flags[curr_instance] = PASISIO_SELECT_IO_READY;

         curr_instance = (curr_instance + 1) % linkage_count;
         found++;
    }
    
    list->count = found;
}


static void
select_servers_load_balanced (pasisio_linkage_t *linkages,
                              unsigned int total_linkage_count, 
                              unsigned int desired_count, 
                              unsigned int type_index,
                              pasisio_select_t *list,
                              int list_offset,
                              int shuffle)
{
    char base_path[64];
    unsigned int found =0 ;
    unsigned int num_passes = 0;
    int i, curr_instance;

#if 1
    /* For load balancing across S-N's, we select the first server in a
     * threshold quorums according to:
     * First server = (Quorum ID) * (Quorum Size) % (Number of Servers) 
     * Where "Quorum ID" is just the lower bits of the object ID. */
    curr_instance = (shuffle * desired_count) % total_linkage_count;
#else

    /* make all clients use the first quorum */
    curr_instance = 0;

#endif

    /* iterate over _all_ servers, and not just |Q|, so in the case of failures
     * we have entries for the other drives.  PasisIO_SelectServers will use the
     * "shuffle" variable our caller saved in order to do the actual shuffling.
     * -mabdelmalek 25/3/2005. */
    for (i = 0; i < total_linkage_count; i++) {
        if (++num_passes > total_linkage_count) break; /* enumerated all S-N's */

        snprintf (base_path, 64, "/%u",
                  linkages[list_offset + curr_instance].partition);

        list->instances[list_offset + curr_instance] = 
            PasisIO_DriveFindInstanceIP (type_index,
                               linkages[list_offset + curr_instance].ip_address,
                               base_path);
                                                                 
         if (list->instances[list_offset + curr_instance] == NULL) {
            PASIS_DPRINTF (PASIS_DEBUG_IO_SEL, 4, 
                           "Select: decode linkage failed to find instance\n");
            --i;
         } else {
             list->flags[list_offset + curr_instance] = PASISIO_SELECT_IO_READY;
             found++;
         }

         curr_instance = (curr_instance + 1) % total_linkage_count;
    }

    PASIS_DASSERT (found >= desired_count);
}


void
PasisIO_SelectThresholdDecodeLinkages (pasisio_linkage_t *linkages,
                                       unsigned int linkage_count, 
                                       unsigned int threshold_count, 
                                       unsigned int type_index,
                                       pasisio_select_t *list,
                                       int shuffle)
{
    PASIS_DASSERT( threshold_count <= list->count );

    memset (list->flags, 0, sizeof (pasisio_select_flags_t) * PASIS_MAX_NUM_SERVERS);
    memset (list->instances, 0,
            sizeof (pasisio_drive_instance_t*) * PASIS_MAX_NUM_SERVERS);

    shuffle = (unsigned int) shuffle;

    select_servers_load_balanced (linkages, linkage_count, threshold_count,
                                  type_index, list, 0, shuffle);

    list->count = linkage_count;
    list->shuffle = shuffle;
}


void
PasisIO_SelectTreeDecodeLinkages (pasisio_linkage_t *linkages,
                                  pasis_md_protocol_t *protocol,
                                  unsigned int type_index,
                                  pasisio_select_t *list,
                                  int shuffle)
{
    PASIS_DASSERT (protocol->quorum_type == PASISIO_QUORUM_TREE);
    PASIS_DASSERT (protocol->B == 1 && protocol->T == 1); /* for now */

    memset (list->flags, 0, sizeof (pasisio_select_flags_t) * PASIS_MAX_NUM_SERVERS);
    memset (list->instances, 0,
            sizeof (pasisio_drive_instance_t*) * PASIS_MAX_NUM_SERVERS);

    if (protocol->delta == 0) {
        /* easy, this is just a threshold quorum with 5-of-6 */
        select_servers_load_balanced (linkages,
                                      6, /* num serverss in group */
                                      5, /* group threshold */
                                      type_index,
                                      list,
                                      0, /* offset into list */
                                      shuffle);
    } else if (protocol->delta == 1) {
        int first_level_shuffle, second_level_shuffle, i;
        PASIS_DASSERT (protocol->N >= 36);

        shuffle = (unsigned int) shuffle;

        /* select "second-level" quorum using top 16 bits of first word of OID */
        second_level_shuffle = (unsigned int)(shuffle >> 16) % 6;
        /* select "first-level" quorum using bottom 16 bits of first word */
        first_level_shuffle = (unsigned int)shuffle % 6;
        
/*         printf("%s: shuffle1 = %d, shuffle2 = %d\n", __FUNCTION__, */
/*                first_level_shuffle, second_level_shuffle); */

        for (i = 0; i < 6; ++i) {
            if (i != first_level_shuffle) {
                select_servers_load_balanced (linkages,
                                              6, /* num serverss in group */
                                              5, /* group threshold */
                                              type_index,
                                              list,
                                              i * 6, /* offset into list */
                                              second_level_shuffle);
            } else {
                //printf("%s: not using server group %d\n", __FUNCTION__, i);
            }
        }
    } else {
        PASIS_DASSERT (0); /* for now */
    }

    list->count = protocol->N;

    PASIS_DASSERT (0); /* need to save second and first level shuffle */
    list->shuffle = shuffle;
}


unsigned int
PasisIO_SelectEncodeLinkages (pasisio_linkage_t *linkages,
                              unsigned int linkage_count,
                              pasisio_select_t *list)
{
    int i;

    PASIS_DASSERT (linkage_count >= list->count);

    for (i = 0; i < list->count; i++) {
        linkages[i].ip_address = list->instances[i]->ip_address;
        sscanf (list->instances[i]->base_path, "/%u", &linkages[i].partition);
    }

    return list->count;
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
