/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#ifndef __PASIS_MD_HIST_H__
#define __PASIS_MD_HIST_H__

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"

#include "ssw_types.h"
#include "ssw_time.h"

#include "pasis_md_prot.h"

#include "pasis_md_s4.h"

/* History version maintenance */

/* The history_oid of the fid_data structure points to a table that contains all
 * history for fid.  Version histories are split into fixed sized history
 * blocks.  The first history block is found at a well-known key
 * {S4_TIME_LATEST}.  Other history blocks are found at {S4 time at history
 * block creation}.
 *
 * Each history block contains:
 * 
 *  History Header (pasis_md_hist_hdr_t) -- valid only on first history block
 *
 *  History Block Header (pasis_md_hist_blk_hdr_t) -- always valid
 *
 *  History entries (pasis_md_ent_t) -- PASIS_FE_HIST_BLK_ENTS/history block
 *    Earliest entry at beginning of block
 *    Latest entry towards end of block
 *    Free space after latest entry
 *
 * History entries are maintained in order.  Latest entries are found in
 * the header block (last block in the chain). 
 *
 * New history blocks are formed when old ones fill up.  The old block is
 * split, resulting in the creation of a new history block.  Half the 
 * history entries (the oldest half) are copied to the new block.  A pointer
 * is set in the old block pointing to the new block, and the new block's
 * pointer is updated.
 *
 */

/* History block size in bytes */
/* There's no point in making this less than 4K due to the way S4 works.
 * For entries >= 1K, S4 put them in 4K blocks...
 * Update: S4 seems happiest with a bit under 4K */
#define PASIS_MD_HIST_BLK_SIZE 2048 /* enough for ~14 hist entries*/

/* Number of bytes in a history block that are available for history entries */
#define PASIS_MD_HIST_BLK_USEFUL_SPACE                              \
  (PASIS_MD_HIST_BLK_SIZE - sizeof (pasis_md_hist_hdr_t) -          \
    sizeof (pasis_md_hist_blk_hdr_t))

/* History header cache number */
#define PASIS_MD_HIST_CACHE_NUM 10007 /* 1009 */


#define PASIS_MD_HIST_KEY_INVALID(key) \
((key)->server_ts.tv_sec==0 && (key)->server_ts.tv_usec==0)

typedef struct pasis_md_hist_key_s {
    s4_timeval_t   server_ts;
} pasis_md_hist_key_t;

typedef struct pasis_md_hist_hdr_s {
    s4_timeval_t     last_write_ts;
#if STORE_AUTHENTICATORS_IN_HIST_TABLE > 0
    p_uint32         auth_num;
    pasis_md_auth_t  authenticators[PASIS_MAX_NUM_SERVERS];    
#endif
} pasis_md_hist_hdr_t;

typedef struct pasis_md_hist_blk_hdr_s {
    p_uint32             num_entries;
    p_uint32             free_space;    /* # of bytes free in block */
    pasis_md_hist_key_t  this_blk_key;
    pasis_md_hist_key_t  next_blk_key;
} pasis_md_hist_blk_hdr_t;

/* History entry */
typedef struct pasis_md_hist_ent_s {
    pasis_md_hts_t        client_ts;
    s4_timeval_t          server_ts;
    pasis_md_rpc_num_t    op_type;
    u_int32_t             is_multiobject_op;
} pasis_md_hist_ent_t;


void pasis_md_hist_stats_init (void);
void pasis_md_hist_stats_dump (int shutdown_flag);

pasis_error_t
pasis_md_hist_init (void);


void
pasis_md_hist_shutdown (void);


/* The "cond_ts" argument serves dual-purposes. First, we need to store it in
 * the version history.  Second, it allows to do automatic garbage collection
 * since we know that the timestamp we conditioned on is the latest complete
 * write.
 */
pasis_error_t
pasis_md_hist_table_insert (pasis_fe_s4_ipc_t         *ipc_op,
                            s4_oid_t                   oid,
                            p_uint32                   N,
                            p_uint32                   node_id,
                            pasis_md_hts_t            *client_ts,
                            pasis_md_hts_t            *cond_ts,
                            pasis_md_rpc_num_t         op_type,
                            u_int32_t                  is_multiobject_op,
                            p_byte                    *create_flag);


/* Creates a history table with the 0 timestamp in it */
pasis_error_t
pasis_md_hist_table_create (pasis_fe_s4_ipc_t       *ipc_op,
                            s4_oid_t                 oid,
                            p_uint32                 N,
                            p_uint32                 node_id);



/* Caller passes in an array whose size is PASIS_MD_HIST_MAX_VERSIONS.
 * num_entries will be set to the number of actual entries. */
pasis_error_t
pasis_md_hist_get_entries (pasis_fe_s4_ipc_t       *ipc_op,
                           s4_oid_t                 oid,
                           p_uint32                *num_entries,   /* Result */
                           pasis_md_history_rec_t  *hist_array);   /* Result */


pasis_error_t
pasis_md_hist_get_latest_entry (pasis_fe_s4_ipc_t *ipc_op,
                                s4_oid_t           oid,
                                pasis_md_hts_t     *hts);   /* Result */


#if STORE_AUTHENTICATORS_IN_HIST_TABLE > 0
/* Caller passes in an array whose size is PASIS_MD_HIST_MAX_VERSIONS.
 * num_entries will be set to the number of actual entries. Also, authenticators
 * are set, since this function is only available if we store the authenticators
 * in the history table. */
pasis_error_t
pasis_md_hist_get_entries_and_auth (pasis_fe_s4_ipc_t       *ipc_op,
                                    s4_oid_t                 oid,
                                    p_uint32                *num_entries,/*Res*/
                                    pasis_md_history_rec_t  *hist_array, /*Res*/
                                    pasis_md_auth_t         *auth);      /*Res*/
#endif


/* Returns info stored in the history table per operation, such as the
 * conditioning timestamp, the op_type, if it's a multi-object op, etc... */
pasis_error_t
pasis_md_hist_get_op_info (pasis_fe_s4_ipc_t       *ipc_op,
                           s4_oid_t                 oid,
                           pasis_md_hts_t          *ts,
                           pasis_md_rpc_num_t      *op_type,     /* Result */
                           p_uint32                *is_multiobject_op,/*Res*/
                           s4_timeval_t            *s4time);


/* Returns the S4 time and flags associated with a (logical) timestamp entry. */
pasis_error_t
pasis_md_hist_get_s4_time (pasis_fe_s4_ipc_t    *ipc_op,
                           s4_oid_t              oid,
                           pasis_md_hts_t       *ts,
                           int                   do_not_use_s4_latest,
                           s4_timeval_t         *s4time);/* Result */

/* Returns PASIS_SUCCESS if we have a specific version */
pasis_error_t
pasis_md_hist_have_entry (pasis_fe_s4_ipc_t    *ipc_op,
                          s4_oid_t              oid,
                          pasis_md_hts_t       *ts);


/* Used by object sync code to update the last s4 time for an object in the
 * case an error occured */
pasis_error_t
pasis_md_hist_update_last_s4_time (pasis_fe_s4_ipc_t *ipc_op,
                                   s4_oid_t           oid);



#endif /* __PASIS_MD_HIST_H__ */


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
