/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/* $cmuPDL: pasis_md_memory_versioning.h,v 1.14.2.3 2007/07/11 21:29:49 mabdelm Exp $ */

/*
 * The Self-* Storage System Project
 * Copyright (c) 2004, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */


/* The PASIS front-end's in-memory versioning interface.  The front-end has a
 * data cache.  Keeping versions in the front-end, and not immediatley sending
 * them to the back-end, allows 1) quicker execution, and 2) efficient garbage
 * collection (GC). As well, we can run in-memory experiments (with a big enough
 * cache, of course) in order to have a more fair comparision with in-memory
 * protocols such as BFT.
 *
 * -mabdelmalek 17/4/2005 */

#ifndef __PASIS_MD_MEMORY_VERSIONING_H__
#define __PASIS_MD_MEMORY_VERSIONING_H__

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_types.h"
#include "pasis_error.h"

#include "pasis_md_prot.h"
#include "ed_c_crc32_iso3309.h"

#include <ext/hash_map>
#include <map>


using namespace std;
using namespace __gnu_cxx;


/* This structure contains object metadata (block size and the distribution) */
typedef struct pasis_md_fid_data_s {
    p_uint32            block_size;    /**< Block size */
    pasis_md_loc_t      distribution;  /**< Protocol and linkages */
} pasis_md_fid_data_t;


// The object container.  Allocated/deallocated in pasis_fe_work.
// It's declared here because a VersionedBlock object needs access to the
// ObjectUniverse's object map, for deallocation purposes.
class MDObjectUniverse;
extern MDObjectUniverse *md_universe;


// Default block cache size.  Can be overridden in the command-line.
#define PASIS_MD_DEFAULT_BLOCK_CACHE_SIZE (256 * 1024 * 1024) /* 512MB */

// Default history cache size.  Can be overridden in the command-line.
#define PASIS_MD_DEFAULT_HISTORY_CACHE_SIZE (32 * 1024 * 1024) /* 64MB */


// A class representing a PASIS timestmap.
class MDVersion
{
 public:
    pasis_md_hts_t timestamp;

    MDVersion () {
        setZero();
    }

    MDVersion (const pasis_md_hts_t* other_ts) {
        md_hts_copy (&timestamp, other_ts);
    }

    MDVersion (const MDVersion& other) {
        md_hts_copy (&timestamp, &other.timestamp);
    }

    bool isLatest() const {
        return PASIS_MD_HTS_IS_LATEST(&timestamp);
    }

    void setLatest() {
        PASIS_MD_HTS_SET_LATEST (&timestamp);
    }

    void setZero() {
        memset (&timestamp, 0, sizeof(pasis_md_hts_t));
    }

    bool isZero() {
        pasis_md_hts_t zero_ts;
        memset(&zero_ts, 0, sizeof(pasis_md_hts_t));
        return memcmp(&timestamp, &zero_ts, sizeof(pasis_md_hts_t)) == 0;
    }

    void zeroObjectHash() {
#if USE_OBJECT_HASHES > 0
        memset(timestamp.object_hash, 0, PASIS_HASH_SIZE);
#endif
    }

    bool operator<(const MDVersion& other) const {
        return md_hts_compare (&timestamp, &other.timestamp) < 0;
    }

    bool operator>(const MDVersion& other) const {
        return md_hts_compare (&timestamp, &other.timestamp) > 0;
    }

    bool operator>=(const MDVersion& other) const {
        return md_hts_compare (&timestamp, &other.timestamp) >= 0;
    }

    bool operator==(const MDVersion& other) const {
        return md_hts_compare (&timestamp, &other.timestamp) == 0;
    }

    MDVersion& operator=(const MDVersion& other) {
        md_hts_copy (&timestamp, &other.timestamp);
        return *this;
    }
};



// A mutex class.
class Mutex
{
public:
    st_mutex_t  mutex;
    st_thread_t lock_owner;
    // st_mutex_lock() fails when a thread tries to lock a mutex that it already
    // has locked.  This lock class provides recursive mutex'es.
    p_uint32    recursive_lock_count;

    Mutex() {
        mutex = st_mutex_new();
        lock_owner = NULL;
        recursive_lock_count = 0;
    }

#if 0
    Mutex(const Mutex& other) {
        mutex = st_mutex_new();
        lock_owner = NULL;
        recursive_lock_count = 0;
    }

    Mutex& operator=(const Mutex& other) {
        mutex = st_mutex_new();
        return *this;
    }
#endif

    ~Mutex() {
        // Destroy the lock.  Be careful if that fails, which indicates that
        // another thread is on the wait queue.
        int res = st_mutex_destroy(mutex);
        PASIS_DASSERT (res == 0);
    }

    void lock(void) {
        if (lock_owner == st_thread_self()) {
            // We already owned the lock; recursive locking happening
            ++recursive_lock_count;
        } else {
            int res = st_mutex_lock(mutex);
            PASIS_DASSERT (res == 0);
            lock_owner = st_thread_self();
        }
    }

    void unlock(void) {
        if (recursive_lock_count) {
            // "Drain" the recursive lock first
            --recursive_lock_count;
        } else {
            int res = st_mutex_unlock(mutex);
            PASIS_DASSERT (res == 0);
            lock_owner = NULL;
        }
    }

    bool is_locked() const {
        return lock_owner != NULL;
    }

    bool is_locked_by_current_thread() const {
        return lock_owner == st_thread_self();
    }
};


// Convenient class for doing locking.  The destructor automatically calls the
// mutex_unlock function.  This simplifies the functions that use this class, as
// they don't have to worry about unlocking the mutex in every return point.
class ScopeLock
{
public:
    Mutex &mutex;

    ScopeLock(Mutex &mutex) : mutex(mutex) {
        mutex.lock();
    }

    ~ScopeLock() {
        mutex.unlock();
    }
};


// A class encapsulating a Q/U object ID.
class MDObjectID
{
 public:
    // The partition number is used for testing. A server can pretend to be
    // multiple servers, by using the incoming partition number as its node ID.
    p_uint32       partition;
    pasis_md_fid_t fid;

    MDObjectID() {
        partition = 0;
        memset(fid.data, 0, PASISIO_FILEID_SIZE);
    }

    MDObjectID(p_uint32 _partition, const pasis_md_fid_t *fid_) {
        partition = _partition;
        memcpy(fid.data, fid_->data, PASISIO_FILEID_SIZE);
    }

    MDObjectID& operator=(const MDObjectID& other) {
        partition = other.partition;
        memcpy(fid.data, other.fid.data, PASISIO_FILEID_SIZE);
        return *this;
    }

    bool operator==(const MDObjectID& other) const {
        return cmp(other) == 0;
    }

    bool operator<(const MDObjectID& other) const {
        return cmp(other) < 0;
    }

    int cmp(const MDObjectID& other) const {
        if (partition < other.partition) {
            return -1;
        } else if (partition > other.partition) {
            return +1;
        } else {
            return memcmp(fid.data, other.fid.data, PASISIO_FILEID_SIZE);
        }
    }

    size_t hash(void) const {
        size_t hash;
        ED_C_CRC32::digest((p_byte*)fid.data, PASISIO_FILEID_SIZE,
                           (p_byte*)&hash);
        return hash + partition;
    }
};


// forward declaration
class MDVersionedBlock;
class MDVersionedObject;
class MDVersionData;



typedef map<MDVersion, MDVersionData> md_history_t;
typedef md_history_t::iterator        md_history_iter_t;
typedef md_history_t::const_iterator  md_history_const_iter_t;


// Each block version has a VersionData structure associated with it
class MDVersionData {
 public:
    void     *data;          // Pointer to actual data
    unsigned  is_dirty  : 1; // Need to flush to back-end?
    unsigned  is_pinned : 1; // Is pinned?
    // if a buffer is pinned and is available for GC, this flag is set
    unsigned  can_gc    : 1; // Can GC, once it's unpinned

    // Only for header block:
    unsigned  calculated_object_hash : 1; // object hash has been calculated?
    unsigned  calculated_auths : 1; // auths has been calculated?

    // Each version needs to keep track of which block it belongs to.  This
    // is because when a block version is evicted, another MDVersionedBlock
    // object may be the one that selected the block to be evicted.  And
    // that other MDVersionedBlock object needs to find the MDVersionedBLock
    // object that corresponds to the evicted block.
    MDVersionedBlock *block_object;

    // LRU list pointers for the data block cache
    md_history_iter_t blockLeft;
    md_history_iter_t blockRight;

    // LRU list pointers for the history block cache
    md_history_iter_t historyLeft;
    md_history_iter_t historyRight;
};



// We allow a service to store some small data in the header block, for
// efficiency.  See write_small_data.  Here's the maximum size of that data.
#define PASIS_MD_MAX_HEADER_DATA_SIZE (100) // in bytes


// A class for storing a versioned block.
//
// Block 0 is special.  The first part contains the object hash.  Second, if we
// save the authenticators in the header, then we store the authenticators.
// Finally, if we're running with object hashes (which reduce # of necessary
// object-sync-servers from b+1 to 1, in common case), then block 0 stores the
// list of block hashes; this lets us quickly recalculate the object hash when a
// particular's block is modified.
class MDVersionedBlock
{
public:
    MDVersionedBlock();
    MDVersionedBlock(const MDVersionedBlock &other);
    ~MDVersionedBlock();

    // Write data at a specific timestamp.
    // Caller allocates memory, but memory ownership taken over by write()
    void write(MDVersion &timestamp, MDVersion &cond_ts, void* data);

    // Some services have small data that they want to store in the header
    // block, for efficiency, rather than devoting a full block for it.  For
    // example, the increment/fetch service can store its small 16 byte data in
    // the header block.  Or the NFS service can store file attributes here.
    // This method is only called on the header block.
    void write_small_data(MDVersion &timestamp, MDVersion &cond_ts, p_uint32 size,
                          void *data);

    void copy_forward(MDVersion &new_ts, MDVersion &old_ts, MDVersion &cond_ts);

    // Returns the data at "timestamp".  If we don't have data at version
    // "timestamp", then we return data for the greatest timestamp that is
    // smaller than "timestamp".
    void* read(MDVersion &timestamp, bool pin = false);

    void* read_small_data(MDVersion &ts, p_uint32 *size = NULL);

    // Returns up to num_hist_in history entries, starting at (but not
    // including) search_ts, going back in time.
    void get_history(MDVersion       &search_ts,
                     p_uint32         num_hist_in,
                     pasis_md_hts_t  *hist_array,    /* Result */
                     p_uint32        *num_hist_out,  /* Result */
                     /* if STORE_AUTHENTICATORS_IN_HIST_TABLE, then caller can
                      * pass in an array of authenticators to be filled in */
                     pasis_md_auth_t *authenticators = NULL);

    // Returns true if this S-N hosts "version"
    bool has_version(MDVersion &version);

    p_uint32 num_versions() const { return history.size(); }

 private:
    bool is_header_block() {return block_num == 0; }

    // Header-specific functions 
    p_uint32 auth_offset();
    p_uint32 block_hashes_offset();
    p_uint32 data_offset();
    p_uint32 object_hash_offset();

    void get_history_from_memory(MDVersion       &search_ts,
                                 p_uint32         num_hist_in,
                                 pasis_md_hts_t  *hist_array,
                                 p_uint32        *num_hist_out,
                                 pasis_md_auth_t *authenticators);

    md_history_iter_t highest_timestamp_less_than(MDVersion &search_ts);

    bool higher_version_exists(MDVersion &timestamp);

    void invalidate_cached_auth(const MDVersion &timestamp);

    md_history_iter_t internal_read(MDVersion &timestamp);

    void do_gc(MDVersion &cond_ts);

    pasis_error_t flush(md_history_iter_t i, bool evict);
    pasis_error_t flush_all();

    void unpin(MDVersion &version);

    void remove_up_to(const MDVersion &timestamp); // up to, but not including
    void remove(MDVersion timestamp);
    void remove(md_history_iter_t i);
    void remove_all();

    char* dump_history();

    void write_to_header(MDVersion &timestamp, MDVersion &cond_ts);
    void write_to_ordinary_block(MDVersion &timestamp, MDVersion &cond_ts,
                                 void* data);

    // Make this function private, and hence inaccessible, as a VersionedBlock
    // object is not meant to be copied.
    MDVersionedBlock& operator=(const MDVersionedBlock& other) {
        return *this;
    }

 private:
    MDVersionedObject *object;    // The object this block belongs to
    p_uint64           block_num; // Block number
    md_history_t       history;   // version history
    bool               possibly_more_history_in_backend;

    friend class MDVersionedObject;
    friend class MDObjectUniverse;
    friend class MDMemoryManager;
};


// Hash function for 64-bit unsigned int.  Used as a template argument to the
// hash_map, in VersionedObject.
struct hash_uint64
{
    size_t operator()(p_uint64 x) const {
        p_uint32 high = x >> 32;
        p_uint32  low = x & 0xffffffff;
        return high ^ low;
    }
};


// An object consists of a set of versioned blocks.
class MDVersionedObject
{
 public:
    // timestamps stored in "private data", as in R/W.
    // block 0 is special.  it will contain the object hash.  even if other
    // chunks are modified, we'd update block 0, since it has the timestmap and
    // object hash.
    // for synthetic objects, object stored in first block as well.  no need for
    // per-block hash, since there's only one block.
    // for nfs objects, first chunk stored in block 1.  object hash in private
    // data for block 0.  attribs + attrib hash in block 0.
    // also, block 0 will contain the "cached" authenticators!
    typedef hash_map<p_uint64, MDVersionedBlock, hash_uint64> block_map_t;

 public:
    virtual ~MDVersionedObject();

    // header size needs to be 104 (private user data) + size(auth * max servers)
    // For max_servers=30, 20 byte hashes => 1024 is enough
    p_uint32 header_size() { return 1024; }
    p_uint32 block_size() { return object_attrs.block_size; }

    void lock(void)   { mutex.lock(); }
    void unlock(void) { mutex.unlock(); }
    bool is_locked() const { return mutex.is_locked(); }

    // See MDVersionedBlock documentation
    void get_history(MDVersion       &search_ts,
                     p_uint32         num_hist_in,
                     pasis_md_hts_t  *hist_array,    /* Result */
                     p_uint32        *num_hist_out,  /* Result */
                     pasis_md_auth_t *authenticators = NULL);

    // Returns the latest version
    void get_latest(MDVersion &result);

    // Returns true if this S-N hosts "version"
    bool has_version(MDVersion &version);

    // Returns a statically-allocated string containing the available versions
    char* dump_history();

    // XXX: for objects with more than one chunk, we may wanna put the chunk
    // count and hashes in the first block.
    // this is OK since we're only in-memory for now.
    // subtract 1 since the header block doesn't count
    p_uint32 num_blocks() const { return blocks.size() - 1; }

    // Memory allocator.  Returns a block's worth of data
    // (object_attrs.block_size).  Does eviction if necessary.
    void* get_memory_for_new_block(void);
    void free_block_memory(void *buf);


    // Writes "data" at block "block_num" at version "ts".  data must be of
    // block size (object_attrs.block_size).
    // cond_ts is passed as it allows the system to do GC.
    void write(MDVersion &ts, MDVersion &cond_ts, p_uint32 block_num,
               void *data);

    // Some services have small data that they want to store in the header
    // block, for efficiency, rather than devoting a full block for it.  For
    // example, the increment/fetch service can store its small 16 byte data in
    // the header block.  Or the NFS service can store file attributes here.
    void write_small_data(MDVersion &ts, MDVersion &cond_ts, p_uint32 size,
                          void *data);

    // Writes a barrier at "ts", condition on "cond_ts".
    void write_barrier(MDVersion &ts, MDVersion &cond_ts);

    // Copies-forward a version at "old_ts" to "new_ts".
    void copy_forward(MDVersion &new_ts, MDVersion &old_ts, MDVersion &cond_ts);

    // Returns the data at "timestamp" for block "block_num".  If we don't have
    // data at version "timestamp", then we return data for the greatest
    // timestamp that is smaller than "timestamp".
    void* read(MDVersion &ts, p_uint32 block_num);

    // Like above, but pins the buffer, so we don't free/evict the block if the
    // caller has a reference.  The caller is responsible for unpinning the
    // buffer.  All the buffers that a thread has pinned are unpinned by a call
    // to unpin.
    void* read_and_pin(MDVersion &ts, p_uint32 block_num);

    // See above
    static void unpin();

    // Reads the small data stored in the header, if anything was stored there.
    // See write_small_data.
    void* read_small_data(MDVersion &ts);

    // Remove all blocks belonging to "timestamp" version
    void remove(MDVersion timestamp);

    MDObjectID          oid;
    pasis_md_fid_data_t object_attrs; // protocol membership, worker lists

 protected:
    // Factory for MDVersionedObjects; only called by MDObjectUniverse::create.
    // Setup a new object's state and insert TS 0 into the version history.
    static MDVersionedObject* create_new_object(MDObjectID        &oid,
                                                p_uint32          block_size,
                                                pasis_md_loc_t   *distribution);
 private:
    MDVersionedObject() {}
    MDVersionedObject(const MDVersionedObject& other) {}
    MDVersionedObject& operator=(const MDVersionedObject& other) {return *this;}

    MDVersionedObject(MDObjectID          &oid,
                      p_uint32             block_size,
                      pasis_md_loc_t      *distribution);


    void* get_memory_for_new_header_block();
    void free_header_block_memory(void *buf);

    void remove_block_object_if_possible(MDVersionedBlock* vb);

    pasis_error_t create_if_necessary(); // creates obj. on back-end if necessary
    // Helper function to initialize an object; it insert TS 0 into the history
    void initialize_new_object();

    //void calculate_object_hash(p_byte *buffer);

    MDVersionedBlock* get_block(p_uint64 block_num, bool doingRead);
    MDVersionedBlock* get_header_block(void) { return get_block(0, false); }

    // The mutex MUST be declared _before_ the blocks hash map.  This is because
    // when the object is being destructed, we need the mutex to be around while
    // the blocks are being destructed.  C++ stipulates that object destruction
    // will occur in _reverse_ order of declaration.
    Mutex               mutex;    // to deal with >1 threads doing create,remove

    block_map_t         blocks;   // map between block number and block versions
    bool                need_to_create_object; // on backend
    bool                deleting_object_now;   // to deal with concurrency

    static p_uint32     num_objects;

    friend class MDVersionedBlock;
    friend class MDObjectUniverse;
    friend class MDBackend;
    friend class MDMemoryManager;
};


// Hash function for FIDs.  Used as a template argument to the hash_map in
// MDObjectUniverse.
struct md_hash_object
{
    size_t operator()(MDObjectID o) const {
        return o.hash();
    }
};

// Equality function for FIDs.  Used as a template argument to the hash_map in
// MDObjectUniverse.
struct md_equal_object
{
    bool operator()(const MDObjectID& x, const MDObjectID& y) const {
        return x == y;
    }
};


// The universe is a set of objects with versioned blocks.
class MDObjectUniverse
{
 public:
    typedef hash_map<MDObjectID, MDVersionedObject*, md_hash_object,
                     md_equal_object> object_map_t;

    MDObjectUniverse ();

    // Flushes new objects/block versions to back-end
    ~MDObjectUniverse ();

    // Returns an object if it exists in the front-end or back-end
    MDVersionedObject* lookup(const MDObjectID &oid);

    // Does a front-end object create (doesn't touch the back-end).  Called for
    // adding a new object in the system (_not_ when simply an object is just in
    // the back-end).
    MDVersionedObject* create(MDObjectID          &oid,
                              p_uint32             block_size,
                              pasis_md_loc_t      *distribution);

    // Deletes all the object's block versions from memory _and_ back-end
    pasis_error_t remove(MDObjectID &oid);

 private:
    object_map_t objects;

    friend class MDMemoryManager;
    friend class MDVersionedBlock;
};



class MDMemoryManager {
 public:
    static void init(p_uint64 block_cache_size, p_uint64 history_cache_size);
    static void shutdown(void);

    // Allocate/deallocate memory for data (blocks)
    static void* get_block_memory(p_uint32 size);
    static void free_block_memory(void* data, p_uint32 size);

    // Account/deaccount for memory for history versions/structures
    // It's easier to make the caller allocate/deallocate the objects, and we
    // just keep track of the amount of memory allocated/deallocated here, and
    // do evication if appropriate
    static void account_history_memory(p_uint32 size);
    static void deaccount_history_memory(p_uint32 size);

    static void wakeup_thread_waiting_for_memory();

    static md_history_iter_t null_iterator() { return dummy_end; }

    static void move_to_block_lru_front(md_history_iter_t &entry);
    static void remove_from_block_lru(md_history_iter_t &entry);
    static void move_to_history_lru_front(md_history_iter_t &entry);
    static void remove_from_history_lru(md_history_iter_t &entry);

 private:
    static void evict_lru_block();
    static void evict_block(md_history_iter_t to_evict);
    static void evict_lru_history();
    static void evict_history(md_history_iter_t to_evict);

    static void wait_for_evictable_block();
    static void wait_for_evictable_history();

    static p_uint64 block_cache_size; // Data cache size, in bytes.
    static p_uint64 history_cache_size; // History cache size, in bytes.
    static p_uint64 total_block_allocation; // Num bytes used for data cache
    static p_uint64 total_history_allocation; // Num bytes used for hist cache

    // We need dummy "NULL" iterators.
    static md_history_t      dummy_map;
    static md_history_iter_t dummy_end;

    // Data LRU list data structures.
    static md_history_iter_t block_lru_first;
    static md_history_iter_t block_lru_last;

    // History LRU list data structures.
    static md_history_iter_t history_lru_first;
    static md_history_iter_t history_lru_last;
    
    // Used for the data buffer cache, when we're low on block memory.
    static st_cond_t block_cache_cond;
    // Number of threads waiting on the above condition variable
    static p_uint32  num_threads_waiting_for_block_memory;

    // Like above, but for history cache
    static st_cond_t history_cache_cond;
    // Number of threads waiting on the above condition variable
    static p_uint32  num_threads_waiting_for_history_memory;

    friend class MDVersionedBlock;
    friend class MDVersionedObject;
};


/* To keep track of which blocks a thread has pinned */

#define PASIS_MD_MAX_WORKER_UNPIN_BLOCKS 32

typedef struct pasis_md_unpin_block_info_s {
    MDVersionedBlock *block;
    MDVersion         version;
} pasis_md_unpin_block_info_t;


/* In order to pin buffers, we associate a list of pinned buffers with each
   thread, to be unpineed after the worker thread replies to the client. */
extern int worker_pin_key; /* points to a pasis_md_unpin_block_info_t array */
extern int worker_pin_count_key; /* count of pinned buffers */




#endif /* __PASIS_MD_MEMORY_VERSIONING_H__ */


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
