/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#ifndef __PASIS_MD_PROT_H__
#define __PASIS_MD_PROT_H__

#include "pasis_config.h"
#include "pasis_types.h"

#include "pasisio.h"
#include "pasis_md_attr.h"

#ifdef __cplusplus
extern "C" {
#endif
#if 0
}
#endif


#define DEBUG_TIMESTAMPS 0 /* Add a string description of operation in TS */
#define DEBUG_OP_DESC_SIZE 20

/* Control whether authentictors are generated on an update operation, or
 * whether we should generate them on every query/update when returning the
 * history.  There's a space-time trade-off. */
#define STORE_AUTHENTICATORS_IN_HIST_TABLE 1

/* Whether to calculate/store object/chunk hashes or not.

 * The advantage of doing so is that during object syncing, a "sync server" only
 * has to receive the object from 1 other server, and the sync server can use
 * the object hash in the timestamp to verify that the other server gave it the
 * right object.
 *
 * The disadvantage is that it's expensive to calculate the object hash every
 * time a "chunk" is updated, and we store the chunk hashes in a separate table
 * that's expensive to update/store on disk.  We can always just make the server
 * contact b other servers to get the object hashes, and have only one server
 * return the _full_ object.
 * It's not even clear that putting the data on the network is more expensive
 * than calculating the hashes.
 *
 */
#define USE_OBJECT_HASHES 0


/* Don't calculate/verify authenticators and hashes.  Useful for benchmarking
 * the cost of authenticators/hashes.  But we still place them on the network.
 * This emulates having a hash/crypto off-loading card.  */
extern p_uint32 DO_NOT_CALCULATE_AUTHENTICATORS;

/* If the above flag is set, we still put dummy authenticators on the network.
 * If this following flag is set, then we don't do this.  This gives us an idea
 * of the network cost of authenticators.  */
extern p_uint32 DO_NOT_TRANSFER_AUTHENTICATORS;


#define PASIS_MD_RPC_PROGNUM   0x20000001
#define PASIS_MD_RPC_VERSION   0x1
#define PASIS_MD_RPC_PORT      34567

#define PASIS_MD_COOKIE_LATEST 0         /* Latest cookie value for read */
#define PASIS_MD_MAX_PARTITION 17        /* Highest partition supported */
#define PASIS_MD_MIN_PARTITION 1         /* Lowest partition supported */

#define PASIS_MD_MAX_NAMELEN   255       /* Maximum file/dir name length */
#define PASIS_MD_MAX_PATHLEN   1024      /* Maximum path name length */

#define PASIS_MD_MAX_LINKAGES PASIS_MAX_NUM_SERVERS

/* Maximum number of FID's that history can be returned for, in one call */
#define MD_MAX_OBJS_PER_REQ 4


/* Max number of history entries per object.  We don't expect more, because GC
 * is automatically performed on complete writes.  */
#define PASIS_MD_HIST_MAX_VERSIONS 64  /* this is pretty pessimistic */


#define PASIS_MD_HTS_SET_LATEST(hts)            \
do {                                            \
    (hts)->time = 0xFFFFFFFFFFFFFFFFULL;        \
} while (0)

#define PASIS_MD_HTS_IS_LATEST(hts) \
  (((hts)->time == 0xFFFFFFFFFFFFFFFFULL))


typedef enum pasis_md_service_s {
    PASIS_MD_GENERIC = 1,      /**< Generic calls (such as read_hist) */
    PASIS_MD_NFS = 2,          /**< NFS metadata service */
    PASIS_MD_EC = 3,           /**< Erasure coded block service */
    PASIS_MD_SYNTH = 4         /**< Synthetic objects service */
} pasis_md_service_t;


/* If you add to this, update pasis_md_work.cpp:md_rpc_to_service and
 * pasis_md_fe.cpp:op_names */
typedef enum pasis_md_rpc_num_s {
    /* Generic service calls */
    PASIS_MD_READ_HISTORY = 1,     /**< Read history vector */
    PASIS_MD_BARRIER = 2,          /**< Repair barrier operation */
    PASIS_MD_OBJECT_SYNC = 3,      /**< Object directory */
    /* NFS metadata service calls */
    PASIS_MD_CREATE = 4,           /**< Create file */
    PASIS_MD_LINK = 5,             /**< Link */
    PASIS_MD_UNLINK = 6,           /**< Unlink */
    PASIS_MD_GETATTR = 7,          /**< Get attributes */
    PASIS_MD_SETATTR = 8,          /**< Set attributes */
    PASIS_MD_RENAME = 9,           /**< Rename */
    PASIS_MD_LOOKUP = 10,          /**< Lookup */
    PASIS_MD_READLINK = 11,        /**< Read link */
    PASIS_MD_READDIR = 12,         /**< Read directory block */
    /* Synthetic object service */
    PASIS_MD_SYNTH_QUERY = 15,     /**< Synthetic query operation */
    PASIS_MD_SYNTH_UPDATE = 16,    /**< Synthetic update operation */
    /* For getting statistics on the server's throughput history */
    PASIS_MD_GET_STATS = 17        /**< Get statistics */
} pasis_md_rpc_num_t;

typedef enum pasis_md_rpc_stat_s {
    PASIS_MD_SUCCESS = 0,                /**< Success */
    PASIS_MD_FAILURE = 100,              /**< Success */
    PASIS_MD_ERR_NOENT = 101,            /**< No entry */
    PASIS_MD_ERR_EXIST = 103,            /**< Entry already exists */
    PASIS_MD_ERR_INVAL_PART = 104,       /**< Invalid partition id */
    PASIS_MD_ERR_INVAL_TIME = 105,       /**< Invalid timestamp */
    PASIS_MD_ERR_INVAL_CKSUMS = 106,     /**< Invalid cross checksums */
    PASIS_MD_ERR_HIST_MATRIX_NO_CW = 107,/**< No complete write in matrix */
    PASIS_MD_ERR_HIST_MATRIX_AUTH = 108, /**< Bad authenticators in matrix */
    PASIS_MD_ERR_HIST_MATRIX_NUM = 109,  /**< Wrong # of matrices provided*/
    PASIS_MD_ERR_HIST_NOT_LATEST = 110,  /**< History vector isn't latest */
    PASIS_MD_ERR_HIST_BARRIER_VAL = 111, /**< Barrier validation failed */
    PASIS_MD_ERR_HIST_BARRIER = 112,     /**< Barrier necessary */
    PASIS_MD_ERR_HIST_NOBARRIER = 113,   /**< No barrier necessary */
    PASIS_MD_ERR_INVAL_READ_FLAG = 114,  /**< Invalid read flag (in read-op) */
    PASIS_MD_ERR_OBJSYNC_NOT_EXIST = 115,/**< Requested version not avail */
    PASIS_MD_ERROR = 255                 /**< Generic error */
} pasis_md_rpc_stat_t;



typedef struct pasis_fe_s4_ipc_s {
#if 0
    int /*s4_ipc_request_t*/  req;
    int /*s4_ipc_reply_t*/    reply;
    st_cond_t         cond;
    int               signal;
    p_byte           *args_buffer;
    p_byte           *data_buffer;

    p_uint32          op_id;     /* Current operation ID */
#endif
} pasis_fe_s4_ipc_t;



typedef struct pasis_md_fid_s {
    char data[PASISIO_FILEID_SIZE];
} pasis_md_fid_t;


static const pasis_md_fid_t PASIS_MD_ROOT_FID_C = {{0xFF, 0xFF, 0xFF, 0xFF,
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}};

static const p_uint32 PASIS_MD_ROOT_FILEID_C = 0xFFFFFFFF;


typedef p_uint64 pasis_md_time_t;
typedef p_uint32 pasis_md_client_t;
typedef p_uint64 pasis_md_cookie_t;
typedef p_uint32 pasis_md_part_t;


typedef struct pasis_md_hts_s {
    pasis_md_time_t     time;                /**< Logical timestamp */
#if USE_OBJECT_HASHES > 0
    p_byte              object_hash[PASIS_HASH_SIZE];/**< Object hash */
#endif
    p_byte              op_hash[PASIS_HASH_SIZE];    /**< Operation hash */
    p_byte              matrix_hash[PASIS_HASH_SIZE];/**< TS matrix hash */
    p_uint32            client_id;           /**< Client ID */

    unsigned            barrier_flag   : 1;  /**< Barrier ? */
    unsigned            multiobject_op : 1;  /**< Multi-object operation? */
    unsigned            op_type        :30;  /**< Op type: pasis_md_rpc_num_t */
#if DEBUG_TIMESTAMPS > 0
    char                op_description[DEBUG_OP_DESC_SIZE]; /**< Op desc. */
#endif
    
    /* MAYBE hash of timestamp we're conditioning on */
} pasis_md_hts_t;

typedef struct pasis_md_protocol_s {
    p_uint32          N;
    p_uint32          Qc;
    p_uint32          B;
    p_uint32          T;
    p_uint32          m;
    p_uint32          scheme_type;
    p_uint32          hash_type;
    p_uint32          cc_type;
    pasisio_quorum_t  quorum_type;
    p_uint32          delta;       /**< Delta for threshold quorums */
} pasis_md_protocol_t;


#define PASIS_MD_DATA_NUM_IOVECS 32

typedef union pasis_md_data_u {
    p_byte        *client;         
    struct {
        unsigned int  iovec_count;
        struct iovec  iovec[PASIS_MD_DATA_NUM_IOVECS];
    } server;
} pasis_md_data_t;


#define pasis_md_linkage_t pasisio_linkage_t

typedef struct pasis_md_loc_s {
    pasis_md_protocol_t  protocol;
    pasis_md_linkage_t   linkages[PASIS_MD_MAX_LINKAGES];
} pasis_md_loc_t;


/* Authenticator */
typedef struct pasis_md_auth_s {
    p_byte  data[PASIS_HASH_SIZE];
} pasis_md_auth_t;



/* History matrix */
/* ts_num        :  number of timestamps 
 * ts_vector     :  vector of timestamps in order, latest at entry 0
 * ts_bitmap     :  bitmap of storage-nodes holding ts, one bitmap per ts
 * auth_num      :  number of authenticators
 * auth_exists   :  = 1 if authenticator[i] is valid; 0 otherwise
 * authenticators:  array of authenticators, one per node
 */
typedef struct pasis_md_hist_matrix_s {
    p_uint32                ts_num;

    pasis_md_hts_t         *ts_vector[PASIS_MD_HIST_MAX_VERSIONS];
    p_uint64                ts_bitmap[PASIS_MD_HIST_MAX_VERSIONS];

    p_uint32                auth_num;
    p_uint32                auth_exists[PASIS_MAX_NUM_SERVERS];
    pasis_md_auth_t        *authenticators[PASIS_MAX_NUM_SERVERS];
} pasis_md_hist_matrix_t;


/* Macros/function for counting bits in the history matrix's bitmap */
#define TWO(c) (0x1ULL << (c))
#define PASIS_MASK(c) (((p_uint64)(-1)) / (TWO(TWO(c)) + 1ULL))
#define COUNT(x,c) ((x) & PASIS_MASK(c)) + (((x) >> (TWO(c))) & PASIS_MASK(c))

static inline int 
bitcount (p_uint64 n)
{
    n = COUNT(n, 0) ;
    n = COUNT(n, 1) ;
    n = COUNT(n, 2) ;
    n = COUNT(n, 3) ;
    n = COUNT(n, 4) ;
    n = COUNT(n, 5) ;   /* for 64-bit integers */
    return n ;
}


typedef struct pasis_md_hist_vector_s {
    p_uint32                node_id;
    p_uint32                history_num;
    pasis_md_hts_t          histories[PASIS_MD_HIST_MAX_VERSIONS];
    p_uint32                auth_num;
    pasis_md_auth_t         authenticators[PASIS_MAX_NUM_SERVERS];
} pasis_md_hist_vector_t;



/* Structure containing the block hashes for a directory.  For now, we have a
 * fixed number of hash entries. */
#define PASIS_MD_DIR_MAX_BLOCKS 128

typedef struct pasis_md_hash_blk_s {
    p_uint32 entries;
    p_uint32 chunk_size;
    p_byte   object_hash[PASIS_HASH_SIZE];
    p_byte   attrib_hash[PASIS_HASH_SIZE];
    p_byte   hash_entries[PASIS_MD_DIR_MAX_BLOCKS][PASIS_HASH_SIZE];
} pasis_md_hash_blk_t;


typedef struct pasis_md_read_history_args_s {
    p_uint32               fid_num;           /**< Number of FID's */
    pasis_md_fid_t        *fid[MD_MAX_OBJS_PER_REQ]; /**< FID */
} pasis_md_read_history_args_t;

typedef struct pasis_md_read_history_res_s {
    p_uint32               result_num;
    pasis_md_rpc_stat_t    statuses[MD_MAX_OBJS_PER_REQ];    
} pasis_md_read_history_res_t;

typedef struct pasis_md_create_args_s {
    pasis_md_fid_t        *parent_dir_fid;  /**< FID of parent directory */
    pasis_md_fid_t        *create_fid;      /**< FID of new file to create */
    char                  *create_name;     /**< Name of new file */
    pasis_md_attr_t        create_attr;     /**< Attributes of new file */
    char                  *symlink_name;    /**< Name of symlink source */
    pasis_md_loc_t         distribution;    /**< Proto/linkages of new file */
    p_uint32               client_id;       /**< Client ID */
} pasis_md_create_args_t;

typedef struct pasis_md_create_res_s {
    pasis_md_attr_t        create_attr;     /**< Attributes of created file */
    pasis_md_attr_t        parent_attr;     /**< New attrs of parent dir */
} pasis_md_create_res_t;


typedef struct pasis_md_link_args_s {
    pasis_md_fid_t        *from_file_fid;  /**< FID of existing file */
    pasis_md_fid_t        *to_dir_fid;     /**< FID of dir to link into */
    char                  *to_name;        /**< Name of to file entry */
    p_uint32               client_id;       /**< Client ID */
} pasis_md_link_args_t;

typedef struct pasis_md_link_res_s {
    pasis_md_attr_t        from_file_attr;  /**< New attributes of from file */
    pasis_md_attr_t        to_dir_attr;     /**< New attributes of to dir */
} pasis_md_link_res_t;


typedef struct pasis_md_unlink_args_s {
    pasis_md_fid_t        *parent_dir_fid;  /**< FID of parent directory */
    pasis_md_fid_t        *name_fid;        /**< FID of name to unlink */
    char                  *name;            /**< Name of file to delete */
    p_uint32               client_id;       /**< Client ID */
} pasis_md_unlink_args_t;

typedef struct pasis_md_unlink_res_s {
    pasis_md_attr_t        parent_dir_attr; /**< New attrs of parent dir */
} pasis_md_unlink_res_t;


typedef struct pasis_md_getattr_args_s {
    pasis_md_fid_t        *fid;             /**< FID */
} pasis_md_getattr_args_t;

typedef struct pasis_md_getattr_res_s {
    pasis_md_attr_t        attr;            /**< Attributes */
    pasis_md_loc_t         distribution;    /**< Proto/linkages of FID */
} pasis_md_getattr_res_t;


typedef struct pasis_md_setattr_args_s {
    pasis_md_fid_t        *fid;             /**< FID */
    pasis_md_sattr_t       attr;            /**< Attributes to set */
    p_uint32               guard_ctime[2];  /**< If !0, check against ctime */
    p_uint32               client_id;       /**< Client ID */
} pasis_md_setattr_args_t;

typedef struct pasis_md_setattr_res_s {
    pasis_md_attr_t        attr;            /**< New attributes */
} pasis_md_setattr_res_t;


typedef struct pasis_md_rename_args_s {
    pasis_md_fid_t        *from_dir_fid;    /**< FID of old dir */
    p_uint32               to_dir_valid;    /**< If invalid, same as from dir*/
    pasis_md_fid_t        *to_dir_fid;      /**< FID of new dir */
    pasis_md_fid_t        *from_name_fid;   /**< FID of old name */
    p_uint32               to_name_valid;   /**< If to_name_fid is valid */
    pasis_md_fid_t        *to_name_fid;     /**< FID of existing file (opt.) */
    char                  *from_name;       /**< Name of old file */
    char                  *to_name;         /**< Name of new file */
    p_uint32               client_id;       /**< Client ID */
} pasis_md_rename_args_t;

typedef struct pasis_md_rename_res_s {
    pasis_md_attr_t        from_dir_attr;   /**< New attributes of old dir */
    pasis_md_attr_t        to_dir_attr;     /**< New attributes of new dir */
} pasis_md_rename_res_t;


typedef struct pasis_md_lookup_args_s {
    pasis_md_fid_t        *parent_dir_fid;  /**< FID of parent directory */
    char                  *name;            /**< Name of file */
} pasis_md_lookup_args_t;

typedef struct pasis_md_lookup_res_s {
    pasis_md_fid_t         fid;             /**< FID */
    pasis_md_attr_t        dir_attr;        /**< Attributes of parent dir */
    pasis_md_loc_t         distribution;    /**< Protocol/linkages of lookup */
} pasis_md_lookup_res_t;


typedef struct pasis_md_readlink_args_s {
    pasis_md_fid_t        *fid;             /**< FID */
} pasis_md_readlink_args_t;

typedef struct pasis_md_readlink_res_s {
    char                   name[PASIS_MD_MAX_PATHLEN]; /**< Name of file */
    pasis_md_attr_t        fid_attr;        /**< Attributes of file */
} pasis_md_readlink_res_t;


typedef struct pasis_md_readdir_args_s {
    pasis_md_fid_t     *fid;                        /**< FID of directory */
    p_uint64            block_num;                  /**< Directory block # */
} pasis_md_readdir_args_t;

typedef struct pasis_md_readdir_res_s {
    pasis_md_attr_t     dir_attr;                   /**< Dir. attributes */
    p_byte              hash[PASIS_HASH_SIZE];  /**< Block hash */
    p_uint32            data_size;                  /**< Size of dir block */
    p_byte             *data;                       /**< Directory block */
} pasis_md_readdir_res_t;


typedef struct pasis_md_object_sync_args_s {
    pasis_md_fid_t  *fid;             /**< FID of object */
    p_uint32         return_attrs;    /**< Flag for returning attrs (only for NFS) */
    p_uint32         start_block_num; /**< Which chunks to begin returning */
    p_uint32         num_blocks_to_ret; /**< Num of blocks to return */
} pasis_md_object_sync_args_t;

typedef struct pasis_md_object_sync_res_s {
    /* Following field is set iff return_attrs in args was set */
    pasis_md_attr_t  attr;                  /**< Directory attributes */

    p_uint32         block_size;            /**< Helps XDR code */
    p_uint32         num_total_blocks;      /**< Total num of blocks */
    p_uint32         num_returned_blocks;   /**< Num of blocks returned */
    pasis_md_data_t  data;                  /**< Blocks */
} pasis_md_object_sync_res_t;


typedef struct pasis_md_barrier_args_s {
    p_uint32               fid_num;             /**< Number of FID's */
    pasis_md_fid_t        *fid[MD_MAX_OBJS_PER_REQ]; /**< FID */
    p_uint32               chunk_size;          /**< Chunk size for synth update */
    p_uint32               client_id;           /**< Client ID */
} pasis_md_barrier_args_t;

typedef struct pasis_md_barrier_res_s {
} pasis_md_barrier_res_t;


typedef struct pasis_md_synth_query_args_s {
    p_uint32               num_fids;  /* <= MD_MAX_OBJS_PER_REQ */
    pasis_md_fid_t        *fids[MD_MAX_OBJS_PER_REQ];
    p_uint32               arg_size;
    p_uint32               res_size;
    p_uint32               chunk_size;
    p_uint32               read_chunks[MD_MAX_OBJS_PER_REQ]; /* bitmaps */
    p_uint32               compute_time;
    p_byte                *args;
} pasis_md_synth_query_args_t;

typedef struct pasis_md_synth_query_res_s {
    p_uint32               res_size;
    p_byte                *res;
} pasis_md_synth_query_res_t;


typedef struct pasis_md_synth_update_args_s {
    p_uint32               num_fids;  /* <= MD_MAX_OBJS_PER_REQ */
    pasis_md_fid_t        *fids[MD_MAX_OBJS_PER_REQ];
    p_uint32               arg_size;
    p_uint32               res_size;
    p_uint32               chunk_size;
    p_uint32               modified_chunks[MD_MAX_OBJS_PER_REQ]; /* bitmaps*/
    p_uint32               compute_time; /* <= in ms */
    /* For testing, useful for client to tell server by how much to incrememnt
     * the timestamp (e.g,, +=2, instead of the normal +=1)*/
    p_uint32               ts_increment;
    pasis_md_loc_t         distribution;    /**< Proto/linkages of new object */
    p_uint32               client_id;       /**< Client ID */
    p_byte                *args;
} pasis_md_synth_update_args_t;

typedef struct pasis_md_synth_update_res_s {
    p_uint32               res_size;
    p_byte                *res;
} pasis_md_synth_update_res_t;



typedef struct pasis_md_get_stats_args_s {
    p_uint32               res_max_num_buckets;
} pasis_md_get_stats_args_t;

typedef struct pasis_md_get_stats_res_s {
    p_uint32              num_buckets;
    p_uint32              bucket_size; /* in ms */
    p_uint32             *res; /**< Statistics */
} pasis_md_get_stats_res_t;



typedef enum pasis_md_read_flag_s {    /**< Flags for non-read-hist ops */
    PASIS_MD_READ_LATEST_DATA = 0x02,  /**< Return latest result */
    PASIS_MD_READ_HTS_DATA    = 0x04,  /**< Return data at specified TS */
} pasis_md_read_flag_t;


typedef union pasis_md_read_op_args_specific_s {
    pasis_md_read_history_args_t   read_history;
    pasis_md_getattr_args_t        getattr;
    pasis_md_lookup_args_t         lookup;
    pasis_md_readlink_args_t       readlink;
    pasis_md_readdir_args_t        readdir;
    pasis_md_object_sync_args_t    object_sync;
    pasis_md_synth_query_args_t    synth_query;
    pasis_md_get_stats_args_t      get_stats;
} pasis_md_read_op_args_specific_t;


typedef struct pasis_md_read_op_args_s {
    pasis_md_part_t      part;
    pasis_md_read_flag_t read_flags;  /* not used in read_history, get_stats */
    pasis_md_hts_t       ts;          /* not used in read_history, get_stats */
    pasis_md_rpc_num_t   op_type;
    pasis_md_read_op_args_specific_t op_args;

    /* Normally, a query is optimistic and doesn't take any OHS arguments.
     * But for the repair cases, the client gives an OHS to the server as
     * proof/info that repair should happen, and the server can then object-
     * sync to perform (part of the) repair. -mabdelmalek 19/3/2005. */
    p_uint32                history_matrices_num;
    pasis_md_hist_matrix_t  history_matrices[MD_MAX_OBJS_PER_REQ];

} pasis_md_read_op_args_t;



typedef union pasis_md_read_op_res_specific_s {
    pasis_md_read_history_res_t   read_history;
    pasis_md_getattr_res_t        getattr;
    pasis_md_lookup_res_t         lookup;
    pasis_md_readlink_res_t       readlink;
    pasis_md_readdir_res_t        readdir;
    pasis_md_object_sync_res_t    object_sync;
    pasis_md_synth_query_res_t    synth_query;
    pasis_md_get_stats_res_t      get_stats;
} pasis_md_read_op_res_specific_t;

typedef struct pasis_md_read_op_res_s {
    pasis_md_rpc_stat_t status;    /* protocol and non-operation errors */
    pasis_md_rpc_stat_t op_status; /* operation execution error */
    pasis_md_rpc_num_t  op_type;

    pasis_md_hts_t      ts; /* Timestamp of the data returned */

    /* We have an array of history elements/authenticators.  We have one 
     * element per each FID specified in the read op.  Read history calls can 
     * specify up to MAX_READ_REQS FID's.  The other calls will just have
     * one FID. 
     */
    p_uint32               history_vec_num; /* # elements in next struct */
    pasis_md_hist_vector_t history_vectors[MD_MAX_OBJS_PER_REQ];

    pasis_md_read_op_res_specific_t op_res;

} pasis_md_read_op_res_t;


typedef union pasis_md_write_op_args_specific_s {
    pasis_md_create_args_t       create;
    pasis_md_link_args_t         link;
    pasis_md_unlink_args_t       unlink;
    pasis_md_setattr_args_t      setattr;
    pasis_md_rename_args_t       rename;
    pasis_md_barrier_args_t      barrier;
    pasis_md_synth_update_args_t synth_update;
} pasis_md_write_op_args_specific_t;

typedef struct pasis_md_write_op_args_s {
    pasis_md_part_t     part;
    pasis_md_rpc_num_t  op_type;
    p_uint32            mtime[2];
    p_uint32            repair_flag;

    pasis_md_write_op_args_specific_t op_args;

    p_uint32                history_matrices_num;
    pasis_md_hist_matrix_t  history_matrices[MD_MAX_OBJS_PER_REQ];

    /* For "deduced-complete" support, we need to send "proof" matrices to the
     * S-N's that prove that an FID is repairable.
     * There can be a proof matrix for each of the FID's in the operation; each
     * proof matrix proves that its corresponding FID is in fact complete.
     * Now, for each FID, the timestamp in question will refer to a multi-object
     * operation which can have more than one other FID.  The fid_indices array
     * tells us which of those indices we're providing proof for.  The indices
     * are set to UINT_MAX If we're not using the i'th proof matrix */
    p_uint32                proof_fid_indices[MD_MAX_OBJS_PER_REQ];
    pasis_md_hist_matrix_t  proof_matrices[MD_MAX_OBJS_PER_REQ];
} pasis_md_write_op_args_t;


typedef union pasis_md_write_op_res_specific_s {
    pasis_md_create_res_t       create;
    pasis_md_link_res_t         link;
    pasis_md_unlink_res_t       unlink;
    pasis_md_setattr_res_t      setattr;
    pasis_md_rename_res_t       rename;
    pasis_md_barrier_res_t      barrier;
    pasis_md_synth_update_res_t synth_update;
} pasis_md_write_op_res_specific_t;

typedef struct pasis_md_write_op_res_s {
    pasis_md_rpc_stat_t status;    /* protocol and non-operation errors */
    pasis_md_rpc_stat_t op_status; /* operation execution error */
    pasis_md_rpc_num_t  op_type;

    p_uint32               history_vec_num; /* # elements in next struct */
    pasis_md_hist_vector_t history_vectors[MD_MAX_OBJS_PER_REQ];

    pasis_md_write_op_res_specific_t op_res;
} pasis_md_write_op_res_t;



static inline int
md_hts_compare (const pasis_md_hts_t *hts1, const pasis_md_hts_t *hts2)
{
    int rc;

    if (hts1->time > hts2->time) return 1;
    if (hts1->time < hts2->time) return -1;

#if USE_OBJECT_HASHES > 0
    rc = memcmp (hts1->object_hash, hts2->object_hash, PASIS_HASH_SIZE);
    if (rc != 0) return rc;
#endif

    rc = memcmp (hts1->op_hash, hts2->op_hash, PASIS_HASH_SIZE);
    if (rc != 0) return rc;

    rc = memcmp (hts1->matrix_hash, hts2->matrix_hash, PASIS_HASH_SIZE);
    if (rc != 0) return rc;

    if (hts1->client_id > hts2->client_id) return 1;
    if (hts1->client_id < hts2->client_id) return -1;

    if (hts1->barrier_flag > hts2->barrier_flag) return 1;
    if (hts1->barrier_flag < hts2->barrier_flag) return -1;

    if (hts1->multiobject_op > hts2->multiobject_op) return 1;
    if (hts1->multiobject_op < hts2->multiobject_op) return -1;

    if (hts1->op_type > hts2->op_type) return 1;
    if (hts1->op_type < hts2->op_type) return -1;

    return 0;
}

static inline void
md_hts_copy (pasis_md_hts_t *dest, const pasis_md_hts_t *src)
{
    dest->time = src->time;
#if USE_OBJECT_HASHES > 0
    memcpy (dest->object_hash, src->object_hash, PASIS_HASH_SIZE);
#endif
    memcpy (dest->op_hash, src->op_hash, PASIS_HASH_SIZE);
    memcpy (dest->matrix_hash, src->matrix_hash, PASIS_HASH_SIZE);
    dest->client_id = src->client_id;
    dest->barrier_flag = src->barrier_flag;
    dest->multiobject_op = src->multiobject_op;
    dest->op_type = src->op_type;

#if DEBUG_TIMESTAMPS > 0
    memcpy (dest->op_description, src->op_description,
            sizeof(src->op_description));
#endif
}


static inline int
md_fid_copy (const pasis_md_fid_t *dest, const pasis_md_fid_t *src)
{
    return memcmp (dest->data, src->data, PASISIO_FILEID_SIZE);
}


static inline int
md_fid_compare (const pasis_md_fid_t *fid1, const pasis_md_fid_t *fid2)
{
    return memcmp (fid1->data, fid2->data, PASISIO_FILEID_SIZE);
}


static inline char *
md_fid_dump (pasis_md_fid_t *fid)
{
    static char buf[PASISIO_FILEID_SIZE*2 + 1];
    int i;
    char *bp = buf;
    for (i=0; i<PASISIO_FILEID_SIZE; i++) {
        snprintf(bp, 3, "%02X", (unsigned char)fid->data[i]);
        bp +=2;
    }

    return buf;
}

#ifdef __cplusplus
}
#endif

#endif /* __PASIS_MD_PROT_H__ */

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */ 
