/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_debug.h"
#include "pasis_timer.h"

#include "pasis_md_prot.h"
#include "pasis_md_auth.h"
#include "pasis_md_log.h"


#define PASIS_MD_TIMER_AUTH_NUM             6

#define PASIS_MD_TIMER_AUTH_GENERATE        0
#define PASIS_MD_TIMER_AUTH_VERIFY          1
#define PASIS_MD_TIMER_AUTH_HASH            2
#define PASIS_MD_TIMER_AUTH_HASH_MATRIX     3
#define PASIS_MD_TIMER_AUTH_KEY_HASH        4
#define PASIS_MD_TIMER_AUTH_HASH_INDIRECT   5


typedef struct pasis_md_auth_stats_s {
    pasis_timer_state_t  *timers;
} pasis_md_auth_stats_t;

static pasis_md_auth_stats_t stats;


/* This is used when calculating hashes over an array of _pointers_ to history
 * entries.  The hash functions need a list of sizes; an entry per each
 * pointer.  Since these are constant, we pre-calculate them. */
unsigned int array_of_hist_entry_sizes[PASIS_MAX_NUM_SERVERS];



/* Takes an array of history entries and returns the (UNkeyed) hash digest. */
inline static void
hash_direct (p_uint32          history_num,
             pasis_md_hts_t   *histories,
             pasis_md_auth_t  *hash)
{
    if (DO_NOT_CALCULATE_AUTHENTICATORS) {
        //memset (hash, 0, ED_HASH_DIGEST_SIZE);
    } else {
        HASH_FUNCTION::digest((p_byte *) histories,
                              history_num * sizeof(pasis_md_hts_t),
                              (p_byte *) hash);
    }
}


/* Takes an array of _pointers_ to history entries and returns the UNkeyed
 * hash. */ 
void
pasis_md_auth_hash_indirect (p_uint32          history_num,
                             pasis_md_hts_t  **histories,
                             pasis_md_auth_t  *hash)
{
    PASIS_TIMER_START ();
    if (DO_NOT_CALCULATE_AUTHENTICATORS) {
        //memset (hash, 0, ED_HASH_DIGEST_SIZE);
    } else {
        HASH_FUNCTION::digest((const p_byte **) histories,
                              array_of_hist_entry_sizes,
                              history_num,
                              (p_byte *) hash);
    }
    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_AUTH_HASH_INDIRECT);
}


/* Takes an array of _pointers_ to history entries. */
inline static void
key_hash_indirect (auth_key_t       *key,
                   p_uint32          history_num,
                   pasis_md_hts_t  **histories,
                   pasis_md_auth_t  *authenticator)
{
    if (DO_NOT_CALCULATE_AUTHENTICATORS) {
        //memset (authenticator, 0, ED_HASH_DIGEST_SIZE);
    } else {
        pasis_md_auth_t hash;

        HASH_FUNCTION::digest((const p_byte **) histories,
                              array_of_hist_entry_sizes,
                              history_num,
                              (p_byte *) &hash);


        HASH_FUNCTION::key_digest((p_byte *) key, sizeof(auth_key_t),
                                  (p_byte *) &hash,
                                  PASIS_HASH_SIZE,
                                  (p_byte *) authenticator);
    }
}


void
pasis_md_auth_generate (p_uint32         local_node_id,
                        p_uint32         history_num,
                        pasis_md_hts_t  *histories,
                        p_uint32         auth_num, /* # of S-N's */
                        pasis_md_auth_t *authenticators)
{
    PASIS_TIMER_START ();

    unsigned int other_node_id;
    pasis_md_auth_t hash;
    auth_key_t   key;

    hash_direct (history_num, histories, &hash);

    key.local_node_id = local_node_id;

    pasis_md_log_msg ("%s: node %u generating authenticators for history vector:",
                      __FUNCTION__, local_node_id);
    for (p_uint32 k = 0; k < history_num; ++k) {
        char temp[256];
        sprintf(temp, "\t [%u] : %"PRIu64, k, histories[k].time);
        if (histories[k].barrier_flag) {
            strcat(temp, " [BARRIER]");
        }
        pasis_md_log_msg(temp);
    }

    for (other_node_id = 0; other_node_id < auth_num; ++other_node_id) {
        key.remote_node_id = other_node_id;

        // Inline the call to pasis_md_auth_key_hash, for performance
        if (DO_NOT_CALCULATE_AUTHENTICATORS) {
            // memset (&authenticators[other_node_id], 0, PASIS_HASH_SIZE);
        } else {
            HASH_FUNCTION::key_digest((p_byte *) &key, sizeof(auth_key_t),
                                      (p_byte *) &hash,
                                      PASIS_HASH_SIZE,
                                      (p_byte *) &authenticators[other_node_id]);
        }

        pasis_md_log_msg ("%s: for_node=%u, auth=%x", __FUNCTION__,
                          other_node_id, 
                          *(p_uint32 *)&authenticators[other_node_id]);
    }

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_AUTH_GENERATE);
}


pasis_error_t
pasis_md_auth_verify (p_uint32          local_node_id,
                      p_uint32          remote_node_id,
                      p_uint32          history_num,
                      pasis_md_hts_t  **histories,
                      pasis_md_auth_t  *authenticator)
{
    pasis_md_auth_t true_authenticator;
    auth_key_t      key;
    int             cmp;

    PASIS_TIMER_START ();

    if (DO_NOT_CALCULATE_AUTHENTICATORS) {
        cmp = 0;
    } else {
        /* Gotta reverse the key order here, since the S-N that generated the key
         * has the opposite point of view regarding who is local/remote.  */
        key.local_node_id  = remote_node_id;
        key.remote_node_id = local_node_id;

        key_hash_indirect (&key, history_num, histories, &true_authenticator);

        cmp = memcmp(&true_authenticator, authenticator, PASIS_HASH_SIZE);

        PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_AUTH_VERIFY);

        if (cmp) {
            printf("%s: node %u: failed check for node %u; got %x, expected %x\n",
                   __FUNCTION__, local_node_id, remote_node_id,
                   *(p_uint32*)authenticator, *(p_uint32*)&true_authenticator);
        }
    }

    // if authenticator is correct, cmp=0 => PASIS_SUCCESS
    return cmp;
}


void
pasis_md_auth_hash (u_int32_t                size,  /* buffer size in bytes */
                    p_byte                  *buffer,
                    p_byte                  *hash)  /* Out arg */
{
    PASIS_TIMER_START ();
    if (DO_NOT_CALCULATE_AUTHENTICATORS) {
        memset (hash, 0, PASIS_HASH_SIZE);
    } else {
        HASH_FUNCTION::digest(buffer, size, hash);
    }
    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_AUTH_HASH);
}


void
pasis_md_auth_hash_matrix (pasis_md_hist_matrix_t  *matrix,
                           p_byte                  *hash)  /* Out arg */
{
    PASIS_TIMER_START ();
    if (DO_NOT_CALCULATE_AUTHENTICATORS) {
        memset (hash, 0, PASIS_HASH_SIZE);
    } else {
        /* 2 intermeditate hashes:  1 for history entries, 1 for bitmap */
        p_byte intermediate_hashes[2][PASIS_HASH_SIZE];

        PASIS_TIMER_START ();

        HASH_FUNCTION::digest((const p_byte**) matrix->ts_vector,
                              array_of_hist_entry_sizes,
                              matrix->ts_num,
                              (p_byte *) &intermediate_hashes[0]);
                          
        HASH_FUNCTION::digest((const p_byte*) matrix->ts_bitmap,
                              matrix->ts_num * sizeof(p_uint64),
                              (p_byte *) &intermediate_hashes[1]);
                          
        HASH_FUNCTION::digest((const p_byte*) intermediate_hashes,
                              PASIS_HASH_SIZE * 2,
                              hash);
    }
    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_AUTH_HASH_MATRIX);
}



pasis_error_t
pasis_md_auth_init (void)
{
    unsigned int i;

    for (i = 0; i < PASIS_MAX_NUM_SERVERS; ++i) {
        array_of_hist_entry_sizes[i] = sizeof(pasis_md_hts_t);
    }

    return PASIS_SUCCESS;
}


void
pasis_md_auth_shutdown (void)
{
}


void
pasis_md_auth_stats_init (void)
{
    memset (&stats, 0, sizeof (pasis_md_auth_stats_t));

    /* Set up timers */
    stats.timers = Pasis_TimerInit ("Auth_Timers", PASIS_MD_TIMER_AUTH_NUM, 0);
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_AUTH_GENERATE,
                         "authentication_generate"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_AUTH_VERIFY,
                         "authentication_verify"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_AUTH_HASH,
                         "authentication_hash"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_AUTH_HASH_MATRIX,
                         "authentication_matrix_hash"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_AUTH_KEY_HASH,
                         "authentication_key_hash"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_AUTH_HASH_INDIRECT,
                         "authentication_hash_indirect"); 
}


void
pasis_md_auth_stats_dump (int shutdown_flag)
{
    if (shutdown_flag) {
        Pasis_TimerShutdown (stats.timers);
    } else {
        Pasis_TimerPrintAndReset (stats.timers);
    }
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */ 
