/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_debug.h"

#include "pasis_md_xdr.h"
#include "pasis_md_prot.h"
#include "pasis_md_clnt.h"

#include "rpc_common.h"



int
pasis_md_clnt_read_op (rpc_handle_t *client, rpc_auth_t *auth,
                       pasis_md_read_op_args_t *argp,
                       pasis_md_read_op_res_t *res)
{
    rpc_req_t *req;
    rpc_tcp_call_hdr_t req_hdr;
    int error;

    req = rpc_req_get (client);
    req->rpc_xid = rpc_req_xid_get ();
    req->rpc_prog = PASIS_MD_RPC_PROGNUM;
    req->rpc_vers = PASIS_MD_RPC_VERSION;
    req->rpc_proc = argp->op_type;
    req->rpc_auth = *auth;

    req->req_iov[0].iov_base = (char *)&req_hdr;
    req->req_iov[0].iov_len = sizeof (rpc_tcp_call_hdr_t);
    req->req_iov_len = 1;

    req->req_iov[1].iov_base = req->rpc_buffer;
    req->req_iov_cur = 1;

    req->reply_iov[0].iov_base = (char *)req->rpc_buffer;
    req->reply_iov[0].iov_len = RPC_MAX_PAYLOAD;
    req->reply_iov_len = 1;

    pasis_md_xdr_enc_read_op_args (req, (u_int32_t *)req->rpc_buffer, argp);

    if (argp->op_type == PASIS_MD_SYNTH_QUERY) {
        req->req_iov[2].iov_base = argp->op_args.synth_query.args;
        req->req_iov[2].iov_len = argp->op_args.synth_query.arg_size;

        req->req_iov_cur += 1;
        req->req_iov_len += 1;
    }

    rpc_encode_tcp_call (client, req, &req_hdr);

    error = rpc_transmit_tcp_req (client, req);
    if (error < 0) {
        rpc_req_put (client, req);
        return error;
    }

    if (req->reply_status != RPC_MSG_ACCEPTED || 
        req->reply_accepted_status != RPC_ACCEPT_SUCCESS) {
        PASIS_PRINT_ERR ("Error: RPC hdr level error\n");
        rpc_req_put (client, req);
        return -1;
    }

    if (pasis_md_xdr_dec_read_op_res ((u_int32_t *)req->rpc_buffer, res) 
        != TRUE) {
        PASIS_PRINT_ERR ("Error: RPC decode error\n");
        rpc_req_put (client, req);
        return -1;
    }

    rpc_req_put (client, req);

    return 0;
}


int
pasis_md_clnt_write_op (rpc_handle_t *client, rpc_auth_t *auth,
                        pasis_md_write_op_args_t *argp,
                        pasis_md_write_op_res_t *res)
{
    rpc_req_t *req;
    rpc_tcp_call_hdr_t req_hdr;
    int error;

    req = rpc_req_get (client);
    req->rpc_xid = rpc_req_xid_get ();
    req->rpc_prog = PASIS_MD_RPC_PROGNUM;
    req->rpc_vers = PASIS_MD_RPC_VERSION;
    req->rpc_proc = argp->op_type;
    req->rpc_auth = *auth;

    req->req_iov[0].iov_base = (char *)&req_hdr;
    req->req_iov[0].iov_len = sizeof (rpc_tcp_call_hdr_t);
    req->req_iov_len = 1;

    req->req_iov[1].iov_base = req->rpc_buffer;
    req->req_iov_cur = 1;

    pasis_md_xdr_enc_write_op_args (req, (u_int32_t *)req->rpc_buffer, argp);

    /* Write w/erasure-coding -- setup data */
    if (argp->op_type == PASIS_MD_SYNTH_UPDATE) {
        req->req_iov[2].iov_base = argp->op_args.synth_update.args;
        req->req_iov[2].iov_len = argp->op_args.synth_update.arg_size;

        req->req_iov_cur += 1;
        req->req_iov_len += 1;
    }

    req->reply_iov[0].iov_base = req->rpc_buffer;
    req->reply_iov[0].iov_len = RPC_MAX_PAYLOAD;
    req->reply_iov_len = 1;

    rpc_encode_tcp_call (client, req, &req_hdr);

    error = rpc_transmit_tcp_req (client, req);
    if (error < 0) {
        rpc_req_put (client, req);
        return error;
    }

    if (req->reply_status != RPC_MSG_ACCEPTED || 
        req->reply_accepted_status != RPC_ACCEPT_SUCCESS) {
        PASIS_PRINT_ERR ("Error: RPC hdr level error\n");
        rpc_req_put (client, req);
        return -1;
    }

    if (pasis_md_xdr_dec_write_op_res ((u_int32_t *)req->rpc_buffer, res) 
        != TRUE) {
        PASIS_PRINT_ERR ("Error: RPC decode error\n");
        rpc_req_put (client, req);
        return -1;
    }

    rpc_req_put (client, req);

    return 0;
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */

