/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_timer.h"
#include "pasis_cache.h"

#include "pasis_md_s4.h"

#include "pasis_md_attr.h"
#include "pasis_md_dir.h"
#include "pasis_md_dir_ops.h"
#include "pasis_md_memory_versioning.h"
#include "pasis_md_err.h"
#include "pasis_md_ops.h"


#if 0
static char *
print_fid (pasis_md_fid_t *fid)
{
    static char buf[2 * PASISIO_FILEID_SIZE + 8];
    char *bp;
    int i;

    snprintf (buf, 3, "0X");
    bp = buf;

    for (i = 0; i < 8; i++) {
        bp += 2;
        snprintf (bp, 3, "%02X", fid->data[i]);
    }
    
    return buf;
}
#endif


/* 
 * Looks up name in blk, returns failure if not found.  
 * Sets fid_out if name is found and fid_out is non-null.
 */
pasis_error_t
pasis_md_dir_lookup (p_byte         *blk,
                     char           *name,
                     unsigned int   *offset_out)
{
   pasis_md_dir_hdr_t   *hdr;
   pasis_md_dir_entry_t *entry;
   unsigned int          name_len, offset;
   
   hdr = (pasis_md_dir_hdr_t *)blk;
   name_len = strlen (name);

   PASIS_DASSERT (name_len <= PASIS_MD_MAX_NAMELEN);

   offset = PASIS_MD_DIR_HDR_SIZE;

   /* Loop through all entries in this block */
   while (offset < hdr->free_offset) {
       entry = (pasis_md_dir_entry_t *)(blk + offset);

       /* First check name lengths, and then compare strings */
       if (entry->name_len == name_len) {
           if (memcmp (blk + offset + PASIS_MD_DIR_ENT_SIZE, 
                       name, name_len) == 0) {
               if (offset_out) {
                   *offset_out = offset;
               }
               return PASIS_SUCCESS;
           }
       }
       
       offset += (entry->name_len + PASIS_MD_DIR_ENT_SIZE);
   } 
   
   return PASIS_ERROR_DIR_NOENT;
}


pasis_error_t
pasis_md_dir_insert (p_byte         *blk,
                     MDObjectID     *file,
                     const char     *name,
                     p_uint32        fileid,
                     unsigned int   *offset_out)
{
    pasis_md_dir_hdr_t   *hdr;
    pasis_md_dir_entry_t *entry;
    unsigned int record_len, name_len;

    hdr = (pasis_md_dir_hdr_t *)blk;

    name_len = strlen (name);
    PASIS_DASSERT (name_len <= PASIS_MD_MAX_NAMELEN);
    record_len = name_len + PASIS_MD_DIR_ENT_SIZE;

    if (record_len + hdr->free_offset > PASIS_MD_DIR_BLK_SIZE) {
        return PASIS_ERROR_DIR_NOSPACE;
    }

    entry = (pasis_md_dir_entry_t *)(blk + hdr->free_offset);
    entry->name_len = name_len;
    entry->fileid = fileid;
    memcpy (&entry->fid, &file->fid, sizeof (pasis_md_fid_t));
    entry->partition = file->partition;
    memcpy (blk + hdr->free_offset + PASIS_MD_DIR_ENT_SIZE,
            name, entry->name_len);

    /* printf ("%s: inserting fileid=%u at offset=%u with len=%u\n",
     *            __FUNCTION__, fileid, hdr->free_offset, record_len); */

    hdr->free_offset += record_len;
    if (offset_out) *offset_out = hdr->free_offset;

    return PASIS_SUCCESS;
}


unsigned int
pasis_md_dir_remove (p_byte        *blk,
                     unsigned int   entry_offset)
{
    pasis_md_dir_hdr_t   *hdr;
    pasis_md_dir_entry_t *entry;
    unsigned int          record_len;

    hdr = (pasis_md_dir_hdr_t *)blk;
    entry = (pasis_md_dir_entry_t *)(blk + entry_offset);
    
    /* Calculate record len */
    record_len = PASIS_MD_DIR_ENT_SIZE + entry->name_len;

    /* If this is the last entry, removal is trivial */
    if (entry_offset + record_len == hdr->free_offset) {
        hdr->free_offset -= record_len;
        return hdr->free_offset;
    }

    /* Compact entries */
    memmove (blk + entry_offset, blk + entry_offset + record_len,
             hdr->free_offset - (entry_offset + record_len));
    
    hdr->free_offset -= record_len;

    return hdr->free_offset;
}


pasis_error_t
pasis_md_dir_create (MDVersionedObject  *parent_dir,
                     pasis_md_attr_t    *attr_in,
                     MDVersionedObject  *create_dir,
                     pasis_md_hts_t     *new_hts,
                     pasis_md_hts_t     *cond_hts)
{
    pasis_error_t error;
    p_byte* dir_data = (p_byte*) create_dir->get_memory_for_new_block();
    
    memset(dir_data, 0, PASIS_MD_DIR_BLK_SIZE);
    pasis_md_dir_init_hdr(dir_data);
    pasis_md_dir_insert(dir_data, &parent_dir->oid, "..", attr_in->fileid, NULL);

    // Write the new chunk!  Damn, the Q/U interface is so clean!
    MDVersion new_version(new_hts), cond_version(cond_hts);
    create_dir->write(new_version, cond_version, 0, dir_data);
    // Update attributes
    create_dir->write_small_data(new_version, cond_version,
                                 sizeof(pasis_md_attr_t), attr_in);
    return OP_OK;
}


pasis_error_t
pasis_md_dir_blk_is_empty (MDVersionedObject   *dir,
                           MDVersion           *version,
                           pasis_md_attr_t     *attr)
{
    pasis_md_dir_hdr_t *dir_hdr;
    pasis_error_t       error;
    p_byte             *blk;
    p_uint64            i;

    /* Loop through all blocks checking that they are empty */
    for (i = 0; i < attr->blocks; i++) {
        blk = (p_byte*) dir->read(*version, i);
        if (blk == NULL) {
            PASIS_PRINT_ERR ("%s: error, reading directory block failed\n", 
                             __FUNCTION__);
            return OPERR_IO;
        }
    
        dir_hdr = (pasis_md_dir_hdr_t *)blk;
        
        /* Remember, the ".." entry is still in first directory block */
        if ( ((i == 0) && dir_hdr->free_offset != 
              (PASIS_MD_DIR_HDR_SIZE + 2 + PASIS_MD_DIR_ENT_SIZE)) ||
             ((i > 0) && (dir_hdr->free_offset != PASIS_MD_DIR_HDR_SIZE)) ) {
            return OPERR_NOTEMPTY;
        }
    }

    return OP_OK;
}


pasis_error_t
pasis_md_dir_find_entry (MDVersionedObject    *dir,
                         MDVersion            &version,
                         char                 *name,
                         p_byte              **blk_out,
                         p_uint64             *blkno_out,
                         unsigned int         *offset_out,
                         pasis_md_attr_t      *dir_attr_out)
{
    pasis_error_t  error;
    p_byte        *blk = NULL;
    p_uint64       blkno;

    void* saved_attr = dir->read_small_data(version);

    if (saved_attr == NULL) {
        // XXX: if saved_attr we have no idea if this is a dir of file; assume
        // it's a directory
        *blk_out = NULL;
        *blkno_out = 0;
        memset (dir_attr_out, 0, sizeof (pasis_md_attr_t));
        return OPERR_NOENT;
    } else {
        memcpy (dir_attr_out, saved_attr, sizeof (pasis_md_attr_t));
    }

    if (!MD_ATTR_IS_DIR (dir_attr_out)) {
        return OPERR_NOTDIR;
    }

    for (blkno = 0; blkno < dir_attr_out->blocks; blkno++) {

        blk = (p_byte*) dir->read(version, blkno);
        if (blk == NULL) {
            PASIS_PRINT_ERR ("%s: dir blk=%"PRIu64" doesn't exist\n", 
                             __FUNCTION__, blkno);
            return OPERR_IO;
        }

        error = pasis_md_dir_lookup (blk, name, offset_out);
        if (error == OP_OK) {
            /* Found the name -- set output values */
            // XXX: pin buffer?
            *blk_out = blk;
            *blkno_out = blkno;
            return OP_OK;
        }       
    }

    // XXX: pin buffer?
    *blk_out = blk;
    *blkno_out = blkno - 1;    

    return OPERR_NOENT;
}


pasis_error_t
pasis_md_dir_rename_parent (MDVersionedObject *dir,
                            MDVersion         &new_version,
                            MDVersion         &cond_version,
                            pasis_md_fid_t    *parent_fid,
                            p_uint32           parent_fileid)
{
    pasis_md_dir_entry_t *entry;
    p_byte               *blk;
    
    /* Read in directory block 0 */
    blk = (p_byte*)dir->read(cond_version, 0);
    if (blk == NULL) {
        return PASIS_ERROR;
    }
    
    /* First entry should always be '..' */
    entry = (pasis_md_dir_entry_t *)(blk + PASIS_MD_DIR_HDR_SIZE);
    
    /* Sanity check */
    if (entry->name_len != 2 || memcmp ("..", blk + PASIS_MD_DIR_HDR_SIZE +
                                        PASIS_MD_DIR_ENT_SIZE, 2) != 0) {
        PASIS_DASSERT (0);
    }

    /* Copy entries if fid's are different */
    if (memcmp (&entry->fid, parent_fid, sizeof (pasis_md_fid_t)) != 0) {    
        pasis_md_dir_hdr_t *hdr = (pasis_md_dir_hdr_t *)blk;
        p_byte* new_blk = (p_byte*) dir->get_memory_for_new_block();
        memmove(new_blk, blk, hdr->free_offset);

        pasis_md_dir_entry_t *new_entry =
            (pasis_md_dir_entry_t *)(new_blk + PASIS_MD_DIR_HDR_SIZE);;

        memcpy (&new_entry->fid, parent_fid, sizeof (pasis_md_fid_t));
        new_entry->fileid = parent_fileid;

        dir->write(new_version, cond_version, 0, new_blk);
    }

    return PASIS_SUCCESS;
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
