/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include <fcntl.h>
#include <sys/stat.h>

#include "pasis_config.h"
#include "pasis_lists.h"
#include "pasis_types.h"
#include "pasis_timer.h"

#include "pasisio.h"

#include "rpc_common.h"

#include "pasis_md_prot.h"
#include "pasis_md_xdr.h"
#include "pasis_md_svc.h"

#include "pasis_md_s4.h"
#include "pasis_md_auth.h"
#include "pasis_md_dir.h"
#include "pasis_md_fid_lock.h"
#include "pasis_md_object_sync.h"
#include "pasis_md_ops.h"
#include "pasis_md_memory_versioning.h"
#include "pasis_md_work.h"
#include "pasis_md_log.h"
#include "pasis_md_delay.h"
#include "pasis_md_validation.h"
#include "pasis_md_stats.h"

#include "pasisio_config_parse.h"
#include "pasisio_select.h"

#define PASIS_MD_TIMER_NUM            18

#define PASIS_MD_TIMER_READ_HISTORY    0
#define PASIS_MD_TIMER_READ_OPERATION  1
#define PASIS_MD_TIMER_CREATE          2
#define PASIS_MD_TIMER_LINK            3
#define PASIS_MD_TIMER_UNLINK          4
#define PASIS_MD_TIMER_GETATTR         5
#define PASIS_MD_TIMER_SETATTR         6
#define PASIS_MD_TIMER_RENAME          7
#define PASIS_MD_TIMER_LOOKUP          8
#define PASIS_MD_TIMER_READLINK        9 
#define PASIS_MD_TIMER_READDIR        10
#define PASIS_MD_TIMER_BARRIER        11
#define PASIS_MD_TIMER_OBJECT_SYNC    12
#define PASIS_MD_TIMER_READ_EC        13
#define PASIS_MD_TIMER_WRITE_EC       14
#define PASIS_MD_TIMER_SYNTH_QUERY    15
#define PASIS_MD_TIMER_SYNTH_UPDATE   16
#define PASIS_MD_TIMER_GET_STATS      17


struct local_stats_t {
    pasis_timer_state_t  *timers;
    unsigned int          total_requests;
    unsigned int          num_requests_in_last_period; /* for periodic stats */
    unsigned int          workers_blocked;
    unsigned int          max_in_progress;
};



typedef struct s4_fe_worker_thread_s {
    st_thread_t   thread_id;
    st_cond_t     cond;
    int           work;

    rpc_req_t    *rpc_req;
    p_byte       *data;
    unsigned int  data_len;

    union {
        pasis_md_read_op_args_t  read_args;
        pasis_md_write_op_args_t write_args;
    } args;

    union {
        pasis_md_read_op_res_t   read_res;
        pasis_md_write_op_res_t  write_res;
    } res;

    pasis_md_log_t    log_info;

    
    // every worker thread can pin some blocks when it reads them.  here's where
    // we keep a record of which entries are pinned
    p_uint32                    num_blocks_to_unpin;
    pasis_md_unpin_block_info_t unpin_buffer_info[PASIS_MD_MAX_WORKER_UNPIN_BLOCKS];


    PASIS_Q_PTRS(struct s4_fe_worker_thread_s) queue;
} s4_fe_worker_thread_t;


typedef PASIS_Q_STRUCT(worker_q_s, s4_fe_worker_thread_t) s4_fe_worker_queue_t;

static p_uint32  op_id = 0;

static local_stats_t stats;

static rpc_handle_t *rpc_handle = NULL;

static s4_fe_worker_queue_t  worker_queue;
static s4_fe_worker_thread_t worker_thread_array[PASIS_MD_WORKERS];
static int workers_shutdown = 0;
static int workers_working = 0;

static int dispatch_waiting = 0;
static st_cond_t dispatch_cond;

static st_cond_t shutdown_cond;
static int shutdown_flag = 0;


static pasis_md_loc_t distribution;

static unsigned int ops_in_progress=0;

const char* op_names[] = {
    "<invalid-op-code>",
    "PASIS_MD_READ_HISTORY",
    "PASIS_MD_BARRIER",
    "PASIS_MD_OBJECT_SYNC",
    "PASIS_MD_CREATE",
    "PASIS_MD_LINK",
    "PASIS_MD_UNLINK",
    "PASIS_MD_GETATTR",
    "PASIS_MD_SETATTR",
    "PASIS_MD_RENAME",
    "PASIS_MD_LOOKUP",
    "PASIS_MD_READLINK",
    "PASIS_MD_READDIR",
    "PASIS_MD_READ_EC",
    "PASIS_MD_WRITE_EC",
    "PASIS_MD_SYNTH_QUERY",
    "PASIS_MD_SYNTH_UPDATE",
    "PASIS_MD_GET_STATS"
};


static void stats_dump (int shutdown_flag);

static void
sig_handler (int signum)
{
    shutdown_flag = 1;
    st_cond_signal(shutdown_cond);
}


static void *
pasis_md_s4_worker_thread (void *arg)
{
    s4_fe_worker_thread_t *worker;

    worker = (s4_fe_worker_thread_t *)arg;
    st_thread_setspecific(worker_pin_key, worker->unpin_buffer_info);
    st_thread_setspecific(worker_pin_count_key, &worker->num_blocks_to_unpin);

    /* Setup log info */
    st_thread_setspecific(log_key, &worker->log_info);

    workers_working++;

    while (1) {
        /* Wait for work or shutdown */
        while (worker->work == 0 && workers_shutdown == 0) {
            st_cond_wait (worker->cond);
        }

        if (workers_shutdown) {
            break;
        }

        /* Do Work */
        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "\n------\n");
        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: proc=%u data=%p\n", 
                       __FUNCTION__, worker->rpc_req->rpc_proc, worker->data);

        ops_in_progress++;
        stats.max_in_progress = MAX (ops_in_progress, stats.max_in_progress);

        // Update statistics about number of outstanding requests
        //pasis_md_stats_callback();

        stats.total_requests++;
        stats.num_requests_in_last_period++;
        worker->log_info.op_id = ++op_id;

        pasis_md_log_msg ("%s from %s", op_names[worker->rpc_req->rpc_proc],
                          pasis_ip_to_hostname(worker->rpc_req->reply_ip));

        switch (worker->rpc_req->rpc_proc) {

        case PASIS_MD_READ_HISTORY: {
            pasis_md_read_op_args_t *args = &worker->args.read_args;
            pasis_md_read_op_res_t  *res  = &worker->res.read_res;
            p_uint32 i;

            PASIS_TIMER_START ();
            
            pasis_md_xdr_dec_read_op_args ((u_int32_t *)worker->data, args);
            
            for (i = 0; i < args->op_args.read_history.fid_num; i++) {
                pasis_md_log_msg ("[%d] FID: %#llx", i, *(p_uint64 *)
                                  args->op_args.read_history.fid[i]);
            }

            pasis_md_work_read_history (NULL, args, res);

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_READ_HISTORY);

            pasis_md_svc_read_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_vectors (res->history_vectors, res->history_vec_num);
            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_CREATE: {
            pasis_md_write_op_args_t *args = &worker->args.write_args;
            pasis_md_write_op_res_t  *res  = &worker->res.write_res;
            u_int32_t                 op_size;
            u_int32_t                *op;

            PASIS_TIMER_START ();

            pasis_md_xdr_dec_write_op_args ((u_int32_t *)worker->data, args,
                                            &op_size, &op);

            pasis_md_log_msg ("[0] FID: %#llx", *(p_uint64 *)
                              args->op_args.create.parent_dir_fid);
            pasis_md_log_msg ("[1] FID: %#llx", *(p_uint64 *)
                              args->op_args.create.create_fid);

            pasis_md_log_matrices (args->history_matrices, 
                                   args->history_matrices_num);

            pasis_md_work_create (args, op_size, op, res);

            //pasis_md_do_sync ();

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_CREATE);

            pasis_md_svc_write_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_vectors (res->history_vectors, res->history_vec_num);
            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_LINK: {
            pasis_md_write_op_args_t *args = &worker->args.write_args;
            pasis_md_write_op_res_t  *res  = &worker->res.write_res;
            u_int32_t                 op_size;
            u_int32_t                *op;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_write_op_args ((u_int32_t *)worker->data, args,
                                            &op_size, &op);

            pasis_md_work_link (args, op_size, op, res);

            //pasis_md_do_sync ();

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_LINK);
                              
            pasis_md_svc_write_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_UNLINK: {
            pasis_md_write_op_args_t *args = &worker->args.write_args;
            pasis_md_write_op_res_t  *res  = &worker->res.write_res;
            u_int32_t                 op_size;
            u_int32_t                *op;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_write_op_args ((u_int32_t *)worker->data, args,
                                            &op_size, &op);

            pasis_md_work_unlink (args, op_size, op, res);

            //pasis_md_do_sync ();

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_UNLINK);
                              
            pasis_md_svc_write_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_GETATTR: {
            pasis_md_read_op_args_t *args = &worker->args.read_args;
            pasis_md_read_op_res_t  *res  = &worker->res.read_res;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_read_op_args ((u_int32_t *)worker->data, args);

            pasis_md_work_getattr (args, res);

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_GETATTR);
                              
            pasis_md_svc_read_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_SETATTR: {
            pasis_md_write_op_args_t *args = &worker->args.write_args;
            pasis_md_write_op_res_t  *res  = &worker->res.write_res;
            u_int32_t                 op_size;
            u_int32_t                *op;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_write_op_args ((u_int32_t *)worker->data, args,
                                            &op_size, &op);

            pasis_md_work_setattr (args, op_size, op, res);

            //pasis_md_do_sync ();

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_SETATTR);
                              
            pasis_md_svc_write_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_RENAME: {
            pasis_md_write_op_args_t *args = &worker->args.write_args;
            pasis_md_write_op_res_t  *res  = &worker->res.write_res;
            u_int32_t                 op_size;
            u_int32_t                *op;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_write_op_args ((u_int32_t *)worker->data, args,
                                            &op_size, &op);

            pasis_md_work_rename (args, op_size, op, res);

            //pasis_md_do_sync ();

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_RENAME);
                              
            pasis_md_svc_write_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_LOOKUP: {
            pasis_md_read_op_args_t *args = &worker->args.read_args;
            pasis_md_read_op_res_t  *res  = &worker->res.read_res;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_read_op_args ((u_int32_t *)worker->data, args);

            pasis_md_work_lookup (args, res);

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_LOOKUP);
                              
            pasis_md_svc_read_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_READLINK: {
            pasis_md_read_op_args_t *args = &worker->args.read_args;
            pasis_md_read_op_res_t  *res  = &worker->res.read_res;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_read_op_args ((u_int32_t *)worker->data, args);

            pasis_md_work_readlink (args, res);

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_READLINK);
                              
            pasis_md_svc_read_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_READDIR: {
            pasis_md_read_op_args_t *args = &worker->args.read_args;
            pasis_md_read_op_res_t  *res  = &worker->res.read_res;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_read_op_args ((u_int32_t *)worker->data, args);

            res->op_res.readdir.data = worker->data;

            pasis_md_work_readdir (args, res);

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_READDIR);
                              
            pasis_md_svc_read_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_OBJECT_SYNC: {
            pasis_md_read_op_args_t *args = &worker->args.read_args;
            pasis_md_read_op_res_t  *res  = &worker->res.read_res;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_read_op_args ((u_int32_t *)worker->data, args);

            pasis_md_object_sync_server (NULL, args, res);

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_OBJECT_SYNC);
                              
            pasis_md_svc_read_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_BARRIER: {
            pasis_md_write_op_args_t *args = &worker->args.write_args;
            pasis_md_write_op_res_t  *res  = &worker->res.write_res;
            u_int32_t                 op_size;
            u_int32_t                *op;

            PASIS_TIMER_START ();

            pasis_md_xdr_dec_write_op_args ((u_int32_t *)worker->data, args,
                                            &op_size, &op);

            for (p_uint32 i = 0; i < args->op_args.barrier.fid_num; i++) {
                pasis_md_log_msg ("[%d] FID: %#llx", i, *(p_uint64 *)
                                  args->op_args.barrier.fid[i]);
            }

            pasis_md_log_matrices (args->history_matrices, 
                                   args->history_matrices_num);

            pasis_md_work_barrier (NULL,
                                   args, res);

            //pasis_md_do_sync ();

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_BARRIER);
                              
            pasis_md_svc_write_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_vectors (res->history_vectors, res->history_vec_num);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        case PASIS_MD_SYNTH_QUERY: {
            pasis_md_read_op_args_t *args = &worker->args.read_args;
            pasis_md_read_op_res_t  *res  = &worker->res.read_res;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_read_op_args ((u_int32_t *)worker->data, args);

            res->op_res.synth_query.res = worker->data;

            pasis_md_work_synth_query (NULL, args, res);

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_SYNTH_QUERY);
                              
            pasis_md_svc_read_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);

            break;
        }

        case PASIS_MD_SYNTH_UPDATE: {
            pasis_md_write_op_args_t *args = &worker->args.write_args;
            pasis_md_write_op_res_t  *res  = &worker->res.write_res;
            u_int32_t                 op_size;
            u_int32_t                *op;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_write_op_args ((u_int32_t *)worker->data, args,
                                            &op_size, &op);

            res->op_res.synth_update.res = worker->data;

            pasis_md_work_synth_update (args, op_size, op, res);

            //pasis_md_do_sync ();

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_SYNTH_UPDATE);
                              
            pasis_md_svc_write_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u",
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);

            break;
        }

        case PASIS_MD_GET_STATS: {
            pasis_md_read_op_args_t *args = &worker->args.read_args;
            pasis_md_read_op_res_t  *res  = &worker->res.read_res;
            
            PASIS_TIMER_START ();

            pasis_md_xdr_dec_read_op_args ((u_int32_t *)worker->data, args);

            /* reuse worker->data to store statistics in there */
            res->op_res.get_stats.res = (p_uint32*)worker->data;

            pasis_md_stats_get (args, res);

            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_GET_STATS);
                              
            pasis_md_svc_read_op_reply (rpc_handle, worker->rpc_req, res);

            pasis_md_log_msg ("%s status=%u, op_status=%u", 
                              op_names[worker->rpc_req->rpc_proc],
                              res->status, res->op_status);
            break;
        }

        default:
            PASIS_PRINT_ERR ("Error: pasis s4 frontend unknown proc=%#x\n",
                             worker->rpc_req->rpc_proc);
        }
        
        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: proc=%u completed\n", 
                       __FUNCTION__, worker->rpc_req->rpc_proc);
        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "------\n\n");

        worker->work = 0;

        ops_in_progress--;

        PASIS_Q_ADD_TAIL (worker_queue, queue, worker);

        // unpin any buffers this thread had pinned while handling this request
        MDVersionedObject::unpin();

        // Another thread could be blocked, waiting to get memory.  So we check
        // for this condition now that this thread has finished its work (it
        // might have performed a write/read that allocated memory).  And we
        // wake up the thread that's waiting, so it can evict some memory and
        // then execute.
        MDMemoryManager::wakeup_thread_waiting_for_memory();

        if (dispatch_waiting) {
            dispatch_waiting = 0;
            st_cond_signal (dispatch_cond);
        }
    }

    workers_working--;

    return NULL;
}


static void
pasis_md_s4_rpc_dispatch (rpc_req_t   *req,
                          p_byte      *data,
                          unsigned int len)
{
    s4_fe_worker_thread_t *worker;
    
    PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: xid=%u, proc=%u\n", 
                   __FUNCTION__, req->rpc_xid, req->rpc_proc);
    
    worker = PASIS_Q_FIRST (worker_queue);
    while (worker == NULL) {

        stats.workers_blocked++;

        dispatch_waiting = 1;
        while (dispatch_waiting) {
            st_cond_wait (dispatch_cond);
        }
        worker = PASIS_Q_FIRST (worker_queue);
    }

    PASIS_Q_REMOVE (worker_queue, queue, worker);
    
    worker->work = 1;
    worker->rpc_req = req;
    worker->data = data;
    worker->data_len = len;

    st_cond_signal (worker->cond);
}


static pasis_error_t
pasis_md_s4_workers_init (void)
{
    int i;

    /* Initialize dispatch waiting condition variable */
    if ((dispatch_cond = st_cond_new ()) == NULL) {
        PASIS_PRINT_ERR("Unable to init condition variable\n");
        return PASIS_ERROR;
    }

    /* Initialize and startup all worker threads */
    PASIS_Q_INIT (worker_queue);

    for (i = 0; i < PASIS_MD_WORKERS; i++) {
        s4_fe_worker_thread_t *worker;

        worker = &worker_thread_array[i];
        worker->work = 0;
        worker->cond = st_cond_new ();
        worker->num_blocks_to_unpin = 0;
        
        /* Increase stack size to 128K */
        worker->thread_id = st_thread_create (pasis_md_s4_worker_thread, 
                                              worker, 1, 128 << 10);

        PASIS_DASSERT (worker->thread_id != NULL);

        PASIS_Q_ADD_TAIL (worker_queue, queue, worker);
    }

    return PASIS_SUCCESS;
}


static void
pasis_md_s4_workers_shutdown (void)
{
    int i;

    workers_shutdown = 1;

    /* Shutdown and cleanup all worker threads */
    for (i = 0; i < PASIS_MD_WORKERS; i++) {
        s4_fe_worker_thread_t *worker;

        worker = &worker_thread_array[i];

        st_cond_signal (worker->cond);
        st_thread_join (worker->thread_id, NULL);
        
        st_cond_destroy (worker->cond);
    }

    PASIS_DASSERT (workers_working == 0);
}


static pasis_error_t
pasis_md_create_root (p_uint32 node_id)
{
    pasis_error_t       error;
    p_uint32            i, start_i, end_i;

    // when we have a back-end, we'll need to check if root directory exists
    // there or not

    // Figure out how many root objects we're going to create
    if (node_id == PASIS_MAX_NUM_SERVERS) {
        // Used when a single server is emulating many servers, by using the
        // incoming partition number as the node ID
        start_i = 0;
        end_i   = PASIS_MAX_NUM_SERVERS;
    } else {
        start_i = node_id;
        end_i   = node_id + 1;
    }

    for (i = start_i; i < end_i; i++) {     
        MDObjectID root_object_id(i, &PASIS_MD_ROOT_FID_C);

        MDVersionedObject* root_object =
            md_universe->create(root_object_id, PASIS_MD_DIR_BLK_SIZE,
                                &distribution);

        error = pasis_md_op_createroot(root_object);
        if (error) return error;
    }

    return PASIS_SUCCESS;
}


static void
stats_init (void)
{
    memset (&stats, 0, sizeof (local_stats_t));

    /* Set up timers */
    stats.timers = Pasis_TimerInit ("Pasis_MD_Timers", PASIS_MD_TIMER_NUM, 0);

    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_READ_HISTORY,
                         "read history");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_READ_OPERATION,
                         "read operation");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_CREATE, "create");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_LINK, "link");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_UNLINK, "unlink");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_GETATTR, "getattr");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_SETATTR, "setattr");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_RENAME, "rename");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_LOOKUP, "lookup");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_READLINK, "readlink");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_READDIR, "readdir");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_BARRIER, "barrier");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_READ_EC, "read_ec");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_WRITE_EC, "write_ec");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_SYNTH_QUERY, "synth_query");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_SYNTH_UPDATE, "synth_update");

    pasis_md_auth_stats_init ();
    pasis_md_validation_stats_init ();
    pasis_md_fid_lock_stats_init ();
    pasis_md_fid_lock_stats_init ();
    //pasis_md_hash_stats_init ();
    pasis_md_object_sync_stats_init ();
    //pasis_md_sync_stats_init ();
    //pasis_md_existence_stats_init ();
    pasis_md_work_stats_init ();
}

static void
stats_dump (int shutdown_flag)
{
    if (shutdown_flag == 0) {
        PASIS_PRINT_INFO ("\n\n\n\n\n******************************************"
                          "*****************************\n");
        PASIS_PRINT_INFO ("Statistics for time = %ld\n", time (NULL));
    }


    PASIS_PRINT_INFO ("\nStatistics:\n");
    PASIS_PRINT_INFO ("-----------\n");
   
    PASIS_PRINT_INFO ("Total Requests       :  %u\n", stats.total_requests);
    PASIS_PRINT_INFO ("Max Workers In Use   :  %u/%u\n", 
                      stats.max_in_progress, PASIS_MD_WORKERS);
    PASIS_PRINT_INFO ("Workers Blocked      :  %u\n", stats.workers_blocked);
                      
    
    if (shutdown_flag) {
        Pasis_TimerShutdown (stats.timers);
    } else {
        Pasis_TimerPrintAndReset (stats.timers);
        stats.total_requests = 0;
        stats.workers_blocked = 0;
    }

    pasis_md_auth_stats_dump (shutdown_flag);
    pasis_md_validation_stats_dump ();
    pasis_md_fid_lock_stats_dump (shutdown_flag);
    pasis_md_fid_lock_stats_dump (shutdown_flag);
    //pasis_md_hash_stats_dump (shutdown_flag);
    pasis_md_object_sync_stats_dump (shutdown_flag);
    //pasis_md_sync_stats_dump (shutdown_flag);
    //pasis_md_existence_stats_dump (shutdown_flag);
    pasis_md_work_stats_dump (shutdown_flag);
}


static void
usage (char *name)
{
    const char* usage =
        "Usage: %s [-n #] -c <config. filename> \n"
        "  -n #             node ID\n"
        "                   This is optional; if not specified, \n"
        "                   incoming partition # will be used as node ID\n"
        "  -c <filename>    configuration filename\n"
        "  -H               history cache size in MB (default %dM)\n"
        "  -M               data block cache size in MB (default %dM)\n"
        "  -A               don't calculate Authenticators; uses memset(0)\n"
        "  -T               don't transfer Authenticators; (implies -A)\n"
        "  -S <# buckets,bucket size in ms> Track # of completed requests in fixed-sized buckets\n\n";

    PASIS_PRINT_INFO (usage, name, PASIS_MD_DEFAULT_HISTORY_CACHE_SIZE / 1024 / 1024,
                      PASIS_MD_DEFAULT_BLOCK_CACHE_SIZE / 1024 / 1024);
}


int
main (int argc, char *argv[])
{
    char *filename = NULL;
    st_thread_t shutdown_thread_id;
    pasis_error_t error;
    pasis_fe_s4_ipc_t ipc_op;
    pasis_config_server_list_t server_list;
    int ret;
    p_uint64 block_cache_size, history_cache_size;
    p_uint32 num_buckets = 1, bucket_size = (p_uint32)-1;
    char opts[64];
    char c;

    strcpy (opts, "c:M:H:n:S:AT");

    // Default cache sizes
    block_cache_size   = PASIS_MD_DEFAULT_BLOCK_CACHE_SIZE;
    history_cache_size = PASIS_MD_DEFAULT_HISTORY_CACHE_SIZE;

    not_using_backend = 1; // for now

    while ((c = getopt (argc, argv, opts)) != -1) {
        switch (c) {
        case 'n':
            sscanf (optarg, "%u", &node_id);
            break;
        case 'c':  /* Configuration filename */
            filename = optarg;
            break;

        case 'H':  /* History cache size */
            history_cache_size = atoi(optarg) * 1024 * 1024;
            break;

        case 'M': { /* Data block cache size */
            char unit_char = optarg[strlen(optarg) - 1];
            block_cache_size = atoi(optarg);
            if (unit_char == 'k' || unit_char == 'K') {
                printf("Block cache specified as %"PRIu64" KB\n", block_cache_size);
                block_cache_size *= 1024;
            } else {
                block_cache_size *= 1024 * 1024;
            }
            break;
        }

        case 'A':
            DO_NOT_CALCULATE_AUTHENTICATORS = 1;
            break;

        case 'T':
            DO_NOT_TRANSFER_AUTHENTICATORS = 1;
            break;

        case 'S': /* Paramets for tracking number of issued requests */
            ret = sscanf(optarg, "%u,%u", &num_buckets, &bucket_size);
            if (ret != 2) {
                fprintf(stderr, "Error specifying arguments for -S option\n");
                usage(argv[0]);
                exit(1);
            }

            break;

        default:
            usage (argv[0]);
            exit (1);
        }
    }

    if (DO_NOT_TRANSFER_AUTHENTICATORS && !DO_NOT_CALCULATE_AUTHENTICATORS) {
        PASIS_PRINT_INFO ("-T implies -A; so I'm turning off authenticators.\n");
        DO_NOT_CALCULATE_AUTHENTICATORS = 1;
    }

    if (node_id >= PASIS_MAX_NUM_SERVERS) {
        PASIS_PRINT_INFO ("Using incoming partition number as node ID.\n");
    }

    if (filename == NULL) {
        usage (argv[0]);
        exit (1);
    }

#if 1
    ret = st_set_eventsys(ST_EVENTSYS_ALT); /* use epoll instead of poll */
    printf("%s: using the %s StateThread event notification system\n",
           __FUNCTION__, st_get_eventsys_name());
#endif

    st_init ();
    st_timecache_set(1);

    shutdown_cond = st_cond_new();

    /* Parse config file, which contains protocol/scheme/storage node list */
    if (PasisIO_ParseConfigurationFile (filename, &distribution.protocol, &server_list,
                                        NULL, NULL)) {
        PASIS_PRINT_ERR ("Error: configuration file parsing failed\n");
        exit (1);
    }

    PASIS_PRINT_INFO_FLUSH ();

    //pasis_debug_print_enable = PASIS_DEBUG_S4_MD_OBJECT_SYNC;
    pasis_debug_print_enable = 0;

    /* Put linkage records into global distribution structure */
    PasisIO_ServerListToLinkageRecords (&server_list,
                                   (pasisio_linkage_t*)&distribution.linkages);

    stats_init ();

    /* Setup signal handler */
    if (signal (SIGINT, sig_handler) == SIG_ERR) {
        PASIS_PRINT_ERR ("Error: installing signal handler for SIGINT\n");
        exit (1);
    }
    if (signal (SIGTERM, sig_handler) == SIG_ERR) {
        PASIS_PRINT_ERR ("Error: installing signal handler for SIGTERM\n");
        exit (1);
    }

#ifdef DEBUG
    /* When in debug compile, SIGUSR2 causes a printout of the stack */
    if (signal (SIGUSR2, _st_print_thread_stacks) == SIG_ERR) {
        fprintf(stderr, "PASIS SHUTTING DOWN...\n");
        PASIS_PRINT_ERR ("Error: installing signal handler for SIGUSR2\n");
        exit (1);
    }
#endif

    /* Open a log file -- needs to be above worker thread creation */
    pasis_md_log_init (node_id);

    /* Setup worker threads */
    error = pasis_md_s4_workers_init ();
    if (error) {
        exit (1);
    }

    /* Initialize object locking/operation serialization module */
    pasis_md_fid_lock_init (1);
    if (error) {
        exit (1);
    }

    /* Initialize authentication module  */
    error = pasis_md_auth_init ();
    if (error) {
        exit (1);
    }

    /* Setup attr cache */
    //error = pasis_md_obj_init ();
    if (error) {
        exit (1);
    }

    /* Setup object syncing code */
    error = pasis_md_object_sync_init (&distribution, node_id);
    if (error) {
        exit (1);
    }

    /* Setup dispatch code */
    error = pasis_md_work_init (&distribution, block_cache_size,
                                history_cache_size);
    if (error) {
        exit (1);
    }

    /* Setup the object-existence code (eliminates unnecessary lookups) */
    //pasis_md_existence_init ();

    /* Create the root directory in case we're hosting the NFS metadata service */
    error = pasis_md_create_root (node_id);
    if (error) {
        exit (1);
    }

    /* Setup the delay-calibration code (used in synthetic operations) */
    error = pasis_md_delay_init ();
    if (error) {
        exit (1);
    }

    /* Setup code for measuring the number of outstanding requests */
    //pasis_md_stats_init(num_buckets, bucket_size);

    /* Setup RPC */
    rpc_init ();
    rpc_auth_server_init ();
    rpc_handle = rpc_server_create_tcp (PASIS_MD_WORKERS, PASIS_MD_RPC_PORT,
                                        PASIS_MD_RPC_PROGNUM, 
                                        PASIS_MD_RPC_VERSION,
                                        pasis_md_s4_rpc_dispatch);

    if (rpc_handle == NULL) {
        exit (1);
    }

    printf("%s: initialized.  Waiting for requests.\n", __FUNCTION__);

    /* Wait for shutdown */
    while (shutdown_flag == 0) {
        st_cond_wait (shutdown_cond);
    }

    /* Do shutdown */
    
    /* Signal rpc to shutdown -- stops incoming connections */
    rpc_signal_shutdown (rpc_handle);

    /* Wait for all workers to complete */
    pasis_md_s4_workers_shutdown ();

    /* Shutdown authentication code */
    pasis_md_auth_shutdown ();

    /* Shutdown attribute cache */
    //pasis_md_obj_shutdown ();

    /* Shutdown object locking */
    pasis_md_fid_lock_shutdown ();

    /* Shutdown object syncing code */
    pasis_md_object_sync_shutdown ();

    /* Shutdown syncer */
    //pasis_md_syncer_shutdown ();

    /* Shutdown dispatch code */
    pasis_md_work_shutdown ();

    /* Shutdown the object-existence code (eliminates unnecessary lookups) */
    //pasis_md_existence_shutdown ();

    /* Shutdown delay code */
    pasis_md_delay_shutdown ();

    /* Shutdown rpc layer */
    rpc_auth_server_shutdown ();
    rpc_shutdown ();

    /* Dump stats */
    stats_dump (1);
    //pasis_md_stats_shutdown();

    /* Close log file */
    pasis_md_log_shutdown ();

    return 0;
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */

