/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/* Does a read history call to the storage node. */

#include <math.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_error.h"
#include "pasis_timer.h"
#include "pasis_macosx_compat.h"

#include "pasisio.h"
#include "pasisio_config_parse.h"
#include "pasisio_drive_md.h"
#include "pasisio_drives.h"
#include "pasisio_select.h"

#include "pasis_md_prot.h"
#include "pasis_md_clnt.h"
#include "pasis_md_hist_matrix.h"

static pasisio_linkage_t       linkages[PASIS_MAX_NUM_SERVERS];
static pasis_md_protocol_t     protocol;

#define                        BLOCK_SIZE   16384

static pasis_config_server_list_t  server_list;

static rpc_auth_t                auth;
static rpc_handle_t             *client;

static unsigned int              shm_region_size; 
static void                     *shm_region;

static p_byte                   *read_buffer;
static p_byte                   *write_buffer;


static pasis_md_read_op_args_t   read_hist_args;
static pasis_md_read_op_res_t    read_hist_res;



static void
dump_params (void)
{
    printf(">>>>>>>>>> Dumping Params\n");
    printf ("Protocol:  N=%u, Qc=%u, B=%u, T=%u\n", protocol.N, 
            protocol.Qc, protocol.B, protocol.T);
}


static int
init (void)
{
    pasisio_drive_instance_t *drive;
    pasis_error_t status;

    st_init ();

    Pasis_TimeInit ();

    /* Read/write buffers must be in shared memory, in case ED is running
     * in different process */
    shm_region_size = PAGE_ALIGN (2 * BLOCK_SIZE);
    shm_region = mmap (0, shm_region_size, PROT_READ | PROT_WRITE,
                       MAP_ANONYMOUS | MAP_SHARED, -1, 0);
    if (shm_region == MAP_FAILED) {
        perror ("Mmap failed");
        return -1;
    }
    read_buffer  = shm_region;
    write_buffer = shm_region + BLOCK_SIZE;

    /* Initialize PASIS */
    status = PasisIO_Init (1, BLOCK_SIZE);
    if (status) {
        PASIS_PRINT_ERR ("PasisIO_Init error: %s\n", PASIS_STRERROR (status));
        return -1;
    }

    /* Connect to the storage nodes */
    status = PasisIO_ConnectToStorageNodes (&server_list, 
                                            PASISIO_DRIVE_TYPE_MD_INDEX,
                                            linkages, NULL);
    if (status) {
        PASIS_PRINT_ERR ("PasisIO_ConnectToStorageNodes failed: %s\n",
                         PASIS_STRERROR (status));
        return -1;
    }
    
    /* Get S4 drive struct */
    drive = PasisIO_DriveFindInstanceName (PASISIO_DRIVE_TYPE_MD_INDEX,
                                           "localhost", "/1");
    if (drive == NULL) {
        PASIS_PRINT_ERR ("Error: PasisIO_DriveFindInstanceName failed\n");
        exit (1);
    }

    client = ((pasisio_drive_md_t *)drive->priv)->md_client;

    /* Setup authorization struct */
    auth.uid = 100;
    auth.gid = MD_DEFAULT_GID;

    return 0;
}


void
dump_history (pasis_md_history_rec_t *history, int num,
              pasis_md_auth_t* authenticators, int auth_num)
{
    int i;

    fprintf(stderr, "--------------------------------------------------\n");

    for (i = 0; i < num; i++) {
        PASIS_PRINT_INFO ("Write TS : %u:%u (%llu) %s",
                          (unsigned)(history[i].ts.time >> 32),
                          (unsigned)(history[i].ts.time),
                          history[i].ts.time,
                          (history[i].ts.barrier_flag) ?
                          "[BARRIER]" : "");

        PASIS_PRINT_INFO ("\n");
    }
}


static void
do_read_root_history (void)
{
    int error;

    read_hist_args.part = 1;
    read_hist_args.read_flags = 0;
    read_hist_args.op_type = PASIS_MD_READ_HISTORY;

    read_hist_args.op_args.read_history.fid_num = 1;
    read_hist_args.op_args.read_history.fid[0] =
        (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C;

    error = pasis_md_clnt_read_op (client, &auth, &read_hist_args,
                                   &read_hist_res);
    if (error || read_hist_res.status != PASIS_FE_SUCCESS) {
        PASIS_PRINT_ERR ("Error: pasis_md_clnt_read_history failed (%d, %d)\n",
                         error, read_hist_res.status);
        exit (1);
    }

    dump_history (read_hist_res.history_vectors[0].histories,
                  read_hist_res.history_vectors[0].history_num,
                  read_hist_res.history_vectors[0].authenticators,
                  read_hist_res.history_vectors[0].auth_num);
}


static void
do_create (void)
{
    pasis_md_write_op_args_t write_args;
    pasis_md_write_op_res_t  write_res;
    pasis_md_fid_t new_fid;
    pasis_md_create_args_t* create = &write_args.op_args.create;
    int i, error;
    static int count = 0;

    write_args.part = 1;
    write_args.op_type = PASIS_MD_CREATE;

    memset (new_fid.data, 9, PASISIO_FILEID_SIZE); 

    /* Setup create directory FID, new file FID, and new filename */
    create->parent_dir_fid = (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C;
    create->create_fid = &new_fid;

    if (count == 0) {
        create->create_name =
            "one small step for pasis, one giant leap for mankind";
    } else {
        create->create_name =
            "two small steps for pasis, two giant leaps for mankind";
    }
    ++count;

    /* Setup create attributes */
    create->create_attr.mode = MD_ATTR_MODE_DIR;
    create->create_attr.ref_cnt = 1;
    create->create_attr.uid = 1;
    create->create_attr.gid = 1;
    create->create_attr.fileid = 5;
    memset(&create->create_attr.mtime, 0, sizeof(create->create_attr.mtime));
    memset(&create->create_attr.atime, 0, sizeof(create->create_attr.atime));
    memset(&create->create_attr.ctime, 0, sizeof(create->create_attr.ctime));
    create->create_attr.blocks = 0;
    create->create_attr.size = 0;

    /* Not a sym link */
    create->symlink_name = NULL;

    /* Set distribution to all 0's, since it's not used */
    memset(&create->distribution, 0, sizeof(pasis_md_loc_t));

    /* Set history matrices */
    write_args.history_matrices_num = 1;
    pasis_md_hist_matrix_init(&write_args.history_matrices[0]);
    for (i = 0; i < 3; ++i) { /* XXX: N = 3 HARDCODED */
        read_hist_res.history_vectors[0].node_id = i;
        pasis_md_hist_matrix_add_vec(&read_hist_res.history_vectors[0],
                                     &write_args.history_matrices[0]);
    }

    error = pasis_md_clnt_write_op (client, &auth, &write_args, &write_res);
    if (error || write_res.status != PASIS_FE_SUCCESS) {
        PASIS_PRINT_ERR ("Error: pasis_md_write_op failed (%d, %d)\n",
                         error, write_res.status);
        exit (1);
    }

    fprintf(stderr, "pasis_md_clnt_write_op succeeded!!!\n");
}


static void
cleanup (void)
{
    munmap (shm_region, shm_region_size);
}


static void
usage (char *name)
{
    PASIS_PRINT_INFO ("Usage: %s [-d] config_filename\n", name);
    PASIS_PRINT_INFO ("  -d             Debug, parse file and exit\n");
}


int
main (int argc, char *argv[])
{
    char                  *filename=NULL;
    char                   c;
    int                    dump=0;

    pasis_debug_print_enable = 0;

    while ((c = getopt (argc, argv, "d")) != -1) {
        switch (c) {
        case 'd':
            dump = 1;
            break;
        default:
            usage (argv[0]);
            exit (1);
        }
    }

    if (optind < argc) {
        filename = argv[optind];
    }

    if (filename == NULL) {
        usage (argv[0]);
        exit (1);
    }

    if (PasisIO_ParseConfigurationFile (filename, &protocol, &server_list,
                                        NULL, NULL)) {
        PASIS_PRINT_ERR ("Error: configuration file parsing failed\n");
        exit (1);
    }

    if (dump) {
        dump_params ();
        exit (0);
    }

    if (init ()) {
        cleanup ();
        exit (1);
    }

    do_read_root_history ();

    do_create ();

    do_read_root_history ();

    do_create ();

    cleanup ();

    pasis_info_print_enable = 0;
    PasisIO_Shutdown ();

    return 0;
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
