/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include <map>
#include <list>
#include <vector>
#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_md_fid_lock.h"
#include "pasis_md_memory_versioning.h"


typedef struct pasis_md_fid_lock_stats_s {
    unsigned int num_order_blocks; /* blocks because of operation ordering */
} pasis_md_fid_lock_stats_t;

static pasis_md_fid_lock_stats_t stats;

static int order_operations_flag = 0;


class Operation;

typedef std::list<Operation*> op_list_t;
typedef std::map<MDObjectID, op_list_t> op_list_map_t;

class Operation
{
 public:
    Operation (p_uint32 n, MDVersionedObject **objects) :
        n (n), objects (objects), num_signals_left(0) {}

    /* Order operation with respect to previous operations.  Here's the scenario
     * we're trying to avoid:
     * 1) Client executes insert ("/", "a")
     * 2) Client executes insert ("/", "b")
     * 3) Client executes setattr ("b")
     * With early termination enabled, a slow S-N could receive operation #2
     * while it's still executing operation #1.  As a result, operation #2 will
     * block waiting to get the lock for the root ("/") directory.  Now,
     * operation #3 arrives at that S-N, and it's able to lock the FID for "b",
     * since operation #2 hasn't tried to lock the FID for "b" yet, as it's
     * waiting to acquire the lock for "/".  As a result, operation #3 is
     * executed by the S-N before operation #2, and the setattr doesn't find the
     * object it's operating on, which causes unnecessary object syncing. 
     *
     * We fix this by ordering operation on the S-N.  Before any FID's are
     * locked, Operation::order records the operation's "intent" to lock its
     * relevant FID's.  If a later operation uses any of these FID's, it will be
     * blocked waiting until the earlier operation finishes with its FID's.
     */
    void order (void) {
        for (p_uint32 i = 0; i < n; ++i) {
            /* Check if we have any pending operations ("pending" in the sense
             * that they are waiting for a lock) that share an object with this
             * operation. */
            op_list_t& op_list = op_lists[objects[i]->oid];

            if (!op_list.empty ()) {
                /* An earlier operation is waiting to lock this object.  We wait
                 * for the earlier operation to complete first, to preserve
                 * operation ordering */
                ++num_signals_left;
            }

            /* Record the fact that this operation is pending. */
            op_list.push_back (this);
        }

        if (num_signals_left > 0) {
            /* There is some operation dependence.  So wait for earlier ops. */
            ++stats.num_order_blocks;
            cond = st_cond_new ();
            while (num_signals_left > 0) {
                st_cond_wait (cond);
            }
        } else {
            cond = NULL;
        }
    }

    void signal_waiters (void) {
        for (p_uint32 i = 0; i < n; ++i) {
            op_list_t& op_list = op_lists[objects[i]->oid];
            PASIS_DASSERT (op_list.empty() == false);

            Operation* front_op = op_list.front();
            PASIS_DASSERT (front_op == this);
            op_list.pop_front();

            if (op_list.empty()) {
                /* Remove this object from the map */
                op_lists.erase(objects[i]->oid);
            } else {
                /* Signal the operation waiting for this object */
                front_op = op_list.front();
                PASIS_DASSERT (front_op->num_signals_left > 0);
                --front_op->num_signals_left;
                st_cond_signal (front_op->cond);
            }
        }

        if (cond) {
            st_cond_destroy (cond);
        }
    }

    p_uint32            n;
    MDVersionedObject **objects;
    st_cond_t           cond;
    p_uint32            num_signals_left;

    static op_list_map_t op_lists;
};


op_list_map_t Operation::op_lists;


int
pasis_md_fid_lock_init (int _order_operations_flag)
{
    order_operations_flag = _order_operations_flag;
    return 0;
}

void
pasis_md_fid_lock_shutdown (void)
{
}


void
pasis_md_fid_lock_stats_init (void)
{
    memset (&stats, 0, sizeof (pasis_md_fid_lock_stats_t));
}


void
pasis_md_fid_lock_stats_dump (int shutdown_flag)
{
    PASIS_PRINT_INFO ("--- Statistics: FID locking ---\n");
    PASIS_PRINT_INFO ("Num order blocks: %u\n", stats.num_order_blocks);

    if (shutdown_flag == 0) {
        stats.num_order_blocks = 0;
    }
}


static int compar(const void *a, const void *b)
{
    MDVersionedObject *one = (MDVersionedObject*)a, *two = (MDVersionedObject*)b;
    return one->oid.cmp(two->oid);
}


/* Lock N objects.  Objects are sorted first, on OID, so we can do 2 phase
 * locking, and avoid deadlocks. */
void
pasis_md_fid_lock (p_uint32            n,
                   MDVersionedObject **objects)
{
    MDVersionedObject *sorted_objects[MD_MAX_OBJS_PER_REQ];

    /* Sort objects; notice our sorting function looks at the object the pointer
     * points to.  for consistency, don't change the order of the caller's
     * objects; hence, we need to work with our own copy of the pointers. */
    PASIS_DASSERT (n <= MD_MAX_OBJS_PER_REQ);
    memcpy(sorted_objects, objects, n * sizeof(MDVersionedObject*));
    qsort(sorted_objects, n, sizeof(MDVersionedObject*), compar);

    Operation op(n, sorted_objects);
    if (order_operations_flag) {
        /* There is a very subtle locking phenomena that you run into when you
         * lock multiple objects at a time.  See comment in the above Operation
         * class. To ensure this problem doesn't happen, we serialize some
         * operations if they depend on each earlier operation who are waiting
         * for locks. */
        op.order();
    } else {
        /* we don't do operation ordering when locking
         * occurs in the client side */
    }

    /* Lock each object */
    for (p_uint32 i = 0; i < n; ++i) {
        sorted_objects[i]->lock();
    }

    /* Now that we've locked this operation's objects, we remove this operation
     * from the list containing which operations are pending lock ownership */
    if (order_operations_flag) {
         /* signal any operations who were serialized after us */
        op.signal_waiters ();
    }
}


void
pasis_md_fid_unlock (p_uint32            n,
                     MDVersionedObject **objects)
{
    for (p_uint32 i = 0; i < n; ++i) {
        objects[i]->unlock();
    }
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
