/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/* This history code used to store the operation in the history table, and hence
 * it supported storing variable-sized history entries.
 * But now, we don't store the operation in the history table anymore, and hence
 * the code should be simplified, since all the entries are fixed-size.
 * For now, stick with the "2-pass" code that tries to find entry boundaries, as
 * I don't have time to change the code.  But that can taken out.
 * -mabdelmalek 20/3/2005
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_debug.h"

#include "pasis_cache.h"

#include "ssw_time.h"


#include "pasis_md_s4.h"
#include "pasis_md_auth.h"
#include "pasis_md_hist.h"
#include "pasis_md_prot.h"
#include "pasis_md_xdr.h"

#include "pasis_freelist.h"

#define PASIS_MD_TIMER_HIST_NUM                 7

#define PASIS_MD_TIMER_HIST_GET                 0
#define PASIS_MD_TIMER_HIST_INSERT              1
#define PASIS_MD_TIMER_HIST_GET_OP              2
#define PASIS_MD_TIMER_HIST_COMPACTION          3
#define PASIS_MD_TIMER_HIST_UPDATE_LAST_S4_TIME 4
#define PASIS_MD_TIMER_HIST_GET_S4_TIME         5
#define PASIS_MD_TIMER_HIST_GET_LATEST          6




typedef struct pasis_md_hist_stats_s {
    pasis_timer_state_t  *timers;
    /* If you add any counters, reset them in pasis_fe_work_stats_dump */
} pasis_md_hist_stats_t;

typedef struct hist_tag_s {
    s4_oid_t     oid;
} hist_tag_t;

static pasis_md_hist_stats_t stats;

/* Temporary history block space */
static pasis_freelist_t *hist_blk_freelist;

static pasis_cache_t *hist_cache;
static unsigned int   cache_hits=0, cache_misses=0;


static unsigned int
hist_cache_hash (void *tag,
                 unsigned int num_entries)
{
    hist_tag_t *hist_tag = (hist_tag_t *)tag;

    PASIS_DASSERT(0); // should use FID, not obsolescent OID
    return (hist_tag->oid.high ^ hist_tag->oid.low) % num_entries;
}


static int
hist_cache_compare (void *tag1,
                    void *tag2)
{
    hist_tag_t *hist_tag1 = (hist_tag_t *)tag1;
    hist_tag_t *hist_tag2 = (hist_tag_t *)tag2;

    if (S4_OID_CMP(&hist_tag1->oid, &hist_tag2->oid, ==)) {
        return 1;
    }

    return 0;
}


static inline pasis_error_t
hist_read_hdr (pasis_fe_s4_ipc_t             *ipc_op,
               s4_oid_t                       oid,
               s4_ipc_args_in_lbykeytable_t  *lkup_args_in,
               s4_ipc_args_out_lbykeytable_t *lkup_args_out,
               p_byte                        *block)
{
    pasis_md_hist_key_t key;
    pasis_error_t       error;
    p_byte             *cache_data=NULL;
    hist_tag_t          cache_tag;

    /* Setup lookup args -- necessary even if lookup hits in cache */
    lkup_args_in->oid = oid;
    S4_TIME_LATEST (&lkup_args_in->when);
    lkup_args_in->keyLen = sizeof (pasis_md_hist_key_t);
    lkup_args_in->getAttr = S4_FALSE;
    lkup_args_in->setAttr = S4_FALSE;
    lkup_args_in->sync = S4_FALSE;

    /* Lookup header */
    S4_TIME_LATEST (&key.server_ts);

    /* Do cache lookup */
    cache_tag.oid = oid;

    cache_data = PasisCache_Lookup (hist_cache, &cache_tag, 0);
    if (cache_data != NULL) {
        cache_hits++;
        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: hist cache hit\n",
                       __FUNCTION__);

        /* Could optimize not to do copy if we track this block and don't
         * issue any ipcs while block is in use */
        memcpy (block, cache_data, PASIS_MD_HIST_BLK_SIZE);

        return PASIS_SUCCESS;
    } else {
        cache_misses++;
    }

#if 1
    /* Not going to backend */
    PASIS_DASSERT (error == PASIS_SUCCESS);
#else
    error = s4_ipc_table_key_lookup (ipc_op, lkup_args_in, lkup_args_out,
                                     (p_byte *)&key, block);
#endif

    if (!error) {
        PasisCache_Insert (hist_cache, &cache_tag, block);
    }

    return error;
}


static inline pasis_error_t
hist_update_block (pasis_fe_s4_ipc_t             *ipc_op,
                   s4_oid_t                       oid,
                   pasis_md_hist_key_t           *key,
                   p_byte                        *block)
{
#if 1
    /* Not going to backend */
    return PASIS_SUCCESS;
#else
    s4_ipc_args_in_insenttable_t  ins_args_in;
    s4_ipc_args_out_insenttable_t ins_args_out;
    pasis_error_t error;

    /* Setup table insert args */
    ins_args_in.oid = oid;
    ins_args_in.keyLen = sizeof (pasis_md_hist_key_t);
    ins_args_in.dataLen = PASIS_MD_HIST_BLK_SIZE;
    ins_args_in.getAttr = S4_FALSE;
    ins_args_in.setAttr = S4_FALSE;
    ins_args_in.sync = S4_FALSE;

    /* Update S4 */
    error = s4_ipc_table_insert (ipc_op, &ins_args_in, &ins_args_out,
                                 (p_byte *)key, block);
    return error;
#endif
}


static void
hist_find_block (p_byte                        *block, 
                 pasis_md_hts_t                *ts,
                 pasis_fe_s4_ipc_t             *ipc_op,
                 s4_ipc_args_in_lbykeytable_t  *lkup_args_in,
                 s4_ipc_args_out_lbykeytable_t *lkup_args_out)
{
    pasis_md_hist_ent_t *ent_ptr;
    pasis_md_hist_blk_hdr_t *blk_hdr;
    pasis_error_t error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: find_block called\n", 
                   __FUNCTION__);

    blk_hdr = (pasis_md_hist_blk_hdr_t *)(block + sizeof(pasis_md_hist_hdr_t));

    ent_ptr = (pasis_md_hist_ent_t *)(block + sizeof(pasis_md_hist_hdr_t) +
                                      sizeof(pasis_md_hist_blk_hdr_t));

    /* Iterate through blocks looking for the place to insert the entry */
    while (1) {
        pasis_md_hist_ent_t *entry;

        /* Find first entry and check if insert should occur in this block */
        entry = &ent_ptr[0];

        /* If no next block exists or TS fits in this block, insert here */ 
        if (PASIS_MD_HIST_KEY_INVALID (&blk_hdr->next_blk_key) ||
            md_hts_compare (ts, &entry->client_ts) >= 0) { 
            break;
        }

        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: checking for next block\n",
                       __FUNCTION__);
        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: doing lookup\n",
                       __FUNCTION__);
        

#if 1
    /* Not going to backend */
    PASIS_DASSERT (0);
#else
        /* Do lookup for next block */
        error = s4_ipc_table_key_lookup (ipc_op, lkup_args_in, 
                                         lkup_args_out,
                                         (p_byte *)&blk_hdr->next_blk_key, 
                                         block);

        PASIS_DASSERT (error == PASIS_SUCCESS);
#endif
    }

    return;
}

#if 0
static void
dump_ents (pasis_md_hist_ent_t    *ents,
           int                     num)
{
    pasis_md_hist_ent_t* entry = ents;
    int i;
    printf ("%s --------------------------------\n", __FUNCTION__);
    //for (i = num-1; i >= 0; i--) {
    for (i = 0; i < num; i++) {
        printf ("[%d], %llu, ", i, entry->client_ts.time);
#if DEBUG_TIMESTAMPS > 0
        printf("op description: %s, ", entry->client_ts.op_description);
#endif
        printf ("\n");
        entry = entry + 1;
    }
}
#endif

static void
hist_insert_block (p_byte                  *block, 
                   pasis_md_hist_ent_t     *new_ent)
{
    pasis_md_hist_ent_t *ent_ptr; /* points to beginning of entries */
    pasis_md_hist_ent_t *entry;   /* points to iteration entry */
    pasis_md_hist_blk_hdr_t *blk_hdr;
    int idx;
    
    PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s\n", __FUNCTION__);
    
    blk_hdr = (pasis_md_hist_blk_hdr_t *)(block + sizeof(pasis_md_hist_hdr_t));
    
    ent_ptr = (pasis_md_hist_ent_t *)(block + sizeof(pasis_md_hist_hdr_t) +
                                      sizeof(pasis_md_hist_blk_hdr_t));
    
    PASIS_DASSERT (blk_hdr->free_space <= PASIS_MD_HIST_BLK_USEFUL_SPACE);
    PASIS_DASSERT (blk_hdr->free_space >= sizeof(pasis_md_hist_ent_t));
    
    PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: num_entries=%d\n", 
                   __FUNCTION__, blk_hdr->num_entries);
    
    entry = ent_ptr;
    idx = 0;
    
    while (idx < blk_hdr->num_entries) {
        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4,
                       "comparing index=%d: %llu >= %llu\n", idx,
                       new_ent->client_ts.time, entry->client_ts.time);

        if (md_hts_compare (&new_ent->client_ts, &entry->client_ts) < 0) {
            /* Insert here -- shift entries down by 1 */
            u_int32_t used =  PASIS_MD_HIST_BLK_USEFUL_SPACE -
                              blk_hdr->free_space;
            u_int32_t unmoved = (char*)entry - (char*)ent_ptr;
            u_int32_t moved = used - unmoved;

            PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4,
                           "%s: inserting at index=%d, shifting down\n",
                           __FUNCTION__, idx);

            memmove (entry + sizeof(pasis_md_hist_ent_t), entry, moved);
            break;
        }
        entry = entry + 1;
        ++idx;
    }

    blk_hdr->num_entries++;
    blk_hdr->free_space -= sizeof(pasis_md_hist_ent_t);

    PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: inserting entry at index=%d\n", 
                   __FUNCTION__, idx);

    /* Copy entry into version history. */
    memmove (entry, new_ent, sizeof(pasis_md_hist_ent_t));

#if 0
    dump_ents (ents, blk_hdr->num_entries);
#endif

    return;
}


static int
hist_compact (pasis_fe_s4_ipc_t             *ipc_op,
              s4_ipc_args_in_lbykeytable_t  *lkup_args_in,
              s4_ipc_args_out_lbykeytable_t *lkup_args_out,
              p_byte                        *block, 
              pasis_md_hts_t                *lcwts,
              int                            is_first_hist_block)
{
    pasis_md_hist_ent_t *ent_ptr;
    pasis_md_hist_blk_hdr_t *blk_hdr;
    pasis_md_hist_ent_t *entry;
    int i, delete_next_block = 0;
    int entry_indices[PASIS_MD_HIST_MAX_VERSIONS];
    pasis_error_t error;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s\n", __FUNCTION__);

    blk_hdr = (pasis_md_hist_blk_hdr_t *)(block + sizeof(pasis_md_hist_hdr_t));
    ent_ptr = (pasis_md_hist_ent_t *)(block + sizeof(pasis_md_hist_hdr_t) +
                                      sizeof(pasis_md_hist_blk_hdr_t));
    PASIS_DASSERT (blk_hdr->free_space <= PASIS_MD_HIST_BLK_USEFUL_SPACE);

    /* We will soon be iterating through the history block in reverse order.  We
     * need to first create "pointers" to the entries, because their
     * variable-size nature makes reverse iteration non-trivial */
    entry = ent_ptr;
    for (i = 0; i < blk_hdr->num_entries; ++i) {
        entry_indices[i] = (u_int32_t*)entry - (u_int32_t*)ent_ptr;
        entry = entry + 1;
    }


#if 1
    /* Be aggressive about removing entries as quickly as possible */
    i = blk_hdr->num_entries - 1;
#else
    /* We don't want to be over-aggressive with GC, in case another client is
     * doing a query at a specific timestamp.  By keeping some extra versions
     * around, we can satisfy their query request. -mabdelmalek 20/3/2005 */
    i = MAX (blk_hdr->num_entries - 1 - 6, 0); // keep 6 versions around
#endif
    /* Delete entries from the current history block */
    for (; i >= 0; --i) {
        entry = (pasis_md_hist_ent_t *)((u_int32_t*)ent_ptr +
                                        entry_indices[i]);

        if (md_hts_compare (lcwts, &entry->client_ts) > 0) {
            /* Remove unnecessary entries above by shifting entries below up */
            pasis_md_hist_ent_t *new_first_entry;
            u_int32_t used, deleted, moved;

            new_first_entry = entry + 1;
            used =  PASIS_MD_HIST_BLK_USEFUL_SPACE - blk_hdr->free_space;
            deleted = (char*)new_first_entry - (char*)ent_ptr;
            moved = used - deleted;

            memmove (ent_ptr, new_first_entry, moved);

            blk_hdr->num_entries -= i + 1;
            blk_hdr->free_space += deleted;

            /* We can delete any successive history blocks.  Since 1) we deleted
             * entries from this block, and 2) history blocks are ordered, the
             * remaining history blocks must have entries less than lcwts. */
            delete_next_block = 1;
            break;
        }
    }

    /* If we have not already decided to delete the successive history blocks,
     * recurse and see which entries we can delete from them. */
    if (!delete_next_block &&
        !PASIS_MD_HIST_KEY_INVALID (&blk_hdr->next_blk_key)) {
        p_byte *next_block;
        int deleted_all_entries;

        /* Get a buffer to hold the next history block */
        next_block = PasisFreelist_Get (hist_blk_freelist);

        /* Read next history block */
#if 1
        /* Not going to backend */
        PASIS_DASSERT (0);
#else
        error = s4_ipc_table_key_lookup (ipc_op, lkup_args_in, lkup_args_out, 
                                         (p_byte *)&blk_hdr->next_blk_key,
                                         next_block);
        PASIS_DASSERT (error == PASIS_SUCCESS);
#endif

        /* Try to delete entries from the next block */
        deleted_all_entries = hist_compact (ipc_op, lkup_args_in, lkup_args_out,
                                            next_block, lcwts, 0);

        /* If next block is now completely empty, "delete" it */
        if (deleted_all_entries) {
            delete_next_block = 1;
        }

        /* Release memory buffer */
        PasisFreelist_Put (hist_blk_freelist, next_block);
    }

    if (delete_next_block) {
        memset (&blk_hdr->next_blk_key, 0, sizeof(pasis_md_hist_key_t));

        if (is_first_hist_block) {
            /* We don't need to write back the history block to S4, because
             * caller will do this (and they'll also updated the cache). */
        } else {
            error = hist_update_block (ipc_op, lkup_args_in->oid,
                                       &blk_hdr->this_blk_key, block);
            PASIS_DASSERT (error == PASIS_SUCCESS);
        }
    }

    /* This is a recursive function; so we count the timer only during the
     * first invocation */
    if (is_first_hist_block) {
        PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_COMPACTION);
    }

    return blk_hdr->num_entries == 0;
}


pasis_error_t
pasis_md_hist_init (void)
{
    hist_blk_freelist = PasisFreelist_CreateInitFixed(PASIS_MD_HIST_BLK_SIZE,
                                                      PASIS_MD_WORKERS * 2, 
                                                      NULL, NULL);

    hist_cache = PasisCache_Create (PASIS_MD_HIST_CACHE_NUM, 1, 
                                    sizeof (hist_tag_t), 
                                    PASIS_MD_HIST_BLK_SIZE,
                                    hist_cache_hash, hist_cache_compare);

    return PASIS_SUCCESS;
}


void
pasis_md_hist_shutdown (void)
{
    PasisFreelist_Destroy (hist_blk_freelist);
    PasisCache_Shutdown (hist_cache);
}


pasis_error_t
pasis_md_hist_update_last_s4_time (pasis_fe_s4_ipc_t *ipc_op,
                                   s4_oid_t           oid)
{
    p_byte                   *blk;
    pasis_md_hist_hdr_t      *hdr;
    pasis_md_hist_blk_hdr_t  *blk_hdr;

    pasis_error_t             error;
    hist_tag_t                cache_tag;

    s4_ipc_args_in_lbykeytable_t  lkup_args_in;
    s4_ipc_args_out_lbykeytable_t lkup_args_out;

    PASIS_TIMER_START ();

    blk = PasisFreelist_Get (hist_blk_freelist);
    hdr = (pasis_md_hist_hdr_t *)(blk);
    blk_hdr = (pasis_md_hist_blk_hdr_t *)(blk + sizeof (pasis_md_hist_hdr_t));

    /* Read in header */
    error = hist_read_hdr (ipc_op, oid, &lkup_args_in, &lkup_args_out, blk);
    
    if (error != PASIS_SUCCESS) {
        PasisFreelist_Put (hist_blk_freelist, blk);
        PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_UPDATE_LAST_S4_TIME);
        return error;
    }

    /* Update timestamp */
    S4Time_GetTime (&hdr->last_write_ts);

    /* Write out hdr */
    cache_tag.oid = oid;
    PasisCache_Insert (hist_cache, &cache_tag, blk);
    error = hist_update_block (ipc_op, oid, &blk_hdr->this_blk_key, blk);

    PasisFreelist_Put (hist_blk_freelist, blk);
    
    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_UPDATE_LAST_S4_TIME);

    return error;
}



pasis_error_t
pasis_md_hist_get_entries_and_auth (pasis_fe_s4_ipc_t       *ipc_op,
                                    s4_oid_t                 oid,
                                    p_uint32                *num_entries,/*Res*/
                                    pasis_md_history_rec_t  *hist_array, /*Res*/
                                    pasis_md_auth_t         *auth)       /*Res*/
{
    pasis_md_hist_ent_t     *ent_ptr;
    pasis_md_hist_blk_hdr_t *blk_hdr;
    pasis_md_hist_hdr_t     *hdr;
    p_byte                  *blk;
    int                      index;
    int                      entry_indices[PASIS_MD_HIST_MAX_VERSIONS];

    s4_ipc_args_in_lbykeytable_t  lkup_args_in;
    s4_ipc_args_out_lbykeytable_t lkup_args_out;

    pasis_error_t error;

    PASIS_TIMER_START ();

    blk = PasisFreelist_Get (hist_blk_freelist);
    PASIS_DASSERT (blk != NULL);
    hdr = (pasis_md_hist_hdr_t *)(blk);
    blk_hdr = (pasis_md_hist_blk_hdr_t *)(blk + sizeof(pasis_md_hist_hdr_t));

    *num_entries = 0;

    error = hist_read_hdr (ipc_op, oid, &lkup_args_in, &lkup_args_out, blk);
    if (error) {
        PasisFreelist_Put (hist_blk_freelist, blk);
        PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_GET);
        return error;
    }

    ent_ptr = (pasis_md_hist_ent_t *)(blk + sizeof(pasis_md_hist_hdr_t) +
                                      sizeof(pasis_md_hist_blk_hdr_t));

#if STORE_AUTHENTICATORS_IN_HIST_TABLE > 0
    if (auth) {
        memcpy (auth, hdr->authenticators,
                hdr->auth_num * sizeof(pasis_md_auth_t));
    }
#endif

    /* Entries are stored in ascending order in history table, but we need to
     * return them in descending order.  This is hard to do now that the history
     * entries are variable-sized.
     * We do a two-pass algorithm.  First, we find the entry offsets in the
     * history block.  Second, we then copy the entries into the hist_array
     * argument.
     * The above algorithm is performed per history block. */

    index = blk_hdr->num_entries - 1;
    while (1) {
        pasis_md_hist_ent_t *entry;


        /* Check if this is the first pass per block.  If so, calculate entry
         * indices. */
        if (index == blk_hdr->num_entries - 1) {
            int i;
            entry = ent_ptr;
            for (i = 0; i < blk_hdr->num_entries; ++i) {
                entry_indices[i] = (u_int32_t*)entry - (u_int32_t*)ent_ptr;
                entry = entry + 1;
            }
        }

        entry = (pasis_md_hist_ent_t *)((u_int32_t*)ent_ptr +
                                        entry_indices[index]);

        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4,
                       "%s: return idx=%u, index=%u, %llu\n",
                       __FUNCTION__, *num_entries, index,
                       entry->client_ts.time);

        /* Return this entry -- copy entry into history array */
        md_hts_copy (&hist_array[*num_entries].ts, &entry->client_ts);
        hist_array[*num_entries].is_multiobject_op = entry->is_multiobject_op;

        (*num_entries)++;
        PASIS_DASSERT (*num_entries <= PASIS_MD_HIST_MAX_VERSIONS);

        index--;

        /* If there are no more entries in this block, goto next block */
        if (index < 0) {
            /* Exhausted entries in this block, move onto next block */
            if (PASIS_MD_HIST_KEY_INVALID (&blk_hdr->next_blk_key)) {
                break;
            }

            PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "looking up block: key=%llu\n", 
                           *(p_uint64 *)&blk_hdr->next_blk_key.server_ts);


#if 1
            /* Not going to backend */
            PASIS_DASSERT (0);
#else
            error = s4_ipc_table_key_lookup (ipc_op, &lkup_args_in, 
                                             &lkup_args_out, 
                                             (p_byte *)&blk_hdr->next_blk_key,
                                             blk);

            PASIS_DASSERT (error == PASIS_SUCCESS);
#endif

            index = blk_hdr->num_entries - 1;
        }
    }
    
    PasisFreelist_Put (hist_blk_freelist, blk);

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_GET);

    return PASIS_SUCCESS;
}

pasis_error_t
pasis_md_hist_get_entries (pasis_fe_s4_ipc_t       *ipc_op,
                           s4_oid_t                 oid,
                           p_uint32                *num_entries,   /* Result */
                           pasis_md_history_rec_t  *hist_array)    /* Result */
{
    return pasis_md_hist_get_entries_and_auth (ipc_op, oid,
                                               num_entries, hist_array,
                                               NULL);
}


/* General-purpose history table iteration function that's useful for read-only
 * operations */
static void
hist_iterate (pasis_fe_s4_ipc_t       *ipc_op,
              s4_oid_t                 oid,
              int (*func) (pasis_md_hist_hdr_t*, pasis_md_hist_ent_t*, void*, int),
              void*                    arg)
{
    pasis_md_hist_ent_t     *ent_ptr;
    pasis_md_hist_blk_hdr_t *blk_hdr;
    pasis_md_hist_hdr_t     *hdr;
    p_byte                  *blk;
    int                      index, is_last;
    pasis_md_hist_ent_t     *entry;

    s4_ipc_args_in_lbykeytable_t  lkup_args_in;
    s4_ipc_args_out_lbykeytable_t lkup_args_out;

    pasis_error_t error;

    blk = PasisFreelist_Get (hist_blk_freelist);
    PASIS_DASSERT (blk != NULL);
    hdr = (pasis_md_hist_hdr_t *)(blk);
    blk_hdr = (pasis_md_hist_blk_hdr_t *)(blk + sizeof(pasis_md_hist_hdr_t));

    error = hist_read_hdr (ipc_op, oid, &lkup_args_in, &lkup_args_out, blk);
    if (error) {
        PasisFreelist_Put (hist_blk_freelist, blk);
        PASIS_PRINT_ERR ("%s: hist_read_hdr failed: %d\n", __FUNCTION__,
                         error);
        return;
    }

    ent_ptr = (pasis_md_hist_ent_t *)(blk + sizeof(pasis_md_hist_hdr_t) +
                                      sizeof(pasis_md_hist_blk_hdr_t));

    entry = ent_ptr;
    index = blk_hdr->num_entries - 1;
    while (1) {
        is_last = (index == 0 &&
                   PASIS_MD_HIST_KEY_INVALID (&blk_hdr->next_blk_key));

        int isDone = func (hdr, entry, arg, is_last);

        if (isDone == 1) {
            break;
        }

        index--;
        /* If there are no more entries in this block, goto next block */
        if (index < 0) {
            /* Exhausted entries in this block, move onto next block */
            if (PASIS_MD_HIST_KEY_INVALID (&blk_hdr->next_blk_key)) {
                break;
            }

            PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, 
                           "looking up block: key=%llu\n", 
                           *(p_uint64 *)&blk_hdr->next_blk_key.server_ts);


#if 1
            /* Not going to backend */
            PASIS_DASSERT (0);
#else
            error = s4_ipc_table_key_lookup (ipc_op, &lkup_args_in, 
                                             &lkup_args_out, 
                                             (p_byte *)&blk_hdr->next_blk_key,
                                             blk);

            PASIS_DASSERT (error == PASIS_SUCCESS);
#endif

            entry = ent_ptr;
            index = blk_hdr->num_entries - 1;
        } else {
            entry = entry + 1;
        }
    }
    
    PasisFreelist_Put (hist_blk_freelist, blk);
}



typedef struct {
    pasis_md_hts_t      *ts;
    pasis_md_rpc_num_t  *op_type;
    p_uint32            *is_multiobject_op;
    s4_timeval_t        *s4time;
    int                  do_not_use_s4_latest;
    int                  checking_for_all_barriers;
    int                  all_barriers_after_match;
    p_byte               error;
} hist_get_op_info_args_t;

static int
hist_get_op_info_iter_func (pasis_md_hist_hdr_t *hdr,
                            pasis_md_hist_ent_t *entry,
                            void                *arg,
                            int                  is_last)
{
    hist_get_op_info_args_t *args = (hist_get_op_info_args_t*)arg;

    /* the logic blow is tricky, but it's correct :)  I thought I had a
     * condition wrong below, and I wrote this code!  So don't touch it unless
     * you really understand it. -mabdelmalek 21/3/2005  */
    if (args->checking_for_all_barriers) {
        if (entry->client_ts.barrier_flag == 0) {
            /* found a non-barrier */
            args->all_barriers_after_match = 0;
            return 1; /* done */
        } else {
            return 0; /* check next timestamp */
        }
    } else if (md_hts_compare (args->ts, &entry->client_ts) == 0) {
        /* found match;  copy operation info */
        *args->op_type = entry->op_type;
        *args->is_multiobject_op = entry->is_multiobject_op;

        /* Optimize for S4 -- better if we can use S4_TIME_LATEST */
        if (S4_TIME_CMP (&entry->server_ts, &hdr->last_write_ts, == )) {
            if (args->do_not_use_s4_latest) {
                *args->s4time = entry->server_ts;
            } else {
                S4_TIME_LATEST (args->s4time);
            }
            args->error = PASIS_SUCCESS;
            return 1;
        } else {
            *args->s4time = entry->server_ts;
            /* If all the entries after us are barriers, then we can
             * still use S4_TIME_LATEST.  So what we'll do now is look at
             * all the following entries and see if they're barriers.*/
            args->checking_for_all_barriers = 1;
            args->error = PASIS_SUCCESS;
            return 0;
        }
    }

    return 0;
}

pasis_error_t
pasis_md_hist_get_op_info (pasis_fe_s4_ipc_t       *ipc_op,
                           s4_oid_t                 oid,
                           pasis_md_hts_t          *ts,
                           pasis_md_rpc_num_t      *op_type,     /* Result */
                           p_uint32                *is_multiobject_op,/*Res*/
                           s4_timeval_t            *s4time)
{
    hist_get_op_info_args_t args = {ts, op_type, is_multiobject_op, s4time,
                                    0, 0, 1};

    PASIS_TIMER_START ();

    s4time->tv_sec = s4time->tv_usec = 0;
    args.error = PASIS_ERROR;

    hist_iterate (ipc_op, oid, hist_get_op_info_iter_func, (void*)&args);

    if (args.checking_for_all_barriers && args.all_barriers_after_match) {
        S4_TIME_LATEST (s4time);
    }

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_GET_OP);

    return args.error;
}



pasis_error_t
pasis_md_hist_get_s4_time (pasis_fe_s4_ipc_t    *ipc_op,
                           s4_oid_t              oid,
                           pasis_md_hts_t       *ts,
                           int                   do_not_use_s4_latest,
                           s4_timeval_t         *s4time) /* Result */
{
    pasis_md_rpc_num_t op_type;
    p_uint32 is_multiobject_op;
    hist_get_op_info_args_t args = {ts, &op_type, &is_multiobject_op, s4time,
                                    0, 0, 1};

    PASIS_TIMER_START ();
    s4time->tv_sec = s4time->tv_usec = 0;
    args.error = PASIS_ERROR;

    hist_iterate (ipc_op, oid, hist_get_op_info_iter_func, (void*)&args);

    if (args.checking_for_all_barriers && args.all_barriers_after_match) {
        S4_TIME_LATEST (s4time);
    }

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_GET_S4_TIME);

    return args.error;    
}


pasis_error_t
pasis_md_hist_have_entry (pasis_fe_s4_ipc_t    *ipc_op,
                          s4_oid_t              oid,
                          pasis_md_hts_t       *ts)
{
    s4_timeval_t s4time;
    return pasis_md_hist_get_s4_time (ipc_op, oid, ts, 0, &s4time);
}



typedef struct {
    pasis_md_hts_t      *result;
    pasis_error_t        error;
} hist_get_latest_entry_args_t;


static int
hist_get_latest_entry_iter_func (pasis_md_hist_hdr_t *hdr,
                                 pasis_md_hist_ent_t *entry,
                                 void                *arg,
                                 int                  is_last)
{
    hist_get_latest_entry_args_t *args = (hist_get_latest_entry_args_t*)arg;
    if (is_last) {
        md_hts_copy (args->result, &entry->client_ts);
        args->error = PASIS_SUCCESS;
    }
    return 0;
}

pasis_error_t
pasis_md_hist_get_latest_entry (pasis_fe_s4_ipc_t   *ipc_op,
                                s4_oid_t             oid,
                                pasis_md_hts_t      *hts)   /* Result */
{
    hist_get_latest_entry_args_t args = {hts, PASIS_ERROR};
    PASIS_TIMER_START ();
    hist_iterate (ipc_op, oid, hist_get_latest_entry_iter_func, (void*)&args);
    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_GET_LATEST);
    return args.error;
}



typedef struct {
    int count;
} hist_dump_entry_args_t;


static int
hist_dump_entry_iter_func (pasis_md_hist_hdr_t *hdr,
                           pasis_md_hist_ent_t *entry,
                           void                *arg,
                           int                  is_last)
{
    hist_dump_entry_args_t* args = (hist_dump_entry_args_t*) arg;
    printf ("[%d], TS %llu, client %x, barrier %x", args->count,
            entry->client_ts.time,
            entry->client_ts.client_id,
            entry->client_ts.barrier_flag);
#if DEBUG_TIMESTAMPS > 0
    printf("op description: %s, ", entry->client_ts.op_description);
#endif
    printf ("\n");
    args->count++;
    return 0;
}


void
pasis_md_hist_print (pasis_fe_s4_ipc_t    *ipc_op,
                     s4_oid_t              oid)
{
    hist_dump_entry_args_t args;
    args.count = 0;
    hist_iterate (ipc_op, oid, hist_dump_entry_iter_func, &args);
}


pasis_error_t
pasis_md_hist_table_insert (pasis_fe_s4_ipc_t       *ipc_op,
                            s4_oid_t                 oid,
                            p_uint32                 N,
                            p_uint32                 node_id,
                            pasis_md_hts_t          *client_ts,
                            pasis_md_hts_t          *cond_ts,
                            pasis_md_rpc_num_t       op_type,
                            u_int32_t                is_multiobject_op,
                            p_byte                  *create_flag)
{
    pasis_md_hist_ent_t       new_ent;
    pasis_md_hist_hdr_t      *hdr1, *hdr2;
    pasis_md_hist_blk_hdr_t  *blk_hdr1, *blk_hdr2;
    pasis_md_hist_ent_t      *ent1;
    pasis_error_t             error;
    p_byte                   *blk1, *blk2;

#if STORE_AUTHENTICATORS_IN_HIST_TABLE > 0
    pasis_md_history_rec_t    hist_array[PASIS_MD_HIST_MAX_VERSIONS];
    int                       num_entries;
    pasis_md_hist_ent_t      *ent_ptr;
    int                       index;
    int                       entry_indices[PASIS_MD_HIST_MAX_VERSIONS];
#endif

    hist_tag_t                cache_tag;

    int                       hdr1_dirty=0, updating_first_block=1;
    int                       do_compaction_flag=1;

    s4_ipc_args_in_lbykeytable_t  lkup_args_in;
    s4_ipc_args_out_lbykeytable_t lkup_args_out;

    s4_ipc_args_in_insenttable_t  ins_args_in;
    s4_ipc_args_out_insenttable_t ins_args_out;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4,
                   "%s: client_ts=%llu, cond_ts=%llu, barrier_flag=%u\n",
                   __FUNCTION__, client_ts->time, cond_ts->time,
                   client_ts->barrier_flag);

    if (cond_ts->barrier_flag != 0 || client_ts->barrier_flag != 0) {
        do_compaction_flag = 0;
    } 

    ins_args_in.oid = oid;
    ins_args_in.keyLen = sizeof (pasis_md_hist_key_t);
    ins_args_in.dataLen = PASIS_MD_HIST_BLK_SIZE;
    ins_args_in.getAttr = S4_FALSE;
    ins_args_in.setAttr = S4_FALSE;
    ins_args_in.sync = S4_FALSE;

    /* Fill in new history entry */
    md_hts_copy (&new_ent.client_ts, client_ts);
    S4Time_GetTime (&new_ent.server_ts);
    new_ent.op_type = op_type;
    new_ent.is_multiobject_op = is_multiobject_op;

    /* Setup some initial pointers */
    blk1 = PasisFreelist_Get (hist_blk_freelist);
    PASIS_DASSERT (blk1 != NULL);
    hdr1 = (pasis_md_hist_hdr_t *)blk1;
    blk_hdr1 = (pasis_md_hist_blk_hdr_t *)(blk1 + sizeof(pasis_md_hist_hdr_t));

    ent1 = (pasis_md_hist_ent_t *)(blk1 + sizeof (pasis_md_hist_hdr_t) +
                                   sizeof (pasis_md_hist_blk_hdr_t));

    /* Read header block */
    error = hist_read_hdr (ipc_op, oid, &lkup_args_in, &lkup_args_out, blk1);

    /* If header doesn't exist, create */
    if (error) {
        if (error == S4_ERROR_SYS_NOT_FOUND) {
            PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4,
                           "%s: header does not exist, creating.\n",
                           __FUNCTION__);
            memset (hdr1, 0, sizeof (pasis_md_hist_hdr_t));
            memset (blk_hdr1, 0, sizeof (pasis_md_hist_blk_hdr_t));
            blk_hdr1->free_space = PASIS_MD_HIST_BLK_USEFUL_SPACE;

            S4_TIME_LATEST (&blk_hdr1->this_blk_key.server_ts);

            /* Set create flag */
            if (create_flag) { *create_flag = 1; }
        } else {
            PasisFreelist_Put (hist_blk_freelist, blk1);
            PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_INSERT);
            return error;
        }
    } else if (create_flag) {
        *create_flag = 0;
    }

    /* Garbage collection */
    /* delete version entries < than LCWTS.  But only do this if we're NOT
     * conditioning on a barrier. */
    if (do_compaction_flag) {
        hist_compact (ipc_op, &lkup_args_in, &lkup_args_out, blk1, cond_ts, 1);
        /* delete data objects < than LCWTS */
        /* pasis_md_gc_collect(fid, fid_data, cond_ts); -unimplemented */
    }

    /* Set last write timestamp -- S4 time */
    memcpy (&hdr1->last_write_ts, &new_ent.server_ts, 
            sizeof (s4_timeval_t));

    /* Mark header dirty */
    hdr1_dirty = 1;

    /* Set up cache lookup */
    cache_tag.oid = oid;

    /* If hdr block is not going to hold new entry, write it out if dirty */
    if (!(PASIS_MD_HIST_KEY_INVALID (&blk_hdr1->next_blk_key) ||
          md_hts_compare (&new_ent.client_ts, &ent1->client_ts) >= 0)) { 
        
        /* If hdr is dirty, write it out */
        if (hdr1_dirty) {
            PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "cache insert.\n");

#if 1
            PASIS_PRINT_ERR ("%s: not going to S4; too many versions\n",
                             __FUNCTION__);
            PASIS_DASSERT (0);
#else
            /* Insert into history cache first */
            PasisCache_Insert (hist_cache, &cache_tag, blk1);

            error = s4_ipc_table_insert (ipc_op, &ins_args_in, &ins_args_out, 
                                         (p_byte *)&blk_hdr1->this_blk_key, 
                                         blk1);

            PASIS_DASSERT (error == PASIS_SUCCESS);
#endif
            updating_first_block = 0;
            hdr1_dirty = 0;
        }

        /* Traverse block list to find which block should hold new entry */
        hist_find_block (blk1, &new_ent.client_ts,
                         ipc_op, &lkup_args_in, &lkup_args_out);
    } else {
        /* Insert into hdr block, mark dirty to insert into cache later */
        hdr1_dirty = 1;
    }

    /* Check if a split is needed -- if so create new block, fixup pointers */
    if (blk_hdr1->free_space < sizeof(pasis_md_hist_ent_t)) {
        p_byte *ents1, *ents2;
        u_int32_t bytes_moved, bytes_shifted;

#if DO_NOT_USE_S4 > 0
        PASIS_PRINT_ERR("%s: not going to S4, hence can't have too many versions\n",
                        __FUNCTION__);
        PASIS_DASSERT(0);
#endif

        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "%s: Splitting block\n",
                       __FUNCTION__);
        
        /* Setup block 2's pointers -- this will be the result of the split */
        blk2 = PasisFreelist_Get (hist_blk_freelist);
        PASIS_DASSERT (blk2 != NULL);
        hdr2 = (pasis_md_hist_hdr_t *)blk2;
        blk_hdr2 = (pasis_md_hist_blk_hdr_t *)(blk2 + 
                                               sizeof (pasis_md_hist_hdr_t));

        /* Setup pointers to beginning of history entries */
        ents1 =  blk1 + sizeof (pasis_md_hist_hdr_t) +
            sizeof (pasis_md_hist_blk_hdr_t);
        ents2 =  blk2 + sizeof (pasis_md_hist_hdr_t) +
            sizeof (pasis_md_hist_blk_hdr_t);

        /* Fixup block 2's key pointers and number of entries */
#if 1
        /* Put all of but two of the latest versions in the old entry block.
         * Very optimistic!  */
        blk_hdr2->num_entries = blk_hdr1->num_entries - 2;
#else
        blk_hdr2->num_entries = blk_hdr1->num_entries / 2;
#endif
        S4Time_GetTime (&blk_hdr2->this_blk_key.server_ts);
        memcpy (&blk_hdr2->next_blk_key, &blk_hdr1->next_blk_key, 
                sizeof (pasis_md_hist_key_t));

        /* Fixup block 1's key pointers and number of entries */
        blk_hdr1->num_entries -= blk_hdr2->num_entries;
        memcpy (&blk_hdr1->next_blk_key, &blk_hdr2->this_blk_key,
                sizeof (pasis_md_hist_key_t));
        
        /* Copy entries from block 1 to block 2. */
        bytes_moved = blk_hdr2->num_entries * sizeof(pasis_md_hist_ent_t);
        memcpy (ents2, ents1, bytes_moved);

        /* Shift block 1 entries from bottom to top of block */
        bytes_shifted = PASIS_MD_HIST_BLK_USEFUL_SPACE - blk_hdr1->free_space -
            bytes_moved;
        memmove (ents1, ents1 + bytes_moved, bytes_shifted);

        /* Update free_space variables */
        blk_hdr1->free_space += bytes_moved;
        blk_hdr2->free_space = PASIS_MD_HIST_BLK_USEFUL_SPACE - bytes_moved;

        /* Check which of the splitted blocks should get the new entry */
        if (md_hts_compare (&new_ent.client_ts, &ent1->client_ts) >= 0) {
            /* Insert into block 1 */
            hist_insert_block (blk1, &new_ent);
        } else {
            /* Insert into block 2 */
            hist_insert_block (blk2, &new_ent);
        }

        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "split block: key=%llu, %llu\n", 
                       *(p_uint64 *)&blk_hdr2->this_blk_key.server_ts,
                       *(p_uint64 *)&blk_hdr1->next_blk_key.server_ts);

        /* Writeout block 2 */
#if 1
        /* Not going to backend */
        PASIS_DASSERT (0);
#else
        error = s4_ipc_table_insert (ipc_op, &ins_args_in, &ins_args_out, 
                                     (p_byte *)&blk_hdr2->this_blk_key, blk2);

        PASIS_DASSERT (error == PASIS_SUCCESS);
#endif

        PasisFreelist_Put (hist_blk_freelist, blk2);

    } else {  /* Insert new entry into block 1 */
        hist_insert_block (blk1, &new_ent);
    }
    
#if STORE_AUTHENTICATORS_IN_HIST_TABLE > 0
    /* Update authenticators
     * The first part of this phase consists of generating the (ordered) history
     * array.  We're repeating the code from hist_get_entries, but this code
     * is efficient about not having to do extra reads/writes. */

    /* Reread the first block (containing the header) if we were using a
     * different block */
    if (updating_first_block == 0) {
#if 1
        PASIS_PRINT_ERR("%s: not going to S4, hence can't have too many versions\n",
                        __FUNCTION__);
        PASIS_DASSERT(0);
#else
        /* First, write out the current block */
        error = s4_ipc_table_insert (ipc_op, &ins_args_in, &ins_args_out, 
                                     (p_byte *)&blk_hdr1->this_blk_key, blk1);
        PASIS_DASSERT (error == PASIS_SUCCESS);

        /* Second, read the first (header) block */
        error = hist_read_hdr (ipc_op, oid, &lkup_args_in, &lkup_args_out, blk1);
        PASIS_DASSERT (error == PASIS_SUCCESS);
        hdr1_dirty = 1;
#endif
   }

    /* Generate the ordered history vector */
    num_entries = 0;
    ent_ptr = (pasis_md_hist_ent_t *)(blk1 + sizeof(pasis_md_hist_hdr_t) +
                                      sizeof(pasis_md_hist_blk_hdr_t));
    index = blk_hdr1->num_entries - 1;
    while (1) {
        pasis_md_hist_ent_t *entry;
        /* Check if this is the first pass per block.  If so, calculate entry
         * indices. */
        if (index == blk_hdr1->num_entries - 1) {
            int i;
            entry = ent_ptr;
            for (i = 0; i < blk_hdr1->num_entries; ++i) {
                entry_indices[i] = (u_int32_t*)entry - (u_int32_t*)ent_ptr;
                entry = entry + 1;
            }
        }
        entry = (pasis_md_hist_ent_t *)((u_int32_t*)ent_ptr +
                                        entry_indices[index]);

        /* Copy entry into history array */
        md_hts_copy (&hist_array[num_entries].ts, &entry->client_ts);
        hist_array[num_entries].is_multiobject_op = entry->is_multiobject_op;
        num_entries++;
        PASIS_DASSERT (num_entries <= PASIS_MD_HIST_MAX_VERSIONS);

        index--;
        /* If there are no more entries in this block, goto next block */
        if (index < 0) {
            /* Exhausted entries in this block, move onto next block */
            if (PASIS_MD_HIST_KEY_INVALID (&blk_hdr1->next_blk_key)) {
                break;
            }

#if 1
            /* Not going to backend */
            PASIS_DASSERT (0);
#else
            error = s4_ipc_table_key_lookup (ipc_op, &lkup_args_in, 
                                             &lkup_args_out, 
                                             (p_byte *)&blk_hdr1->next_blk_key,
                                             blk1);
            PASIS_DASSERT (error == PASIS_SUCCESS);
#endif

            index = blk_hdr1->num_entries - 1;
        }
    }

    /* Generate the new authenticators */
    hdr1->auth_num = N;
    pasis_md_auth_generate (node_id,
                            num_entries,
                            hist_array,
                            hdr1->auth_num,
                            hdr1->authenticators);
#endif

    /* If hdr is dirty, insert into cache */
    if (hdr1_dirty) {
        PASIS_DPRINTF (PASIS_DEBUG_S4_FE, 4, "cache insert\n");
        PasisCache_Insert (hist_cache, &cache_tag, blk1);
    }

#if 1
    /* We want to stay in cache for our BFT-comparison experiments, so don't
     * write the history to S4, since it causes it to write to disk */
    error = PASIS_SUCCESS;
#else
    /* Write block 1 */
    error = s4_ipc_table_insert (ipc_op, &ins_args_in, &ins_args_out, 
                                 (p_byte *)&blk_hdr1->this_blk_key, blk1);
    PASIS_DASSERT (error == PASIS_SUCCESS);
#endif

    PasisFreelist_Put (hist_blk_freelist, blk1);

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_HIST_INSERT);

    return error;
}


pasis_error_t
pasis_md_hist_table_create (pasis_fe_s4_ipc_t       *ipc_op,
                            s4_oid_t                 oid,
                            p_uint32                 N,
                            p_uint32                 node_id)
{
    pasis_error_t  error;
    pasis_md_hts_t zero_ts;
    
    memset (&zero_ts, 0, sizeof (pasis_md_hts_t));

#if DEBUG_TIMESTAMPS > 0
    snprintf(zero_ts.op_description, DEBUG_OP_DESC_SIZE, "Init");
#endif

    error = pasis_md_hist_table_insert (ipc_op, oid, N, node_id, &zero_ts,
                                        &zero_ts, 0, 0, NULL);
    return error;
}


void
pasis_md_hist_stats_init (void)
{
    memset (&stats, 0, sizeof (pasis_md_hist_stats_t));

    /* Set up timers */
    stats.timers = Pasis_TimerInit ("Hist_Timers", PASIS_MD_TIMER_HIST_NUM, 0);
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_HIST_GET,
                         "history_getentries"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_HIST_INSERT,
                         "history_insert"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_HIST_GET_OP,
                         "history_get_op"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_HIST_COMPACTION,
                         "history_compaction"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_HIST_UPDATE_LAST_S4_TIME,
                         "history_update_last_s4_time"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_HIST_GET_S4_TIME,
                         "history_get_s4_time"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_HIST_GET_LATEST,
                         "history_get_latest"); 
}


void
pasis_md_hist_stats_dump (int shutdown_flag)
{
    PASIS_PRINT_INFO ("Hist cache hits:    %u\n", cache_hits);
    PASIS_PRINT_INFO ("Hist cache misses:  %u\n", cache_misses);

    if (shutdown_flag) {
        Pasis_TimerShutdown (stats.timers);
    } else {
        Pasis_TimerPrintAndReset (stats.timers);
        cache_hits = cache_misses = 0;
    }

}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */ 
