/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/* $cmuPDL: pasis_md_memory_versioning.cpp,v 1.18.2.5 2008/01/13 01:15:27 mabdelm Exp $ */

/*
 * The Self-* Storage System Project
 * Copyright (c) 2004, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */


#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_debug.h"

#include "pasis_md_prot.h"
// #include "pasis_md_backend.h"
#include "pasis_md_auth.h"
#include "pasis_md_log.h"
#include "pasis_md_memory_versioning.h"
#include "pasis_md_work.h"


// The object container.
MDObjectUniverse *md_universe = NULL;

int worker_pin_key       = 0;
int worker_pin_count_key = 0;

p_uint32 not_using_backend = 1;

p_uint32 node_id = PASIS_MAX_NUM_SERVERS;

p_uint32 DO_NOT_CALCULATE_AUTHENTICATORS = 0;
p_uint32 DO_NOT_TRANSFER_AUTHENTICATORS = 0;


// Static member variables
p_uint64          MDMemoryManager::block_cache_size;
p_uint64          MDMemoryManager::history_cache_size;
md_history_t      MDMemoryManager::dummy_map;
md_history_iter_t MDMemoryManager::dummy_end;
md_history_iter_t MDMemoryManager::block_lru_first;
md_history_iter_t MDMemoryManager::block_lru_last;
md_history_iter_t MDMemoryManager::history_lru_first;
md_history_iter_t MDMemoryManager::history_lru_last;
p_uint64          MDMemoryManager::total_block_allocation = 0;
p_uint64          MDMemoryManager::total_history_allocation = 0;
st_cond_t         MDMemoryManager::block_cache_cond = NULL;
st_cond_t         MDMemoryManager::history_cache_cond = NULL;
p_uint32          MDMemoryManager::num_threads_waiting_for_block_memory = 0;
p_uint32          MDMemoryManager::num_threads_waiting_for_history_memory = 0;
p_uint32          MDVersionedObject::num_objects = 0;

bool oracle_gc_mode = false;
bool verify_during_write_mode = false;
p_uint32 max_unverified_versions = 0;



////////////////////////////////////////////////////////////////////////////////
// MDMemoryManager members
////////////////////////////////////////////////////////////////////////////////

void MDMemoryManager::init(p_uint64 _block_cache_size,
                           p_uint64 _history_cache_size)
{
    block_cache_size = _block_cache_size;
    history_cache_size = _history_cache_size;
    
    dummy_end = dummy_map.end();
    block_lru_first = block_lru_last = dummy_end;
    history_lru_first = history_lru_last = dummy_end;
    block_cache_cond = st_cond_new();
    history_cache_cond = st_cond_new();

    PASIS_DASSERT(st_key_create(&worker_pin_key, NULL) == 0);
    PASIS_DASSERT(st_key_create(&worker_pin_count_key, NULL) == 0);

    md_universe = new MDObjectUniverse();
}


void MDMemoryManager::shutdown(void)
{
    delete md_universe;
    md_universe = NULL;

    // Check that we've deallocated all memory
    PASIS_DASSERT(total_block_allocation == 0);
    PASIS_DASSERT(total_history_allocation == 0);

    // Check that there no threads waiting for memory
    PASIS_DASSERT(num_threads_waiting_for_block_memory == 0);
    PASIS_DASSERT(num_threads_waiting_for_history_memory == 0);
    st_cond_destroy(block_cache_cond);
    st_cond_destroy(history_cache_cond);
    block_cache_cond = NULL;
    history_cache_cond = NULL;
}


void* MDMemoryManager::get_block_memory(p_uint32 size)
{
    while (total_block_allocation + size > block_cache_size) {
        if (not_using_backend) {
            PASIS_PRINT_ERR ("%s: not using backend; hence we can't evict "
                              "blocks.\n", __FUNCTION__);
            PASIS_DASSERT(0);
        } else {
            evict_lru_block();
        }
    }

    total_block_allocation += size;

    // Allocate an extra 4 bytes for a magic value at the end
    void *data = malloc (size + 4);
    *(p_uint32*)((char*)data + size) = 314159265;
        
    return data;
}

void MDMemoryManager::free_block_memory(void* data, p_uint32 size)
{
    if (data) {
        // printf("%s: IN IN IN\n", __FUNCTION__);
        total_block_allocation -= size;
        p_uint32 *magic = (p_uint32*)((char*)data + size);
        PASIS_DASSERT(*magic == 314159265);
        *magic = 0xdeadbeaf;
        free(data);
    }
}


void MDMemoryManager::account_history_memory(p_uint32 size)
{
    while (total_history_allocation + size > history_cache_size) {
        if (not_using_backend) {
            PASIS_PRINT_ERR ("%s: not using backend; hence we can't evict "
                              "history.\n", __FUNCTION__);
            PASIS_DASSERT(0);
        } else {
            evict_lru_history();
        }
    }

    total_history_allocation += size;
    // printf("%s: size = %u, new total = %llu\n", __FUNCTION__,
    //        size, total_history_allocation);
}


void MDMemoryManager::deaccount_history_memory(p_uint32 size)
{
    total_history_allocation -= size;
    // printf("%s: size = %u, new total = %llu\n", __FUNCTION__,
    //        size, total_history_allocation);
}


// Sometimes, a thread will need memory for its data block cache but we're out
// of cache memory _and_ there are no blocks to be evicted.  The latter
// condition occurs when other threads have allocated data blocks that are
// currently busy (for writes/reads).  So in this case, the thread has to wait
// until the other threads are finished with their memory and we can evict their
// memory.
void MDMemoryManager::wait_for_evictable_block()
{
    ++num_threads_waiting_for_block_memory;
    // printf("%s: waiting for memory; thread %p, num_waiters = %u\n",
    //        __FUNCTION__, st_thread_self(), num_threads_waiting_for_memory);
    st_cond_wait(block_cache_cond);
    --num_threads_waiting_for_block_memory;
    // printf("%s: got memory; thread %p, num_waiters = %u\n",
    //        __FUNCTION__, st_thread_self(), num_threads_waiting_for_memory);
}


void MDMemoryManager::wait_for_evictable_history()
{
    ++num_threads_waiting_for_history_memory;
    // printf("%s: waiting for memory; thread %p, num_waiters = %u\n",
    //        __FUNCTION__, st_thread_self(), num_threads_waiting_for_memory);
    st_cond_wait(history_cache_cond);
    --num_threads_waiting_for_history_memory;
    // printf("%s: got memory; thread %p, num_waiters = %u\n",
    //        __FUNCTION__, st_thread_self(), num_threads_waiting_for_memory);
}


void MDMemoryManager::wakeup_thread_waiting_for_memory()
{
    if (num_threads_waiting_for_block_memory > 0) {
        st_cond_signal(block_cache_cond);
    }

    if (num_threads_waiting_for_history_memory > 0) {
        st_cond_signal(history_cache_cond);
    }
}


// XXX: Right now, we have a simple policy in which we select the least-recently
// used block for eviction.  We may also want to take into account whether the
// block is clean or not.
void MDMemoryManager::evict_lru_block()
{
 try_again:
    md_history_iter_t to_evict = block_lru_last;
    while (to_evict != dummy_end &&
           to_evict->second.is_pinned == false &&
           to_evict->second.block_object->object->is_locked()) {
        to_evict = to_evict->second.blockLeft;
    }

    if (to_evict == dummy_end) {
        // Need to block this thread, because either all the blocks in the data
        // cache have been evicted by other threads.  So we have wait until some
        // read/write operation is done and hence we can evict its data block.
        wait_for_evictable_block();
        goto try_again;
    }

    // Put brace brackets around the following block, to make sure that the
    // lock's destructor gets called if we goto to try_again.
    {
        ScopeLock auto_lock(to_evict->second.block_object->object->mutex);

        // Ensure the block is still "alive" by checking to see if it's on the
        // LRU list.
        //  We _will_ not be going through the loop (a lot) times on average,
        //  because, in the common case, after we lock the block, it's still on
        //  the end of the LRU list, and hence we'll find it right away.  It's
        //  only in the uncommon case when the block has disappeared that we'll
        //  have to do a lot of iterations.
        md_history_iter_t to_evict_refind = block_lru_last;
        while ((to_evict_refind != dummy_end) && (to_evict_refind != to_evict)) {
            to_evict_refind = to_evict_refind->second.blockLeft;
        }

        if (to_evict_refind == dummy_end) {
            // block isn't alive anymore.  repeat search for evictable block.
            // above locked object _will_ get unlocked!
            goto try_again;
        }

        if (to_evict->second.is_pinned) {
            // Some other thread has pinned this block.  can't evict it
            goto try_again;
        }

        evict_block(to_evict);
    }
}


void MDMemoryManager::evict_block(md_history_iter_t to_evict)
{
    // printf("%s: evicting a block, due to cache pressure.  Block %llu, "
    //        "version %llu\n", __FUNCTION__,
    //        to_evict->second.block_object->block_num,
    //        to_evict->first.timestamp.time);

    // flush() will deallocate the memory and do a write-back if the version is
    // dirty.
    PASIS_DASSERT(to_evict->second.block_object->object != NULL);
    to_evict->second.block_object->flush(to_evict, true);
}


// XXX: Right now, we have a simple policy in which we select the least-recently
// used history entry for eviction.  We may also want to take into account
// whether the history entry is clean or not.
void MDMemoryManager::evict_lru_history()
{
    // printf("%s: IN IN IN\n", __FUNCTION__);
 try_again:
    md_history_iter_t to_evict = history_lru_last;
    while (to_evict == dummy_end ||
           to_evict->second.is_pinned == false &&
           to_evict->second.block_object->object->is_locked()) {
        to_evict = to_evict->second.historyLeft;
    }

    if (to_evict == dummy_end) {
        // Need to block this thread, because either all the blocks in the data
        // cache have been evicted by other threads.  So we have wait until some
        // read/write operation is done and hence we can evict its history
        wait_for_evictable_history();
        goto try_again;
    }

    // Put brace brackets around the following block, to make sure that the
    // lock's destructor gets called if we goto to try_again.
    {
        ScopeLock auto_lock(to_evict->second.block_object->object->mutex);

        // Ensure the block is still "alive" by checking to see if it's on the
        // LRU list.
        //  We _will_ not be going through the loop (a lot) times on average,
        //  because, in the common case, after we lock the block, it's still on
        //  the end of the LRU list, and hence we'll find it right away.  It's
        //  only in the uncommon case when the block has disappeared that we'll
        //  have to do a lot of iterations.
        md_history_iter_t to_evict_refind = history_lru_last;
        while ((to_evict_refind != dummy_end) && (to_evict_refind != to_evict)) {
            to_evict_refind = to_evict_refind->second.historyLeft;
        }
        
        if (to_evict_refind == dummy_end) {
            // block isn't alive anymore.  repeat search for evictable history.
            // above locked object _will_ get unlocked!
            goto try_again;
        }

        if (to_evict->second.is_pinned) {
            // Some other thread has pinned this block.  can't evict it
            goto try_again;
        }

        evict_history(to_evict);
    }
    // exit scope to unlock object's mutex, then we can safely destruct object
    // Get rid of the block object, from the MDVersionedObject object, if
    // possible

    MDVersionedBlock  *blockObject = to_evict->second.block_object;
    MDVersionedObject *objectObject = blockObject->object;
    objectObject->remove_block_object_if_possible(blockObject);
    PASIS_DASSERT(0); // we probably want two chunk counts: one for real (in mem
    // and on disk) count(), and another one, just in memory, that can enable an
    // object to be deleted when it's 0
    if (objectObject->blocks.size() == 0) {
        MDObjectUniverse::object_map_t::iterator i =
            md_universe->objects.find(objectObject->oid);
        // Can get rid of the MDVersionedObject object
        // printf("%s: getting rid of an object\n", __FUNCTION__);
        md_universe->objects.erase(i);
    }
}

void MDMemoryManager::evict_history(md_history_iter_t to_evict)
{
    MDVersionedBlock *blockObject = to_evict->second.block_object;

    // printf("%s: evicting a history entry, due to cache pressure.  Block %llu, "
    //        "version %llu\n", __FUNCTION__,
    //        to_evict->second.block_object->block_num,
    //        to_evict->first.timestamp.time);

    // Remove history entry from the history LRU list
    remove_from_history_lru(to_evict);

    // In case we have block data in front-end (i.e., it's dirty), flush it.
    // If we don't have dirty data, then that version entry has been flushed
    // already and there's no need to write anything to the back-end.
    evict_block(to_evict);

    // Memory allocation accounting
    MDMemoryManager::deaccount_history_memory(sizeof(MDVersionData) +
                                              sizeof(MDVersion));
    
    // When removing the version entry, remember to use the right
    // MDVersionedBlock object.
    // Now, time to 1) erase the history entry, 2) check if we can get rid of
    // the MDVersionedBlock object, and 3) check if we can get rid of the
    // MDVersionedObject object.
    // Get rid of the history entry
    blockObject->history.erase(to_evict);

    // We'll do step #2, #3 in evict_lru_history, to ensure that we've unlocked
    // the block's mutex before we potentially destruct it
}


void MDMemoryManager::move_to_block_lru_front(md_history_iter_t &entry)
{
    // In case we already had an entry for this block version
    remove_from_block_lru(entry);

    MDVersionData& vd = entry->second;
    vd.blockLeft      = dummy_end;
    vd.blockRight     = block_lru_first;
    if (block_lru_first != dummy_end) {
        block_lru_first->second.blockLeft = entry;
    }
    block_lru_first = entry;
    if (block_lru_last == dummy_end) {
        block_lru_last = entry;
    }

//     printf("%s: block %llu data block LRU list.  First = %llu, Last = %llu:\n",
//            __FUNCTION__, block_lru_first->second.block_object->block_num,
//            block_lru_first->first.timestamp.time,
//            block_lru_last->first.timestamp.time);
//     for (history_iter_t i = block_lru_first; i != dummy_end;) {
//         printf("%llu, ", i->first.timestamp.time);
//         i = i->second.blockRight;
//     }
//     printf("\n");
}


void MDMemoryManager::remove_from_block_lru(md_history_iter_t &entry)
{
    MDVersionData& vd = entry->second;

    if (vd.blockLeft != dummy_end) {
        vd.blockLeft->second.blockRight = vd.blockRight;
    }
    if (vd.blockRight != dummy_end) {
        vd.blockRight->second.blockLeft = vd.blockLeft;
    }

    if (entry == block_lru_first) {
        block_lru_first = vd.blockRight;
    }
    if (entry == block_lru_last) {
        block_lru_last = vd.blockLeft;
    }
}


// Note: this function can sometimes get called more times than it needs to be
// called.  But don't remove the calls to this function!  While they may seem
// extra, they at least guarantee that a history entry is placed/moved in the
// LRU list whenever necessary.  In the face of the modular code + sometimes
// having to touch disk and reading more (intermediate) versions than we return,
// we end up calling this function a few more times than is needed.  But it's
// OK, since it's a cheap function.
void MDMemoryManager::move_to_history_lru_front(md_history_iter_t &entry)
{
    // In case we already had an entry for this history version
    remove_from_history_lru(entry);

    MDVersionData& vd = entry->second;
    vd.historyLeft    = dummy_end;
    vd.historyRight   = history_lru_first;
    if (history_lru_first != dummy_end) {
        history_lru_first->second.historyLeft = entry;
    }
    history_lru_first = entry;
    if (history_lru_last == dummy_end) {
        history_lru_last = entry;
    }

//     printf("%s: block %llu history LRU list. First = %llu, Last = %llu:\n",
//            __FUNCTION__, history_lru_first->second.block_object->block_num,
//            history_lru_first->first.timestamp.time,
//            history_lru_last->first.timestamp.time);
//     for (history_iter_t i = history_lru_first; i != dummy_end;) {
//         printf("%llu, ", i->first.timestamp.time);
//         i = i->second.historyRight;
//     }
//     printf("\n");
}


void MDMemoryManager::remove_from_history_lru(md_history_iter_t &entry)
{
    MDVersionData& vd = entry->second;

    if (vd.historyLeft != dummy_end) {
        vd.historyLeft->second.historyRight = vd.historyRight;
    }
    if (vd.historyRight != dummy_end) {
        vd.historyRight->second.historyLeft = vd.historyLeft;
    }

    if (entry == history_lru_first) {
        history_lru_first = vd.historyRight;
    }
    if (entry == history_lru_last) {
        history_lru_last = vd.historyLeft;
    }
}



////////////////////////////////////////////////////////////////////////////////
// MDVersionedBlock members
////////////////////////////////////////////////////////////////////////////////


MDVersionedBlock::MDVersionedBlock() :
    object(NULL), block_num(0), possibly_more_history_in_backend(false)
{
    MDMemoryManager::total_history_allocation += sizeof(*this);
}


MDVersionedBlock::MDVersionedBlock(const MDVersionedBlock &other) :
    object(other.object), block_num(other.block_num), history(other.history),
    possibly_more_history_in_backend(other.possibly_more_history_in_backend)
{
    MDMemoryManager::total_history_allocation += sizeof(*this);
}


MDVersionedBlock::~MDVersionedBlock()
{
    // First, flush all dirty versions (and create the object, if necessary)
    flush_all();
    // Second, deallocate the blocks from memory
    remove_all();

    MDMemoryManager::deaccount_history_memory(sizeof(*this));
}


// Invalidates the cached authenticators for any version >= timestamp
void MDVersionedBlock::invalidate_cached_auth(const MDVersion &timestamp)
{
    md_history_iter_t i = history.lower_bound(timestamp);
    for (; i != history.end(); ++i) {
        i->second.calculated_auths = false;
    }
}


// Writing to the header means adding a history entry
void MDVersionedBlock::write_to_header(MDVersion &timestamp, MDVersion &cond_ts)
{
    md_history_iter_t i = history.find(timestamp);
    if (i != history.end()) {
        // We've already recorded this timestmap in the version history.  Just
        // move to "front" of LRU list
        MDMemoryManager::move_to_history_lru_front(i);
        MDMemoryManager::move_to_block_lru_front(i);
    } else {
        // First write to any block at this version.  Add to history.

        pasis_md_log_msg ("%s: adding %llu to RH.", __FUNCTION__,
                          timestamp.timestamp.time);

        // need a magic header size, to ensure enough space in header for
        // authenticator and hashes, even for small objects with block size of 4
        // bytes say
        void* data = object->get_memory_for_new_header_block();
        memset(data, 0, object->header_size());

#if USE_OBJECT_HASHES > 0
        PASIS_DASSERT(0); // copy cond_ts's hashes, if timestamp != 0 and we're
        // using object hashes
#endif
        // copy "private user data" from previous version, if we had any
        p_uint32 user_data_size;
        void *existing_small_data = read_small_data(cond_ts, &user_data_size);
        if (existing_small_data) {
            p_byte *user_data = (p_byte*)data + data_offset();
            *(p_uint32*)user_data = user_data_size;
            memmove(user_data + 4, existing_small_data, user_data_size);
        }
        

        // XXX: any other header initialization?

        // Tricky tricky...  We need to "invalidate" any authenticators that
        // have been calculated for larger versions.
        // Imagine we have a complete at time X, a complete (at all but one
        // servers) at X+1, and an incomplete (at the Nth server) at X+2.  The
        // client then writes a barrier at X+3.  And say it accesses the Nth
        // server, so it has to cause object syncing at the Nth server to get
        // X+1.  In this case, the Nth server has to recalculate the
        // authenticators when it returns history from X+3 downwards, since a
        // new version has entered the system.
        invalidate_cached_auth(timestamp);

        // Insert header into version history
        write_to_ordinary_block(timestamp, cond_ts, data);
    }
}


void MDVersionedBlock::write_to_ordinary_block(MDVersion &timestamp,
                                               MDVersion &cond_ts, void* data)
{
    md_history_iter_t null_iter = MDMemoryManager::null_iterator();
    MDVersionData new_vd = {data, true, false, false, false, false, this,
                            null_iter, null_iter, null_iter, null_iter};
    pair<md_history_iter_t, bool> p =
        history.insert(pair<MDVersion, MDVersionData>(timestamp, new_vd));

    if (p.second == false) {
        // duplicate entry at TS; remove existing data and then reinsert
        remove(p.first);
        p = history.insert(pair<MDVersion, MDVersionData>(timestamp, new_vd));
        PASIS_DASSERT(p.second == true);
    }

    MDMemoryManager::account_history_memory(sizeof(MDVersionData) +
                                            sizeof(MDVersion));
    MDMemoryManager::move_to_history_lru_front(p.first);
    MDMemoryManager::move_to_block_lru_front(p.first);

#if USE_OBJECT_HASHES > 0
    // XXX: if object hashes are used, then update block's hash in the header
    // block
    PASIS_DASSERT(0);
#endif

    // GC time!  (But not when performing or conditioning on barriers.)
    if (timestamp.timestamp.barrier_flag == 0 &&
        cond_ts.timestamp.barrier_flag == 0) {
        do_gc(cond_ts);
    }
}


void MDVersionedBlock::write(MDVersion &timestamp, MDVersion &cond_ts, void* data)
{
    if (is_header_block()) {
        PASIS_DASSERT(data == NULL); // no user data for header
        write_to_header(timestamp, cond_ts);
    } else {
        write_to_ordinary_block(timestamp, cond_ts, data);
    }
}


void MDVersionedBlock::write_small_data(MDVersion &timestamp, MDVersion &cond_ts,
                                        p_uint32 size, void *data)
{
    PASIS_DASSERT(is_header_block());
    md_history_iter_t i = history.find(timestamp);
    if (i == history.end()) {
        // this should never happen, as VersionedObject::write_small_data
        // creates the header block if it's not there already
        PASIS_DASSERT(0);
    } else {
        PASIS_DASSERT(size <= PASIS_MD_MAX_HEADER_DATA_SIZE);
        p_byte *user_data = ((p_byte*)i->second.data) + data_offset();
        *(p_uint32*)user_data = size;
        memcpy(user_data + 4, data, size);
#if USE_OBJECT_HASHES > 0
        // XXX: if object hashes are used, then update hash in header block
        PASIS_DASSERT(0);
#endif
    }
}


bool MDVersionedBlock::higher_version_exists(MDVersion &timestamp)
{
    if (num_versions() == 0) {
        return false;
    }

    md_history_iter_t j = history.lower_bound(timestamp);
    // j points to smallest element that's equal or greater than search_ts.

    if (j == history.end()) {
        return false;
    } else if (j->first == timestamp) {
        return false;
    } else {
        return true;
    }
}


void MDVersionedBlock::copy_forward(MDVersion &new_ts, MDVersion &old_ts,
                                    MDVersion &cond_ts)
{
    if (higher_version_exists(old_ts)) {
        void* data = read(old_ts);
        write(new_ts, cond_ts, data);
    } else {
        // no need to copy forward, since nothing exists in the "middle"
    }
}


void MDVersionedBlock::do_gc(MDVersion &cond_ts)
{
    remove_up_to(cond_ts);
}


void MDVersionedBlock::unpin(MDVersion &version)
{
    md_history_iter_t i = internal_read(version);
    if (i == history.end()) {
        PASIS_DASSERT(0);
    } else {
        PASIS_DASSERT(i->second.is_pinned);
        i->second.is_pinned = 0;
        if (i->second.can_gc) {
            // now we can delete the block, once it has been unpinned
            remove(i);
        }
    }
}


void MDVersionedBlock::remove(md_history_iter_t i)
{
    // Destruct existing data at this timestamp
    //printf("%s: node %u removing block %llu at timestamp %lld\n",
    //       __FUNCTION__, object->oid.partition, block_num,
    //       i->first.timestamp.time);

    // Any cached authenticators have to be invalidated, since the history has
    // changed.
    invalidate_cached_auth(i->first);

    MDMemoryManager::remove_from_block_lru(i);
    MDMemoryManager::remove_from_history_lru(i);
    if (is_header_block()) {
        object->free_header_block_memory(i->second.data);
    } else {
        object->free_block_memory(i->second.data);
    }
    i->second.data = NULL;
    i->second.is_dirty = false;
    MDMemoryManager::deaccount_history_memory(sizeof(MDVersionData) +
                                              sizeof(MDVersion));
    history.erase(i);
}


void MDVersionedBlock::remove(MDVersion timestamp)
{
    MDVersion timestamp_copy(timestamp);
    // our history doesn't have object hashes, as we create the version entry
    // before the object is completely updated (and hence we can't calculate the
    // object hash on version-insertion).  So zero out the object hash when
    // looking for versions.  It's OK to do this since we can't have two
    // versions with the same logical timestamp, but different object hashes,
    // due to the way we assign (logical) timestamps.
    timestamp_copy.zeroObjectHash();

    md_history_iter_t i = history.find(timestamp_copy);
    if (i != history.end()) {
        remove(i);
    }
}


void MDVersionedBlock::remove_up_to(const MDVersion& timestamp)
{
    // ScopeLock auto_lock(mutex);

    MDVersion timestamp_copy(timestamp);
    // our history doesn't have object hashes, as we create the version entry
    // before the object is completely updated (and hence we can't calculate the
    // object hash on version-insertion).  So zero out the object hash when
    // looking for versions.  It's OK to do this since we can't have two
    // versions with the same logical timestamp, but different object hashes,
    // due to the way we assign (logical) timestamps.
    timestamp_copy.zeroObjectHash();

    md_history_iter_t j = highest_timestamp_less_than(timestamp_copy);
    if (j == history.end()) {
        return; // nothing to delete
    }

    // Imagine we have version 1 of some block.  then version 2 of the _object_
    // is written, which does _not_ update this particular block.  now, a new
    // update, version 3, comes in, and updates the block.  so we have versions
    // 1 and 3 of the block.  the last known good version is 2.  so we can't
    // actually delete 1 yet though, since we don't know if 3 is complete or
    // not.  this problem happens when there are "holes" in a block's version
    // history.
    if (history.find(timestamp_copy) == history.end()) {
        // ok we don't have the condition-on timestamp (2 in the above example).
        // so we keep along the newest version less than the cond ts.  i.e.: 1

        if (j == history.begin()) {
            // we had only one version to delete, so if we were to skip that, we
            // have nothing to delete
           return;
        } else {
            // don't delete this newest-known-complete version
            --j;
        }
    }

    for (md_history_iter_t i = history.begin(); ;) {
        // Get pointer to next iterator, before we erase current iterator
        md_history_iter_t next_i = i, i_backup = i;
        ++next_i;

        if (i->second.is_pinned) {
            // if buffer is pinned, don't GC it.  It will get GC'ed next time we
            // unpin the buffer; see MDVersionedBlock::unpin
            i->second.can_gc = true;
        } else {
            // printf("%s: removing version %llu for block %llu\n", __FUNCTION__,
            //        i->first.timestamp.time, block_num);
            remove(i);
        }

        if (i_backup == j) {
            // we delete up to, and including, j
            break;
        }

        i = next_i;
    }
}


void MDVersionedBlock::remove_all()
{
    // ScopeLock auto_lock(mutex);
    for (md_history_iter_t i = history.begin(); i != history.end(); ) {
        // Get pointer to next iterator, before we erase current iterator
        md_history_iter_t next_i = i; ++next_i;
        remove(i);
        i = next_i;
    }
}


char* MDVersionedBlock::dump_history()
{
    static char buffer[512];
    char temp[16];
    buffer[0] = '\0';
    for (md_history_iter_t i = history.begin(); i != history.end(); ++i) {
        if (i->first.timestamp.barrier_flag) {
            sprintf(temp, "%"PRIu64"B, ", i->first.timestamp.time);
        } else {
            sprintf(temp, "%"PRIu64", ", i->first.timestamp.time);
        }
        strcat(buffer, temp);
    }
    return buffer;
}

 
p_uint32 MDVersionedBlock::data_offset()
{
    PASIS_DASSERT(is_header_block());
    return 0;
}


p_uint32 MDVersionedBlock::object_hash_offset()
{
    // Another 4 bytes as we store the size of the small data
    return data_offset() + PASIS_MD_MAX_HEADER_DATA_SIZE + 4;
}



p_uint32 MDVersionedBlock::auth_offset()
{
#if USE_OBJECT_HASHES > 0
    return object_hash_offset() + PASIS_HASH_SIZE;
#else
    // not storing object hash in header
    return object_hash_offset();
#endif
}


p_uint32 MDVersionedBlock::block_hashes_offset()
{
#if STORE_AUTHENTICATORS_IN_HIST_TABLE > 0
    p_uint32 auth_num = object->object_attrs.distribution.protocol.N;
    return auth_offset() + auth_num * sizeof(pasis_md_auth_t);
#else
    // not storing authenticators in header
    return auth_offset();
#endif
}


md_history_iter_t
MDVersionedBlock::highest_timestamp_less_than(MDVersion &search_ts)
{
    if (num_versions() == 0) {
        return history.end();
    }

    md_history_iter_t j = history.lower_bound(search_ts);
    // j points to smallest element that's equal or greater than search_ts.

    if (j == history.end()) {
        // all the history entries we have are smaller than the search_ts.
        // start returning from highest entry.
        --j; // guaranteed to work, since we've checked we have > 1 versions
    } else if (j->first >= search_ts) {
        // We want highest version that is < search_ts.  Hence, we
        // ignore first element that's >= search_ts.

        // But first, check if we have no entries to return, once we skip this
        // entry.
        if (j == history.begin()) {
            /* no entries to return after we skip this entry; return now */
            return history.end();
        }
        --j; // skip past this entry
    }
    return j;
}


// Returns num_hist_in history entries from the in-memory history versions.
void
MDVersionedBlock::get_history_from_memory(MDVersion       &search_ts,
                                          p_uint32         num_hist_in,
                                          pasis_md_hts_t  *hist_array,
                                          p_uint32        *num_hist_out,
                                          pasis_md_auth_t *authenticators)
{
    *num_hist_out = 0;

    MDVersion search_ts_copy(search_ts);
    // our history doesn't have object hashes, as we create the version entry
    // before the object is completely updated (and hence we can't calculate the
    // object hash on version-insertion).  So zero out the object hash when
    // looking for versions.  It's OK to do this since we can't have two
    // versions with the same logical timestamp, but different object hashes,
    // due to the way we assign (logical) timestamps.
    search_ts_copy.zeroObjectHash();

    md_history_iter_t j, j_save;
    j = j_save = highest_timestamp_less_than(search_ts_copy);
    if (j == history.end()) {
        return; // nothing to return
    }

    // We return history in descending order (e.g., 5, 4, 2).  So use the
    // reverse iterators.
    // So we have to "convert" forward iterator to reverse.
    // constructing a reverse iterator from a forward iterator is tricky.
    // see section 19.2.5 of Stroustrup.  we need to add one, since the
    // reverse iterator code is "off by one" from the normal forward
    // iterator that's the implementation.
    ++j;
    md_history_t::reverse_iterator i(j);

    for (; (*num_hist_out < num_hist_in) && (i != history.rend()); ++i) {
        md_hts_copy(&hist_array[*num_hist_out], &i->first.timestamp);
        *num_hist_out = *num_hist_out + 1;

#if USE_OBJECT_HASHES > 0
        PASIS_DASSERT(0);
        // XXX: if using object hashes, updated object hash in TS entry by
        // calling block->get_object_hash, which will calculate hash if the
        // MDVersionData flag says it needs to be calculated
#endif

        // Use forward iterator for the history LRU list operation
        --j;
        MDMemoryManager::move_to_history_lru_front(j);
    }

#if STORE_AUTHENTICATORS_IN_HIST_TABLE > 0
    // after the hist records have the correct object hash, if we're using
    // object hashes, we can now calculate the authenticators, if they haven't
    // been calculated before
    if (authenticators) { // we are being asked for the authenticators
        p_uint32 auth_num  = object->object_attrs.distribution.protocol.N;
        p_uint32 auth_size = auth_num * sizeof(pasis_md_auth_t);
        MDVersionData *vd = &j_save->second;
        PASIS_DASSERT(vd->data != NULL); // xxx: running only in memory for now
        if (vd->calculated_auths == false) { // need to calculate authenticators?
            // XXX: if we have to calculate the authenicators, then we must have the
            // full version history.  we can ensure we have the full version history
            // here by regenerating it, in case the user just asked for a subset.
            // but for simplicity, we make the unessential requirement that the user
            // asked us for the full version history, so we can simply use the above
            // version history array.
            PASIS_DASSERT(num_hist_in == PASIS_MD_HIST_MAX_VERSIONS);

            p_uint32 real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ?
                                     object->oid.partition : node_id);
            PASIS_DASSERT(auth_offset() + auth_size <= object->header_size());
            pasis_md_auth_generate(real_node_id, *num_hist_out, hist_array,
                                   auth_num,
                                   (pasis_md_auth_t *) ((p_byte*)vd->data +
                                                        auth_offset()));
            vd->calculated_auths = true;
        } else {
            // pasis_md_log_msg ("%s: not recalculating auths", __FUNCTION__);
        }
        memcpy(authenticators, (p_byte*)vd->data + auth_offset(), auth_size);
    }
#endif    
}


void
MDVersionedBlock::get_history(MDVersion       &search_ts,
                              p_uint32         num_hist_in,
                              pasis_md_hts_t  *hist_array,
                              p_uint32        *num_hist_out,
                              pasis_md_auth_t *authenticators)
{
    // get_history() should only be called on the header block
    PASIS_DASSERT(is_header_block());

    // ScopeLock auto_lock(mutex);

    // 1) Make sure we have version search_ts in-memory, if it exists.
    // XXX: we're only operating in-memory for now.  -mabdelmalek 7/23/2005
    // loadHistoryUpToVersion(search_ts);

    // 2) See if the history versions we have in memory satisfy the request
    get_history_from_memory(search_ts, num_hist_in, hist_array, num_hist_out,
                            authenticators);

    // 3) If the in-memory history versions weren't enough _and_ there's a
    // chance there's more history in the back-end, read more history from the
    // back-end.
    // XXX: we're only operating in-memory for now.  -mabdelmalek 7/23/2005
#if 0
    if (*num_hist_out < num_hist_in && possibly_more_history_in_backend) {
        // Get more history from the back-end.
        loadMoreHistoryFromBackEnd(num_hist_in - *num_hist_out);

        // Retry the get history from memory operation with the new history
        get_history_from_memory(search_ts, num_hist_in, hist_array,
                                num_hist_out, authenticators);
    }
#endif
}


// Returns true if this S-N hosts "version"
bool MDVersionedBlock::has_version(MDVersion &version)
{
    // get_history() should only be called on the header block
    PASIS_DASSERT(is_header_block());

    MDVersion version_copy(version);
    // our history doesn't have object hashes, as we create the version entry
    // before the object is completely updated (and hence we can't calculate the
    // object hash on version-insertion).  So zero out the object hash when
    // looking for versions.  It's OK to do this since we can't have two
    // versions with the same logical timestamp, but different object hashes,
    // due to the way we assign (logical) timestamps.
    version_copy.zeroObjectHash();

    // ScopeLock auto_lock(mutex);

    // 1) Make sure we have version search_ts in-memory, if it exists.
    // XXX: we're only operating in-memory for now.  -mabdelmalek 7/23/2005
    // loadHistoryUpToVersion(version_copy);

    return history.find(version_copy) != history.end();
}



md_history_iter_t MDVersionedBlock::internal_read(MDVersion &timestamp)
{
    MDVersion timestamp_copy(timestamp);
    // our history doesn't have object hashes, as we create the version entry
    // before the object is completely updated (and hence we can't calculate the
    // object hash on version-insertion).  So zero out the object hash when
    // looking for versions.  It's OK to do this since we can't have two
    // versions with the same logical timestamp, but different object hashes,
    // due to the way we assign (logical) timestamps.
    timestamp_copy.zeroObjectHash();

    // 1) Make sure we have version "timestamp" in-memory.
    // XXX: we're only operating in-memory for now.  -mabdelmalek 7/23/2005
    // md_history_iter_t i = loadHistoryUpToVersion(timestamp);

    md_history_iter_t i = history.find(timestamp_copy);
    if (i == history.end()) {
        // if we don't have version at "timestamp", then return previous version
        i = highest_timestamp_less_than(timestamp_copy);
    }
    if (i == history.end()) {
        return i;
    }

    // We have the version.  But data may not be in front-end.
    MDVersionData *vd = &i->second;

    if (vd->data != NULL) {
        // nothing to do; we already have the data in-memory
    } else {
#if 1
        // XXX: we're only operating in-memory for now.  -mabdelmalek 7/23/2005
        PASIS_DASSERT(0);
#else
        PASIS_DASSERT (backend);

        // Load block into memory from back-end */
        void* new_data;
        if (is_header()) {
            new_data = object->get_memory_for_new_header_block();
        } else {
            new_data = object->get_memory_for_new_block();
        }
        pasis_error_t error = backend->read(object, block_num,
                                            &timestamp.timestamp,
                                            data_size, new_data);
        if (error) {
            if (is_header()) {
                free_header_block_memory(new_data);
            } else {
                free_block_memory(new_data);
            }
            return NULL;
        }

        // Change the history version to point to the data block we just
        // read
        vd->data = new_data;
#endif
    }

    MDMemoryManager::move_to_history_lru_front(i);
    MDMemoryManager::move_to_block_lru_front(i);
    return i;
}


void* MDVersionedBlock::read(MDVersion &timestamp, bool pin /* = false */)
{
    md_history_iter_t i = internal_read(timestamp);
    if (i == history.end()) {
        return NULL;
    } else {
        i->second.is_pinned = pin;
        return i->second.data;
    }
}


void* MDVersionedBlock::read_small_data(MDVersion &ts, p_uint32 *size)
{
    void* block_data = read(ts);
    if (block_data == NULL) {
        if (size) {
            *size = 0;
        }
        return NULL;
    } else {
        p_byte *user_data = ((p_byte*)block_data) + data_offset();
        p_uint32 user_data_size = *(p_uint32*)user_data;
        if (size) {
            *size = user_data_size;
        }
        PASIS_DASSERT(user_data_size >= 0);
        if (user_data_size == 0) {
            return NULL;
        } else {
            return user_data + 4;
        }
    }
}


pasis_error_t MDVersionedBlock::flush(md_history_iter_t i, bool evict)
{
    PASIS_DASSERT (object != NULL);

    // If object is being deleted (by the client), then don't create the object
    // in the back-end or flush any versions.
    if (object->deleting_object_now) {
        // printf("%s: not flushing as object is being deleted %s\n", __FUNCTION__,
        //        my_dump_fid(&object->oid.fid));
        return PASIS_SUCCESS;
    }

    // First, create the object in the back-end, if necessary.
    pasis_error_t error = object->create_if_necessary();
    if (error) {
        printf("%s: create object failed\n", __FUNCTION__);
        return error;
    }

    // Only flush dirty versions
    if (i->second.is_dirty) {
        //   printf("%s: block %lld, flushing version at timestamp %lld\n",
        //          __FUNCTION__, block_num, i->first.timestamp.time);
#if 1
        // XXX: we're only operating in-memory for now.  -mabdelmalek 7/23/2005
        error = PASIS_SUCCESS;
#else
        error = backend->write(object, block_num, &i->first.timestamp,
                               i->second.data_size, i->second.data);
#endif
        if (error) {
            return error;
        } else {
            i->second.is_dirty = false;
        }
    }
    
    if (evict) {
        // Remove block from the blocks LRU list
        MDMemoryManager::remove_from_block_lru(i);

        void *data_to_flush = i->second.data;
        // Set the version's pointer to NULL, to indicate the data isn't in the
        // front-end anymore.
        i->second.data = NULL;
        // Reset the LRU pointers
        i->second.blockRight = i->second.blockLeft =
            MDMemoryManager::null_iterator();
        
        // Free the memory
        if (is_header_block()) {
            object->free_header_block_memory(data_to_flush);
        } else {
            object->free_block_memory(data_to_flush);
        }
    }

    return PASIS_SUCCESS;    
}


pasis_error_t MDVersionedBlock::flush_all()
{
    if (not_using_backend) {
        // Not using back-end, so nowhere to flush.
        return PASIS_SUCCESS;
    } else {
        // Iterate over the block versions we have and write them to backend
        for (md_history_iter_t i = history.begin(); i != history.end(); ++i) {
            pasis_error_t error = flush(i, false);
            if (error) {
                return error;
            }
        }
        return PASIS_SUCCESS;
    }
}


////////////////////////////////////////////////////////////////////////////////
// MDVersionedObject members
////////////////////////////////////////////////////////////////////////////////


// Factory for MDVersionedObjects
MDVersionedObject*
MDVersionedObject::create_new_object(MDObjectID          &oid,
                                     p_uint32             block_size,
                                     pasis_md_loc_t      *distribution)
{
    MDVersionedObject *object = new MDVersionedObject(oid, block_size,
                                                      distribution);
    object->need_to_create_object = true;

    // Add version 0 to version history
    object->initialize_new_object();

    return object;
}


MDVersionedObject::MDVersionedObject(MDObjectID          &oid,
                                     p_uint32             block_size,
                                     pasis_md_loc_t      *distribution) :
    oid(oid)
{
    object_attrs.block_size = block_size;
    memcpy(&object_attrs.distribution, distribution, sizeof(pasis_md_loc_t));
    need_to_create_object  = deleting_object_now = false;
    // subclasses's object size shouldn't be larger than MDVersionedObject
    // (i.e., they don't add any new members variables)
    MDMemoryManager::total_history_allocation += sizeof(*this);
    ++num_objects;
}


MDVersionedObject::~MDVersionedObject()
{
    // Explicitly erase all the blocks.  This way, the MDVersionedBlock::flush
    // function gets called while this object is still live.
    blocks.clear();
    MDMemoryManager::total_history_allocation -= sizeof(*this);
    --num_objects;
}


void MDVersionedObject::initialize_new_object()
{
    MDVersion zero_ts; // constructor inits to 0

#if DEBUG_TIMESTAMPS > 0
    snprintf(zero_ts.timestamp.op_description, DEBUG_OP_DESC_SIZE, "Init");
#endif

    MDVersionedBlock* header = get_header_block();
    header->write(zero_ts, zero_ts, NULL); // null data as we're updating hist
}


void* MDVersionedObject::get_memory_for_new_block()
{
    return MDMemoryManager::get_block_memory(block_size());
}


void* MDVersionedObject::get_memory_for_new_header_block()
{
    return MDMemoryManager::get_block_memory(header_size());
}


void MDVersionedObject::free_block_memory(void *buf)
{
    MDMemoryManager::free_block_memory(buf, block_size());
}


void MDVersionedObject::free_header_block_memory(void *buf)
{
    MDMemoryManager::free_block_memory(buf, header_size());
}



void MDVersionedObject::remove_block_object_if_possible(MDVersionedBlock* vb)
{
    if (vb->num_versions() == 0) {
        blocks.erase(vb->block_num);
    }
}


void MDVersionedObject::write(MDVersion &ts, MDVersion &cond_ts,
                              p_uint32 block_num, void *data)
{
    MDVersion ts_copy(ts);
    // our history doesn't have object hashes, as we create the version entry
    // before the object is completely updated (and hence we can't calculate the
    // object hash on version-insertion).  So zero out the object hash when
    // looking for versions.  It's OK to do this since we can't have two
    // versions with the same logical timestamp, but different object hashes,
    // due to the way we assign (logical) timestamps.
    ts_copy.zeroObjectHash();

    // 1) write new version info to header.  If we have already written another
    // chunk at the same timestamp, then header won't get updated
    MDVersionedBlock* header = get_header_block();
    header->write(ts_copy, cond_ts, NULL); // null data since we're just updating hist

    // 2) update block-specific hash (set "global" object hash to dirty; update
    // whenever requested; simplifies object hash interface!).  *Only if we're
    // hashing objects.*
    
    // 3) do actual write
    // increment block_num, as block 0 is header
    MDVersionedBlock* block = get_block(block_num + 1, false);
    block->write(ts_copy, cond_ts, data);
}


void MDVersionedObject::write_barrier(MDVersion &ts, MDVersion &cond_ts)
{
    MDVersion ts_copy(ts);
    // our history doesn't have object hashes, as we create the version entry
    // before the object is completely updated (and hence we can't calculate the
    // object hash on version-insertion).  So zero out the object hash when
    // looking for versions.  It's OK to do this since we can't have two
    // versions with the same logical timestamp, but different object hashes,
    // due to the way we assign (logical) timestamps.
    ts_copy.zeroObjectHash();

    // 1) write new version info to header.
    MDVersionedBlock* header = get_header_block();
    header->write(ts_copy, cond_ts, NULL); // null data since we're just updating hist
}


// Copies-forward a version at "old_ts" to "new_ts".
void MDVersionedObject::copy_forward(MDVersion &new_ts, MDVersion &old_ts,
                                     MDVersion &cond_ts)
{
    MDVersion new_ts_copy(new_ts), old_ts_copy(old_ts);
    // our history doesn't have object hashes, as we create the version entry
    // before the object is completely updated (and hence we can't calculate the
    // object hash on version-insertion).  So zero out the object hash when
    // looking for versions.  It's OK to do this since we can't have two
    // versions with the same logical timestamp, but different object hashes,
    // due to the way we assign (logical) timestamps.
    new_ts_copy.zeroObjectHash();
    old_ts_copy.zeroObjectHash();

    // 1) write new version info to header.
    MDVersionedBlock* header = get_header_block();
    header->write(new_ts_copy, cond_ts, NULL); // null data since we're just updating hist

    // 2) Write the small data, if we had any
    p_uint32 user_data_size;
    void *small_data = header->read_small_data(old_ts_copy, &user_data_size);
    if (small_data) {
        header->write_small_data(new_ts, cond_ts, user_data_size, small_data);
    }

    // 3) Write the blocks
    p_uint32 n = num_blocks();
    for (p_uint32 i = 0; i < n; ++i) {
        // + 1 since we skip header block
        MDVersionedBlock* block = get_block(i + 1, true);
        if (block) {
            block->copy_forward(new_ts, old_ts_copy, cond_ts);
        }
    }
}


void MDVersionedObject::write_small_data(MDVersion &ts, MDVersion &cond_ts,
                                         p_uint32 size, void *data)
{
    MDVersion ts_copy(ts);
    // our history doesn't have object hashes, as we create the version entry
    // before the object is completely updated (and hence we can't calculate the
    // object hash on version-insertion).  So zero out the object hash when
    // looking for versions.  It's OK to do this since we can't have two
    // versions with the same logical timestamp, but different object hashes,
    // due to the way we assign (logical) timestamps.
    ts_copy.zeroObjectHash();

    // 1) write new version info to header.  If we have already written another
    // chunk at the same timestamp, then header won't get updated
    MDVersionedBlock* header = get_header_block();
    header->write(ts_copy, cond_ts, NULL); // null data since we're just updating hist

    // 2) update block-specific hash (set "global" object hash to dirty; update
    // whenever requested; simplifies object hash interface!).  *Only if we're
    // hashing objects.*
    
    // 3) do actual write of the small data
    header->write_small_data(ts, cond_ts, size, data);
}


void* MDVersionedObject::read(MDVersion &ts, p_uint32 block_num)
{
    // incrememnt block_num, as block 0 is header
    MDVersionedBlock* block = get_block(block_num + 1, true);
    if (block == NULL) {
        return NULL;
    } else {
        return block->read(ts);
    }
}


static void pin_block(MDVersionedBlock *block, MDVersion &ts)
{
    pasis_md_unpin_block_info_t *unpin_array;
    p_uint32                    *num_blocks_to_unpin;

    unpin_array = (pasis_md_unpin_block_info_t*)
        st_thread_getspecific(worker_pin_key);
    
    num_blocks_to_unpin = (p_uint32 *)
        st_thread_getspecific(worker_pin_count_key);

    PASIS_DASSERT(unpin_array);
    PASIS_DASSERT(*num_blocks_to_unpin < PASIS_MD_MAX_WORKER_UNPIN_BLOCKS);

    unpin_array[*num_blocks_to_unpin].block   = block;
    unpin_array[*num_blocks_to_unpin].version = ts;
    *num_blocks_to_unpin = *num_blocks_to_unpin + 1;
}


void* MDVersionedObject::read_and_pin(MDVersion &ts, p_uint32 block_num)
{
    // incrememnt block_num, as block 0 is header
    MDVersionedBlock* block = get_block(block_num + 1, true);
    if (block == NULL) {
        return NULL;
    } else {
        void* ret = block->read(ts, true);
        if (ret) {
            pin_block(block, ts);
        }
        return ret;
    }
}

void MDVersionedObject::unpin()
{
    pasis_md_unpin_block_info_t *unpin_array;
    p_uint32                    *num_blocks_to_unpin;

    unpin_array = (pasis_md_unpin_block_info_t*)
        st_thread_getspecific(worker_pin_key);
    
    num_blocks_to_unpin = (p_uint32 *)
        st_thread_getspecific(worker_pin_count_key);

    PASIS_DASSERT(unpin_array);
    PASIS_DASSERT(*num_blocks_to_unpin < PASIS_MD_MAX_WORKER_UNPIN_BLOCKS);

    for (p_uint32 i = 0; i < *num_blocks_to_unpin; ++i) {
        unpin_array[i].block->unpin(unpin_array[i].version);
    }

    *num_blocks_to_unpin = 0;
}


// Reads the small data stored in the header, if anything was stored there.
// See write_small_data.
void* MDVersionedObject::read_small_data(MDVersion &ts)
{
    MDVersionedBlock* header = get_header_block();
    return header->read_small_data(ts);
}


void MDVersionedObject::remove(MDVersion timestamp)
{
    p_uint32 n = num_blocks();
    for (p_uint32 i = 0; i < n; ++i) {
        // + 1 since we skip header block
        MDVersionedBlock* block = get_block(i + 1, true);
        if (block) {
            block->remove(timestamp);
        }
    }
    get_header_block()->remove(timestamp);
}


MDVersionedBlock* MDVersionedObject::get_block(p_uint64 block_num,
                                               bool     doingRead)
{
    // First, Check if we have this block in memory.
    block_map_t::iterator i = blocks.find(block_num);
    if (i != blocks.end()) {
        return &i->second; // Found object in memory; return it
    } else {
        if (doingRead && need_to_create_object) {
            // If we don't have the block in memory, the block could still exist
            // in the back-end.  Thus, ordinarily we have to check the back-end.
            // But there is an optimization: if the object is new and hasn't
            // been created on the back-end yet, then we know the back-end
            // doesn't have more versions than in-memory.
            // We can only do this optimization if we're not being asked to
            // create the block object, in case of a write.
            return NULL;
        }

        // Create the block object
        MDVersionedBlock& block = blocks[block_num];
        // Initialize some member variables
        block.block_num = block_num;
        block.object    = this;

        if (need_to_create_object == true) {
            // If object is new and only exists in-memory, then there can't
            // possibly be more versions in the back-end.
            block.possibly_more_history_in_backend = false;
        } else {
            block.possibly_more_history_in_backend = true;
        }

        return &block;
    }
}


void MDVersionedObject::get_history(MDVersion       &search_ts,
                                    p_uint32         num_hist_in,
                                    pasis_md_hts_t  *hist_array,
                                    p_uint32        *num_hist_out,
                                    pasis_md_auth_t *authenticators)
{
    // History is recorded at header block (block 0)
    MDVersionedBlock* header = get_header_block();
    PASIS_DASSERT(header != NULL);

    header->get_history(search_ts, num_hist_in, hist_array, num_hist_out,
                        authenticators);
}


// Returns the latest version
void MDVersionedObject::get_latest(MDVersion &result)
{
    p_uint32       num_hist_out;
    pasis_md_hts_t hist_array;

    MDVersion search_ts;
    search_ts.setLatest();

    get_history(search_ts, 1, &hist_array, &num_hist_out);

    if (num_hist_out == 0) {
        result.setZero();
    } else {
        result = MDVersion(&hist_array);
    }
}


bool MDVersionedObject::has_version(MDVersion &version)
{
    MDVersionedBlock* header = get_header_block();
    return header->has_version(version);
}


char* MDVersionedObject::dump_history()
{
    MDVersionedBlock* header = get_header_block();
    return header->dump_history();
}


// creates obj. on back-end if necessary
pasis_error_t MDVersionedObject::create_if_necessary()
{
    PASIS_DASSERT(0); // XXX: only in-memory for now
}


////////////////////////////////////////////////////////////////////////////////
// MDObjectUniverse members
////////////////////////////////////////////////////////////////////////////////

MDObjectUniverse::MDObjectUniverse()
{
}


MDObjectUniverse::~MDObjectUniverse()
{
    object_map_t::iterator i;
    for (i = objects.begin(); i != objects.end(); ) {
        // Get pointer to next iterator, before we erase current iterator
        object_map_t::iterator next_i = i; ++next_i;
        delete i->second;
        objects.erase(i);
        i = next_i;
    }

    // Check that we've deallocated all the objects
    PASIS_DASSERT(MDVersionedObject::num_objects == 0);
}


MDVersionedObject* MDObjectUniverse::lookup(const MDObjectID &oid)
{
    // First, check to see if we have object in front-end cache.
    object_map_t::iterator i = objects.find(oid);
    if (i != objects.end()) {
        // Check if the object is currently being deleted.  If so, pretend it's
        // already gone.
        if (i->second->deleting_object_now) {
            return NULL;
        } else {
            // Cache hit.  Return object
            return i->second;
        }
    } else if (not_using_backend) {
        // If not using back-end and object isn't in front-end cache, then it
        // doesn't exist.
        return NULL;
    } else {
        // Cache miss.  Need to check with the back-end to see if the object
        // exists or not.

#if 1
        // XXX: we're only operating in-memory for now.  -mabdelmalek 7/23/2005
        return NULL;
#else
        // back-end lookup.  Note: this code works fine even if multiple threads
        // do a back-end lookup concurrently for the same object.
        MDVersionedObject new_object(oid, 0, yyy);
        vo.oid = oid;
        pasis_error_t error = backend->lookup (&vo);
        if (error) {
            return NULL; // error on lookup; object does not exist.
        } else {
            // Object existed on back-end; add to object cache
            
            VersionedObject& new_object = objects[oid];
            PASIS_DASSERT(0); // XXX: set/get object hash
            new_object.need_to_create_object = false;
            // "vo.fid_Data" has been initialized from fid_table_lookup, and so
            // we init new object from "vo".
            memcpy(&new_object.fid_data, &vo.fid_data,
                   sizeof(pasis_fe_fid_data_t));
            // no need to create object in backend, since it already exists
            new_object.oid = oid;

            return &new_object;
        }
#endif
    }
}


// Does a front-end object create (doesn't touch the back-end).  Called for
// adding a new object in the system (_not_ when simply an object is just in the
// back-end).
MDVersionedObject* MDObjectUniverse::create(MDObjectID          &oid,
                                            p_uint32             block_size,
                                            pasis_md_loc_t      *distribution)
{
    // Imagine that multiple threads are doing writes to the same (uncreated)
    // object concurrently, and their lookups went to the backend and "failed".
    // So now we can have multiple threads calling create().  Fix this by
    // checking once again if the object exists or not.
    object_map_t::iterator i = objects.find(oid);
    if (i != objects.end()) {
        return i->second;
    }

    // Create object, inits it (inserts version 0), and adds it to object map
    MDVersionedObject *new_object =
        MDVersionedObject::create_new_object(oid, block_size, distribution); 
    objects[oid] = new_object;

    return new_object;
}


pasis_error_t MDObjectUniverse::remove(MDObjectID &oid)
{
    pasis_error_t error;

    // First, lookup object (in memory, then on disk) to see if it exists
    MDVersionedObject* object = lookup(oid);
    if (object == NULL) {
        return PASIS_MD_ERR_NOENT;
    }

    // Tricky!  Imagine that another thread was also doing a remove.  So it did
    // the lookup, got the object, and then was stuck in I/O with the backend
    // while deleting the object.  Our thread comes along, and does a lookup,
    // gets the object, and tries to double-delete it!
    // To avoid this, we set a flag early on to declare that we're deleting the
    // object.  lookup() will not return an object if it's currently being
    // deleted.  (lookup() will pretend the object doesn't exist.)
    // Also, this flag tells the MDVersionedBlock objects not to flush the data.
    object->deleting_object_now = true;

    // Remove object from the back-end, if necessary
    if (object->need_to_create_object) {
        // If the object's need-to-create flag is set, it means we haven't
        // created the object on the back-end yet.  Hence, we don't have to go
        // to the back-end to remove it.
        error = PASIS_SUCCESS;
    } else {
#if 1
        // XXX: we're only operating in-memory for now.  -mabdelmalek 7/23/2005
        error = PASIS_SUCCESS;
#else
        // Remove the object from the back-end
        ScopeLock auto_lock(object->mutex);
        error = backend->remove (object);
#endif
    }

    // Remove the object from the object hash map
    if (error == PASIS_SUCCESS) {
        delete object;
        objects.erase(oid);
    } else {
        //printf("%s: not deleting object\n", __FUNCTION__);
    }
 
    return error;
}
  
 
/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
