/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_debug.h"
#include "pasis_timer.h"

#include "pasis_md_s4.h"
#include "pasis_md_attr.h"
#include "pasis_md_auth.h"
#include "pasis_md_clnt.h"
#include "pasis_md_dir.h"
#include "pasis_md_hist_matrix.h"
#include "pasis_md_log.h"
#include "pasis_md_memory_versioning.h"
#include "pasis_md_object_sync.h"
#include "pasisio_drives.h"
#include "pasisio_drive_md.h"
#include "pasisio_md_protocol.h"

#define PASIS_MD_TIMER_OBJECT_SYNC_NUM    2
#define PASIS_MD_TIMER_OBJECT_SYNC_CLIENT 0
#define PASIS_MD_TIMER_OBJECT_SYNC_SERVER 1


typedef struct pasis_md_object_sync_stats_s {
    pasis_timer_state_t  *timers;
} pasis_md_object_sync_stats_t;

static pasis_md_object_sync_stats_t stats;



/* Protocol/linkage records; helps us locate other metadata S-N's */
static pasis_md_loc_t *distribution = NULL;

static p_uint32 node_id = PASIS_MAX_NUM_SERVERS;



static bool
is_server_up (p_uint32            server_node_id,
              pasisio_linkage_t  *linkage) /* Result */
{
    pasis_md_linkage_t          *server_linkage;
    pasisio_drive_instance_t    *server_drive;
    char                         base_path[64];

    server_linkage = &distribution->linkages[server_node_id];
    snprintf (base_path, 64, "/%u", server_linkage->partition);
    server_drive = PasisIO_DriveFindInstanceIP (PASISIO_DRIVE_TYPE_MD_INDEX,
                                                server_linkage->ip_address,
                                                base_path);

    if (server_drive == NULL || server_drive->status == PASISIO_DRIVE_STATUS_DOWN) {
        return false;
    }

    linkage->ip_address = server_linkage->ip_address;
    linkage->partition  = server_linkage->partition;

    return true;
}


/* Selects another S-N to perform object syncing with.  */
static pasis_error_t
client_select_servers (pasis_md_hist_matrix_t   *hist_matrix,
                       pasis_md_hts_t           *sync_ts,
                       p_uint32                  N,
                       p_uint32                  B,
                       pasisio_linkage_t        *servers) /* result */
{
    p_uint32 server_id;
    p_uint32 num_found = 0;

    /* For load-balancing, we start our search of S-N's at an offset. */
    server_id = (node_id + 1) % distribution->protocol.N;

    /* Find other storage-node with the timestamp we're syncing to */
    while (1) {
        if (pasis_md_hist_matrix_node_has_ts (server_id, hist_matrix, sync_ts) &&
            is_server_up (server_id, &servers[num_found])) {
            PASIS_DPRINTF (PASIS_DEBUG_S4_MD_OBJECT_SYNC, 4,
                           "%s: syncing with %u\n", __FUNCTION__, server_id);
            // We need B+1 servers to sync with
            if (++num_found == B + 1) {
                return PASIS_SUCCESS;
            }
        }

        server_id = (server_id + 1) % N;

        if (server_id == node_id) {
            break; /* already iterated over all S-N's */
        }
    }

    /* we didn't find enough other S-N's with the timestamp we're syncing to */
    return PASIS_ERROR;
}



pasis_error_t
pasis_md_object_sync_client (p_uint32                  node_id,
                             MDVersionedObject        *object,
                             pasis_md_hts_t           *sync_ts,
                             pasis_md_hts_t           *lcwts_hts,
                             pasis_md_hts_t           *new_hts,
                             pasis_md_hist_matrix_t   *hist_matrix)
{
    pasisio_linkage_t server_linkages[PASIS_MAX_NUM_SERVERS];
    pasis_error_t error;
    p_uint32 block_size = object->object_attrs.block_size;
    pasis_md_protocol_t *protocol = &object->object_attrs.distribution.protocol;

    PASIS_TIMER_START ();

    pasis_md_log_msg ("%s: object syncing to get %llu and make it %llu",
                      __FUNCTION__, sync_ts->time, new_hts->time);

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_OBJECT_SYNC, 3,
                   "%s: node %u objct syncing FID %"PRIx64", TS %"PRIu64", barrier = %d",
                   __FUNCTION__, node_id, *(p_uint64*)&object->oid.fid,
                   sync_ts->time, sync_ts->barrier_flag);
#if DEBUG_TIMESTAMPS > 0
    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_OBJECT_SYNC, 3,
                   ", op description: %s", sync_ts->op_description);
#endif
    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_OBJECT_SYNC, 3, "\n");

    // make sure caller always sets object's chunk_size to correct value
    PASIS_DASSERT(block_size > 0);

    // Check if we should not be doing object syncing
    MDVersion sync_version(sync_ts);
    PASIS_DASSERT(object->has_version(sync_version) == false);

    /* 1) Select which other S-Ns to contact. */
    error = client_select_servers(hist_matrix, sync_ts, protocol->N, protocol->B,
                                  server_linkages);
    if (error) {
        PASIS_DASSERT (0); /* other S-Ns should have this timestamp */
    }


    // Transfer up to MIN(128K, PASIS_MD_DATA_NUM_IOVECS blocks) at a time
    p_uint32 starting_block_num = 0;
    p_uint32 num_blocks_per_transfer = MIN((128*1024*1024) / block_size,
                                           PASIS_MD_DATA_NUM_IOVECS);
    p_uint32 buffer_size = (protocol->B + 1) * num_blocks_per_transfer * block_size;
    p_byte *buffer = (p_byte*)malloc(buffer_size + 8);
    p_byte *useable_buffer = buffer + 4;
    *(p_uint32*)buffer = 314159265;
    *(p_uint32*)(useable_buffer + buffer_size) = 314159265;
    p_uint32 num_total_blocks, num_returned_blocks;
    
    while (1) {
        error = PasisIO_MDObjectSync(protocol, &object->oid.fid, server_linkages,
                                     sync_ts, block_size, useable_buffer,
                                     starting_block_num, num_blocks_per_transfer,
                                     &num_total_blocks, &num_returned_blocks);
        if (error) {
            // probably failed because other server doesn't have the version we want
            // printf("%s: object syncing failed (%u)\n", __FUNCTION__, error);
            goto exit;
        }

        // check that we didn't overrun "buffer"
        PASIS_DASSERT(num_returned_blocks <= num_blocks_per_transfer);

        MDVersion new_version(new_hts), cond_version(lcwts_hts);
        for (p_uint32 i = 0; i < num_returned_blocks; ++i) {
            void *block_buffer = object->get_memory_for_new_block();
            memmove(block_buffer, useable_buffer + i * block_size, block_size);
            object->write(new_version, cond_version, starting_block_num + i,
                          block_buffer);
        }

        // If we got less blocks than we asked for, then we're done object syncing
        if (num_returned_blocks < num_blocks_per_transfer) {
            break; // done
        } else {
            // more object syncing to do
            starting_block_num += num_blocks_per_transfer;
        }
    }

    if (starting_block_num + num_returned_blocks < num_total_blocks) {
        // we didn't get all the blocks
        PASIS_DASSERT(0); // unexpected for now
        error = PASIS_ERROR;
    }

 exit:
    pasis_md_log_msg ("%s: object syncing finished with error %u",
                      __FUNCTION__, error);


    PASIS_DASSERT(*(p_uint32*)buffer == 314159265);
    PASIS_DASSERT(*(p_uint32*)(useable_buffer + buffer_size) == 314159265);
    free(buffer);

    if (error) {
        // make sure we remove all blocks belonging to sync_ts, in case we fail
        // halfway through
        object->remove(sync_ts);
    }

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_OBJECT_SYNC_CLIENT);

    return error;
}


static pasis_error_t
server_do_some_syncing (MDVersionedObject           *object,
                        MDVersion                   &sync_version,
                        pasis_md_object_sync_args_t *args,
                        pasis_md_object_sync_res_t  *res)
{
    p_uint32      i;
    pasis_error_t error;

    res->block_size          = object->object_attrs.block_size;
    PASIS_DASSERT(res->block_size > 0);
    res->num_total_blocks    = object->num_blocks();

    for (i = 0; i < args->num_blocks_to_ret; ++i) {
        p_uint32 current_block_num = args->start_block_num + i;
        if (current_block_num >= res->num_total_blocks) {
            // client asked for more blocks than we have
            break;
        }

        void *data = object->read_and_pin(sync_version, current_block_num);
        PASIS_DASSERT(data != NULL);
        res->data.server.iovec[i].iov_base = data;
    }

    res->data.server.iovec_count = i;
    res->num_returned_blocks     = i;

    error = PASIS_SUCCESS;
 exit:
    return error;
}


void
pasis_md_object_sync_server (pasis_fe_s4_ipc_t       *ipc_op,
                             pasis_md_read_op_args_t *args,
                             pasis_md_read_op_res_t  *res)
{
    MDVersionedObject           *object;
    MDVersion                    sync_version;
    pasis_md_object_sync_args_t *sync_args = &args->op_args.object_sync;
    pasis_md_object_sync_res_t  *sync_res = &res->op_res.object_sync;
    p_uint32                     real_node_id;    
    pasis_error_t                error;
    
    PASIS_TIMER_START ();

    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1: node_id);

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_OBJECT_SYNC, 3,
                   "%s: node %u received sync request for FID %"PRIx64", TS %"PRIu64", ",
                   __FUNCTION__, real_node_id, *(p_uint64*)sync_args->fid,
                   args->ts.time);
#if DEBUG_TIMESTAMPS > 0
    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_OBJECT_SYNC, 3, ", op description: %s",
                   sync_args->ts->op_description);
#endif
    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_OBJECT_SYNC, 3, "\n");

    /* Setup crucial return fields right away */
    res->op_type = PASIS_MD_OBJECT_SYNC;
    res->op_status = (pasis_md_rpc_stat_t)PASIS_ERROR;
    res->history_vec_num = 0;
    sync_res->data.server.iovec_count = 0;
    sync_res->num_total_blocks = 0;
    sync_res->num_returned_blocks = 0;

    /* To avoid deadlocks, we can't lock the objcts here.  Consider a
     * multi-object update operation involving FIDs 1 and 2.  Imagine S-N A
     * needs to object-sync FID 1, while S-N B needs to object-sync FID 2.  Both
     * S-N's first lock the FID's and then send the object-sync request.
     * However, since both S-N's have locked the FID's, neither of them can
     * handle the object syncing request.
     *
     * Hence, we can NOT lock the FID's here.  Consistency is still fine.
     * Update: why is consistency still fine?!  What if I/O happens and we get
     * swapped out?  For now, this is OK since we're only in-memory. */
    /* pasis_md_fid_lock (args->part, 1, &sync_args->fid, 0, &fid_lock); */

    MDObjectID oid(real_node_id, sync_args->fid);
    object = md_universe->lookup(oid);

    if (object == NULL) {
        printf("%s: we don't have the object to begin with!\n", __FUNCTION__);
        PASIS_DASSERT(0);
        error = PASIS_MD_ERR_OBJSYNC_NOT_EXIST;
        goto exit;
    }

#if 0
    printf("%s: here are the versions we have: %s\n", __FUNCTION__,
           object->dump_history());
#endif

    sync_version = MDVersion(&args->ts);
    if (object->has_version(sync_version) == false) {
        //printf("%s: we don't have the requested version %"PRIu64".\n", __FUNCTION__,
        //       args->ts.time);
        //printf("%s: here are the versions we have: %s\n", __FUNCTION__,
        //       object->dump_history());
        /* Tell object sync client that we don't have the version they want */
        error = PASIS_MD_ERR_OBJSYNC_NOT_EXIST;
        goto exit;
    }

    if (sync_args->return_attrs) {
        // for NFS service, return attrs
        PASIS_DASSERT(0);
    } else {
        // zero out attributes so that all servers agree in voting stage
        memset(&sync_res->attr, 0, sizeof(pasis_md_attr_t));
    }

    error = server_do_some_syncing (object, sync_version,
                                    sync_args, sync_res);
    if (error) goto exit;

    error = PASIS_SUCCESS;

 exit:
    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_OBJECT_SYNC_SERVER);
    /* pasis_md_fid_unlock (1, &fid_lock); */
    res->status = (pasis_md_rpc_stat_t)error;
    res->op_status = (pasis_md_rpc_stat_t)error;
}


pasis_error_t
pasis_md_object_sync_init (pasis_md_loc_t  *_distribution,
                           p_uint32         _node_id)
{
    distribution = _distribution;
    node_id = _node_id;

    /* Initialize PASIS protocol module, since we use PasisIO_MDObjectSync.
     * Potentially every request can cause object syncing, so use
     * PASIS_MD_WORKERS as the number of outstanding requests. */
    /* how big should max block size be?  probably only important for EC service*/
    return PasisIO_Init(PASIS_MD_WORKERS, 4096);
}


void
pasis_md_object_sync_shutdown (void)
{
    PasisIO_Shutdown();
}


void
pasis_md_object_sync_stats_init (void)
{
    memset (&stats, 0, sizeof (pasis_md_object_sync_stats_t));

    /* Set up timers */
    stats.timers = Pasis_TimerInit ("Object_Sync_Timers",
                                    PASIS_MD_TIMER_OBJECT_SYNC_NUM, 0);
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_OBJECT_SYNC_CLIENT,
                         "object_sync_client");
    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_OBJECT_SYNC_SERVER,
                         "object_sync_server");
}


void
pasis_md_object_sync_stats_dump (int shutdown_flag)
{
    if (shutdown_flag) {
        Pasis_TimerShutdown (stats.timers);
    } else {
        Pasis_TimerPrintAndReset (stats.timers);
    }
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
