/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_timer.h"
#include "pasis_cache.h"

#include "pasis_md_s4.h"

#include "pasis_md_attr.h"
#include "pasis_md_dir_ops.h"
#include "pasis_md_memory_versioning.h"
#include "pasis_md_ops.h"
#include "pasis_md_err.h"

typedef struct pasis_md_op_sym_blk_s {
    unsigned int   len;
    p_byte         path[PASIS_MD_MAX_PATHLEN];
} pasis_md_op_sym_blk_t;


static inline void
dump_attrs (const char *str,
            pasis_md_attr_t *attr)
{
#if 0
    printf ("%s attrs: mode=%o, size=%llu, blocks=%llu, ref_cnt=%u, fileid=%u\n",
            str, attr->mode, attr->size, attr->blocks, attr->ref_cnt, attr->fileid);
#endif
}


static inline pasis_error_t
attr_write (MDVersionedObject *file,
            MDVersion         &new_version,
            MDVersion         &cond_version,
            pasis_md_attr_t   *attr)
{
    pasis_error_t error = PASIS_SUCCESS;
    
#if USE_OBJECT_HASHES > 0
    /* Write a hash of the chunk */
    /* error = pasis_md_hash_insert (ipc_op, fid, 0, NULL, 0,
       (p_byte *)attr, sizeof (pasis_md_attr_t)); */
    assert(0); /* unimplemnted */
    if (error) return error;
#else
    /* Not storing object hash in the timestamp */
#endif


    file->write_small_data(new_version, cond_version,
                           sizeof(pasis_md_attr_t), attr);

    return error;
}


static pasis_error_t
symlink_create (MDVersionedObject    *symfile,
                char                 *path,
                MDVersion            &new_version,
                MDVersion            &cond_version)
{
    pasis_md_op_sym_blk_t *sym_blk;
    p_byte                *sblk;
    pasis_error_t          error;
    
    sblk = (p_byte*) symfile->get_memory_for_new_block();
    sym_blk = (pasis_md_op_sym_blk_t *)(sblk);
    
    sym_blk->len = MIN (strlen (path) + 1, PASIS_MD_MAX_PATHLEN);
    memcpy (sym_blk->path, path, sym_blk->len);

    symfile->write(new_version, cond_version, 0, sblk);

    return OP_OK;
}


static pasis_error_t
internal_link (p_uint32             mtime[2],
               MDVersionedObject   *from,
               MDVersionedObject   *to_dir,
               char                *to_name,
               p_uint64             blkno,
               p_byte              *blk,
               MDVersion           &new_from_version,
               MDVersion           &cond_from_version,
               MDVersion           &new_to_version,
               MDVersion           &cond_to_version,
               pasis_md_attr_t     *from_attr_out,      /* Out args */
               pasis_md_attr_t     *to_dir_attr_out,    /* Out args */
               p_byte             **new_blk_out = NULL) /* Out arg */
{
    p_byte *new_blk;
    pasis_error_t  error;

    if (MD_ATTR_IS_DIR (from_attr_out)) {
        /* Need to fixup the '..' entry and incr parents ref_cnt */
        error = pasis_md_dir_rename_parent (from, new_from_version,
                                            cond_from_version, &to_dir->oid.fid,
                                            to_dir_attr_out->fileid);
        if (error) {
            fprintf (stderr, "Error: rename parent\n");
            return error;
        }
        
        to_dir_attr_out->ref_cnt++;
    }

    /* Insert to_name in to_dir */
    /* blk points to last block, try and insert entry */
    new_blk = (p_byte*) to_dir->get_memory_for_new_block();
    pasis_md_dir_hdr_t *hdr = (pasis_md_dir_hdr_t *)blk;
    memmove(new_blk, blk, hdr->free_offset);

    error = pasis_md_dir_insert (new_blk, &from->oid, to_name, 
                                 from_attr_out->fileid, NULL);
    if (error) {
        /* Not enough free space, allocate/init a new block */
        pasis_md_dir_init_hdr (new_blk);
        pasis_md_dir_insert (new_blk, &from->oid, to_name,
                             from_attr_out->fileid, NULL);
        to_dir_attr_out->blocks++;
        to_dir_attr_out->size += PASIS_MD_DIR_BLK_SIZE;
        blkno++;
    }

    /* Incr from ref_cnt */
    from_attr_out->ref_cnt++;

    /* Set mtime of to_dir and from */
    pasis_md_attr_update_time (to_dir_attr_out->mtime, mtime);
    pasis_md_attr_update_time (from_attr_out->mtime, mtime);

    /* Write out from attributes */
    error = attr_write (from, new_from_version, cond_from_version,
                        from_attr_out);
    if (error) {
        PASIS_PRINT_ERR ("%s: error setting attributes\n", __FUNCTION__);
        to_dir->free_block_memory(new_blk);
        return OPERR_IO;
    }

    /* pasis_md_dump_dir_block (blk); */

    /* Write out dir block */
    to_dir->write(new_to_version, cond_to_version, blkno, new_blk);

    if (new_blk_out) {
        *new_blk_out = new_blk;
    }
    
    /* printf ("\nOUT %s  (OK)\n", __FUNCTION__); */
    
    return OP_OK;
}


pasis_error_t
internal_unlink (p_uint32             mtime[2],
                 MDVersionedObject   *parent_dir,
                 MDVersionedObject   *from_fid,
                 char                *from_name,
                 unsigned int         offset,
                 p_uint64             blkno,
                 p_byte              *blk,
                 MDVersion           &new_parent_version,
                 MDVersion           &cond_parent_version,
                 MDVersion           &new_file_version,
                 MDVersion           &cond_file_version,
                 pasis_md_attr_t     *from_attr,
                 pasis_md_attr_t     *parent_attr_out,    /* Out args */
                 p_byte             **new_blk_out = NULL) /* Out arg */
{
    pasis_error_t error;

    /* Decrement parent ref_cnt if from is a directory */
    if (MD_ATTR_IS_DIR (from_attr)) {
        parent_attr_out->ref_cnt--;
    }

    /* Decrement from_attr's ref_cnt */
    from_attr->ref_cnt--;

    /* XXX if the ref_cnt is 0 for files, or 1 for directories this obj
     * should be deleted */

    // Create a new buffer for the soon-to-be-modified directory block
    p_byte *new_blk = (p_byte*) parent_dir->get_memory_for_new_block();
    pasis_md_dir_hdr_t *hdr = (pasis_md_dir_hdr_t *)blk;
    memmove(new_blk, blk, hdr->free_offset);

    /* Unlink from_name */
    pasis_md_dir_remove (new_blk, offset);

    /* Set mtimes of parent and from */
    pasis_md_attr_update_time (from_attr->mtime, mtime);
    pasis_md_attr_update_time (parent_attr_out->mtime, mtime);

    /* Write out from attrs */
    error = attr_write (from_fid, new_file_version, cond_file_version,
                        from_attr);
    if (error) {
        PASIS_PRINT_ERR ("%s: error setting attributes\n", __FUNCTION__);
        return OPERR_IO;
    }

    /* Write out parent */
    parent_dir->write(new_parent_version, cond_parent_version, blkno, new_blk);
    attr_write (parent_dir, new_parent_version, cond_parent_version,
                parent_attr_out);

    if (new_blk_out) {
        *new_blk_out = new_blk;
    }

    /* printf ("\nOUT %s  (OK)\n", __FUNCTION__); */

    return OP_OK;
}


/* Creates the root directory */
pasis_error_t
pasis_md_op_createroot (MDVersionedObject *root_object)
{
    pasis_error_t   error;
    pasis_md_attr_t root_attr;
    pasis_md_hts_t  one_hts, zero_hts;

    memset(&zero_hts, 0, sizeof(pasis_md_hts_t));
    memset(&one_hts,  0, sizeof(pasis_md_hts_t));
    one_hts.time = 1; /* First version has logical time 1 */
#if DEBUG_TIMESTAMPS > 0
    snprintf(one_hts.op_description, DEBUG_OP_DESC_SIZE, "Root create");
#endif

    /* Set up create attributes */
    memset(&root_attr, 0, sizeof(pasis_md_attr_t));
    root_attr.mode    = MD_ATTR_MODE_DIR | 01777;  /* Sticky bit set */
    root_attr.ref_cnt = 2;
    root_attr.uid     = 0;
    root_attr.gid     = 0;
    root_attr.fileid  = PASIS_MD_ROOT_FILEID_C;
    root_attr.blocks  = 1;
    root_attr.size    = PASIS_MD_DIR_BLK_SIZE;

#if 0
    /* The following line causes different S-N's to have different attributes
     * for the root FID.  This is bad, as the classification code won't find a
     * complete classification (since timestamps include the attribute hash) */
    pasis_md_attr_set_time(root_attr.ctime);
#else
    /* Instead, we just zero the create time, to ensure that all root FID's have
     * the same starting timestamp */
    root_attr.ctime[0] = root_attr.ctime[1] = 0;
#endif
    root_attr.ctime[1] = 0;  /* Zero usec field mtime, ctime, atime */
    root_attr.atime[0] = root_attr.mtime[0] = root_attr.ctime[0];
    root_attr.atime[1] = root_attr.mtime[1] = root_attr.ctime[1];

    error = pasis_md_dir_create(root_object, &root_attr, root_object,
                                &one_hts, &zero_hts);
    if (error) {
        return error;
    }

    return OP_OK;
}



/* Covers: create, mkdir, mknod, symlink */
pasis_error_t
pasis_md_op_create(p_uint32             mtime[2],
                   MDVersionedObject   *parent,
                   MDVersionedObject   *create,
                   pasis_md_hts_t      *new_hts,
                   pasis_md_hts_t     **cond_hts,
                   char                *create_name,
                   pasis_md_attr_t     *create_attr,
                   char                *symlink_path,
                   pasis_md_attr_t     *create_attr_out,   /* Out args */
                   pasis_md_attr_t     *parent_attr_out)   /* Out args */
{
    pasis_error_t error;
    p_uint64      blkno;
    p_byte       *old_blk, *new_blk;
    unsigned int  blk_len;
    pasis_md_dir_hdr_t *hdr;
    MDVersion new_parent_version(&new_hts[0]), cond_parent_version(cond_hts[0]);
    MDVersion new_create_version(&new_hts[1]), cond_create_version(cond_hts[1]);


    /** Perform validation -- read-only phase **/

    /* Get parent's attributes, and lookup create_name in parent */
    error = pasis_md_dir_find_entry (parent, cond_parent_version, create_name,
                                     &old_blk, &blkno, NULL, parent_attr_out);
    if (error != OPERR_NOENT) {
        printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
        return (error == OP_OK) ? OPERR_EXIST : error;
    }

    new_blk = (p_byte*) create->get_memory_for_new_block();

    /** Perform operation **/
    if (old_blk == NULL) {
        // Don't have any directory blocks
        goto add_to_new_dir_block;
    }

    hdr = (pasis_md_dir_hdr_t *)old_blk;
    memmove(new_blk, old_blk, hdr->free_offset);

    /* Insert create_name in parent_dir */
    /* blk points to last block, try and insert entry */
    error = pasis_md_dir_insert (new_blk, &create->oid, create_name, 
                                 create_attr->fileid, &blk_len);
    if (error) {
add_to_new_dir_block:
        /* Not enough free space, allocate/init a new block */
        pasis_md_dir_init_hdr (new_blk);
        pasis_md_dir_insert (new_blk, &create->oid, create_name, 
                             create_attr->fileid, &blk_len);
        parent_attr_out->blocks++;
        blkno++;
    }

    /* Set up create attributes */
    create_attr->blocks = 0;
    create_attr->ref_cnt = 1;

    pasis_md_attr_copy_time (create_attr->ctime, mtime);
    pasis_md_attr_copy_time (create_attr->atime, mtime);
    pasis_md_attr_copy_time (create_attr->mtime, mtime);
    
    /* Set parents mtime */
    pasis_md_attr_update_time (parent_attr_out->mtime, mtime);

    if (MD_ATTR_IS_DIR (create_attr)) {
        create_attr->size = PASIS_MD_DIR_BLK_SIZE;
        create_attr->blocks = 1;
        create_attr->ref_cnt++;
    }

    /* Set attributes on create FID */
    error = attr_write (create, new_create_version, cond_create_version,
                        create_attr);
    if (error) {
        PASIS_PRINT_ERR ("%s: error creating data object\n", __FUNCTION__);
        create->free_block_memory(new_blk);
        return OPERR_IO;
    }

    /* Create symlink and dirs data blocks, update parent link count */
    if (MD_ATTR_IS_DIR (create_attr)) {
        parent_attr_out->ref_cnt++;
        error = pasis_md_dir_create (parent, create_attr, create,
                                     &new_hts[1], cond_hts[1]);
        if (error) {
            PASIS_PRINT_ERR ("%s: error dir write failed\n", __FUNCTION__);
            create->free_block_memory(new_blk);
            return error;
        }
    } else if (MD_ATTR_IS_LNK (create_attr)) {
        error = symlink_create (create, symlink_path, new_create_version,
                                cond_create_version);
        if (error) {
            PASIS_PRINT_ERR ("%s: error symlink write failed\n", __FUNCTION__);
            create->free_block_memory(new_blk);
            return error;
        }
    }

    /* Write out modified parent dir */
    parent->write(new_parent_version, cond_parent_version, blkno, new_blk);
    attr_write (parent, new_parent_version, cond_parent_version, parent_attr_out);
    
    /* Copy output create attrs */
    memcpy (create_attr_out, create_attr, sizeof (pasis_md_attr_t));

    /* printf ("\nOUT %s  (OK)\n", __FUNCTION__); */

    return OP_OK;
}


pasis_error_t
pasis_md_op_link (p_uint32             mtime[2],
                  MDVersionedObject   *from,
                  MDVersionedObject   *to_dir,
                  pasis_md_hts_t      *new_hts,
                  pasis_md_hts_t     **cond_hts,
                  char                *to_name,
                  pasis_md_attr_t     *from_attr_out,   /* Out args */
                  pasis_md_attr_t     *to_dir_attr_out) /* Out args */
{
    pasis_error_t  error;
    p_uint64       blkno;
    p_byte        *blk;

    MDVersion new_from_version(&new_hts[0]), cond_from_version(cond_hts[0]);
    MDVersion new_to_version(&new_hts[1]), cond_to_version(cond_hts[1]);

    /** Perform validation -- read-only phase **/

    /* Lookup to_name in to_dir, get to_dir's attrs while at it */
    error = pasis_md_dir_find_entry (to_dir, cond_to_version, to_name,
                                     &blk, &blkno, NULL, to_dir_attr_out);
    if (error != OPERR_NOENT) {
        printf ("link failed (%d)\n", __LINE__);
        return OPERR_EXIST;
    }

    /* Get from attrs */
    void* local_from_attr = from->read_small_data(cond_from_version);
    if (local_from_attr == NULL) {
        PASIS_PRINT_ERR ("%s: error reading attributes\n", __FUNCTION__);
        printf ("link failed (%d)\n", __LINE__);
        return OPERR_IO;
    } else {
        memcpy (from_attr_out, local_from_attr, sizeof (pasis_md_attr_t));
    }

    /* Fail if from is a directory */
    if (MD_ATTR_IS_DIR (from_attr_out)) {
        printf ("link failed (%d)\n", __LINE__);
        return OPERR_ISDIR;
    }

    /** Do real link **/
    error =  internal_link (mtime, from, to_dir, to_name, blkno, blk,
                            new_from_version, cond_from_version,
                            new_to_version, cond_to_version,
                            from_attr_out, to_dir_attr_out);

    if (error) {
        printf ("link failed (%d)\n", __LINE__);
    }
    return error;
}


/* Doubles as unlink and rmdir if from_fid specifies a directory */
pasis_error_t
pasis_md_op_unlink(p_uint32             mtime[2],
                   MDVersionedObject   *parent_dir,
                   MDVersionedObject   *from_fid,
                   pasis_md_hts_t      *new_hts,
                   pasis_md_hts_t     **cond_hts,
                   char                *from_name,
                   pasis_md_attr_t     *parent_attr_out) /* Out args */
{
    pasis_error_t  error;
    unsigned int   offset;
    p_uint64       blkno;
    p_byte        *old_blk;

    pasis_md_dir_entry_t *dent;
    pasis_md_attr_t       from_attr;

    MDVersion new_parent_version(&new_hts[0]), cond_parent_version(cond_hts[0]);
    MDVersion new_from_version(&new_hts[1]), cond_from_version(cond_hts[1]);

    /** Perform validation -- read-only phase **/

    /* Get parent's attributes, and lookup from_name in parent */
    error = pasis_md_dir_find_entry (parent_dir, cond_parent_version, from_name,
                                     &old_blk, &blkno, &offset, parent_attr_out);
    if (error) {
        printf ("unlink failed for %s (%d, %d)\n", from_name, error, __LINE__);
        //printf ("\tDirectory contents:\n");
        //pasis_md_dump_dir (ipc_op, parent_dir);
        return error;
    }

    /* Set directory entry pointer */
    dent = (pasis_md_dir_entry_t *)(old_blk + offset);

    /* Check from_fid's fid matches name's fid */
    if (md_fid_compare (&from_fid->oid.fid, &dent->fid) != 0) {
        printf ("unlink failed (%d)\n", __LINE__);
        return OPERR_STALE;
    }
    
    /* Get attributes of from_fid */
    void* local_from_attr = from_fid->read_small_data(cond_from_version);
    if (local_from_attr == NULL) {
        PASIS_PRINT_ERR ("%s: error getting attributes\n", __FUNCTION__);
        printf ("unlink failed (%d, %d)\n", error, __LINE__);
        return OPERR_IO;
    } else {
        memcpy (&from_attr, local_from_attr, sizeof (pasis_md_attr_t));
    }

    /* If unlinking a dir, and not part of a rename make sure dir is empty */
    if (MD_ATTR_IS_DIR (&from_attr)) {
        /* Make sure directory is empty */
        if (pasis_md_dir_blk_is_empty(from_fid, &cond_from_version, &from_attr)
            != OP_OK) {
            printf ("unlink failed (%d, %d)\n", error, __LINE__);
            //printf ("\tDirectory contents:\n");
            //pasis_md_dump_dir (ipc_op, from_fid);
            return OPERR_NOTEMPTY;
        }
        PASIS_DASSERT (from_attr.ref_cnt == 2);
    }


    /** Do real unlink **/
    error = internal_unlink (mtime, parent_dir, from_fid, from_name, 
                             offset, blkno, old_blk, new_parent_version,
                             cond_parent_version, new_from_version,
                             cond_from_version, &from_attr, parent_attr_out);
    
    if (error) {
        printf ("unlink failed (%d, %d)\n", error, __LINE__);
        return error;
    }

    return OP_OK;
}


pasis_error_t
pasis_md_op_setattr (MDVersionedObject    *fid,
                     pasis_md_hts_t       *new_hts,
                     pasis_md_hts_t      **cond_hts,
                     pasis_md_sattr_t     *sattr,
                     p_uint32              guard_ctime[2],
                     pasis_md_attr_t      *attr_out) /* Out args */
{
    MDVersion new_version(&new_hts[0]), cond_version(cond_hts[0]);
    pasis_error_t error;

    /* printf ("\nIN %s\n", __FUNCTION__); */

    //if (error) return error;
    
    /** Perform operation **/
    
    /* Get FID attrs */
    void* local_attr = fid->read_small_data(cond_version);
    if (local_attr == NULL) {
        PASIS_PRINT_ERR ("%s: error getting attributes\n", __FUNCTION__);
        return OPERR_IO;
    } else {
        memcpy (attr_out, local_attr, sizeof (pasis_md_attr_t));
    }

    dump_attrs ("Setattr before", attr_out);

    /* Check guard to make sure setattr should actually occur */
    if (guard_ctime[0] != 0 && guard_ctime[1] != 0) {
        if (guard_ctime[0] != attr_out->ctime[0] || 
            guard_ctime[1] != attr_out->ctime[1]) {
            /* NFS3ERR_NOT_SYNC error */
            return OPERR_NOT_SYNC;
        }
    }

    /* Set attrs */
    if (sattr->set_mode == 1) {

        if (((sattr->mode & MD_ATTR_MODE_MASK) != 
             (attr_out->mode & MD_ATTR_MODE_MASK)) && (attr_out->mode != 0)) {
            /* Can't change mode type once set */
            return OPERR_INVAL;
        }
        attr_out->mode = sattr->mode;
    }

    if (sattr->set_uid == 1) {
        attr_out->uid = sattr->uid;
    }

    if (sattr->set_gid == 1) {
        attr_out->gid = sattr->gid;
    }

    if (sattr->set_size == 1) {
        /* May result in truncation or sparse growing */
        attr_out->size = sattr->size;
    }

    if (sattr->set_blocks == 1) {
        attr_out->blocks = sattr->blocks;
    }

    if (sattr->set_atime == 1) {
        attr_out->atime[0] = sattr->atime[0];
        attr_out->atime[1] = sattr->atime[1];
    }

    if (sattr->set_mtime == 1) {
        attr_out->mtime[0] = sattr->mtime[0];
        attr_out->mtime[1] = sattr->mtime[1];
    }

    /* Write out attrs */
    error = attr_write (fid, new_version, cond_version, attr_out);
    if (error) {
        PASIS_PRINT_ERR ("%s: error setting attributes\n", __FUNCTION__);
        return OPERR_IO;
    }

    dump_attrs ("Setattr after", attr_out);

    /* printf ("\nOUT %s  (OK)\n", __FUNCTION__); */

    return OP_OK;
}


pasis_error_t
pasis_md_op_rename(p_uint32             mtime[2],
                   MDVersionedObject   *from_dir,
                   MDVersionedObject   *to_dir,
                   MDVersionedObject   *from_fid,
                   MDVersionedObject   *to_fid,
                   pasis_md_hts_t      *new_hts,
                   pasis_md_hts_t     **cond_hts,
                   char                *from_name,
                   char                *to_name,
                   pasis_md_attr_t     *from_dir_attr_out, /* Out args */
                   pasis_md_attr_t     *to_dir_attr_out)   /* Out args */
{
    pasis_md_dir_entry_t *dent;
    pasis_md_attr_t       from_attr, to_attr;
    pasis_error_t         error;

    p_byte               *from_blk, *to_blk;
    p_uint64              from_blkno, to_blkno;
    unsigned int          from_offset, to_offset;
    unsigned int          to_from_dir_same=0;

    MDVersion new_from_dir_version(&new_hts[0]),
        cond_from_dir_version(cond_hts[0]);
    MDVersion new_to_dir_version, cond_to_dir_version,
        new_from_fid_version, cond_from_fid_version,
        new_to_fid_version, cond_to_fid_version;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_RENAME, 4,
                   "\nIN %s (from=%s, to=%s)\n", __FUNCTION__,
                   from_name, to_name);

    /** Perform validation -- read-only phase **/

    int to_dir_index, from_fid_index, to_fid_index;

    if (to_dir == NULL) {
        to_from_dir_same = 1;
        to_dir = from_dir;

        to_dir_index = 0;
        from_fid_index = 1;
        to_fid_index = 2;
    } else {
        to_dir_index = 1;
        from_fid_index = 2;
        to_fid_index = 3;
    }

    new_to_dir_version = MDVersion(&new_hts[to_dir_index]);
    cond_to_dir_version = MDVersion(cond_hts[to_dir_index]);
        
    new_from_fid_version = MDVersion(&new_hts[from_fid_index]);
    cond_from_fid_version = MDVersion(cond_hts[from_fid_index]);

    new_to_fid_version = MDVersion(&new_hts[to_fid_index]);
    cond_to_fid_version = MDVersion(cond_hts[to_fid_index]);


    /* Get attributes of from_fid */
    void* from_fid_attr = from_fid->read_small_data(cond_from_fid_version);
    if (from_fid_attr == NULL) {
        PASIS_PRINT_ERR ("%s: error reading attributes\n", __FUNCTION__);
        printf ("rename failed (%d)\n", __LINE__);
        return OPERR_IO;
    } else {
        memcpy (&from_attr, from_fid_attr, sizeof (pasis_md_attr_t));
    }

    /* Check from_dir for from_name -- verify against from_fid passed in */
    error = pasis_md_dir_find_entry (from_dir, cond_from_dir_version,
                                     from_name, &from_blk, 
                                     &from_blkno, &from_offset, 
                                     from_dir_attr_out);
    if (error) {
        printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
        return error;
    }
    
    /* Set directory entry pointer */
    dent = (pasis_md_dir_entry_t *)(from_blk + from_offset);

    /* Check from_fid's fid matches from_name's fid */
    if (md_fid_compare (&from_fid->oid.fid, &dent->fid) != 0) {
        printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
        return OPERR_STALE;
    }

    /* Check to_dir for to_name -- verify against to_fid passed in */
    error = pasis_md_dir_find_entry (to_dir, cond_to_dir_version,
                                     to_name, &to_blk, 
                                     &to_blkno, &to_offset, to_dir_attr_out);

    if ((error == OPERR_NOENT && to_fid != NULL) ||
        (error == OP_OK && to_fid == NULL)) {
        /* NFS server should retry op */
        PASIS_PRINT_ERR ("%s: error to_name not in sync with to_fid\n", 
                         __FUNCTION__);
        printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
        return OPERR_NOT_SYNC;
    } else if (error != OPERR_NOENT && error != OP_OK) {
        printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
        return error;
    }

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_RENAME, 4,
                   "to_blk=%p, to_blkno=%"PRIu64", to_offset=%u\n",
                   to_blk, to_blkno, to_offset);
    
    /* Validity checking on to_fid if to_name exists */
    if (to_fid != NULL) {

        /* Verify that if to_name exists that to_fid matches its fid */
        dent = (pasis_md_dir_entry_t *)(to_blk + to_offset);
        if (md_fid_compare (&to_fid->oid.fid, &dent->fid) != 0) {
            printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
            return OPERR_STALE;
        }
        
        /* Get to_fid's attributes */
        void* to_fid_attr = to_fid->read_small_data(cond_to_fid_version);
        if (to_fid_attr == NULL) {
            printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
            return OPERR_IO;
        } else {
            memcpy (&to_attr, to_fid_attr, sizeof (pasis_md_attr_t));
        }
        
        /* Mode mismatch of src and destination names */
        if ((MD_ATTR_MODE_MASK & to_attr.mode) != 
            (MD_ATTR_MODE_MASK & from_attr.mode)) {
            printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
            return OPERR_EXIST;
        }

        /* Check that if to_fid is a directory, that it is empty */
        if (MD_ATTR_IS_DIR (&to_attr) &&
            pasis_md_dir_blk_is_empty(to_fid, &cond_to_fid_version, &to_attr) != OP_OK) {
            printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
            return OPERR_EXIST;
        }
    }

    /** Perform rename operation now **/


    // need the new block, because we want to use that in the
    // unlink below, so we don't "forget" this link
    p_byte *new_to_blk = to_blk;

    /* Unlink to_fid from to_dir if it exists */
    if (to_fid != NULL) {
        p_byte *new_to_blk2;
        error = internal_unlink (mtime, to_dir, to_fid, to_name, 
                                 to_offset, to_blkno, to_blk,
                                 new_to_dir_version, cond_to_dir_version,
                                 new_to_fid_version, cond_to_fid_version,
                                 &to_attr, to_dir_attr_out, &new_to_blk2);
        if (error) {
            printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
            return error;
        }

        if (to_from_dir_same) {
            // so that the link, unlink calls below modify the directory version
            // which had to_fid removed
            new_to_blk = new_to_blk2;
        }
    }

    /* Link in from_fid in to_dir */

    /* Use the to_blk to pass into link, it is either last block of to_dir,
     * or has just had the to_name removed from it */
    error = internal_link (mtime, from_fid, to_dir, to_name, to_blkno,
                           new_to_blk, new_from_fid_version, cond_from_fid_version,
                           new_to_dir_version, cond_to_dir_version,
                           &from_attr, to_dir_attr_out, &new_to_blk);
    if (error) {
        printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
        return error;
    }

    /* Unlink from_fid from from_dir */

    /* We need to pass in the block containing from_name, however if to_dir and
     * from_dir are the same, and to_name and from_name both existed in the
     * same block, things get complicated, since from_name's offset may not
     * be correct anymore. */
    if (to_blkno == from_blkno && to_from_dir_same &&
        pasis_md_dir_lookup (new_to_blk, from_name, &to_offset) == PASIS_SUCCESS) {

        error = internal_unlink (mtime, from_dir, from_fid, from_name,
                                 to_offset, to_blkno, new_to_blk,
                                 new_from_dir_version, cond_from_dir_version,
                                 new_from_fid_version, cond_from_fid_version,
                                 &from_attr, from_dir_attr_out);
    } else {
        error = internal_unlink (mtime, from_dir, from_fid, from_name,
                                 from_offset, from_blkno, from_blk,
                                 new_from_dir_version, cond_from_dir_version,
                                 new_from_fid_version, cond_from_fid_version,
                                 &from_attr, from_dir_attr_out);
    }

    if (error) {
        printf ("%s: error line=%d\n", __FUNCTION__, __LINE__);
        return error;
    }


    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_RENAME, 4,
                   "\nFROMDIR after unlink of %s: \n", from_name);
    /* pasis_md_dump_dir (ipc_op, from_dir); */

    return OP_OK;
}


/***************** Read only operations ******************/

pasis_error_t
pasis_md_op_getattr (MDVersionedObject   *file,
                     MDVersion           *version,
                     pasis_md_attr_t     *attr_out) /* Out args */
{
    /* printf ("\nIN %s\n", __FUNCTION__); */

    /* Get FID attrs */
    void* saved_attr = file->read_small_data(*version);
    if (saved_attr == NULL) {
        // XXX: object sync attrs??
        memset (attr_out, 0, sizeof (pasis_md_attr_t));
    } else {
        memcpy (attr_out, saved_attr, sizeof (pasis_md_attr_t));
    }

    // dump_attrs ("Getattr", attr_out);
    /* printf ("\nOUT %s  (OK)\n", __FUNCTION__); */

    return OP_OK;
}


pasis_error_t
pasis_md_op_lookup (MDVersionedObject   *directory,
                    MDVersion           *version,
                    char                *name,
                    pasis_md_fid_t      *fid_out,      /* Out args */
                    pasis_md_attr_t     *dir_attr_out) /* Out args */
{
    pasis_error_t error;
    unsigned int  offset;
    p_uint64      blkno;
    p_byte       *blk;

    pasis_md_dir_entry_t *dent;

    /* printf ("\nIN %s\n", __FUNCTION__); */

    /** Lookup to_name in to_dir, get to_dir's attrs while at it **/
    error = pasis_md_dir_find_entry (directory, *version, name, &blk, &blkno, 
                                     &offset, dir_attr_out);
    if (error) {
        return error;
    }

    dent = (pasis_md_dir_entry_t *)(blk + offset);

    /* Lookup name in dir -- find fid_out */
    memcpy (fid_out, &dent->fid, sizeof (pasis_md_fid_t));

    /* Remainder of output args: fid_attr, and location should be filled in 
     * at the upper level (through getattr, and fid_lookup) */

    dump_attrs ("Lookup dir", dir_attr_out);
    /* printf ("\nOUT %s  (OK)\n", __FUNCTION__); */

    return OP_OK;
}


pasis_error_t
pasis_md_op_readlink (MDVersionedObject   *fid,
                      MDVersion           *version,
                      unsigned int        *path_len_out, /* Out args */
                      char                *path_out,     /* Out args */
                      pasis_md_attr_t     *fid_attr_out) /* Out args */
{
    pasis_md_op_sym_blk_t *sym_blk;

    /* printf ("\nIN %s\n", __FUNCTION__); */

    /* Getattrs first to make sure this is a symlink */
    void* saved_attr = fid->read_small_data(*version);
    if (saved_attr == NULL) {
        return OPERR_IO;
    } else {
        memcpy (fid_attr_out, saved_attr, sizeof (pasis_md_attr_t));
    }

    if (!MD_ATTR_IS_LNK (fid_attr_out)) {
        return OPERR_INVAL;
    }

    /* Read symlink and fetch fid attrs */
    sym_blk = (pasis_md_op_sym_blk_t *)fid->read(*version, 0);
    if (sym_blk == NULL) {
        PASIS_PRINT_ERR ("%s: error reading symlink\n", __FUNCTION__);
        return OPERR_IO;
    }

    *path_len_out = sym_blk->len;
    memcpy (path_out, sym_blk->path, sym_blk->len);

    /* printf ("\nOUT %s  (OK)\n", __FUNCTION__); */

    return OP_OK;
}

  
/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */             
