/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_lists.h"
#include "pasis_types.h"
#include "pasis_timer.h"

#include "pasisio.h"

#include "pasis_md_prot.h"

#include "pasis_md_s4.h"
#include "pasis_md_auth.h"
#include "pasis_md_hist_matrix.h"
#include "pasis_md_work.h"
#include "pasis_md_validation.h"
#include "pasis_md_memory_versioning.h"
#include "pasis_md_log.h"
#include "pasis_md_xdr.h"
#include "pasisio_md_protocol.h"


#define PASIS_MD_TIMER_VALIDATION_NUM  1

#define PASIS_MD_TIMER_VALIDATION_OHS_AUTH  0


typedef struct pasis_md_validation_stats_s {
    pasis_timer_state_t  *timers;
    unsigned int          latest_time_reads;
    unsigned int          num_ohs_validations;
    unsigned int          num_ohs_quick_validations;
} pasis_md_validation_stats_t;

static pasis_md_validation_stats_t stats;


static int
already_executed_op (p_uint32                node_id,
                     MDVersionedObject      *object,
                     pasis_md_hist_matrix_t *hist_matrix,
                     pasis_md_hts_t         *ts)
{
    /* First, check if the OHS says we have executed the operation.  This may
     * sound dumb, but in the fast repair case (for example), the client sends
     * the operation to _all_ servers, even the ones that already have the op.
     * In that case, the OHS will show that we have the operation. */
    if (pasis_md_hist_matrix_node_has_ts (node_id, hist_matrix, ts)) {
        return 1;
    }

    /* Second, we need to look at our local replica history and see if we have
     * executed the op.  Hence, we could've executed the op after the client
     * constructed its OHS. */
    MDVersion version(ts);
    return object->has_version(version);
}


static pasis_error_t
validate_matrix (p_uint32                  node_id,
                 pasis_md_protocol_t      *protocol,
                 MDVersionedObject        *object,
                 pasis_md_hist_matrix_t   *hist_matrix,
                 p_uint32                  proof_fid_index,
                 pasis_md_hist_matrix_t   *proof_matrix,
                 p_byte                   *op_hash,
                 unsigned int              writing_barrier,
                 pasis_md_hts_t          **latest_time,      /* Out */
                 pasis_md_hts_t          **ts_cond_out,      /* Out */
                 pasis_md_hts_t          **ts_candidate,     /* Out */
                 int                      *already_executed, /* Out */
                 int                      *doing_repair,     /* Out */
                 int                      *doing_fast_repair)/* Out */
{
    pasis_md_hts_t  *ts_latest, *ts_complete, *ts_barrier;
    pasis_error_t    latest_status, candidate_status, barrier_status;
    int              dummy1, dummy2;
    pasis_error_t    error;

    *already_executed = 0;
    *doing_repair = 0;
    *doing_fast_repair = 0;

    /* Perform classification */
    pasis_md_hist_matrix_classify_node (hist_matrix, protocol, &ts_barrier,
                                        &barrier_status, &ts_latest,
                                        &latest_status, &dummy1, ts_candidate,
                                        &candidate_status, &dummy2,
                                        &ts_complete);

    *latest_time = (ts_barrier == NULL) ? ts_latest : ts_barrier;
    *ts_cond_out = ts_complete;

    pasis_md_log_msg ("%s: latest complete=%llu, latest=%llu, barrier=%llu",
                      __FUNCTION__, ts_complete->time, (*latest_time)->time,
                      (ts_barrier == NULL) ? 0 : ts_barrier->time);

    /* Check if a barrier is needed and not being written */
    /* Note: this differs from pseudo-code in that ts_candidate is used
     * instead of ts_complete.  This is ok if we don't require barriers
     * ontop of repairables.  Must just validate that the op being performed
     * is the correct repair op.
     *
     * Also, need only check ts_barrier == NULL, instead of checking that the
     * barrier is the latest -- since it is non-NULL if it is the latest.
     */

    /* ts_candidate holds latest non-incomp, ts_latest the latest non-barrier */
    if (md_hts_compare (*ts_candidate, ts_latest) != 0 &&
        ((ts_barrier == NULL) || 
         (barrier_status != PASIS_ERROR_PROTO_COMPLETE))) {
        
        if (writing_barrier == 0) {
            /* Need a barrier */
            error = PASIS_MD_ERR_HIST_BARRIER;
            return error;
        }

    } else {
        if (writing_barrier) {
            /* Don't need a barrier */
            if (ts_barrier && barrier_status == PASIS_ERROR_PROTO_COMPLETE) {
                return PASIS_MD_ERR_HIST_NOBARRIER;
            }
            //error = PASIS_MD_ERR_HIST_NOBARRIER;
            //return error;
        }
    }
    
    /* If the candidate is repairable, it may be complete due to "deduction".
     * If the client sent us a "proof" matrix as a proof that this is indeed a
     * complete version, we'll TRUST them.  It's easy to validate this proof
     * matrix but it's just ugly code.  It doesn't take much time to do the
     * check, so we can skip it and not worry about "being too fast". */
    if ((candidate_status == PASIS_ERROR_PROTO_REPAIR) &&
        proof_fid_index != UINT_MAX) {
        printf("%s: got a repairable, assuming there's a deducing proof that this is "
               "complete, yadda yadda yadda\n", __FUNCTION__);
        *ts_cond_out = *ts_candidate;
        return PASIS_SUCCESS;
    }

    /* If we should be performing a repair, validate that the op being performed
     * is the correct op to do the repair.  We've already checked if we need a
     * barrier or not */
    if (candidate_status == PASIS_ERROR_PROTO_REPAIR && writing_barrier == 0) {
        *doing_repair = 1;
        if (md_hts_compare (*ts_candidate, ts_latest) == 0) {
            *doing_fast_repair = 1;
        }

        if (already_executed_op (node_id, object, hist_matrix, *ts_candidate)) {
            *already_executed = 1;
        } else {
            //printf ("%s: repair case; executing op that we haven't seen "
            //        "before (ts = %llu), fast_repair = %d\n", __FUNCTION__,
            //        (*ts_candidate)->time, *doing_fast_repair);
            /* execute op normally */
        }
    }

    return PASIS_SUCCESS;
}


inline static int
replica_histories_are_identical (pasis_md_hist_matrix_t *hist_matrix)
{
    p_uint64 first_bitmap;

    PASIS_DASSERT (hist_matrix->ts_num > 0);

    first_bitmap = hist_matrix->ts_bitmap[0];

    for (p_uint32 i = 1; i < hist_matrix->ts_num; ++i) {
        if (hist_matrix->ts_bitmap[i] != first_bitmap) {
            return 0;
        }
    }

    return 1;
}


/* Optimization when checking the authenticators.  If the RHs are all the same,
 * then we can just hash once over a single RH, and use a keyed hash of that
 * hash for each S-N auth check.  This makes a big difference! */
inline static pasis_error_t
do_quick_auth_validation (pasis_md_protocol_t     *protocol,
                          pasis_md_hist_matrix_t  *hist_matrix,
                          p_uint32                 node_id)
{
    p_uint32                num_correct_verifiers = 0, repairable_threshold, got_first_rh;
    int                     cmp;
    pasis_md_auth_t         rh_hash;
    auth_key_t              key;

    /* Gotta reverse the key order here, since the S-N that generated the key
     * has the opposite point of view regarding who is local/remote.  */
    key.remote_node_id = node_id;

    got_first_rh = 0;
    for (p_uint32 i = 0; i < protocol->N; ++i) {
        pasis_md_hts_t   *remote_histories[PASIS_MD_HIST_MAX_VERSIONS];
        p_uint32          remote_history_num;
        pasis_md_auth_t  *rh_auth, true_auth;

        pasis_md_hist_matrix_get_vec (i, hist_matrix, &remote_history_num,
                                      &rh_auth, remote_histories);

        if (remote_history_num) {

            if (got_first_rh == 0) {
                pasis_md_auth_hash_indirect (remote_history_num,
                                             remote_histories, &rh_hash);
                got_first_rh = 1;
            }

            key.local_node_id  = i;

            pasis_md_auth_key_hash (&key, &rh_hash, &true_auth);

            if (DO_NOT_CALCULATE_AUTHENTICATORS) {
                cmp = 0;
            } else {
                cmp = memcmp(&true_auth, rh_auth, PASIS_HASH_SIZE);
            }

            if (cmp == 0) {
                ++num_correct_verifiers;
            } else {
#if 1
                printf("%s: node %u: authenticator-check failed for node %u. "
                       "true auth. = %x, RH auth = %x\n", __FUNCTION__, node_id,
                       i, *(p_uint32*)&true_auth, *(p_uint32*)rh_auth);

                printf("%s: History vector:\n", __FUNCTION__);
                for (p_uint32 k = 0; k < remote_history_num; ++k) {
                    printf("\t[%u] : %"PRIu64"", k, remote_histories[k]->time);
                    if (remote_histories[k]->barrier_flag) {
                        printf(" [BARRIER]");
                    }
                    printf("\n");
                }

                printf("Authenticators:\n");
                for (p_uint32 j = 0; j < protocol->N; ++j) {
                    pasis_md_hist_matrix_get_vec (j, hist_matrix, &remote_history_num,
                                                  &rh_auth, remote_histories);
                    if (rh_auth) 
                        printf("\t [%d]: %x\n", j, *(p_uint32*)rh_auth);
                    else
                        printf("\t [%d]: NULL\n", j);
                }
#endif
                PASIS_DASSERT (0); /* should never happen */
            }
        }
    }

    repairable_threshold = PasisIO_GetRepairableThreshold (protocol);

    if (num_correct_verifiers < repairable_threshold) {
        printf ("%s: authenticator-checking failed; (%u)\n",
                __FUNCTION__, num_correct_verifiers);
        return PASIS_MD_ERR_HIST_MATRIX_AUTH;
    }
    
    return PASIS_SUCCESS;
}


static pasis_error_t
validate_auth (pasis_md_protocol_t     *protocol,
               MDVersionedObject       *object,
               pasis_md_hist_matrix_t  *hist_matrix,
               p_uint32                 node_id) 
{
    pasis_error_t           error;
    pasis_md_auth_t        *auth;
    p_uint32                num_correct_verifiers = 0, repairable_threshold;

    PASIS_TIMER_START ();

    error = PASIS_SUCCESS;

    ++stats.num_ohs_validations;

    if (replica_histories_are_identical (hist_matrix)) {
        ++stats.num_ohs_quick_validations;
        error = do_quick_auth_validation(protocol, hist_matrix, node_id);
    } else {
        for (p_uint32 i = 0; i < protocol->N; ++i) {
            pasis_md_hts_t *remote_histories[PASIS_MD_HIST_MAX_VERSIONS];
            p_uint32        remote_history_num;

            pasis_md_hist_matrix_get_vec (i, hist_matrix, &remote_history_num,
                                          &auth, remote_histories);

            if (remote_history_num) {
                error = pasis_md_auth_verify (node_id, i, remote_history_num,
                                              remote_histories, auth);
                if (!error) {
                    ++num_correct_verifiers;
                } else {
                    printf("%s: node %u: authenticator-check failed for node %u\n",
                           __FUNCTION__, node_id, i);

                    printf("%s: History vector from OHS:\n", __FUNCTION__);
                    for (p_uint32 k = 0; k < remote_history_num; ++k) {
                        printf("\t[%u] : %"PRIu64"", k, remote_histories[k]->time);
                        if (remote_histories[k]->barrier_flag) {
                            printf(" [BARRIER]");
                        }
                        printf("\n");
                    }

                    printf("%s: local history: %s\n", __FUNCTION__, object->dump_history());

                    printf("Authenticators:\n");
                    for (p_uint32 j = 0; j < protocol->N; ++j) {
                        pasis_md_hist_matrix_get_vec (j, hist_matrix, &remote_history_num,
                                                      &auth, remote_histories);
                        if (auth) 
                            printf("\t [%d]: %x\n", j, *(p_uint32*)auth);
                        else
                            printf("\t [%d]: NULL\n", j);
                    }

                    PASIS_DASSERT (0); /* should never happen */
                }
            }
        }

        repairable_threshold = PasisIO_GetRepairableThreshold (protocol);

        if (num_correct_verifiers < repairable_threshold) {
            printf ("%s: authenticator-checking failed; (%u)\n",
                    __FUNCTION__, num_correct_verifiers);
            error = PASIS_MD_ERR_HIST_MATRIX_AUTH;
        }
    }

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_TIMER_VALIDATION_OHS_AUTH);

    return error;
}


static pasis_error_t
check_for_stale_ohs (MDVersionedObject *object, pasis_md_hts_t *ohs_latest_ts)
{
    MDVersion local_latest;

    // If we don't have any versions, local_latest will be set to 0
    object->get_latest(local_latest);

    /* Fail validation if replica's latest TS is greater than the OHSs */
    if (local_latest > ohs_latest_ts) {
        //printf ("%s: stale history vector (local=%"PRIu64" > ohs=%"PRIu64")\n",
        //      __FUNCTION__, local_latest.timestamp.time,
        //      ohs_latest_ts->time);
        return PASIS_MD_ERR_HIST_NOT_LATEST;
    } else {
      return PASIS_SUCCESS;
    }
}


#if PASIS_DEBUG > 0
static void
check_for_duplicate_ts(pasis_md_hist_matrix_t *hist_matrix)
{
    for (p_uint32 i = 0; i < hist_matrix->ts_num; ++i) {
        for (p_uint32 j = i + 1; j < hist_matrix->ts_num; ++j) {
            if (md_hts_compare(hist_matrix->ts_vector[i],
                               hist_matrix->ts_vector[j]) == 0) {
                fprintf(stderr, "%s: found duplicate TS in OHS", __FUNCTION__);
                PASIS_DASSERT(0);
            }
        }
    }
}
#endif


pasis_error_t
pasis_md_validate_hist_matrix (p_uint32                node_id,
                               MDVersionedObject      *object,
                               pasis_md_protocol_t    *protocol,
                               pasis_md_hist_matrix_t *hist_matrix,
                               p_uint32                proof_fid_index,
                               pasis_md_hist_matrix_t *proof_matrix,
                               p_byte                 *op_hash,
                               unsigned int            writing_barrier,
                               pasis_md_hts_t        **cond_hts,   /* Out args */
                               pasis_md_hts_t        **latest_hts,/* Out args */
                               pasis_md_hts_t        **candidate_hts,/* Out*/
                               int                    *already_executed,/* Out*/
                               int                    *doing_repair,    /* Out*/
                               int                    *doing_fast_repair)/* Out*/
{
    pasis_error_t error, classify_error;

#if PASIS_DEBUG > 0
    check_for_duplicate_ts(hist_matrix);
#endif

    /* Perform OHS validation.  Classify OHS and check what method (update,
     * barrier, copy-forward) is required and whether the client is performing
     * that method. */
    classify_error = validate_matrix (node_id, protocol, object, hist_matrix,
                                      proof_fid_index, proof_matrix,
                                      op_hash, writing_barrier, latest_hts,
                                      cond_hts, candidate_hts, already_executed,
                                      doing_repair, doing_fast_repair);

    /* Validate authenticators */
    error = validate_auth (protocol, object, hist_matrix, node_id);
    if (error) goto exit;
    
    /* Check that RH in OHS is not stale. */
    error = check_for_stale_ohs (object, *latest_hts);
    if (error) goto exit;

    /* Delay returning error on classification until now. This aids the
     * barrier write operation. */
 exit:
    if (classify_error) {
        return classify_error;
    } else {
      return error;
    }
}


void
pasis_md_validation_stats_init (void)
{
    memset (&stats, 0, sizeof (pasis_md_validation_stats_t));

    /* Set up timers */
    stats.timers = Pasis_TimerInit ("Validation_Timers", 
                                    PASIS_MD_TIMER_VALIDATION_NUM, 0);

    Pasis_TimerRegister (stats.timers, PASIS_MD_TIMER_VALIDATION_OHS_AUTH,
                         "OHS auth. validation");
}


void
pasis_md_validation_stats_dump (void)
{
    Pasis_TimerShutdown (stats.timers);

    PASIS_PRINT_INFO ("\nLatest Time Reads    :  %u\n", 
                      stats.latest_time_reads);
    PASIS_PRINT_INFO ("OHS validations      :  %u\n", 
                      stats.num_ohs_validations);
    PASIS_PRINT_INFO ("OHS FAST validations :  %u\n", 
                      stats.num_ohs_quick_validations);
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
